package jp.riken.brain.ni.samuraigraph.application;

import java.awt.EventQueue;
import java.awt.Point;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfoSet;
import jp.riken.brain.ni.samuraigraph.base.SGDrawingWindow;
import jp.riken.brain.ni.samuraigraph.base.SGFigure;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElement;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementAxis;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementAxisBreak;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementGraph;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementLegend;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementShape;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementSignificantDifference;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementString;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementTimingLine;
import jp.riken.brain.ni.samuraigraph.base.SGPropertyMap;
import jp.riken.brain.ni.samuraigraph.base.SGPropertyResults;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.data.SGDataTypeConstants;
import jp.riken.brain.ni.samuraigraph.data.SGDataUtility;
import jp.riken.brain.ni.samuraigraph.data.SGIDataColumnTypeConstants;
import jp.riken.brain.ni.samuraigraph.data.SGNetCDFDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.data.SGNetCDFFile;
import jp.riken.brain.ni.samuraigraph.figure.SGIAxisBreakConstants;
import jp.riken.brain.ni.samuraigraph.figure.SGILegendConstants;
import jp.riken.brain.ni.samuraigraph.figure.SGIShapeConstants;
import jp.riken.brain.ni.samuraigraph.figure.SGISignificantDifferenceConstants;
import jp.riken.brain.ni.samuraigraph.figure.SGIStringConstants;
import jp.riken.brain.ni.samuraigraph.figure.SGITimingLineConstants;
import ucar.nc2.NetcdfFile;

/**
 * The property manager.
 * 
 * @author okumura
 */
class SGCommandManager implements SGIApplicationCommandConstants, 
		SGIDataColumnTypeConstants,
		SGILegendConstants, SGIStringConstants,
		SGISignificantDifferenceConstants, SGIAxisBreakConstants,
		SGITimingLineConstants, SGIShapeConstants {

	private SGMainFunctions mMain = null;

	private SGWindowManager mWindowManager = null;
	
    SGCommandManager(final SGMainFunctions main) {
    	super();
    	this.mMain = main;
    	this.mWindowManager = main.mWindowManager;
    }

    private SGPropertyMap getPropertiesMap(String command, List<String> strList) {
    	SGPropertyMap map = new SGPropertyMap();
        for (String str : strList) {
            final int eqIndex = str.indexOf('=');
            if (eqIndex == -1 || eqIndex == 0) {
                return null;
            }
            final int len = str.length();
            final String key = str.substring(0, eqIndex);
            String value = (eqIndex == len - 1) ? "" : str
                    .substring(eqIndex + 1, len);
            
            // Searches the alias objects.
        	Alias alias = this.mAliasMap.get(value);
            if (alias != null) {
            	String aCommand = alias.getCommand();
            	if (aCommand == null || command.equalsIgnoreCase(aCommand)) {
            		// When the command of the alias is null, the alias is applied to all commands.
            		String aKey = alias.getKey();
            		if (aKey == null || key.equalsIgnoreCase(aKey)) {
                		// When the property key of the alias is null, the alias is applied to 
            			// all properties.
            			if (alias.getRepr().equals(value)) {
            				// Replaces the value.
            				value = alias.getValue();
            			}
            		}
            	}
            }
            
            map.putValue(key, value);
        }
        return map;
    }
    
    private boolean isCommandBringToFront(final String command) {
        return command.toUpperCase().startsWith(COM_HEADER_BRING_TO_FRONT.toUpperCase());
    }

    private boolean isCommandBringForward(final String command) {
        return command.toUpperCase().startsWith(COM_HEADER_BRING_FORWARD.toUpperCase());
    }

    private boolean isCommandSendBackward(final String command) {
        return command.toUpperCase().startsWith(COM_HEADER_SEND_BACKWARD.toUpperCase());
    }

    private boolean isCommandSendToBack(final String command) {
        return command.toUpperCase().startsWith(COM_HEADER_SEND_TO_BACK.toUpperCase());
    }
    
    private String getCommandEndToMove(final String command, final String start) {
        return command.substring(start.length(), command.length());
    }
    
    private SGIFigureElement getFigureElementToMove(final String name, final SGFigure f) {
        final String uc = name.toUpperCase();
        Class<? extends SGIFigureElement> cl = null;
        if (uc.endsWith(COM_DATA.toUpperCase())) {
            cl = SGIFigureElementGraph.class;
        } else if (uc.endsWith(COM_LABEL.toUpperCase())) {
            cl = SGIFigureElementString.class;
        } else if (uc.endsWith(COM_SIGNIFICANT_DIFFERECE.toUpperCase())) {
            cl = SGIFigureElementSignificantDifference.class;
        } else if (uc.endsWith(COM_AXIS_BREAK.toUpperCase())) {
            cl = SGIFigureElementAxisBreak.class;
        } else if (uc.endsWith(COM_TIMING_LINE.toUpperCase())) {
            cl = SGIFigureElementTimingLine.class;
        } else if (uc.endsWith(COM_SHAPE.toUpperCase())) {
            cl = SGIFigureElementShape.class;
        } else {
            return null;
        }
        SGIFigureElement el = f.getIFigureElement(cl);
        return el;
    }

    private Integer getId(final List<String> argsList) {

        if (argsList.size() < 1) {
            return null;
        }
        String str1 = (String) argsList.get(0);

        final Integer id = SGUtilityText.getInteger(str1);
        if (id == null) {
            return null;
        }
        if (id <= 0) {
        	return null;
        }
        
		BigInteger bIntMax = new BigInteger(Integer.toString(Integer.MAX_VALUE));
		BigInteger bnum = new BigInteger(str1);
		if (bnum.compareTo(bIntMax) > 0) {
			return null;
		}

        return id;
    }
    
    // parse given string into the command and arguments
    private String parseCommand(final String line, final List<String> argsList) {
        final int sIndex = line.indexOf('(');
        final int eIndex = line.lastIndexOf(')');
        if (sIndex == -1 || eIndex == -1) {
            return line;
        }
        if (sIndex > eIndex) {
            return line;
        }
        if (sIndex == 0) {
            return line;
        }

        String command = line.substring(0, sIndex);
        String args = line.substring(sIndex + 1, eIndex);

/*        
        boolean inDq = false;   // true between two double quotation marks
        int cntBracket = 0;		// a counter for brackets
        List<Character> cList = new ArrayList<Character>();		// a temporary character list
        for (int ii = 0; ii < args.length(); ii++) {
            final char c = args.charAt(ii);
            if (c == '"') {
            	if (ii != args.length() - 1) {
                	// the next character
            		final char cNext = args.charAt(ii + 1);
            		if (cNext == '"' && !inDq) {
            			// two succesive double quotations outside the double quotations
            			cList.add('"');
            			cList.add('"');
                        char[] cArray = new char[cList.size()];
                        for (int jj = 0; jj < cArray.length; jj++) {
                            cArray[jj] = cList.get(jj).charValue();
                        }
                        String str = new String(cArray);
                        str = str.trim();
                        argsList.add(str);
                        cList.clear();
                        
                        // to the next character other than spaces or to the end of line
                        if (ii == args.length() - 2) {
                        	// last two characters are both double quotations
                        	// break the loop
                        	break;
                        } else {
                        	// search the next character other than spaces
                        	for (int jj = ii + 2; jj < args.length(); jj++) {
                        		final char cc = args.charAt(jj);
                        		if (cc != ' ') {
                        			ii = jj;
                        			break;
                        		}
                        	}
                        	continue;
                        }
            		}
            	}
            	
                if (!inDq) {
                    // from outside of two double quotation marks
                    inDq = true;
                } else {
                    if (args.charAt(ii - 1) != '"') {
                        inDq = false;
                    }
                }
            } else if (c == '(') {
                cntBracket++;
            } else if (c == ')') {
                cntBracket--;
            }

            final boolean end = (ii == args.length() - 1);
            if ((c == ',' && !inDq && cntBracket == 0) || end) {
                if (end) {
                    cList.add(Character.valueOf(c));
                }
                
                // a comma at outside of two double quotation marks
                // or at the end of the string
                char[] cArray = new char[cList.size()];
                for (int jj = 0; jj < cArray.length; jj++) {
                    cArray[jj] = cList.get(jj).charValue();
                }
                String str = new String(cArray);
                str = str.trim();
                argsList.add(str);
                cList.clear();
            } else if (c == '"' && args.charAt(ii - 1) == '"') {
                continue;
            } else if (c == '\\' && args.charAt(ii - 1) == '\\') {
                continue;
            } else {
                cList.add(Character.valueOf(c));
            }
        }
*/

        final int len = args.length();
        
        // find commas
        List<Integer> cIndexList = new ArrayList<Integer>();
        int cntDq = 0;
        int cntBracket = 0;
        for (int ii = 0; ii < len; ii++) {
        	final char c = args.charAt(ii);
        	if (c == '"') {
        		cntDq++;
        	} else if (c == '(') {
                if (cntDq % 2 == 0) {
                    // outside a string
                    cntBracket++;
                }
            } else if (c == ')') {
                if (cntDq % 2 == 0) {
                    // outside a string
                    cntBracket--;
                }
            } else if (c == ',') {
                if (cntDq % 2 == 0 && cntBracket == 0) {
                    // outside a string and a bracket
                    cIndexList.add(Integer.valueOf(ii));
                }
            }
        }

        // split the argument to tokens
        int curIndex = 0;
        List<String> tokenList = new ArrayList<String>();
        for (int ii = 0; ii < cIndexList.size(); ii++) {
        	final int index = cIndexList.get(ii);
        	String token = args.substring(curIndex, index);
            token = token.trim();
        	tokenList.add(token);
        	curIndex = index + 1;
        	if (curIndex >= len) {
        		tokenList.add("");
        		break;
        	}
        }
		tokenList.add(args.substring(curIndex, len).trim());
		
		// reduce successive commas in each token
		for (int ii = 0; ii < tokenList.size(); ii++) {
			String token = tokenList.get(ii);
			final int fIndex = token.indexOf('"');
			if (fIndex == -1) {
				argsList.add(token);
				continue;
			}
			final int lIndex = token.lastIndexOf('"');
			if (lIndex == fIndex) {
				return null;
			}
            if (lIndex != token.length() - 1) {
                return null;
            }
			String head = token.substring(0, fIndex);
			String body = token.substring(fIndex + 1, lIndex);

            // check whether double quotations are successive
            for (int jj = 0; jj < body.length(); jj++) {
                final char c = body.charAt(jj);
                if (c == '"') {
                    if (jj == body.length() - 1) {
                        return null;
                    }
                    final char cNext = body.charAt(jj + 1);
                    if (cNext != '"') {
                        return null;
                    }
                    jj++;
                }
            }
            
			List<Character> cList = new ArrayList<Character>();
			boolean dq = false;
			for (int jj = 0; jj < body.length(); jj++) {
				final char c = body.charAt(jj);
				if (c == '"') {
                    if (!dq) {
                        dq = true;
                    } else {
                        cList.add(Character.valueOf(c));
                        dq = false;
                    }
				} else {
                    cList.add(Character.valueOf(c));
				}
			}
			char[] cArray = new char[cList.size()];
			for (int jj = 0; jj < cArray.length; jj++) {
				cArray[jj] = cList.get(jj).charValue();
			}
			StringBuffer sb = new StringBuffer();
			sb.append(head);
			sb.append('"');
			sb.append(new String(cArray));
			sb.append('"');
			argsList.add(sb.toString());
		}
        
        return command;
    }
    
    // A temporary attribute to store the returned value.
    private int mReturnedValue;

    /**
     * Execute a command.
     * 
     * @param line
     *            the command line
     * @return the status
     */
    int exec(final String line) {
    	
    	// parses the command
        final List<String> argsList = new ArrayList<String>();
        final String command = this.parseCommand(line, argsList);
        if (command == null) {
            return STATUS_FAILED;
        }

        if (COM_COMMAND_SLEEP.equalsIgnoreCase(command)) {
            final int argn = argsList.size();
            if (argn == 0) {
                return STATUS_FAILED;
            }
            int[] dataIdArray = new int[argn];
            List<Double> dataList = new ArrayList<Double>();
            for (int ii = 0; ii < dataIdArray.length; ii++) {
                String str = argsList.get(ii);
                Double num = SGUtilityText.getDouble(str);
                if (null != num) {
                    dataList.add(num);
                }
            }
            if (dataList.size()!=1) {
                return STATUS_FAILED;
            }
            if (argn!=dataList.size()) {
                return STATUS_FAILED;
            }
            try {   // sleep
                Thread.sleep(Math.round(Math.abs(dataList.get(0).doubleValue())*1000));
            } catch (InterruptedException e) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        }
/*        
        // basic commands
        try {
            EventQueue.invokeAndWait(new Runnable() {
                public void run() {
                    mReturnedValue = execBasicCommand(
                            command, argsList);
                }
            });
        } catch (InterruptedException e1) {
        } catch (InvocationTargetException e1) {
        }
        if (mReturnedValue != STATUS_NOT_FOUND) {
            return mReturnedValue;
        }
        
        int ret;
        // commands in the menu-bar
        ret = this.execMenuBarCommand(command, argsList);
        if (ret != STATUS_NOT_FOUND) {
            return ret;
        }
*/
        
        // commands for exporting an image
        try {
            EventQueue.invokeAndWait(new Runnable() {
                public void run() {
                    mReturnedValue = execImageCommand(
                            command, argsList);
                }
            });
        } catch (InterruptedException e1) {
        } catch (InvocationTargetException e1) {
        }
        if (mReturnedValue != STATUS_NOT_FOUND) {
            return mReturnedValue;
        }
        
        // commands for the windows
        try {
            EventQueue.invokeAndWait(new Runnable() {
                public void run() {
                    mReturnedValue = execWindowCommand(
                            command, argsList);
                }
            });
        } catch (InterruptedException e1) {
        } catch (InvocationTargetException e1) {
        }
        if (mReturnedValue != STATUS_NOT_FOUND) {
            return mReturnedValue;
        }
        
        // commands for the window menu bar
        try {
            EventQueue.invokeAndWait(new Runnable() {
                public void run() {
                    mReturnedValue = execMenuBarCommandFile(
                            command, argsList);
                }
            });
        } catch (InterruptedException e1) {
        } catch (InvocationTargetException e1) {
        }
        if (mReturnedValue != STATUS_NOT_FOUND) {
            return mReturnedValue;
        }

        // commands for the figures
        try {
            EventQueue.invokeAndWait(new Runnable() {
                public void run() {
                    mReturnedValue = execFigureCommand(
                            command, argsList);
                }
            });
        } catch (InterruptedException e1) {
        } catch (InvocationTargetException e1) {
        }
        if (mReturnedValue != STATUS_NOT_FOUND) {
            return mReturnedValue;
        }
        
        // commands for the figure elements
        try {
            EventQueue.invokeAndWait(new Runnable() {
                public void run() {
                    mReturnedValue = execFigureElementCommand(
                            command, argsList);
                }
            });
        } catch (InterruptedException e) {
        } catch (InvocationTargetException e) {
        }
        if (mReturnedValue != STATUS_NOT_FOUND) {
            return mReturnedValue;
        }
        
        return STATUS_NOT_FOUND;
    }

    // commands in the menu-bar File
    private int execMenuBarCommandFile(final String command, final List<String> argsList) {
        final int argn = argsList.size();
        final SGDrawingWindow wnd = this.mWindowManager.getCurrentWindow();
        if (wnd == null) {
            return STATUS_FAILED;
        }
        

/*    
        // load property
        if (COM_LOAD_PROPERTY.equalsIgnoreCase(command)) {
            if (argn < 1) {
                return STATUS_FAILED;
            }
            if (aWnd == null) {
                return STATUS_FAILED;
            }
            String str1 = (String) argsList.get(0); // file path

            File file = new File(str1);
            if (file.exists() == false) {
                return STATUS_FAILED;
            }
//            String path = SGUtility.getCanonicalPath(file.getPath());
//            if (path == null) {
//                return STATUS_FAILED;
//            }
//
//            // check validity of the file
//            URL url = null;
//            try {
//                url = file.toURI().toURL();
//            } catch (MalformedURLException ex) {
//                return STATUS_FAILED;
//            }
//
//            // create a Document object
//            Document doc = SGUtilityText.getDocument(url);
//            if (doc == null) {
//                return STATUS_FAILED;
//            }
//
//            //
//            Element elWnd = this.mPropertyFileManager.getWindowElement(doc);
//            NodeList nListFigure = elWnd
//                    .getElementsByTagName(SGFigure.TAG_NAME_FIGURE);
//            final int figureNum = nListFigure.getLength();
//            int cnt = 0;
//            final int[] dataNumArray = new int[figureNum];
//            for (int ii = 0; ii < figureNum; ii++) {
//                Node node = nListFigure.item(ii);
//                if ((node instanceof Element) == false) {
//                    return STATUS_FAILED;
//                }
//                Element elFigure = (Element) node;
//                NodeList nListData = elFigure
//                        .getElementsByTagName(SGIFigureElementGraph.TAG_NAME_DATA);
//                dataNumArray[ii] = nListData.getLength();
//                cnt += dataNumArray[ii];
//            }
//            if (cnt != argn - 1) {
//                return STATUS_FAILED;
//            }
//
//            // list of path-names of data files
//            ArrayList pathList = new ArrayList(argsList.subList(1, argn));
//
//            // create a data map
//            Map dataMap = new TreeMap();
//            cnt = 0;
//            for (int ii = 0; ii < figureNum; ii++) {
//                final int dataNum = dataNumArray[ii];
//                ArrayList dataList = new ArrayList();
//                for (int jj = 0; jj < dataNum; jj++) {
//                    dataList.add(pathList.get(cnt + jj));
//                }
//                dataMap.put(Integer.valueOf(ii), dataList);
//                cnt += dataNum;
//            }
//
//            // set properties
//            if (!this.mPropertyFileManager
//                    .setPropertyFileFromEventDispatchThread(aWnd, doc, dataMap)) {
//                return STATUS_FAILED;
//            }

//            SGDrawingWindow wnd = this.mWindowManager.getActiveWindow();
//            if (this.mPropertyFileManager.showMultiDataFileChooserDialog(
//                    file, wnd) == false) {
//                return STATUS_FAILED;
//            }

            return STATUS_SUCCEEDED;
        }
        // save property
        else if (COM_SAVE_PROPERTY.equalsIgnoreCase(command)) {
            if (argn < 1) {
                return STATUS_FAILED;
            }
            if (aWnd == null) {
                return STATUS_FAILED;
            }
            if (aWnd.getVisibleFigureList().size() == 0) {
                return STATUS_FAILED;
            }
            final String str1 = (String) argsList.get(0); // path
            final List<Integer> retList = new ArrayList<Integer>();
            try {
                SwingUtilities.invokeAndWait(new Runnable() {
                    public void run() {
                        final int ret;
                        if (mPropertyFileManager
                                .saveProperties(aWnd, str1, SAVE_PROPERTIES_TO_PROPERTY_FILE) == OK_OPTION) {
                            ret = STATUS_SUCCEEDED;
                            aWnd.setSaved(true);
                        } else {
                            ret = STATUS_FAILED;
                            aWnd.setSaved(false);
                        }
                        retList.add(ret);
                    }
                });
            } catch (InterruptedException e) {
                return STATUS_FAILED;
            } catch (InvocationTargetException e) {
                return STATUS_FAILED;
            }
            return retList.get(0);
        }
*/
        // load data set
        if (COM_LOAD_DATA_SET.equalsIgnoreCase(command)) {
            if (argn < 1) {
                return STATUS_FAILED;
            }
            String str1 = (String) argsList.get(0); // path
            if (str1.startsWith("\"")) {
                str1 = str1.substring(1);
            }
            if (str1.endsWith("\"")) {
                str1 = str1.substring(0, str1.length()-1);
            }
            final File file = new File(str1);
            if (file.exists() == false) {
                return STATUS_FAILED;
            }
            if (SGCommandManager.this.mMain.mDataSetManager.loadDataSetFromEventDispatchThread(wnd, file) == false) {
                return STATUS_FAILED;
            } else {
                return STATUS_SUCCEEDED;
            }
        }
        // save data set
        else if (COM_SAVE_DATA_SET.equalsIgnoreCase(command)) {
            if (argn < 2) {
                return STATUS_FAILED;
            }
            if (wnd.getVisibleFigureList().size() == 0) {
                return STATUS_FAILED;
            }
            String filePath = null;
            String fileType = null;
            SGPropertyMap map = this.getPropertiesMap(command, argsList);
            if (map == null) {
                return STATUS_FAILED;
            }
            Iterator<String> itr = map.getKeyIterator();
            while (itr.hasNext()) {
                String key = itr.next();
                if (COM_SAVE_FILE_PATH.equalsIgnoreCase(key)) {
                    filePath = map.getValueString(key);
                } else if (COM_SAVE_TYPE.equalsIgnoreCase(key)) {
                    fileType = map.getValueString(key);
                }
            }
            final String[] archiveTypes = {
                    SGIArchiveFileConstants.ARCHIVE_FILETYPE_SGA,
                    SGIArchiveFileConstants.ARCHIVE_FILETYPE_SGA107,
                    SGIArchiveFileConstants.ARCHIVE_FILETYPE_NETCDF
            };
            fileType = SGApplicationUtility.getExactTypeString(archiveTypes, fileType);
            if (filePath == null || fileType == null) {
                return STATUS_FAILED;
            }
            if (SGCommandManager.this.mMain.mDataSetManager.saveDataSet(wnd, filePath, fileType) == OK_OPTION) {
                wnd.setSaved(true);
                return STATUS_SUCCEEDED;
            } else {
                wnd.setSaved(false);
                return STATUS_FAILED;
            }
        }
        // exit
        else if (COM_EXIT.equalsIgnoreCase(command)) {
            if (argn != 0 && (argn==1 && argsList.get(0).trim().length()==0)==false) {
                return STATUS_FAILED;
            }
            this.mWindowManager.closeAllWindowWithoutConfirmation();
            return STATUS_SUCCEEDED;
        }

        return STATUS_NOT_FOUND;
    }
        
    /**
     * Executes a command for a window.
     * 
     * @param command
     *           a command
     * @param argsList
     *           the list of arguments
     * @return the status
     */
    private int execWindowCommand(final String command, 
            final List<String> argsList) {
        
        final int argn = argsList.size();            

        if (COM_WINDOW.equalsIgnoreCase(command)) {
            Integer wndId = this.getId(argsList);
            if (wndId == null) {
                return STATUS_FAILED;
            }
            final List<String> paramsList = new ArrayList<String>(
                    argsList.subList(1, argn));

            // create a new window if it does not exist
            if (mWindowManager.getWindow(wndId) == null) {
                SGDrawingWindow w = mWindowManager.createNewWindow(wndId);
                if (w == null) {
                    return STATUS_FAILED;
                }
                w.setVisible(true);
            }
            
            // sets the current window
            if (mWindowManager.setCurrentWindow(wndId) == false) {
                return STATUS_FAILED;
            }
            SGDrawingWindow wnd = mWindowManager.getCurrentWindow();
            
            SGPropertyMap map = this.getPropertiesMap(command, paramsList);
            if (map == null) {
                return STATUS_FAILED;
            }

            // set the background image
            String imageFilePath = map.getValueString(COM_IMAGE_FILE_PATH);
            if (!"".equals(imageFilePath)) {
                if (map.isDoubleQuoted(COM_IMAGE_FILE_PATH) == false) {
                    return STATUS_FAILED;
                }
                File imageFile = new File(imageFilePath);
                if (!imageFile.exists()) {
                    return STATUS_FAILED;
                }
                
                // add image to the window
            	byte[] imageByteArray = SGApplicationUtility.toByteArray(imageFile);
            	String ext = SGApplicationUtility.getExtension(imageFile);
            	if (ext == null) {
                    return STATUS_FAILED;
            	}
                if (wnd.setImage(imageByteArray, ext, true) == false) {
                    return STATUS_FAILED;
                }
                
                // remove keys
                map.removeValue(COM_IMAGE_FILE_PATH);
            }
            
            SGPropertyResults result = wnd.setProperties(map);
            if (result == null) {
                return STATUS_FAILED;
            }

            // after setting properties
            return this.afterSetProperties(map, result);
            
        } else if (COM_CLOSE_WINDOW.equalsIgnoreCase(command)) {
            Integer wndId = this.getId(argsList);
            if (wndId == null) {
                return STATUS_FAILED;
            }
            if (this.mMain.closeWindowWithoutConfirmation(wndId) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        }

        return STATUS_NOT_FOUND;
    }

    /**
     * Executes a command for a figure.
     * 
     * @param command
     *           a command
     * @param argsList
     *           the list of arguments
     * @return the status
     */
    private int execFigureCommand(final String command, 
            final List<String> argsList) {
        
        final int argn = argsList.size();
        SGDrawingWindow wnd = mWindowManager.getCurrentWindow();
        if (wnd == null) {
            return STATUS_FAILED;
        }

        if (COM_FIGURE.equalsIgnoreCase(command)) {
            Integer figureId = this.getId(argsList);
            if (figureId == null) {
                return STATUS_FAILED;
            }
            final List<String> paramsList = new ArrayList<String>(
                    argsList.subList(1, argn));

            // create a new figure if it does not exist
            SGFigure figure = wnd.getFigure(figureId);
            if (figure == null) {
                figure = this.mMain.mFigureCreator.createEmptyFigure(figureId, wnd);
                if (figure == null) {
                    wnd.endProgress();
                    return STATUS_FAILED;
                }

                // add a new figure to the window
                if (wnd.addFigure(figure, new Point()) == false) {
                    return STATUS_FAILED;
                }
                figure.setFigureX(DEFAULT_FIGURE_X, FIGURE_LOCATION_UNIT);
                figure.setFigureY(DEFAULT_FIGURE_Y, FIGURE_LOCATION_UNIT);

                // set visible the figure
                figure.setVisible(true);

                // init the history of the properties
                SGIFigureElement[] array = figure.getIFigureElementArray();
                for (int ii = 0; ii < array.length; ii++) {
                    if (array[ii].initPropertiesHistory() == false) {
                        wnd.endProgress();
                        return STATUS_FAILED;
                    }
                }
                figure.initPropertiesHistory();

                // update the items after the figure is set visible
                wnd.updateItemsByFigureNumbers();

                wnd.setChanged(true);
            }
            
            // sets the current figure
            if (mWindowManager.setCurrentFigure(wnd.getID(), figure) == false) {
                return STATUS_FAILED;
            }
            
            SGPropertyMap map = this.getPropertiesMap(command, paramsList);
            if (map == null) {
                return STATUS_FAILED;
            }
            SGPropertyResults result = figure.setProperties(map);
            if (result == null) {
                return STATUS_FAILED;
            }

            // after setting properties
            return this.afterSetProperties(map, result);
            
        } else if (COM_BRING_TO_FRONT_FIGURE.equalsIgnoreCase(command)) {
            Integer figureId = this.getId(argsList);
            if (figureId == null) {
                return STATUS_FAILED;
            }
            if (wnd.moveFigureToEnd(figureId, true) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (COM_BRING_FORWARD_FIGURE.equalsIgnoreCase(command)) {
            Integer figureId = this.getId(argsList);
            if (figureId == null) {
                return STATUS_FAILED;
            }
            if (wnd.moveFigure(figureId, true) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (COM_SEND_BACKWARD_FIGURE.equalsIgnoreCase(command)) {
            Integer figureId = this.getId(argsList);
            if (figureId == null) {
                return STATUS_FAILED;
            }
            if (wnd.moveFigure(figureId, false) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (COM_SEND_TO_BACK_FIGURE.equalsIgnoreCase(command)) {
            Integer figureId = this.getId(argsList);
            if (figureId == null) {
                return STATUS_FAILED;
            }
            if (wnd.moveFigureToEnd(figureId, false) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        }

        return STATUS_NOT_FOUND;
    }

    /**
     * Executes a command for a figure elements.
     * 
     * @param command
     *           a command
     * @param argsList
     *           the list of arguments
     * @return the status
     */
    private int execFigureElementCommand(final String command, 
            final List<String> argsList) {
        
        SGFigure figure = mWindowManager.getCurrentFigure();
        if (figure == null) {
            return STATUS_FAILED;
        }
        
        if (COM_DATA.equalsIgnoreCase(command)) {
            // create a data object
            final int argn = argsList.size();
            if (argn < 2) {
                return STATUS_FAILED;
            }
            Integer dataId = this.getId(argsList);
            if (dataId == null) {
                return STATUS_FAILED;
            }

            final List<String> paramsList = new ArrayList<String>(
                    argsList.subList(1, argn));
            SGPropertyMap map = this.getPropertiesMap(command, paramsList);
            if (map == null) {
                return STATUS_FAILED;
            }
            
        	// if data object does not exist or is invisible, draw the new graph
            if (!figure.isDataVisible(dataId)) {
                // create new data object
                String path = null;
                String dataType = null;
                Iterator<String> itr = map.getKeyIterator();
                while (itr.hasNext()) {
                    String key = itr.next();
                    if (COM_DATA_FILE_PATH.equalsIgnoreCase(key)) {
                        path = map.getValueString(key);
                    } else if (COM_DATA_TYPE.equalsIgnoreCase(key)) {
                        dataType = map.getValueString(key);
                    }
                }
                final String[] dataTypes = {
                        SGDataTypeConstants.SXY_DATA, 
                        SGDataTypeConstants.SXY_MULTIPLE_DATA, 
                        SGDataTypeConstants.SXY_SAMPLING_DATA,
                        SGDataTypeConstants.VXY_DATA,
                        SGDataTypeConstants.SXY_DATE_DATA,
                        SGDataTypeConstants.SXYZ_DATA,
                        
                        SGDataTypeConstants.SXY_NETCDF_DATA, 
                        SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DATA, 
                        SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DIMENSION_DATA,
                        SGDataTypeConstants.VXY_NETCDF_DATA,
                        SGDataTypeConstants.SXYZ_NETCDF_DATA
                };
                dataType = SGApplicationUtility.getExactTypeString(dataTypes, dataType);
                if (path == null || dataType == null) {
                    return STATUS_FAILED;
                }
                if (map.isDoubleQuoted(COM_DATA_FILE_PATH) == false) {
                    return STATUS_FAILED;
                }

                SGDrawingWindow wnd = figure.getWindow();

                // check the data type
                if (SGDataUtility.isValidData(dataType) == false) {
                    return STATUS_FAILED;
                }
                final boolean isNetCDF = SGDataUtility.isNetCDFData(dataType);

                // get information map
                final Map<String, Object> infoMap = this.mMain.createInfoMap(dataType, map);
                if (infoMap == null) {
                    return STATUS_FAILED;
                }

            	NetcdfFile ncFile = null;
                SGNetCDFFile nc = null;
                if (isNetCDF) {
                    try {
                        ncFile = NetcdfFile.open(path);
                    } catch (IOException e) {
                        return STATUS_FAILED;
                    }
                    nc = new SGNetCDFFile(ncFile);
                    infoMap.put(KEY_DATA_SOURCE, nc);
                }

                // column information
                SGDataColumnInfoSet colInfoSet = null;
                if (isNetCDF) {
                    colInfoSet = this.mMain.getNetCDFDefaultDataColumnInfo(nc,
                            dataType, infoMap);
                } else {
                    colInfoSet = this.mMain.getDefaultDataColumnInfo(path, dataType,
                            infoMap, false, null);
                }
                if (colInfoSet == null) {
                    return STATUS_FAILED;
                }

                if (isNetCDF) {
                    if (SGDataUtility.isNetCDFDimensionData(dataType)) {
                        SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();
                        SGNetCDFDataColumnInfo[] nCols = new SGNetCDFDataColumnInfo[colInfo.length];
                        for (int i = 0; i < nCols.length; i++) {
                            nCols[i] = (SGNetCDFDataColumnInfo)colInfo[i];
                        }
                        if (SGDataUtility.updatePickupParameters(infoMap, nCols)==false) {
                            return STATUS_FAILED;
                        }
                    }
                }

// TODO				
                // draw the graph
                if (this.mMain.drawGraph(wnd, figure.getID(), colInfoSet, 
                		infoMap, path, new Integer[] { dataId }, false, null) == false) {
                    return STATUS_FAILED;
                }
            	
                // set the changed flag and notify to the root
                wnd.setChanged(true);
                wnd.notifyToRoot();
                
                // remove keys
                map.removeValue(COM_DATA_FILE_PATH);
                map.removeValue(COM_DATA_TYPE);
                map.removeValue(COM_DATA_SAMPLING_RATE);
                map.removeValue(COM_DATA_POLAR);
                
                if (ncFile != null) {
                    try {
                        ncFile.close();
                    } catch (IOException e) {
                        return STATUS_FAILED;
                    }
                }
            }

            // set data properties
            SGPropertyResults result = figure.getGraphElement().setChildProperties(
                    dataId, map);
            if (result == null) {
                return STATUS_FAILED;
            }

            // after setting properties
            return this.afterSetProperties(map, result);
            
        } else if (this.isCommandBringToFront(command)) {
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            final String end = this.getCommandEndToMove(command, 
                    COM_HEADER_BRING_TO_FRONT);
            SGIFigureElement el = this.getFigureElementToMove(end, figure);
            if (el.moveChildToEnd(id.intValue(), true) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (this.isCommandBringForward(command)) {
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            final String end = this.getCommandEndToMove(command, 
                    COM_HEADER_BRING_FORWARD);
            SGIFigureElement el = this.getFigureElementToMove(end, figure);
            if (el.moveChild(id.intValue(), true) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (this.isCommandSendBackward(command)) {
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            final String end = this.getCommandEndToMove(command, 
                    COM_HEADER_SEND_BACKWARD);
            SGIFigureElement el = this.getFigureElementToMove(end, figure);
            if (el.moveChild(id.intValue(), false) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (this.isCommandSendToBack(command)) {
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            final String end = this.getCommandEndToMove(command, 
                    COM_HEADER_SEND_TO_BACK);
            SGIFigureElement el = this.getFigureElementToMove(end, figure);
            if (el.moveChildToEnd(id.intValue(), false) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (COM_MOVE_TO_TOP_DATA.equalsIgnoreCase(command)) {
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            SGIFigureElementLegend el = figure.getLegendElement();
            if (el.moveLegendToEnd(id.intValue(), true) == false) {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (COM_MOVE_TO_UPPER_DATA.equalsIgnoreCase(command)) {
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            SGIFigureElementLegend el = figure.getLegendElement();
            if (el.moveLegend(id.intValue(), true) == false) {
                return STATUS_FAILED;                    
            }
            return STATUS_SUCCEEDED;
        } else if (COM_MOVE_TO_LOWER_DATA.equalsIgnoreCase(command)) {
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            SGIFigureElementLegend el = figure.getLegendElement();
            if (el.moveLegend(id.intValue(), false) == false) {
                return STATUS_FAILED;                    
            }
            return STATUS_SUCCEEDED;
        } else if (COM_MOVE_TO_BOTTOM_DATA.equalsIgnoreCase(command)) {
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            SGIFigureElementLegend el = figure.getLegendElement();
            if (el.moveLegendToEnd(id.intValue(), false) == false) {
                return STATUS_FAILED;                    
            }
            return STATUS_SUCCEEDED;
        } else if (COM_DELETE_DATA.equalsIgnoreCase(command)) {
            final int argn = argsList.size();
            int[] dataIdArray = new int[argn];
            List<Integer> dataIdList = new ArrayList<Integer>();
            for (int ii = 0; ii < dataIdArray.length; ii++) {
                String str = argsList.get(ii);
                Integer num = SGUtilityText.getInteger(str);
                if (null != num) {
                    dataIdList.add(num);
                }
            }
            if (argn!=dataIdList.size()) {
                return STATUS_FAILED;
            }
            for (int i = 0; i < argn; i++) {
                dataIdArray[i] = dataIdList.get(i).intValue();
            }

            // hide data
            if (figure.hideData(dataIdArray) == false) {
                return STATUS_FAILED;
            }
            
            if (dataIdArray.length!=0) {
                SGDrawingWindow wnd = figure.getWindow();
                wnd.setChanged(true);
                wnd.notifyToRoot();
            }

            return STATUS_SUCCEEDED;
        } else if (COM_FIT_AXES.equalsIgnoreCase(command)) {
            final int argn = argsList.size();
            if (argn == 0) {
                return STATUS_FAILED;
            }
            int[] dataIdArray = new int[argn];
            List<Integer> dataIdList = new ArrayList<Integer>();
            for (int ii = 0; ii < dataIdArray.length; ii++) {
            	String str = argsList.get(ii);
            	Integer num = SGUtilityText.getInteger(str);
            	if (null != num) {
            	    dataIdList.add(num);
            	}
            }
            if (dataIdList.size()==0) {
                if (figure.fitAxisRangeToVisibleData() == false) {
                    return STATUS_FAILED;
                }
                return STATUS_SUCCEEDED;
            }
            if (argn!=dataIdList.size()) {
                return STATUS_FAILED;
            }
            for (int i = 0; i < argn; i++) {
                dataIdArray[i] = dataIdList.get(i).intValue();
            }

            // fit axes
            if (figure.fitAxisRangeToData(dataIdArray) == false) {
                return STATUS_FAILED;
            }

            return STATUS_SUCCEEDED;
        } else if (COM_ALIGN_BARS.equalsIgnoreCase(command)) {
            final int argn = argsList.size();
            if (argn != 1) {
                return STATUS_FAILED;
            }
            String str = argsList.get(0);
            if (str.trim().equals("")) {
                if (figure.alignVisibleBars() == false) {
                    return STATUS_FAILED;
                }
            } else {
                return STATUS_FAILED;
            }
            return STATUS_SUCCEEDED;
        } else if (COM_SPLIT.equalsIgnoreCase(command)) {
            final int argn = argsList.size();
            if (argn == 0) {
                return STATUS_FAILED;
            }
            int[] dataIdArray = new int[argn];
            List<Integer> dataIdList = new ArrayList<Integer>();
            for (int ii = 0; ii < dataIdArray.length; ii++) {
                String str = argsList.get(ii);
                Integer num = SGUtilityText.getInteger(str);
                if (null != num) {
                    dataIdList.add(num);
                }
            }
            if (dataIdList.size()==0) {
                return STATUS_FAILED;
            }
            if (argn!=dataIdList.size()) {
                return STATUS_FAILED;
            }
            for (int i = 0; i < argn; i++) {
                dataIdArray[i] = dataIdList.get(i).intValue();
            }

            // split
            List<SGData> dataList = SGMainFunctionsSplitMerge.splitDataInFigure(figure, dataIdArray);
            if (dataList == null || dataList.size()==0) {
                return STATUS_FAILED;
            }

            return STATUS_SUCCEEDED;
        } else if (COM_MERGE.equalsIgnoreCase(command)) {
            final int argn = argsList.size();
            if (argn == 0) {
                return STATUS_FAILED;
            }
            int[] dataIdArray = new int[argn];
            List<Integer> dataIdList = new ArrayList<Integer>();
            for (int ii = 0; ii < dataIdArray.length; ii++) {
                String str = argsList.get(ii);
                Integer num = SGUtilityText.getInteger(str);
                if (null != num) {
                    dataIdList.add(num);
                }
            }
            if (dataIdList.size()<=1) {
                return STATUS_FAILED;
            }
            if (argn!=dataIdList.size()) {
                return STATUS_FAILED;
            }
            for (int i = 0; i < argn; i++) {
                dataIdArray[i] = dataIdList.get(i).intValue();
            }

            // merge
            List<SGData> dataList = SGMainFunctionsSplitMerge.mergeDataInFigure(figure, dataIdArray);
            if (dataList == null || dataList.size()==0) {
                return STATUS_FAILED;
            }

            return STATUS_SUCCEEDED;
        } else if (COM_INSERT_NETCDF_DATA_LABEL.equalsIgnoreCase(command)) {
            final int argn = argsList.size();
            if (argn == 0) {
                return STATUS_FAILED;
            }
            int[] dataIdArray = new int[argn];
            List<Integer> dataIdList = new ArrayList<Integer>();
            for (int ii = 0; ii < dataIdArray.length; ii++) {
                String str = argsList.get(ii);
                Integer num = SGUtilityText.getInteger(str);
                if (null != num) {
                    dataIdList.add(num);
                }
            }
            if (dataIdList.size()==0) {
                return STATUS_FAILED;
            }
            if (argn!=dataIdList.size()) {
                return STATUS_FAILED;
            }
            for (int i = 0; i < argn; i++) {
                dataIdArray[i] = dataIdList.get(i).intValue();
            }

            // insert labels
            if (figure.insertNetCDFLabels(dataIdArray)==false) {
                return STATUS_FAILED;
            }

            return STATUS_SUCCEEDED;
        } else if (COM_AXIS.equalsIgnoreCase(command)) {                
            if (argsList.size() < 1) {
                return STATUS_FAILED;
            }
            String args0 = argsList.get(0);
            final int location = SGUtility.getAxisLocation(args0);

            SGPropertyResults result = null;
            SGPropertyMap map = null;
            if (location != -1) {
                // for each axis
                List<String> paramsList = new ArrayList<String>(
                        argsList.subList(1, argsList.size()));
                map = this.getPropertiesMap(command, paramsList);
                if (map == null) {
                    return STATUS_FAILED;
                }
                
                // set properties for a axis
                result = figure.getAxisElement().setChildProperties(location, map);
                if (result == null) {
                    return STATUS_FAILED;
                }
                
            } else {
                // for all axes
                map = this.getPropertiesMap(command, argsList);
                if (map == null) {
                    return STATUS_FAILED;
                }
                
                // set properties of axes
                result = figure.getAxisElement().setProperties(map);
                if (result == null) {
                    return STATUS_FAILED;
                }
            }
            
            // after setting properties
            return this.afterSetProperties(map, result);
            
        } else if (COM_COLOR_BAR.equalsIgnoreCase(command)) {
            SGPropertyMap map = this.getPropertiesMap(command, argsList);
            if (map == null) {
                return STATUS_FAILED;
            }
            
            // set properties
            SGPropertyResults result = figure.getAxisElement().setChildProperties(
            		SGIFigureElementAxis.AXIS_NORMAL, map);

            if (result == null) {
                return STATUS_FAILED;
            }
            
            // after setting properties
            return this.afterSetProperties(map, result);
            
        } else if (COM_LEGEND.equalsIgnoreCase(command)) {
            SGPropertyMap map = this.getPropertiesMap(command, argsList);
            if (map == null) {
                return STATUS_FAILED;
            }
            
            // set properties
            SGPropertyResults result = figure.getLegendElement().setProperties(map);
            if (result == null) {
                return STATUS_FAILED;
            }
            
            // after setting properties
            return this.afterSetProperties(map, result);

        } else if (COM_LABEL.equalsIgnoreCase(command)
        		|| COM_SIGNIFICANT_DIFFERECE.equalsIgnoreCase(command)
        		|| COM_AXIS_BREAK.equalsIgnoreCase(command)
        		|| COM_TIMING_LINE.equalsIgnoreCase(command)
        		|| COM_SHAPE.equalsIgnoreCase(command)) {

            Class<? extends SGIFigureElement> cl = null;
            if (COM_LABEL.equalsIgnoreCase(command)) {
                cl = SGIFigureElementString.class;
            } else if (COM_SIGNIFICANT_DIFFERECE.equalsIgnoreCase(command)) {
                cl = SGIFigureElementSignificantDifference.class;
            } else if (COM_AXIS_BREAK.equalsIgnoreCase(command)) {
                cl = SGIFigureElementAxisBreak.class;
            } else if (COM_TIMING_LINE.equalsIgnoreCase(command)) {
                cl = SGIFigureElementTimingLine.class;
            } else if (COM_SHAPE.equalsIgnoreCase(command)) {
                cl = SGIFigureElementShape.class;
            } else {
                return STATUS_FAILED;
            }
        	
            Integer id = this.getId(argsList);
            if (id == null) {
                return STATUS_FAILED;
            }
            final int argn = argsList.size();
            final List<String> paramsList = new ArrayList<String>(argsList.subList(1, argn));
            SGPropertyMap map = this.getPropertiesMap(command, paramsList);
            if (map == null) {
                return STATUS_FAILED;
            }
            
            SGIFigureElement el = (SGIFigureElement) figure.getIFigureElement(cl);
            if (el == null) {
                return STATUS_FAILED;
            }
            
            // set properties
            SGPropertyResults result = el.setChildProperties(id.intValue(), map);
            if (result == null) {
                return STATUS_FAILED;
            }
            
            // after setting properties
            return this.afterSetProperties(map, result);
        }
        
        return STATUS_NOT_FOUND;
    }

    /**
     * Executes a command to export as an image.
     * 
     * @param command
     *           a command
     * @param argsList
     *           the list of arguments
     * @return the status
     */
    private int execImageCommand(final String command, 
            final List<String> argsList) {
        
        if (COM_EXPORT_IMAGE.equalsIgnoreCase(command)) {
            final SGDrawingWindow wnd = mWindowManager.getCurrentWindow();
            if (wnd == null) {
                return STATUS_FAILED;
            }
            if (wnd.getVisibleFigureList().size() == 0) {
                return STATUS_FAILED;
            }

            SGPropertyMap map = this.getPropertiesMap(command, argsList);
            if (map == null) {
                return STATUS_FAILED;
            }
            
            // create new data object
            String path = null;
            String imgType = null;
            Iterator<String> itr = map.getKeyIterator();
            while (itr.hasNext()) {
                String key = itr.next();
                if (COM_EXPORT_IMAGE_FILE_PATH.equalsIgnoreCase(key)) {
                    path = map.getValueString(key);
                } else if (COM_EXPORT_IMAGE_TYPE.equalsIgnoreCase(key)) {
                    imgType = map.getValueString(key);
                }
            }
            final String[] imageTypes = {
                    "EMF", "EPS", "PS", "PDF",
                    "SVG", "CGM", "SWF",
                    "JPG", "JPEG", "GIF", "PNG", "RAW", "PPM"
            };
            imgType = SGApplicationUtility.getExactTypeString(imageTypes, imgType);
            if (path == null || imgType == null) {
                return STATUS_FAILED;
            }
        	if (map.isDoubleQuoted(COM_EXPORT_IMAGE_FILE_PATH) == false) {
                return STATUS_FAILED;
        	}

        	// check the parent
        	File f = new File(path);
            File parent = new File(f.getParent());
            if (parent.exists() == false) {
            	return STATUS_FAILED;
            }

            // remove keys
            map.removeValue(COM_EXPORT_IMAGE_FILE_PATH);
            map.removeValue(COM_EXPORT_IMAGE_TYPE);

        	// export
            SGPropertyResults result = wnd.exportAsImage(map, imgType, path, true);
            if (result == null) {
                return STATUS_FAILED;
            }

            // after setting properties
            return this.afterSetProperties(map, result);
        }
        
        return STATUS_NOT_FOUND;
    }

    // Called after setting properties.
    private int afterSetProperties(SGPropertyMap map, SGPropertyResults result) {
    	
        // check "not found" keys
        Iterator<String> itr = map.getKeyIterator();
        while (itr.hasNext()) {
            String key = itr.next();
        	Integer res = result.getResult(key);
        	if (res == null) {
        		result.putResult(key, SGPropertyResults.NOT_FOUND);
        	}
        }

        // shows the results
        result.showResults();
        
        // checks setting error
        itr = map.getKeyIterator();
        while (itr.hasNext()) {
            String key = itr.next();
        	Integer res = result.getResult(key);
        	if (res.intValue() == SGPropertyResults.INVALID_INPUT_VALUE
        			|| res.intValue() == SGPropertyResults.NOT_FOUND) {
        		return STATUS_PARTIALLY_FAILED;
        	}
        }
        return STATUS_SUCCEEDED;
    }
    
    // The map of alias objects. Keys of this map are given by the string representation of each alias.
    private Map<String, Alias> mAliasMap = new HashMap<String, Alias>();

    /**
     * Adds an alias.
     * 
     * @param command
     *           the command applying the new alias
     * @param key
     *           the property key applying the new alias
     * @param repr
     *           the string representation of the new alias
     * @param value
     *           the value for the new alias
     */
    public void addAlias(String command, String key, String repr, String value) {
    	Alias a = new Alias(command, key, repr, value);
    	this.mAliasMap.put(repr, a);
    }

    /**
     * Removes the alias.
     * 
     * @param repr
     *           the string representation of the alias to remove
     */
    public void removeAlias(String repr) {
    	this.mAliasMap.remove(repr);
    }
    
    /**
     * The alias class.
     */
    static class Alias {
    	
    	// The command applying this alias.
    	private String mCommand = null;
    	
    	// The property key applying this alias.
    	private String mKey = null;
    	
    	// The string representation of the alias.
    	private String mRepr = null;
    	
    	// The value of this alias.
    	private String mValue = null;
    	
    	/**
    	 * Builds an alias object.
    	 * The command and key can be set null value.
    	 * 
	     * @param command
	     *           the command applying the new alias
	     * @param key
	     *           the property key applying the new alias
	     * @param repr
	     *           the string representation of the new alias
	     * @param value
	     *           the value of the new alias
    	 */
    	public Alias(String command, String key, String repr, String value) {
    		if (repr == null || value == null) {
    			throw new IllegalArgumentException("repr == null || value == null");
    		}
    		this.mCommand = command;
    		this.mKey = key;
    		this.mRepr = repr;
    		this.mValue =  value;
    	}
    	
    	/**
    	 * Returns the command applying this alias.
    	 * 
    	 * @return the command applying this alias
    	 */
    	public String getCommand() {
    		return this.mCommand;
    	}
    	
    	/**
    	 * Returns the property key applying this alias.
    	 * 
    	 * @return the property key applying this alias
    	 */
    	public String getKey() {
    		return this.mKey;
    	}
    	
    	/**
    	 * Returns the string representation of this alias.
    	 * 
    	 * @return the string representation of this alias
    	 */
    	public String getRepr() {
    		return this.mRepr;
    	}
    	
    	/**
    	 * Returns the value of this alias.
    	 * 
    	 * @return the value of this alias
    	 */
    	public String getValue() {
    		return this.mValue;
    	}
    }
    
}
