package jp.riken.brain.ni.samuraigraph.application;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.JOptionPane;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfoSet;
import jp.riken.brain.ni.samuraigraph.base.SGDrawingWindow;
import jp.riken.brain.ni.samuraigraph.base.SGDrawingWindow.BackgroundImage;
import jp.riken.brain.ni.samuraigraph.base.SGFigure;
import jp.riken.brain.ni.samuraigraph.base.SGIConstants;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElement;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementAxis;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementGraph;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementLegend;
import jp.riken.brain.ni.samuraigraph.base.SGIProgressControl;
import jp.riken.brain.ni.samuraigraph.base.SGIPropertyFileConstants;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.data.SGArrayData;
import jp.riken.brain.ni.samuraigraph.data.SGDataColumn;
import jp.riken.brain.ni.samuraigraph.data.SGDataTypeConstants;
import jp.riken.brain.ni.samuraigraph.data.SGDataUtility;
import jp.riken.brain.ni.samuraigraph.data.SGDate;
import jp.riken.brain.ni.samuraigraph.data.SGIDataColumnNameConstants;
import jp.riken.brain.ni.samuraigraph.data.SGIDataColumnTypeConstants;
import jp.riken.brain.ni.samuraigraph.data.SGNetCDFFile;
import jp.riken.brain.ni.samuraigraph.data.SGNumberDataColumn;
import jp.riken.brain.ni.samuraigraph.data.SGSXYMultipleDimensionNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYMultipleVariableNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYZNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGVXYNetCDFData;
import jp.riken.brain.ni.samuraigraph.figure.SGIFigureTypeConstants;
import jp.riken.brain.ni.samuraigraph.figure.SGXYFigure;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import ucar.ma2.Array;
import ucar.ma2.ArrayByte;
import ucar.ma2.DataType;
import ucar.ma2.Index;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.Group;
import ucar.nc2.NetcdfFile;
import ucar.nc2.NetcdfFileWriteable;
import ucar.nc2.Variable;

class SGNetCDFDataSetManager implements SGIArchiveFileConstants {
	
    /**
     * The main functions.
     */
    private final SGMainFunctions mMain;

    static final String ATTRIBUTE_KEY_LONG_NAME = "long_name";
    
    static final String SERIAL_NUMBER_DIMENSION_NAME = SGIDataColumnTypeConstants.SERIAL_NUMBER;
    
    static final String ATTRIBUTE_KEY_IMAGE_FILE_EXTENSION = "extension";
    
    static final String DIMENSION_NAME_BACKGROUND_IMAGE_BYTE_INDEX = "byte_index";
    
    static final String VARIABLE_NAME_BACKGROUND_IMAGE = "background_image";

    static final String ATTRIBUTE_NAME_PROPERTY = "samurai_graph_property";

    SGNetCDFDataSetManager(SGMainFunctions main) {
        super();
        this.mMain = main;
    }

    private String createGroupName(final int figureIndex, final int dataIndex) {
        StringBuffer sb = new StringBuffer();
        sb.append("id");
        sb.append(figureIndex);
        sb.append('-');
        sb.append(dataIndex);
        return sb.toString();
    }

    private String getGroupVariableName(final String groupName, final String name) {
    	StringBuffer sb = new StringBuffer();
    	sb.append(groupName);
    	sb.append('/');
    	sb.append(name);
    	return sb.toString();
    }

    private String getGroupNameFromDataFile(File file) {
        return file.getName().substring(0, file.getName().length() - 3);
    }

    /**
     * Vector-XY
     * @param data
     * @param file
     * @param groupName
     * @return
     * @throws IOException
     * @throws InvalidRangeException
     */
    private boolean saveArrayDataVXYToTempNetCDFFiles(
            final SGArrayData data, final File file, final String groupName, final SGDrawingWindow wnd)
    throws IOException, InvalidRangeException {

        final int dataNum = data.getPointsNumber();

        SGDataColumn[] colArray = data.getDataSetColumnsClone();
        if (colArray.length < 4) {
            return false;
        }
        SGNumberDataColumn colx = (SGNumberDataColumn) colArray[0];
        SGNumberDataColumn coly = (SGNumberDataColumn) colArray[1];
        SGNumberDataColumn colc1 = (SGNumberDataColumn) colArray[2];
        SGNumberDataColumn colc2 = (SGNumberDataColumn) colArray[3];
        
        double[] xvalues = colx.getNumberArray();
        double[] yvalues = coly.getNumberArray();
        double[] c1values = colc1.getNumberArray();
        double[] c2values = colc2.getNumberArray();

        NetcdfFileWriteable ncfile = null;
        try {
            ncfile = NetcdfFileWriteable.createNew(file.getAbsolutePath(), true);

            // Creates the dimension and variable of serial numbers.
            Dimension serialNumberDim = this.addSerialNumberDimension(ncfile, dataNum);
            String serialNumberDimName = serialNumberDim.getName();
            Variable serialNumberVar = this.addSerialNumberVarialbe(ncfile, serialNumberDim);
            serialNumberVar.addAttribute(SGDataUtility.getValueTypeAttribute(SGIDataColumnTypeConstants.VALUE_TYPE_NUMBER));
            
            // Add data columns as variables.
            Variable varx = new Variable(ncfile, null, null, "column0", DataType.DOUBLE, serialNumberDimName);
            Variable vary = new Variable(ncfile, null, null, "column1", DataType.DOUBLE, serialNumberDimName);
            Variable varc1 = new Variable(ncfile, null, null, "column2", DataType.DOUBLE, serialNumberDimName);
            Variable varc2 = new Variable(ncfile, null, null, "column3", DataType.DOUBLE, serialNumberDimName);
            String title = colx.getTitle();
            if (null!=title && "".equals(title.trim())==false) {
                Attribute attr = new Attribute(ATTRIBUTE_KEY_LONG_NAME, title);
                varx.addAttribute(attr);
            }
            title = coly.getTitle();
            if (null!=title && "".equals(title.trim())==false) {
                Attribute attr = new Attribute(ATTRIBUTE_KEY_LONG_NAME, title);
                vary.addAttribute(attr);
            }
            title = colc1.getTitle();
            if (null!=title && "".equals(title.trim())==false) {
                Attribute attr = new Attribute(ATTRIBUTE_KEY_LONG_NAME, title);
                varc1.addAttribute(attr);
            }
            title = colc2.getTitle();
            if (null!=title && "".equals(title.trim())==false) {
                Attribute attr = new Attribute(ATTRIBUTE_KEY_LONG_NAME, title);
                varc2.addAttribute(attr);
            }

            ncfile.addVariable(null, varx);
            ncfile.addVariable(null, vary);
            ncfile.addVariable(null, varc1);
            ncfile.addVariable(null, varc2);

            String[] varNames = new String[colArray.length];
            varNames[0] = varx.getName();
            varNames[1] = vary.getName();
            varNames[2] = varc1.getName();
            varNames[3] = varc2.getName();
            
            varx.addAttribute(SGDataUtility.getValueTypeAttribute(colx.getValueType()));
            vary.addAttribute(SGDataUtility.getValueTypeAttribute(coly.getValueType()));
            varc1.addAttribute(SGDataUtility.getValueTypeAttribute(colc1.getValueType()));
            varc2.addAttribute(SGDataUtility.getValueTypeAttribute(colc2.getValueType()));

            ncfile.create();

            // add values

            Array xarray = Array.factory(DataType.DOUBLE, new int[] { dataNum });
            for (int i = 0; i < xvalues.length; i++) {
                xarray.setDouble(i, xvalues[i]);
            }
            ncfile.write(varNames[0], xarray);

            Array yarray = Array.factory(DataType.DOUBLE, new int[] { dataNum });
            for (int i = 0; i < yvalues.length; i++) {
                yarray.setDouble(i, yvalues[i]);
            }
            ncfile.write(varNames[1], yarray);

            Array c1array = Array.factory(DataType.DOUBLE, new int[] { dataNum });
            for (int i = 0; i < c1values.length; i++) {
                c1array.setDouble(i, c1values[i]);
            }
            ncfile.write(varNames[2], c1array);

            Array c2array = Array.factory(DataType.DOUBLE, new int[] { dataNum });
            for (int i = 0; i < c2values.length; i++) {
                c2array.setDouble(i, c2values[i]);
            }
            ncfile.write(varNames[3], c2array);

            // write serial numbers
            this.writeSerialNumberVarialbe(ncfile, serialNumberVar);
            
        } finally {
            if (null!=ncfile) {
                ncfile.close();
            }
        }

        return true;
    }

    /**
     * Scalar-XYZ
     * @param data
     * @param file
     * @param groupName
     * @return
     * @throws IOException
     * @throws InvalidRangeException
     */
    private boolean saveArrayDataSXYZToTempNetCDFFiles(
            final SGArrayData data, final File file, final String groupName, final SGDrawingWindow wnd)
    throws IOException, InvalidRangeException {

        final int dataNum = data.getPointsNumber();

        SGDataColumn[] colArray = data.getDataSetColumnsClone();
        if (colArray.length<3) {
            return false;
        }
        SGNumberDataColumn colx = (SGNumberDataColumn) colArray[0];
        SGNumberDataColumn coly = (SGNumberDataColumn) colArray[1];
        SGNumberDataColumn colz = (SGNumberDataColumn) colArray[2];
        
        double[] xvalues = colx.getNumberArray();
        double[] yvalues = coly.getNumberArray();
        double[] zvalues = colz.getNumberArray();

        NetcdfFileWriteable ncfile = null;
        try {
            ncfile = NetcdfFileWriteable.createNew(file.getAbsolutePath(), true);

            // Creates the dimension and variable of serial numbers.
            Dimension serialNumberDim = this.addSerialNumberDimension(ncfile, dataNum);
            String serialNumberDimName = serialNumberDim.getName();
            Variable serialNumberVar = this.addSerialNumberVarialbe(ncfile, serialNumberDim);
            serialNumberVar.addAttribute(SGDataUtility.getValueTypeAttribute(SGIDataColumnTypeConstants.VALUE_TYPE_NUMBER));
            
            // Add data columns as variables.
            Variable varx = new Variable(ncfile, null, null, "column0", DataType.DOUBLE, serialNumberDimName);
            Variable vary = new Variable(ncfile, null, null, "column1", DataType.DOUBLE, serialNumberDimName);
            Variable varz = new Variable(ncfile, null, null, "column2", DataType.DOUBLE, serialNumberDimName);
            String title = colx.getTitle();
            if (null!=title && "".equals(title.trim())==false) {
                Attribute attr = new Attribute(ATTRIBUTE_KEY_LONG_NAME, title);
                varx.addAttribute(attr);
            }
            title = coly.getTitle();
            if (null!=title && "".equals(title.trim())==false) {
                Attribute attr = new Attribute(ATTRIBUTE_KEY_LONG_NAME, title);
                vary.addAttribute(attr);
            }
            title = colz.getTitle();
            if (null!=title && "".equals(title.trim())==false) {
                Attribute attr = new Attribute(ATTRIBUTE_KEY_LONG_NAME, title);
                varz.addAttribute(attr);
            }

            ncfile.addVariable(null, varx);
            ncfile.addVariable(null, vary);
            ncfile.addVariable(null, varz);
            
            String[] varNames = new String[colArray.length];
            varNames[0] = varx.getName();
            varNames[1] = vary.getName();
            varNames[2] = varz.getName();

            varx.addAttribute(SGDataUtility.getValueTypeAttribute(colx.getValueType()));
            vary.addAttribute(SGDataUtility.getValueTypeAttribute(coly.getValueType()));
            varz.addAttribute(SGDataUtility.getValueTypeAttribute(colz.getValueType()));

            ncfile.create();

            // add values

            Array xarray = Array.factory(DataType.DOUBLE, new int[] { dataNum });
            for (int i = 0; i < xvalues.length; i++) {
                xarray.setDouble(i, xvalues[i]);
            }
            ncfile.write(varNames[0], xarray);

            Array yarray = Array.factory(DataType.DOUBLE, new int[] { dataNum });
            for (int i = 0; i < yvalues.length; i++) {
                yarray.setDouble(i, yvalues[i]);
            }
            ncfile.write(varNames[1], yarray);

            Array zarray = Array.factory(DataType.DOUBLE, new int[] { dataNum });
            for (int i = 0; i < zvalues.length; i++) {
                zarray.setDouble(i, zvalues[i]);
            }
            ncfile.write(varNames[2], zarray);

            // write serial numbers
            this.writeSerialNumberVarialbe(ncfile, serialNumberVar);
            
        } finally {
            if (null!=ncfile) {
                ncfile.close();
            }
        }

        return true;
    }
    
    // Creates the dimension of serial numbers.
    private Dimension addSerialNumberDimension(NetcdfFileWriteable ncfile, int len) {
		Dimension serialNumberDim = new Dimension(SERIAL_NUMBER_DIMENSION_NAME, len);
		ncfile.addDimension(null, serialNumberDim);
		return serialNumberDim;
    }
    
	// Creates the variable of serial numbers.
    private Variable addSerialNumberVarialbe(NetcdfFileWriteable ncfile, Dimension dim) {
		Variable var = new Variable(ncfile, null, null,
				SERIAL_NUMBER_DIMENSION_NAME, DataType.INT, dim.getName());
		ncfile.addVariable(null, var);
		return var;
    }
    
	private void writeSerialNumberVarialbe(NetcdfFileWriteable ncfile,
			Variable var) throws IOException, InvalidRangeException {
		final int len = var.getDimension(0).getLength();
		Array indexArray = Array.factory(DataType.INT, new int[] { len });
		for (int ii = 0; ii < len; ii++) {
			indexArray.setInt(ii, ii);
		}
		ncfile.write(var.getName(), indexArray);
	}
	
	private static class ByteData {
		Variable var = null;
		byte[] byteArray = null;
	}
	
	private ByteData createImageVariable(NetcdfFileWriteable ncfile,
			SGDrawingWindow wnd) throws IOException, InvalidRangeException {

		BackgroundImage bgImg = wnd.getBackgroundImage();
		if (bgImg == null) {
			return null;
		}
		byte[] imageByteArray = bgImg.getByteArray();
		if (imageByteArray == null) {
			return null;
		}
		
		// create dimension and variable
		Dimension dim = new Dimension(DIMENSION_NAME_BACKGROUND_IMAGE_BYTE_INDEX, imageByteArray.length);
		ncfile.addDimension(null, dim);
		Variable var = new Variable(ncfile, null, null, VARIABLE_NAME_BACKGROUND_IMAGE,
				DataType.BYTE, dim.getName());
		ncfile.addVariable(null, var);

		// add an attribute
		var.addAttribute(SGDataUtility
				.getValueTypeAttribute(SGIDataColumnTypeConstants.VALUE_TYPE_BYTE_DATA));
		var.addAttribute(new Attribute(ATTRIBUTE_KEY_IMAGE_FILE_EXTENSION, bgImg.getExtension()));

		// set to returned value
		ByteData ret = new ByteData();
		ret.var = var;
		ret.byteArray = imageByteArray;

		ncfile.create();

		return ret;
	}
	
	private void writeImageData(NetcdfFileWriteable ncfile, ByteData imageData)
			throws IOException, InvalidRangeException {
		Variable var = imageData.var;
		byte[] byteArray = imageData.byteArray;
		ArrayByte array = new ArrayByte(new int[] { byteArray.length });
		Index index = array.getIndex();
		for (int ii = 0; ii < byteArray.length; ii++) {
			array.setByte(index.set(ii), byteArray[ii]);
		}
		ncfile.write(var.getName(), array);
	}
    
    /**
     * Scalar-XY
     * @param data
     * @param file
     * @param groupName
     * @return
     * @throws IOException
     * @throws InvalidRangeException
     */
    private boolean saveArrayDataSXYToTempNetCDFFiles(
            final SGArrayData data, final File file, final String groupName, final SGDrawingWindow wnd)
    throws IOException, InvalidRangeException {

        NetcdfFileWriteable ncfile = null;
        try {
            ncfile = NetcdfFileWriteable.createNew(file.getAbsolutePath(), true);

            final int dataNum = data.getPointsNumber();
            
            // Creates the dimension and variable of serial numbers.
            Dimension serialNumberDim = this.addSerialNumberDimension(ncfile, dataNum);
            String serialNumberDimName = serialNumberDim.getName();
            Variable serialNumberVar = this.addSerialNumberVarialbe(ncfile, serialNumberDim);
            serialNumberVar.addAttribute(SGDataUtility.getValueTypeAttribute(SGIDataColumnTypeConstants.VALUE_TYPE_NUMBER));

            // Add data columns as variables.
            SGDataColumn[] colArray = data.getDataSetColumnsClone();
            
            // Add dimensions for text string.
            String[] textDimensionName = new String[colArray.length];
            int[] maxTextLength = new int[colArray.length];
            Arrays.fill(maxTextLength, 0);
            for (int i = 0; i < colArray.length; i++) {
                SGDataColumn col = colArray[i];
                if (SGIDataColumnTypeConstants.VALUE_TYPE_TEXT.equals(col.getValueType())
                		|| SGIDataColumnTypeConstants.VALUE_TYPE_DATE.equals(col.getValueType())) {
                    String[] textString = ((SGDataColumn)col).getStringArray();
                    int maxLength = 0;
                    for (int j = 0; j < textString.length; j++) {
                    	byte[] byteArray = textString[j].getBytes("UTF-8");
                    	if (maxLength < byteArray.length) {
                    		maxLength = byteArray.length;
                    	}
                    }
                    
                    String dimName = "clen" + i;
                    Dimension dimc = new Dimension(dimName, maxLength);
                    ncfile.addDimension(null, dimc);
                    
                    textDimensionName[i] = dimName;
                    maxTextLength[i] = maxLength;
                }
            }

            String[] varNames = new String[colArray.length];
            for (int i = 0; i < colArray.length; i++) {
                SGDataColumn col = colArray[i];
                String colValueType = col.getValueType();
                String varName = "column" + i;
                Variable var = null;
                if (SGIDataColumnTypeConstants.VALUE_TYPE_NUMBER.equals(colValueType) ||
                        SGIDataColumnTypeConstants.VALUE_TYPE_SAMPLING_RATE.equals(colValueType)) {
                    var = new Variable(ncfile, null, null, varName, DataType.DOUBLE, serialNumberDimName);
                    var.addAttribute(SGDataUtility.getValueTypeAttribute(col.getValueType()));
                } else if (SGIDataColumnTypeConstants.VALUE_TYPE_TEXT.equals(colValueType)
                		|| SGIDataColumnTypeConstants.VALUE_TYPE_DATE.equals(colValueType)) {
                	StringBuffer sb = new StringBuffer();
                	sb.append(serialNumberDimName);
                	sb.append(' ');
                	sb.append(textDimensionName[i]);
                	String dims = sb.toString();
                    var = new Variable(ncfile, null, null, varName, DataType.CHAR, dims);
                    String attrValueType;
                    if (SGIDataColumnTypeConstants.VALUE_TYPE_TEXT.equals(colValueType)) {
                    	attrValueType = SGIDataColumnTypeConstants.VALUE_TYPE_TEXT;
                    } else {
                    	attrValueType = SGIDataColumnTypeConstants.VALUE_TYPE_DATE;
                    }
                    var.addAttribute(SGDataUtility.getValueTypeAttribute(attrValueType));
                } else {
					throw new Error("Illegal value type=" + col.getValueType()
							+ ", (" + col.getTitle() + ", "
							+ col.getColumnType() + ")");
                }

                String title = col.getTitle();
                if (null != title && !"".equals(title.trim())) {
                    Attribute attr = new Attribute(ATTRIBUTE_KEY_LONG_NAME, title);
                    var.addAttribute(attr);
                }

    			varNames[i] = var.getName();
                ncfile.addVariable(null, var);
            }
            
            ncfile.create();

            for (int i = 0; i < colArray.length; i++) {
                SGDataColumn col = colArray[i];
                String colValueType = col.getValueType();
                if (SGIDataColumnTypeConstants.VALUE_TYPE_NUMBER.equals(colValueType) ||
                        SGIDataColumnTypeConstants.VALUE_TYPE_SAMPLING_RATE.equals(colValueType)) {
                    Array array = Array.factory(DataType.DOUBLE, new int[] { dataNum });
                    for (int j = 0; j < dataNum; j++) {
                        Object obj = col.getValue(j);
                        if (obj instanceof Double) {
                            array.setDouble(j, ((Double)obj).doubleValue());
                        } else {
                            array.setDouble(j, Double.NaN);
                        }
                    }
                    ncfile.write(varNames[i], array);
                    
                } else if (SGIDataColumnTypeConstants.VALUE_TYPE_TEXT.equals(colValueType)
                		|| SGIDataColumnTypeConstants.VALUE_TYPE_DATE.equals(colValueType)) {
                	int maxLength = 0;
                	byte[][] byteArrays = new byte[dataNum][];
                    for (int j = 0; j < dataNum; j++) {
                        Object obj = col.getValue(j);
                        String str;
                        if (obj instanceof String) {
                        	str = (String) obj;
                        } else if (obj instanceof SGDate) {
                           	str = ((SGDate) obj).toString();
                        } else {
                        	str = "";
                        }
                    	byteArrays[j] = str.getBytes("UTF-8");
                    	int len = byteArrays[j].length;
                    	if (len > maxLength) {
                    		maxLength = len;
                    	}
                    }
                	
                    ArrayByte array = new ArrayByte(new int[] { dataNum, maxLength });
            		Index index = array.getIndex();
                    for (int j = 0; j < dataNum; j++) {
                    	for (int k = 0; k < byteArrays[j].length; k++) {
                    		array.setByte(index.set(j, k), byteArrays[j][k]);
                    	}
                    }
                    ncfile.write(varNames[i], array);
                }
            }

            // write serial numbers
            this.writeSerialNumberVarialbe(ncfile, serialNumberVar);

        } finally {
            if (null!=ncfile) {
                ncfile.close();
            }
        }

        return true;
    }
    
    // save background image
    private File saveImageToFile(
            final SGDrawingWindow wnd, final File datasetTempDir)
    throws IOException, InvalidRangeException {
    	StringBuffer sb = new StringBuffer();
    	sb.append(datasetTempDir);
    	sb.append(SGIConstants.FILE_SEPARATOR);
    	sb.append(ARCHIVE_IMAGE_NAME);
    	sb.append(".nc");
        String fname = sb.toString();
        File file = new File(fname);
        NetcdfFileWriteable ncfile = null;
        try {
            ncfile = NetcdfFileWriteable.createNew(file.getAbsolutePath(), true);
            
            // create image variable
            ByteData imageData = this.createImageVariable(ncfile, wnd);

            // write image data
            if (imageData != null) {
            	this.writeImageData(ncfile, imageData);
            }

        } finally {
            if (ncfile != null) {
                ncfile.close();
            }
        }
    	
        return file;
    }

    private List<File> saveDataToFiles(
            final SGDrawingWindow wnd, final File datasetTempDir)
            throws IOException, InvalidRangeException {
        List<File> flist = new ArrayList<File>();
        ArrayList<SGFigure> figures = wnd.getVisibleFigureList();
        for (int ii = 0; ii < figures.size(); ii++) {
            SGFigure figure = (SGFigure) figures.get(ii);
            List<SGData> dataList = figure.getVisibleDataList();
            for (int jj = 0; jj < dataList.size(); jj++) {
                SGData data = (SGData) dataList.get(jj);
                String groupName = this.createGroupName(ii, jj);

                StringBuffer sb = new StringBuffer();
                sb.append(datasetTempDir.getPath());
                sb.append(SGIConstants.FILE_SEPARATOR);
                sb.append(groupName);
                sb.append(".nc");
                String fname = sb.toString();
                File file = new File(fname);

                boolean result = false;
                if (SGDataUtility.isArrayData(data)) {
                    if (SGUtilityText.isEqualString(data.getDataType(), SGDataTypeConstants.VXY_DATA)) {
                        result = this.saveArrayDataVXYToTempNetCDFFiles((SGArrayData)data, file, groupName, wnd);
                    } else if (SGUtilityText.isEqualString(data.getDataType(), SGDataTypeConstants.SXYZ_DATA)) {
                        result = this.saveArrayDataSXYZToTempNetCDFFiles((SGArrayData)data, file, groupName, wnd);
                    } else {
                        // SXY
                        result = this.saveArrayDataSXYToTempNetCDFFiles((SGArrayData)data, file, groupName, wnd);
                    }
                } else if (SGDataUtility.isNetCDFData(data)) {
                    result = data.saveData(file, SGIConstants.SAVE_PROPERTIES_TO_DATA_SET_NETCDF);
                }
                if (result==false) {
                    file.delete();
                    file = null;
                }
                if (null!=file) {
                    flist.add(file);
                }
            }
        }

        return flist;
    }

    private void copyNetcdfTempFileHeaderToOutFile(final File[] files, final NetcdfFileWriteable outNcfile)
    throws IOException {
        for (int i = 0; i < files.length; i++) {
            File file = files[i];
            String groupName = this.getGroupNameFromDataFile(file);

            NetcdfFile ncfile = null;
            Group group = new Group(outNcfile, null, groupName);
            try {
                ncfile = NetcdfFile.open(file.getAbsolutePath());

                List<Attribute> gattrs = ncfile.getGlobalAttributes();
                List<Dimension> dims = ncfile.getDimensions();
                List<Variable> vars = ncfile.getVariables();

                for (Dimension dim : dims) {
                    String name = dim.getName();
                    dim.setName(this.getGroupVariableName(groupName, name));
                    outNcfile.addDimension(null, dim);
                }
                for (Attribute gattr : gattrs) {
                    group.addAttribute(gattr);
                }
                for (Variable v : vars) {
                    String varName = v.getName();
                    String[] groupNames = varName.split("/");
                    if (groupNames.length>=2) {
                        Group g = group;
                        for (int j = 0; j < groupNames.length-1; j++) {
                            Group g2 = g.findGroup(groupNames[j]);
                            if (g2==null) {
                                g2 = new Group(outNcfile, g, groupNames[j]);
                                g.addGroup(g2);
                            }
                            g = g2;
                        }
                        v.setName(groupNames[groupNames.length-1]);
                        g.addVariable(v);
                    } else {
                        group.addVariable(v);
                    }
                }

                outNcfile.addGroup(null, group);

            } finally {
                if (null!=ncfile) {
                    ncfile.close();
                }
            }
        }

    }

    private void copyNetcdfTempFileArrayToOutFile(final File[] files, final NetcdfFileWriteable outNcfile)
    throws IOException, InvalidRangeException {
        for (int i = 0; i < files.length; i++) {
            File file = files[i];
            String groupName = this.getGroupNameFromDataFile(file);

            NetcdfFile ncfile = null;
            try {
                ncfile = NetcdfFile.open(file.getAbsolutePath());

                List<Variable> vars = ncfile.getVariables();
                for (Variable v : vars) {
                    Array array = v.read();
                    outNcfile.write(this.getGroupVariableName(groupName, v.getName()), array);
                }
            } finally {
                if (null!=ncfile) {
                    ncfile.close();
                }
            }
        }
    }

    private void copyNetcdfTempFileToOutFile(
            final String propertyString,
            final File datasetTempDir, final File outFile)
    throws IOException, InvalidRangeException {
        File[] files = datasetTempDir.listFiles();

        NetcdfFileWriteable outNcfile = null;
        try {
            outNcfile = NetcdfFileWriteable.createNew(outFile.getAbsolutePath());
            outNcfile.addGlobalAttribute(ATTRIBUTE_NAME_PROPERTY, propertyString);

            this.copyNetcdfTempFileHeaderToOutFile(files, outNcfile);

            outNcfile.create();

            this.copyNetcdfTempFileArrayToOutFile(files, outNcfile);

        } finally {
            if (null!=outNcfile) {
                outNcfile.close();
            }
        }

    }

    /**
     * Returns property string.
     * @param wnd
     * @param versionString
     * @return
     * @throws IOException
     */
    private String getPropertyString(final SGDrawingWindow wnd, final String versionString)
    throws IOException {
        SGPropertyFileCreator propFileCreator = new SGPropertyFileCreator();

        final int type = SGIConstants.SAVE_PROPERTIES_TO_DATA_SET_NETCDF;

        ByteArrayOutputStream baos = null;
        try {
            baos = new ByteArrayOutputStream();
            propFileCreator.create(wnd, baos, type, versionString);
            return baos.toString("UTF-8");
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        } finally {
            if (null!=baos) {
                baos.close();
            }
        }

    }

    /**
     * Do save to netCDF dataset file.
     * @param wnd
     * @param versionString
     * @param datasetTempDir temporary directory that contains files which dataset file takes on.
     * @param outFile output file
     * @return SUCCESSFUL_COMPLETION if succeeds. other if failed.
     */
    public int save(
            final SGDrawingWindow wnd, final String versionString,
            final File datasetTempDir, final File outFile) {

//        List<File> files;
        try {
        	
        	// save data
        	this.saveDataToFiles(wnd, datasetTempDir);
        	
//            files = saveDataToFiles(wnd, datasetTempDir);
//            if (files.size()>0) {
//                String propertyString = this.getPropertyString(wnd, versionString);
////                System.out.println(propertyString);
//
//                this.copyNetcdfTempFileToOutFile(propertyString, datasetTempDir, outFile);
//            } else {
//                return SGIConstants.FILE_SAVE_FAILURE;
//            }
        	
        	// save background image
        	this.saveImageToFile(wnd, datasetTempDir);
        	
        	// get properties
            String propertyString = this.getPropertyString(wnd, versionString);
            
            // output to the file
            this.copyNetcdfTempFileToOutFile(propertyString, datasetTempDir, outFile);
            
        } catch (IOException e1) {
            return SGIConstants.FILE_SAVE_FAILURE;
        } catch (InvalidRangeException e1) {
            return SGIConstants.FILE_SAVE_FAILURE;
        } catch (Exception e1) {
            return SGIConstants.FILE_SAVE_FAILURE;
        }

        return SGIConstants.SUCCESSFUL_COMPLETION;

    }

    private void setupNetCDFGroup(final NetcdfFile ncfile) {
        List<Dimension> dims = ncfile.getDimensions();
        Set<String> set = new HashSet<String>();
        for (int i = 0; i < dims.size(); i++) {
        	Dimension dim = dims.get(i);
        	String name = dim.getName();
        	String[] names = name.split("/");
//            String[] names = dims.get(i).getName().split("/");
        	String groupName = names[0].trim();
            if (!"".equals(groupName)) {
                set.add(groupName);
            }
        }
        for (String dimName : set) {
            Group g = new Group(ncfile, null, dimName);
            ncfile.addGroup(null, g);
        }
    }

    private int getNumberOfDataInNetcdfFile(final NetcdfFile ncfile, final int figureIndex) {
        Group root = ncfile.getRootGroup();
        List<Group> groups = root.getGroups();

//        String dataNamePrefix = "id"+figureIndex+"-";
        StringBuffer sb = new StringBuffer();
        sb.append("id");
        sb.append(figureIndex);
        sb.append('-');
        String dataNamePrefix = sb.toString();

        int count = 0;
        for (int i = 0; i < groups.size(); i++) {
            if (groups.get(i).getName().startsWith(dataNamePrefix)) {
                count++;
            }
        }
        return count;
    }


    private boolean setupFigureFromProperty(
            final Element elFigure, final SGDrawingWindow wnd, final SGFigure figure) {

        //
        // create a figure
        //

        final int figureID = wnd.assignFigureId();
        figure.setID(figureID);

        //
        // create SGIFigureElement objects
        //

        if (figure.readProperty(elFigure) == false) {
            return false;
        }

        if (this.mMain.mFigureCreator.createFigureElementFromPropertyFile(
                figure, elFigure) == SGIConstants.PROPERTY_FILE_INCORRECT) {
            return false;
        }

        //
        // called after created SGIAxisElement object
        //
        String str = null;
        Number num = null;

        // space axis line and numbers
        str = elFigure.getAttribute(SGFigure.KEY_SPACE_AXIS_LINE_AND_NUMBER);
        if (str.length() != 0) {
            StringBuffer uSpaceLN = new StringBuffer();
            num = SGUtilityText.getNumber(str, uSpaceLN);
            if (num == null) {
                return false;
            }
            final float spaceLN = num.floatValue();
            SGIFigureElementAxis aElement = figure.getAxisElement();
            aElement.setSpaceAxisLineAndNumber(spaceLN, uSpaceLN.toString());
            if (figure.setSpaceAxisLineAndNumber(spaceLN, uSpaceLN.toString()) == false) {
                return false;
            }
        }

        // space numbers and title
        str = elFigure.getAttribute(SGFigure.KEY_SPACE_NUMBER_AND_TITLE);
        if (str.length() != 0) {
            StringBuffer uSpaceNT = new StringBuffer();
            num = SGUtilityText.getNumber(str, uSpaceNT);
            if (num == null) {
                return false;
            }
            final float spaceNT = num.floatValue();
            SGIFigureElementAxis aElement = figure.getAxisElement();
            aElement.setSpaceNumberAndTitle(spaceNT, uSpaceNT.toString());
            if (figure.setSpaceNumberAndTitle(spaceNT, uSpaceNT.toString()) == false) {
                return false;
            }
        }

        return true;
    }

    private SGData getDataInstanceFromProperty(final String dataType, final Element elData) {
        Class<?> dataClass = null;

        if (SGDataUtility.isArrayData(dataType)) {
            // use netCDF data class for array data type.
            if (SGDataUtility.isSXYTypeData(dataType)) {
                String pickupDimensionName = elData.getAttribute(SGIDataColumnNameConstants.KEY_PICKUP_DIMENSION_NAME);
                if (null==pickupDimensionName || pickupDimensionName.trim().length()==0) {
                    dataClass = SGSXYMultipleVariableNetCDFData.class;
                } else {
                    dataClass = SGSXYMultipleDimensionNetCDFData.class;
                }
            } else if (SGDataTypeConstants.VXY_DATA.equals(dataType)) {
                dataClass = SGVXYNetCDFData.class;
            } else if (SGDataTypeConstants.SXYZ_DATA.equals(dataType)) {
                dataClass = SGSXYZNetCDFData.class;
            } else {
                return null;
            }
        } else if (SGDataUtility.isNetCDFData(dataType)) {
            if (SGDataUtility.isSXYTypeData(dataType)) {
                if (SGDataUtility.isNetCDFDimensionData(dataType)) {
                    dataClass = SGSXYMultipleDimensionNetCDFData.class;
                } else {
                    String pickupDimensionName = elData.getAttribute(SGIDataColumnNameConstants.KEY_PICKUP_DIMENSION_NAME);
                    if (null==pickupDimensionName || pickupDimensionName.trim().length()==0) {
                        dataClass = SGSXYMultipleVariableNetCDFData.class;
                    } else {
                        dataClass = SGSXYMultipleDimensionNetCDFData.class;
                    }
                }
            } else if (SGDataTypeConstants.VXY_NETCDF_DATA.equals(dataType)) {
                dataClass = SGVXYNetCDFData.class;
            } else if (SGDataTypeConstants.SXYZ_NETCDF_DATA.equals(dataType)) {
                dataClass = SGSXYZNetCDFData.class;
            } else {
                return null;
            }
        } else {
            return null;
        }

        String className = dataClass.getName();

        // get the Class object
        Class<?> cl = null;
        try {
            cl = Class.forName(className);
        } catch (ClassNotFoundException ex) {
            return null;
        }
        if (cl == null) {
            return null;
        }

        // create a new SGData object
        SGData data = null;
        try {
            Object obj = cl.newInstance();
            if (obj instanceof SGData) {
                data = (SGData)obj;
            }
        } catch (Exception ex) {
            return null;
        }

        return data;
    }

    /*
	private String[] getVariableNameInDataAttribute(final Element elData,
			final String attributeName) {
		String attributeValues = elData.getAttribute(attributeName);
		if (attributeValues.length() == 0) {
			return null;
		}

		String[] values = SGUtilityText.parseStrings(attributeValues);
		return values;
	}
	
    private String[] appendGroupNameWithVariableNames(final String groupName, 
    		final String[] variableNames) {
        if (variableNames == null) {
            return null;
        }
        String[] resultNames = new String[variableNames.length];
        for (int i = 0; i < variableNames.length; i++) {
        	StringBuffer sb = new StringBuffer();
        	sb.append(groupName);
        	sb.append('/');
        	sb.append(variableNames[i]);
        	resultNames[i] = sb.toString();
        }
        return resultNames;
    }
    */

    private int createDataObjectsFromPropertyFile(final Element elFigure,
            final SGFigure figure, final int figureIndex, final NetcdfFile ncfile,
            final SGIProgressControl progress, final boolean readDataProperty,
            final String versionNumber) {

        final int ic = SGIConstants.PROPERTY_FILE_INCORRECT;
        final int di = SGIConstants.DATA_FILE_INVALID;

        NodeList nList = elFigure.getElementsByTagName(SGIFigureElementGraph.TAG_NAME_DATA);
        final int len = nList.getLength();
        if (len != this.getNumberOfDataInNetcdfFile(ncfile, figureIndex)) {
            return ic;                
        }

        int[] indexArray = new int[len];
        SGData[] dataArray = new SGData[len];
        boolean indexValid = true;
        for (int ii = 0; ii < len; ii++) {
            String groupName = this.createGroupName(figureIndex, ii);

            Node node = nList.item(ii);
            if ((node instanceof Element) == false) {
                continue;
            }
            Element elData = (Element) node;

            // data type
            String dataType = elData.getAttribute(SGIFigureElement.KEY_DATA_TYPE);
            if (dataType == null) {
                return ic;
            }

            // replaces the data type
            // because sampling SXY-data and date SXY-date is to be eliminated at some stage.
            dataType = SGApplicationUtility.getArrayDataType(dataType);
            
            if (SGDataUtility.isArrayData(dataType)) {
            	if (SGDataUtility.isSXYTypeData(dataType)) {
            		dataType = SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DATA;
            	} else if (SGDataUtility.isVXYTypeData(dataType)) {
            		dataType = SGDataTypeConstants.VXY_NETCDF_DATA;
            	} else if (SGDataTypeConstants.SXYZ_DATA.equals(dataType)) {
            		dataType = SGDataTypeConstants.SXYZ_NETCDF_DATA;
            	}
            }

            // set the class type for backward compatibility between 1.0.7
            if (SGIFigureTypeConstants.FIGURE_TYPE_XY.equals(figure.getClassType())) {
                String type = null;
                if (SGDataUtility.isVXYTypeData(dataType)) {
                    type = SGIFigureTypeConstants.FIGURE_TYPE_VXY;
                } else {
                    type = SGIFigureTypeConstants.FIGURE_TYPE_SXY;
                }
                figure.setClassType(type);
            }

            SGData data = this.getDataInstanceFromProperty(dataType, elData);
            if (data == null) {
                return ic;
            }

            // create information map
            Map<String, Object> infoMap = this.mMain.createInfoMap(dataType, elData);
            infoMap.put(SGIFigureElementGraph.KEY_GROUP_NAME, groupName);
//            SGDataUtility.dumpInfoMap(infoMap);

            if (this.getNumberOfDataInNetcdfFile(ncfile, figureIndex) == 0) {
                return SGIConstants.DATA_NUMBER_SHORTAGE;
            }

            // get an object from dataList

            // get the data column information
            SGDataColumnInfoSet colInfoSet = null;
            SGNetCDFFile nc = null;
            try {
                NetcdfFile newNcfile = NetcdfFile.open(ncfile.getLocation());
                nc = new SGNetCDFFile(newNcfile);
            } catch (IOException e) {
                return di;
            }
            infoMap.put(SGIDataColumnTypeConstants.KEY_DATA_SOURCE, nc);
            colInfoSet = this.mMain.getNetCDFDefaultDataColumnInfo(nc, dataType, infoMap);
            if (colInfoSet == null) {
                return di;
            }

/*            
            if (SGDataUtility.isArrayData(dataType)) {
                SGDataColumnInfo[] columns = colInfoSet.getDataColumnInfoArray();
                for (int i = 0; i < columns.length; i++) {
                    columns[i].setColumnType("");
                }

                // serial numbers
                String[] snIndices = this.appendGroupNameWithVariableNames(groupName, 
                		new String[] { SERIAL_NUMBER_DIMENSION_NAME });
                if (snIndices.length != 1) {
                	return di;
                }
                int snVarIndex = nc.getVariableIndex(snIndices[0]);
                if (snVarIndex != -1) {
                    columns[snVarIndex].setColumnType(SGIDataColumnTypeConstants.SERIAL_NUMBER);
                }
                
                if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXY_DATA)
                		|| SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXY_MULTIPLE_DATA)) {
                	
//                if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXY_DATA)) {
//                    String[] xIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_VALUE_COLUMN_INDEX);
//                    String[] yIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_VALUE_COLUMN_INDEX);
//                    String[] leIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_LOWER_ERROR_BAR_COLUMN_INDEX);
//                    String[] ueIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_UPPER_ERROR_BAR_COLUMN_INDEX);
//                    String[] ehIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_ERROR_BAR_HOLDER_COLUMN_INDEX);
//                    String[] tlIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_TICK_LABEL_COLUMN_INDEX);
//                    String[] thIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_TICK_LABEL_HOLDER_COLUMN_INDEX);
//                    xIndices = this.appendGroupNameWithVariableNames(groupName, xIndices);
//                    yIndices = this.appendGroupNameWithVariableNames(groupName, yIndices);
//                    leIndices = this.appendGroupNameWithVariableNames(groupName, leIndices);
//                    ueIndices = this.appendGroupNameWithVariableNames(groupName, ueIndices);
//                    ehIndices = this.appendGroupNameWithVariableNames(groupName, ehIndices);
//                    tlIndices = this.appendGroupNameWithVariableNames(groupName, tlIndices);
//                    thIndices = this.appendGroupNameWithVariableNames(groupName, thIndices);
//                    if (xIndices==null || yIndices==null) {
//                        return di;
//                    }
//                    
//                    SGDataUtility.updateColumnTypeOfSXYMultipleVariableNetCDFDataFromVariableNames(
//                            nc, columns, xIndices, yIndices,
//                            leIndices, ueIndices, ehIndices,
//                            tlIndices, thIndices);
//                    
//                } else if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXY_MULTIPLE_DATA)) {
                    String[] xIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_VALUE_COLUMN_INDICES);
                    String[] yIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_VALUE_COLUMN_INDICES);
                    String[] leIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_LOWER_ERROR_BAR_COLUMN_INDICES);
                    String[] ueIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_UPPER_ERROR_BAR_COLUMN_INDICES);
                    String[] ehIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_ERROR_BAR_HOLDER_COLUMN_INDICES);
                    String[] tlIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_TICK_LABEL_COLUMN_INDICES);
                    String[] thIndices = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_TICK_LABEL_HOLDER_COLUMN_INDICES);
                    xIndices = this.appendGroupNameWithVariableNames(groupName, xIndices);
                    yIndices = this.appendGroupNameWithVariableNames(groupName, yIndices);
                    leIndices = this.appendGroupNameWithVariableNames(groupName, leIndices);
                    ueIndices = this.appendGroupNameWithVariableNames(groupName, ueIndices);
                    ehIndices = this.appendGroupNameWithVariableNames(groupName, ehIndices);
                    tlIndices = this.appendGroupNameWithVariableNames(groupName, tlIndices);
                    thIndices = this.appendGroupNameWithVariableNames(groupName, thIndices);
                    if (xIndices==null || yIndices==null) {
                        return di;
                    }
                    
                    SGDataUtility.updateColumnTypeOfSXYMultipleVariableNetCDFDataFromVariableNames(
                            nc, columns, xIndices, yIndices,
                            leIndices, ueIndices, ehIndices,
                            tlIndices, thIndices);
                } else if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.VXY_DATA)) {
                    String[] xValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_COORDINATE_COLUMN_INDEX);
                    String[] yValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_COORDINATE_COLUMN_INDEX);
                    String[] c1ValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_COMPONENT_1_COLUMN_INDEX);
                    String[] c2ValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_COMPONENT_2_COLUMN_INDEX);
                    xValueName = this.appendGroupNameWithVariableNames(groupName, xValueName);
                    yValueName = this.appendGroupNameWithVariableNames(groupName, yValueName);
                    c1ValueName = this.appendGroupNameWithVariableNames(groupName, c1ValueName);
                    c2ValueName = this.appendGroupNameWithVariableNames(groupName, c2ValueName);
                    if (xValueName==null || yValueName==null || c1ValueName==null ||c2ValueName==null) {
                        return di;
                    }
                    
                    Boolean isPolar = Boolean.valueOf(elData.getAttribute(SGIFigureElementGraph.KEY_POLAR));
                    if (xValueName.length==1 && yValueName.length==1 && c1ValueName.length==1 && c2ValueName.length==1) {
                        int xVarIndex = nc.getVariableIndex(xValueName[0]);
                        int yVarIndex = nc.getVariableIndex(yValueName[0]);
                        int c1VarIndex = nc.getVariableIndex(c1ValueName[0]);
                        int c2VarIndex = nc.getVariableIndex(c2ValueName[0]);
                        if (xVarIndex<0 || yVarIndex<0 || c1VarIndex<0 || c2VarIndex<0) {
                            return di;
                        }
                        columns[xVarIndex].setColumnType(SGIDataColumnTypeConstants.X_COORDINATE);
                        columns[yVarIndex].setColumnType(SGIDataColumnTypeConstants.Y_COORDINATE);
                        if (isPolar) {
                            columns[c1VarIndex].setColumnType(SGIDataColumnTypeConstants.MAGNITUDE);
                            columns[c2VarIndex].setColumnType(SGIDataColumnTypeConstants.ANGLE);
                        } else {
                            columns[c1VarIndex].setColumnType(SGIDataColumnTypeConstants.X_COMPONENT);
                            columns[c2VarIndex].setColumnType(SGIDataColumnTypeConstants.Y_COMPONENT);
                        }
                    } else {
                        return di;
                    }
                    
                } else if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXYZ_DATA)) {
                    String[] xValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_VALUE_COLUMN_INDEX);
                    String[] yValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_VALUE_COLUMN_INDEX);
                    String[] zValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Z_VALUE_COLUMN_INDEX);
                    xValueName = this.appendGroupNameWithVariableNames(groupName, xValueName);
                    yValueName = this.appendGroupNameWithVariableNames(groupName, yValueName);
                    zValueName = this.appendGroupNameWithVariableNames(groupName, zValueName);
                    
                    if (xValueName==null || yValueName==null || zValueName==null) {
                        return di;
                    }
                    if (xValueName.length==1 && yValueName.length==1 && zValueName.length==1) {
                        int xVarIndex = nc.getVariableIndex(xValueName[0]);
                        int yVarIndex = nc.getVariableIndex(yValueName[0]);
                        int zVarIndex = nc.getVariableIndex(zValueName[0]);
                        if (xVarIndex<0 || yVarIndex<0 || zVarIndex<0) {
                            return di;
                        }
                        columns[xVarIndex].setColumnType(SGIDataColumnTypeConstants.X_VALUE);
                        columns[yVarIndex].setColumnType(SGIDataColumnTypeConstants.Y_VALUE);
                        columns[zVarIndex].setColumnType(SGIDataColumnTypeConstants.Z_VALUE);
                    } else {
                        return di;
                    }
                    
                } else {
                    throw new Error("UNKNOWN DATA TYPE="+dataType);
                }
                
            } else if (SGDataUtility.isNetCDFData(dataType)) {
                SGDataColumnInfo[] columns = colInfoSet.getDataColumnInfoArray();
                for (int i = 0; i < columns.length; i++) {
                    columns[i].setColumnType("");
                }

                if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DATA)
                		|| SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXY_NETCDF_DATA)) {
                    // SXY MultipleVariableNetCDF
                    
                    String[] xValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_VALUE_NAMES);
                    if (xValueNames == null) {
                    	xValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_VALUE_NAME);
                    }
                    String[] yValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_VALUE_NAMES);
                    if (yValueNames == null) {
                    	yValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_VALUE_NAME);
                    }
                    String[] leValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_LOWER_ERROR_VALUE_NAMES);
                    String[] ueValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_UPPER_ERROR_VALUE_NAMES);
                    String[] ehValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_ERROR_BAR_HOLDER_NAMES);
                    String[] tlValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_TICK_LABEL_NAMES);
                    String[] thValueNames = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_TICK_LABEL_HOLDER_NAMES);
                    xValueNames = this.appendGroupNameWithVariableNames(groupName, xValueNames);
                    yValueNames = this.appendGroupNameWithVariableNames(groupName, yValueNames);
                    leValueNames = this.appendGroupNameWithVariableNames(groupName, leValueNames);
                    ueValueNames = this.appendGroupNameWithVariableNames(groupName, ueValueNames);
                    ehValueNames = this.appendGroupNameWithVariableNames(groupName, ehValueNames);
                    tlValueNames = this.appendGroupNameWithVariableNames(groupName, tlValueNames);
                    thValueNames = this.appendGroupNameWithVariableNames(groupName, thValueNames);
                    if (xValueNames==null || yValueNames==null) {
                        return di;
                    }
                    
                    SGDataUtility.updateColumnTypeOfSXYMultipleVariableNetCDFDataFromVariableNames(
                            nc, columns, xValueNames, yValueNames,
                            leValueNames, ueValueNames, ehValueNames,
                            tlValueNames, thValueNames);
                    
                } else if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DIMENSION_DATA)) {
                    // SXY MultipleDimensionNetCDF
                    String pickupDimensionName = elData.getAttribute(SGIDataColumnNameConstants.KEY_PICKUP_DIMENSION_NAME);
                    
                    if (null==pickupDimensionName || pickupDimensionName.trim().length()==0) {
                        return di;
                    } else {
                        String[] xValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_VALUE_NAME);
                        String[] yValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_VALUE_NAME);
                        String[] leValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_LOWER_ERROR_VALUE_NAMES);
                        String[] ueValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_UPPER_ERROR_VALUE_NAMES);
                        String[] tlValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_TICK_LABEL_NAMES);
                        xValueName = this.appendGroupNameWithVariableNames(groupName, xValueName);
                        yValueName = this.appendGroupNameWithVariableNames(groupName, yValueName);
                        leValueName = this.appendGroupNameWithVariableNames(groupName, leValueName);
                        ueValueName = this.appendGroupNameWithVariableNames(groupName, ueValueName);
                        tlValueName = this.appendGroupNameWithVariableNames(groupName, tlValueName);
                        if (xValueName==null || yValueName==null || xValueName.length!=1 || yValueName.length!=1) {
                            return di;
                        }
                        
                        int xVarIndex = nc.getVariableIndex(xValueName[0]);
                        columns[xVarIndex].setColumnType(SGIDataColumnTypeConstants.X_VALUE);
                        int yVarIndex = nc.getVariableIndex(yValueName[0]);
                        columns[yVarIndex].setColumnType(SGIDataColumnTypeConstants.Y_VALUE);

                        boolean errorbarFlag = false;
                        boolean ticklabelFlag = false;
                        if (leValueName!=null && leValueName.length>0 && 
                                ueValueName!=null && ueValueName.length>0 &&
                                leValueName.length==ueValueName.length) {
                            errorbarFlag = true;
                        }
                        if (tlValueName!=null && tlValueName.length>0) {
                            ticklabelFlag = true;
                        }
                        if (errorbarFlag) {
                            int leVarIndex = nc.getVariableIndex(leValueName[0]);
                            int ueVarIndex = nc.getVariableIndex(ueValueName[0]);
                            if (leVarIndex!=ueVarIndex) {
                                columns[leVarIndex].setColumnType(SGIDataColumnTypeConstants.LOWER_ERROR_VALUE);
                                columns[ueVarIndex].setColumnType(SGIDataColumnTypeConstants.UPPER_ERROR_VALUE);
                            } else {
                                columns[leVarIndex].setColumnType(SGIDataColumnTypeConstants.LOWER_UPPER_ERROR_VALUE);
                            }
                        }
                        if (ticklabelFlag) {
                            int tlVarIndex = nc.getVariableIndex(tlValueName[0]);
                            columns[tlVarIndex].setColumnType(SGIDataColumnTypeConstants.TICK_LABEL);
                        }
                        
                        pickupDimensionName = groupName + "/" + pickupDimensionName;
                        int pickupDimIndex = nc.getVariableIndex(pickupDimensionName);
                        if (pickupDimIndex<0) {
                            return di;
                        }
                        columns[pickupDimIndex].setColumnType(SGIDataColumnTypeConstants.PICKUP);
                    }
                    
                } else if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.VXY_NETCDF_DATA)) {
                    // VXY
                    String[] xValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_LOCATION_VARIABLE_NAME);
                    String[] yValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_LOCATION_VARIABLE_NAME);
                    String[] c1ValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_FIRST_COMPONENT_VARIABLE_NAME);
                    String[] c2ValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_SECOND_COMPONENT_VARIABLE_NAME);
                    xValueName = this.appendGroupNameWithVariableNames(groupName, xValueName);
                    yValueName = this.appendGroupNameWithVariableNames(groupName, yValueName);
                    c1ValueName = this.appendGroupNameWithVariableNames(groupName, c1ValueName);
                    c2ValueName = this.appendGroupNameWithVariableNames(groupName, c2ValueName);
                    if (xValueName==null || yValueName==null || c1ValueName==null ||c2ValueName==null) {
                        return di;
                    }
                    
                    Boolean isPolar = Boolean.valueOf(elData.getAttribute(SGIFigureElementGraph.KEY_POLAR));
                    if (xValueName.length==1 && yValueName.length==1 && c1ValueName.length==1 && c2ValueName.length==1) {
                        int xVarIndex = nc.getVariableIndex(xValueName[0]);
                        int yVarIndex = nc.getVariableIndex(yValueName[0]);
                        int c1VarIndex = nc.getVariableIndex(c1ValueName[0]);
                        int c2VarIndex = nc.getVariableIndex(c2ValueName[0]);
                        if (xVarIndex<0 || yVarIndex<0 || c1VarIndex<0 || c2VarIndex<0) {
                            return di;
                        }
                        columns[xVarIndex].setColumnType(SGIDataColumnTypeConstants.X_COORDINATE);
                        columns[yVarIndex].setColumnType(SGIDataColumnTypeConstants.Y_COORDINATE);
                        if (isPolar) {
                            columns[c1VarIndex].setColumnType(SGIDataColumnTypeConstants.MAGNITUDE);
                            columns[c2VarIndex].setColumnType(SGIDataColumnTypeConstants.ANGLE);
                        } else {
                            columns[c1VarIndex].setColumnType(SGIDataColumnTypeConstants.X_COMPONENT);
                            columns[c2VarIndex].setColumnType(SGIDataColumnTypeConstants.Y_COMPONENT);
                        }
                    } else {
                        return di;
                    }
                    
                } else if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXYZ_NETCDF_DATA)) {
                    // SXYZ
                    String[] xValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_X_VALUE_NAME);
                    String[] yValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Y_VALUE_NAME);
                    String[] zValueName = this.getVariableNameInDataAttribute(elData, SGIDataColumnNameConstants.KEY_Z_VALUE_NAME);
                    xValueName = this.appendGroupNameWithVariableNames(groupName, xValueName);
                    yValueName = this.appendGroupNameWithVariableNames(groupName, yValueName);
                    zValueName = this.appendGroupNameWithVariableNames(groupName, zValueName);
                    
                    if (xValueName==null || yValueName==null || zValueName==null) {
                        return di;
                    }
                    if (xValueName.length==1 && yValueName.length==1 && zValueName.length==1) {
                        int xVarIndex = nc.getVariableIndex(xValueName[0]);
                        int yVarIndex = nc.getVariableIndex(yValueName[0]);
                        int zVarIndex = nc.getVariableIndex(zValueName[0]);
                        if (xVarIndex<0 || yVarIndex<0 || zVarIndex<0) {
                            return di;
                        }
                        columns[xVarIndex].setColumnType(SGIDataColumnTypeConstants.X_VALUE);
                        columns[yVarIndex].setColumnType(SGIDataColumnTypeConstants.Y_VALUE);
                        columns[zVarIndex].setColumnType(SGIDataColumnTypeConstants.Z_VALUE);
                    } else {
                        return di;
                    }
                }

                // TIME
				String[] timeValueNames = this.getVariableNameInDataAttribute(
						elData, SGIDataColumnNameConstants.KEY_TIME_VARIABLE_NAME);
				if (timeValueNames != null && timeValueNames.length == 1
						&& !"".equals(timeValueNames[0])) {
					timeValueNames = this.appendGroupNameWithVariableNames(groupName, timeValueNames);
					int timeVarIndex = nc.getVariableIndex(timeValueNames[0]);
					columns[timeVarIndex].setColumnType(SGIDataColumnTypeConstants.TIME);
				}
                
                // serial numbers
				String[] serialValueNames = this.getVariableNameInDataAttribute(
						elData, SGIDataColumnNameConstants.KEY_SERIAL_NUMBER_VARIABLE_NAME);
				if (serialValueNames != null && serialValueNames.length == 1
						&& !"".equals(serialValueNames[0])) {
					serialValueNames = this.appendGroupNameWithVariableNames(groupName, serialValueNames);
					int serialNumberVarIndex = nc.getVariableIndex(serialValueNames[0]);
					columns[serialNumberVarIndex].setColumnType(SGIDataColumnTypeConstants.SERIAL_NUMBER);
				}

                // ORIGIN
                String originString = elData.getAttribute(SGNetCDFData.KEY_ORIGIN_MAP);
                if (originString!=null && originString.trim().equals("")==false) {
                    String[][] originMaps = SGUtilityText.readStringMaps(originString);
                    StringBuffer sb = new StringBuffer();
                    for (int i = 0; i < originMaps.length; i++) {
//                        sb.append(groupName+"/")
//                        .append(originMaps[i][0])
//                        .append("=")
//                        .append(originMaps[i][1]);
//                        if (i!=originMaps.length-1) {
//                            sb.append(",");
//                        }
                        sb.append(groupName);
                        sb.append('/');
                        sb.append(originMaps[i][0]);
                        sb.append('=');
                        sb.append(originMaps[i][1]);
                        if (i != originMaps.length - 1) {
                            sb.append(',');
                        }
                    }
                    elData.setAttribute(SGNetCDFData.KEY_ORIGIN_MAP, sb.toString());

//					for (int i = 0; i < originMaps.length; i++) {
//						try {
//							int originValue = Integer
//									.parseInt(originMaps[i][1]);
//							int varIndex = nc.getVariableIndex(groupName + "/"
//									+ originMaps[i][0]);
//							if (varIndex >= 0 && varIndex < columns.length) {
//								((SGNetCDFDataColumnInfo) columns[varIndex])
//										.setOrigin(originValue);
//							}
//						} catch (NumberFormatException e) {
//						}
//					}
					for (int i = 0; i < originMaps.length; i++) {
						Integer originValue = SGUtilityText.getInteger(originMaps[i][1]);
						if (originValue == null) {
							return di;
						}
						StringBuffer sbTmp = new StringBuffer();
						sbTmp.append(groupName);
						sbTmp.append('/');
						sbTmp.append(originMaps[i][0]);
						final int varIndex = nc.getVariableIndex(sbTmp.toString());
						if (varIndex >= 0 && varIndex < columns.length) {
							((SGNetCDFDataColumnInfo) columns[varIndex])
									.setOrigin(originValue);
						}
					}
                }

                colInfoSet = new SGDataColumnInfoSet(columns);
            }
*/            

            // create data object
            SGDataCreator.CreatedDataSet cdSet = null;

            cdSet = this.mMain.mDataCreator.createForNetCDFDataSet(nc, dataType, colInfoSet, infoMap, progress);
            if (cdSet == null) {
                return di;
            }
            if (cdSet.getDataLength() == 0) {
                return di;
            }
            SGDataCreator.CreatedData cd = cdSet.getData(0);
            SGData firstData = cd.getData();
            try {
                if (data.setData(firstData) == false) {
                    return di;
                }
            } finally {
            	// disposes data
            	firstData.dispose();
            }

            // get the index in the legend
            String str = elData.getAttribute(SGIFigureElement.KEY_INDEX_IN_LEGEND);
            if (str.length() != 0) {
                Number num = SGUtilityText.getInteger(str);
                if (num == null) {
                    return ic;
                }
                indexArray[ii] = num.intValue();
                dataArray[ii] = data;
            } else {
                indexValid = false;
            }

            // create data objects
            if (figure.createDataObjectFromPropertyFile(elData, data, readDataProperty) == false) {
                return ic;
            }
        }

        // sort the order of data objects in legend
        if (indexValid) {
            SGIFigureElementLegend lElement = figure.getLegendElement();
            lElement.sortLegend(dataArray, indexArray);
        }

        SGIFigureElement[] array = figure.getIFigureElementArray();
        for (int ii = 0; ii < array.length; ii++) {
            array[ii].initPropertiesHistory();
        }

        return SGIConstants.SUCCESSFUL_COMPLETION;
    }

    private int createSingleFigureFromPropertyFile(final Element elFigure,
            final SGDrawingWindow wnd, final int figureIndex, final NetcdfFile ncfile, 
            final boolean readDataProperty, final String versionNumber) {

        final int ic = SGIConstants.PROPERTY_FILE_INCORRECT;

        String str = null;

        str = elFigure.getAttribute( SGFigure.KEY_FIGURE_TYPE );
        if (str.length() == 0) {
            return ic;
        }

        SGFigure figure = null;
        if (SGIFigureTypeConstants.FIGURE_TYPE_SXY.equals(str)
                || SGIFigureTypeConstants.FIGURE_TYPE_VXY.equals(str)
                || SGIFigureTypeConstants.FIGURE_TYPE_XY.equals(str)) {
            figure = new SGXYFigure(wnd);
            figure.setClassType(str);
        } else {
            return ic;
        }

        if (this.setupFigureFromProperty(elFigure, wnd, figure)==false) {
            return ic;
        }

        // create data objects
        final int ret = this.createDataObjectsFromPropertyFile(elFigure,
                figure, figureIndex, ncfile, wnd, readDataProperty, versionNumber);
        if (ret != SGIConstants.SUCCESSFUL_COMPLETION) {
            return ret;
        }

        // add the figure to the window
        wnd.addFigure(figure);

        // initialize properties of the figure and figure elements
        figure.initPropertiesHistory();

        return SGIConstants.SUCCESSFUL_COMPLETION;
    }

    int createFiguresFromPropertyFile(final Element elWnd,
            final SGDrawingWindow wnd, final NetcdfFile ncfile,
            final boolean readDataProperty, final String versionNumber) {

        NodeList nList = elWnd.getElementsByTagName(SGFigure.TAG_NAME_FIGURE);
        final int len = nList.getLength();

        for (int ii = 0; ii < len; ii++) {
            Node node = nList.item(ii);
            if (node instanceof Element) {
                Element el = (Element) node;
                int figureIndex = ii;
                final int ret = this.createSingleFigureFromPropertyFile(el,
                        wnd, figureIndex, ncfile, readDataProperty, versionNumber);
                if (ret != SGIConstants.SUCCESSFUL_COMPLETION) {
                    return ret;
                }
            }
        }

        return SGIConstants.SUCCESSFUL_COMPLETION;
    }

    /**
     * 
     * @param wnd
     * @param elWnd element of the window
     * @return
     */
    private boolean setWindowProperty(final SGDrawingWindow wnd, final Element elWnd, final String versionNumber, final NetcdfFile ncfile) {
        // start progress
        wnd.startProgress();
        wnd.setProgressMessage("Read Property");

        // set the property to window from property file
        final boolean result = wnd.readProperty(elWnd, 0.0f, 1.0f);

        // end progress
        wnd.endProgress();

        int errcode;
        if (!result) {
            errcode = SGIConstants.PROPERTY_FILE_INCORRECT;
        } else {
            // create figure objects in a window
            final boolean readDataProperty = true;
            errcode = this.createFiguresFromPropertyFile(
                    elWnd, wnd, ncfile, readDataProperty, versionNumber);

            if (errcode == SGIConstants.SUCCESSFUL_COMPLETION) {
                // add history
                wnd.initPropertiesHistory();

                // initialize the history of save
                wnd.initSavedHistory();

                // set the saved flag
                wnd.setSaved(true);
            }
        }

        // set the message
        String msg = null;
        switch (errcode) {
        case SGIConstants.SUCCESSFUL_COMPLETION:
            msg = SGIApplicationTextConstants.MSG_SUCCESSFUL_COMPLETION;
            break;
        case SGIConstants.DATA_NUMBER_SHORTAGE:
            msg = SGIApplicationTextConstants.MSG_DATA_NUMBER_SHORTAGE;
            break;
        case SGIConstants.DATA_NUMBER_EXCESS:
            msg = SGIApplicationTextConstants.MSG_DATA_NUMBER_EXCESS;
            break;
        case SGIConstants.FILE_OPEN_FAILURE:
            msg = SGIApplicationTextConstants.MSG_FILE_OPEN_FAILURE;
            break;
        case SGIConstants.PROPERTY_FILE_INCORRECT:
            msg = SGIApplicationTextConstants.MSG_PROPERTY_FILE_INVALID;
            break;
        case SGIConstants.DATA_FILE_INVALID:
            msg = SGIApplicationTextConstants.MSG_DATA_FILE_INVALID;
            break;
        default:
            msg = SGIApplicationTextConstants.MSG_UNKNOWN_ERROR_OCCURED;
        }

        // show the message dialog
        if (msg != SGIApplicationTextConstants.MSG_SUCCESSFUL_COMPLETION) {
            SGUtility.showMessageDialog(null, msg, "Property file",
                    JOptionPane.ERROR_MESSAGE);
            return false;
        }
        return true;
    }

    private boolean loadSub(final SGDrawingWindow wnd, final Document doc, final NetcdfFile ncfile) {
        // get root element
        Element root = doc.getDocumentElement();

        // get version number
        String versionNumber = root.getAttribute(SGIPropertyFileConstants.KEY_VERSION_NUMBER);

        // get window element
        Element elWnd = this.mMain.mPropertyFileManager.getWindowElement(doc);
        NodeList nListFigure = elWnd.getElementsByTagName(SGFigure.TAG_NAME_FIGURE);
        final int figureNum = nListFigure.getLength();
        int cnt = 0;
        final int[] dataNumArray = new int[figureNum];
        for (int ii = 0; ii < figureNum; ii++) {
            Node node = nListFigure.item(ii);
            if ((node instanceof Element) == false) {
                return false;
            }
            Element elFigure = (Element) node;
            NodeList nListData = elFigure.getElementsByTagName(SGIFigureElementGraph.TAG_NAME_DATA);
            dataNumArray[ii] = nListData.getLength();
            cnt += dataNumArray[ii];
        }

    	// finds the variable for the background image
    	Variable imgVar = null;
        for (Variable var : ncfile.getVariables()) {
        	if (DataType.BYTE.equals(var.getDataType())) {
            	if (var.getName().startsWith(ARCHIVE_IMAGE_NAME)) {
            		imgVar = var;
            		break;
            	}
        	}
        }

        boolean imgExists = true;
        byte[] imgBytes = null;
        String ext = null;
        if (imgVar != null) {
            // read byte array from the netCDF file
            ByteArrayOutputStream bos = null;
            try {
            	bos = new ByteArrayOutputStream();
    			Array array = imgVar.read();
    			if (!(array instanceof ArrayByte)) {
    				return false;
    			}
    			ArrayByte bArray = (ArrayByte) array;
    			Index index = bArray.getIndex();
    			for (int ii = 0; ii < bArray.getSize(); ii++) {
    				final byte b = bArray.get(index.set(ii));
    				bos.write(b);
    			}
    			imgBytes = bos.toByteArray();
    			
    		} catch (IOException e) {
    			return false;
    		} finally {
    			if (bos != null) {
    				try {
    					bos.close();
    				} catch (IOException e) {
    				}
    			}
    		}
    		
    		// get image type
    		Attribute attr = imgVar.findAttribute(ATTRIBUTE_KEY_IMAGE_FILE_EXTENSION);
    		if (attr == null) {
    			return false;
    		}
    		String value = attr.getStringValue();
    		if (value == null) {
    			return false;
    		}
    		ext = SGApplicationUtility.findExtension(value);
    		if (ext == null) {
    			return false;
    		}
    		
        } else {
        	imgExists = false;
        }

		// increment the counter
        if (imgExists) {
        	cnt++;
        }
        
        // setup netCDF group to the netCDF file
        this.setupNetCDFGroup(ncfile);

        Group rootGroup = ncfile.getRootGroup();
        List<Group> groups = rootGroup.getGroups();
        if (cnt != groups.size()) {
            return false;
        }

        // clear old objects
        wnd.deleteImage();
        wnd.clearUndoBuffer();
        wnd.removeAllFigures();

        // set the background image
        if (imgExists) {
        	if (!wnd.setImage(imgBytes, ext, false)) {
        		return false;
        	}
        }

        // set properties
        if (this.setWindowProperty(wnd, elWnd, versionNumber, ncfile)==false) {
            return false;
        }

        // update the client rectangle
        wnd.updateClientRect();
        
        return true;
    }
    
    /**
     * Load netCDF dataset.
     * 
     * @param wnd
     * @param file filename of netCDF dataset.
     * @return 0 if succeeds. -1 if file is invalid.
     * @throws IOException 
     */
    public int load(final SGDrawingWindow wnd, final File file)
    throws IOException {
        NetcdfFile ncfile = null;
        try {
            ncfile = NetcdfFile.open(file.getAbsolutePath());
            Attribute attr = ncfile.findGlobalAttribute(ATTRIBUTE_NAME_PROPERTY);
            if (null==attr) {
                return -1;
            }
            String propertyString = attr.getStringValue();
            if (null==propertyString || propertyString.length()==0) {
                return -1;
            }

            // create a Document object
            Document doc = SGUtilityText.getDocumentFromString(propertyString);
            if (doc == null) {
                return -1;
            }

            if (this.loadSub(wnd, doc, ncfile)==false) {
                return -1;
            }
        } finally {
            if (null!=ncfile) {
                ncfile.close();
            }
        }

        return SGIConstants.OK_OPTION;
    }
    
    /**
     * 
     * @param location
     * @return
     */
    public static boolean isNetCDFDatasetFile(final String location) {
        NetcdfFile ncfile = null;
        try {
            ncfile = NetcdfFile.open(location);
            List<Attribute> attrList = ncfile.getGlobalAttributes();
            for (Attribute attr : attrList) {
                if (ATTRIBUTE_NAME_PROPERTY.equals(attr.getName()) &&
                        attr.isString() && attr.getStringValue().length()>0) {
                    return true;
                }
            }
            return false;
        } catch (Exception e) {
            return false;
        } finally {
            if (ncfile!=null) {
                try {
                    ncfile.close();
                } catch (IOException e) {
                }
            }
        }
    }
}
