package jp.riken.brain.ni.samuraigraph.base;

import java.awt.Cursor;
import java.awt.Frame;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.swing.JComponent;


import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * An interface for objects on a figure.
 */
public interface SGIFigureElement extends SGIUndoable, SGIConstants,
        SGIPaintable, SGINode, SGIDisposable, SGIDrawingElementConstants {

    /**
     * @return
     * @uml.property name="component"
     */
    public JComponent getComponent();

    /**
     * @param com
     * @uml.property name="component"
     */
    public void setComponent(JComponent com);

    /**
     * Adds a data object with given name.
     * 
     * @param data
     *            a data object
     * @param name
     *            name of the data object
     * @return true if succeeded
     */
    public boolean addData(final SGData data, final String name);

    /**
     * Adds data objects with given name.
     * 
     * @param data
     *            an array of data objects
     * @param name
     *            an array of names of data objects
     * @param infoMap
     *            information map of data objects
     * @return true if succeeded
     */
    public boolean addData(final SGData[] data, final String[] name, final Map<String, Object> infoMap);

    /**
     * Adds a data object with given name and given ID.
     * 
     * @param data
     *            a data object
     * @param name
     *            name of the data object
     * @param id
     *            the ID of the data object
     * @param infoMap
     *            information map of data object
     * @return true if succeeded
     */
    public boolean addData(final SGData data, final String name, final int id, final Map<String, Object> infoMap);

    /**
     * Adds data objects with given name and given ID.
     * 
     * @param data
     *            an array of data objects
     * @param name
     *            an array of names of data objects
     * @param id
     *            the ID of data objects
     * @param infoMap
     *            information map of data objects
     * @return true if succeeded
     */
    public boolean addData(final SGData[] data, final String[] name, final int[] id, final Map<String, Object> infoMap);

    /**
     * Add a data object with given name and properties.
     * 
     * @param data
     *            data to add
     * @param name
     *            name of data
     * @param p
     *            properties to be set the data
     * @return true if succeeded
     */
    public boolean addData(final SGData data, final String name,
            final SGProperties p);

    /**
     * Remove a data object.
     * 
     * @param data
     *             data to be removed
     * @return
     *             true if succeeded
     */
    public boolean removeData(SGData data);
    
    /**
     * Returns a list of useless data objects in this figure element.
     * 
     * @return a list of useless data objects in this figure element
     */
    public List<SGData> getUselessDataList();

    /**
     * Returns a list of data.
     * 
     * @return a list of data
     */
    public List<SGData> getDataList();

    /**
     * Returns properties of the given data.
     */
    public SGProperties getDataProperties(SGData data);

    /**
     * Sets the magnification.
     * 
     * @param mag
     *           the magnification to set
     * @return true if succeeded
     */
    public boolean setMagnification(float mag);

    /**
     * Returns the magnification.
     * 
     * @return the magnification
     */
    public float getMagnification();

    /**
     * Synchronize this component to the other component. <BR>
     * 
     * @param element -
     *            the SGFigureElement object whose property has chenged.
     * @param msg -
     *            message
     * @return true:succeeded, false:failed
     */
    public boolean synchronize(SGIFigureElement element, String msg);

    /**
     * 
     */
    public boolean synchronizeArgument(SGIFigureElement element, String msg);

    /**
     * 
     */
    public boolean setViewBounds(Rectangle2D rect);

    /**
     * 
     * @param topAndBottom
     * @param leftAndRight
     * @return
     */
    public boolean getMarginAroundGraphRect(final SGTuple2f topAndBottom,
            final SGTuple2f leftAndRight);

    /**
     * 
     * @return
     */
    public float getGraphRectX();

    /**
     * 
     * @return
     */
    public float getGraphRectY();

    /**
     * 
     * @return
     */
    public float getGraphRectWidth();

    /**
     * 
     * @return
     */
    public float getGraphRectHeight();

    /**
     * ]
     * 
     * @uml.property name="graphRect"
     */
    public Rectangle2D getGraphRect();

    /**
     * 
     * @return
     */
    public boolean setGraphRect(float x, float y, float width, float height);

    /**
     * @param rect
     * @return
     * @uml.property name="graphRect"
     */
    public boolean setGraphRect(Rectangle2D rect);

    /**
     * 
     */
    public boolean setGraphRectLocation(float x, float y);

    /**
     * 
     */
    public boolean setGraphRectSize(float width, float height);

    /**
     * 
     */
    public boolean setDialogOwner(Frame frame);

    /**
     * 
     */
    public Cursor getFigureElementCursor();

    /**
     * key event
     */
    public boolean onKeyPressed(KeyEvent e);

    /**
     * 
     * @param e
     */
    public boolean onMouseClicked(MouseEvent e);

    /**
     * 
     * @param e
     */
    public boolean onMousePressed(MouseEvent e);

    /**
     * 
     * @param e
     */
    public boolean onMouseDragged(MouseEvent e);

    /**
     * 
     * @param e
     */
    public boolean onMouseReleased(MouseEvent e);

    /**
     * 
     * @param e
     */
    public boolean onDrawingElement(int x, int y);

    /**
     * 
     */
    public void addActionListener(ActionListener listener);

    /**
     * 
     */
    public void removeActionListener(ActionListener listener);

    /**
     * Creates an array of Element objects.
     * 
     * @param document
     *           an Document objects to append elements
     * @return an array of Element objects
     */
    public Element[] createElement(final Document document);

    /**
     * 
     * @param element
     * @return
     */
    public boolean readProperty(final Element element);

    /**
     * initialize compatible properties for previous property file.
     * 
     * @return
     */
    public boolean initCompatibleProperty();

    /**
     * Create objects related to the data with information given by an Element
     * object.
     * 
     * @param el
     *            an Element object
     * @param data
     *            a data object
     * @return true:succeeded, false:failed
     */
    public boolean createDataObject(Element el, SGData data, final boolean readDataProperty);

    /**
     * Clear all focused objects.
     * @return
     *         true if succeeded
     */
    public boolean clearFocusedObjects();

    /**
     * Clear focused objects.
     * @param ori
     *            an origin of this clearance
     * @return
     *           true if succeeded
     */
    public boolean clearFocusedObjects(SGIFigureElement ori);
    
    /**
     * 
     * 
     */
    public boolean hideSelectedObjects();

    /**
     * Set the visibility of symbols around the focused objects. This method
     * just hide the symbol, and does not clear the list of the focused objects.
     * 
     * @param b
     *            visibility
     */
    public void setSymbolsVisibleAroundFocusedObjects(final boolean b);

    /**
     * Set the visibility of symbols around all objects.
     * 
     * @param b
     *            visibility
     */
    public void setSymbolsAroundAllObjectsVisible(final boolean b);

    /**
     * 
     * @param dx
     * @param dy
     */
    public void translateFocusedObjects(final int dx, final int dy);

    /**
     * 
     * @return
     */
    public boolean setTemporaryPropertiesOfFocusedObjects();

    /**
     * Set changed flag to the focused objects.
     * @return
     *         true if succeeded
     */
    public boolean setChangedFocusedObjects();

    /**
     * Returns whether any focused objects are changed.
     * @return
     *         true if succeeded
     */
    public boolean isFocusedObjectsChanged();
    
    /**
     * Returns a list of the focused objects.
     * @return
     *         a list of the focused objects.
     */
    public ArrayList getFocusedObjectsList();

    /**
     * Duplicate the focused objects.
     * 
     * @return true:succeeded, false:failed
     */
    public boolean duplicateFocusedObjects();

    /**
     * Returns the list of copied objects.
     * 
     * @return list of copied objects
     */
    public List<SGICopiable> getCopiedObjectsList();

    /**
     * Cut focused copiable objects.
     * 
     * @return a list of cut objects
     */
    public List<SGICopiable> cutFocusedObjects();

    /**
     * Paste the objects.
     * 
     * @param list
     *            of the objects to be pasted
     * @return true:succeeded, false:failed
     */
    public boolean paste(List<SGICopiable> list);

    /**
     * Move the focused objects to the head or the tail of the list
     * 
     * @param toTail
     *                flag whether to move focused objects to the tail of the list
     * @return true if succeeded
     */
    public boolean moveFocusedObjects(final boolean toTail);

    /**
     * Move the focused objects to forward or backward for given steps.
     * 
     * @param num
     *           the number of levels to move
     * @return true if succeeded
     */
    public boolean moveFocusedObjects(int num);
    
    /**
     * Return list of the visible child objects.
     * @return list of the visible child objects
     */
    public List<SGIChildObject> getVisibleChildList();
    
    /**
     * 
     * @return
     */
    public ArrayList getPropertyDialogObserverList();

    public static final String KEY_DATA_TYPE = "Type";

    public static final String KEY_DATA_NAME = "Name";

    public static final String KEY_VISIBLE_IN_LEGEND = "VisibleInLegend";

    public static final String KEY_AXIS_POSITION = "AxisPosition";

    public static final String KEY_X_AXIS_POSITION = "XAxisPosition";

    public static final String KEY_Y_AXIS_POSITION = "YAxisPosition";

    public static final String KEY_INDEX_IN_LEGEND = "IndexInLegend";
    
    /**
     * The effective digit for the axis range.
     */
    public static final int AXIS_SCALE_EFFECTIVE_DIGIT = 4;

    public void setMode(final int mode);

    /**
     * Returns the child object of a given ID.
     * 
     * @param id
     *           the ID number of the child
     * @return the child object if it exists
     */
    public SGIChildObject getChild(final int id);

    /**
     * Hides the object of given ID.
     * 
     * @param id
     *           the ID of an object
     * @return true if succeeded
     */
    public boolean hideChildObject(final int id);

    /**
     * Brings to front or sends to back the object of given ID.
     * 
     * @param id
     *           the ID of an object
     * @param toFront
     *           true to bring to front
     * @return true if succeeded
     */
    public boolean moveChildToEnd(final int id, final boolean toFront);

    /**
     * Brings forward or sends backward the object of given ID.
     * 
     * @param id
     *           the ID of an object
     * @param toFront
     *           true to bring forward
     * @return true if succeeded
     */
    public boolean moveChild(final int id, final boolean toFront);

    
    public static final String NOTIFY_CHANGE = "Notify the change";

    public static final String NOTIFY_CHANGE_ON_UNDO = "Notify the change on undo";

    public static final String NOTIFY_CHANGE_ON_COMMIT = "Notify the change on commit";
    
    public static final String SET_PROPERTIES_OF_SELECTED_OBJECTS = "Set properties of selected objects";

    public static final String CLEAR_FOCUSED_OBJECTS = "Clear focused objects";

    public static final String NOTIFY_CHANGE_TO_ROOT = "Notify the change to root";

    public static final String NOTIFY_CHANGE_CURSOR = "Notify the change of cursor";

    public static final String NOTIFY_DATA_SELECTION = "Notify the selection of data";
    
    public static final String NOTIFY_DATA_HIDDEN = "Data objects are hidden";
    
    public static final String NOTIFY_DATA_WILL_BE_HIDDEN = "Data objects will be hidden";
    
    public static final String NOTIFY_NETCDF_DATALABEL_WILL_BE_HIDDEN = "Data label will be hidden";
    
    public static final String NOTIFY_DATA_COLUMN_CHANGE_ON_PREVIEW = "Data columns are changed on preview";

    public static final String NOTIFY_DATA_COLUMN_CHANGE_ON_CANCEL = "Data columns are changed on cancel";

    public static final String NOTIFY_DATA_COLUMN_CHANGE_ON_COMMIT = "Data columns are changed on commit";
    
    public static final String NOTIFY_DATA_CONVERSION = "Notify the conversion of data";
    
    public static final String NOTIFY_UNDO_DATA_CONVERSION = "Notify undoing the conversion of data";
    
    public static final String NOTIFY_UNKNOWN_DATA_ERROR = "Notify unknown data error occurred";

    /**
     * Synchronize the data properties with given data properties.
     * 
     * @param p
     *          the properties
     * @param dp
     *          data properties
     * @return synchronized properties
     */
    public SGProperties synchronizeDataProperties(SGProperties p, SGProperties dp);
    
    /**
     * Called when menu items in the menu bar is selected.
     *
     */
    public void onMenuSelected();
    
    /**
     * Sets the properties.
     * 
     * @param map
     *           a map of properties
     * @return the result of setting properties
     */
    public SGPropertyResults setProperties(SGPropertyMap map);

    /**
     * Sets the properties of child object.
     * 
     * @param id
     *           the ID of child object
     * @param map
     *           a map of properties
     * @return the result of setting properties
     */
    public SGPropertyResults setChildProperties(final int id,
            SGPropertyMap map);

    /**
     * Closes the text field.
     * 
     * @return true if succeed
     */
    public boolean closeTextField();
    
    /**
     * Hides the data objects of given IDs.
     * 
     * @param dataIdArray
     * @return
     */
    public boolean hideData(final int[] dataIdArray);
}
