package jp.riken.brain.ni.samuraigraph.base;

import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * A class with utility methods for the calculation of numbers.
 * 
 */
public class SGUtilityNumber implements SGIConstants {

    /**
     * Returns the order of number. For example: 35.5 -> 1, -121.1 -> 2, 0.025 ->
     * -2.
     * 
     * @param d -
     *            number to get order
     * @return the order of the number
     * @throws IllegalArgumentException
     *             input value is equal to 0.0
     */
    public static int getOrder(final double d) {
        if (d == 0.0) {
            throw new IllegalArgumentException("d == 0.0");
        }
        if (Double.isInfinite(d)) {
            throw new IllegalArgumentException("d is infinite value");
        }
        if (Double.isNaN(d)) {
            throw new IllegalArgumentException("d is not a number");
        }

        final double absd = Math.abs(d);
        final double log = Math.log10(absd);
//        final double log = Math.log(absd) / Math.log(10.0);
        int order = (int) Math.floor(log);

/*        
        // When the input value "d" is completely equal to the power of ten such
        // as 10^n
        // where n is an integer, the value "log" cannot always be equal to n.
        // To avoid this error, we compare the input value to a value of the
        // power of ten.
        double p = getPowersOfTen(order + 1);
        if (p == absd) {
            order++;
        }
*/
        return order;
    }

    /**
     * Returns ten to the power of given ordinal number.
     * 
     * @param order -
     *            ordinal number
     * @return ten to the power of given ordinal
     */
    public static double getPowersOfTen(final int order) {
        BigDecimal bd = getBigDecimalPowersOfTen(order);
        return bd.doubleValue();
    }

    /**
     * Returns ten to the power of given ordinal number in BigDecimal form.
     * 
     * @param order -
     *            ordinal number
     * @return ten to the power of given ordinal
     */
    public static BigDecimal getBigDecimalPowersOfTen(final int order) {
        BigDecimal bd = new BigDecimal(1.0);
        bd = bd.movePointRight(order);
        return bd;
    }

    /**
     * Truncate the number at given digit. For example: value=8715.61, digit=2 ->
     * 8700.0
     * 
     * @param value -
     *            number to be truncated
     * @param digit -
     *            digit
     * @return truncated number
     */
    public static double truncateNumber(final double value, final int digit) {
        return truncateNumber(Double.toString(value), digit);
    }

    /**
     * Truncate the number at given digit.
     * 
     * @param value -
     *            a text string of the number to be truncated
     * @param digit -
     *            digit
     * @return truncated number
     */
    public static double truncateNumber(final String value, final int digit) {
        BigDecimal bd = new BigDecimal(value);
        bd = bd.movePointLeft(digit);
        final int num = (int) bd.doubleValue();
        bd = new BigDecimal(num);
        bd = bd.movePointRight(digit);
        return bd.doubleValue();
    }

    /**
     * Round off the number at given digit. For example: value=8715.61, digit=1 ->
     * 8700.0 and value=8765.61, digit=1 -> 8800.0
     * 
     * @param value -
     *            number to be rounded off
     * @param digit -
     *            digit
     * @return the result of rounded off
     */
    public static double roundOffNumber(final double value, final int digit) {
        return roundOffNumber(Double.toString(value), digit);
    }

    /**
     * Round off the number at given digit. For example: value=8715.61, digit=1 ->
     * 8700.0 and value=8765.61, digit=1 -> 8800.0
     * 
     * @param value -
     *            a text string of the number to be rounded off
     * @param digit -
     *            digit
     * @return the result of rounded off
     */
    public static double roundOffNumber(final String value, final int digit) {
        BigDecimal bd = new BigDecimal(value);
	double sign = bd.doubleValue() >= 0.0 ? 1.0 : -1.0;
        bd = bd.movePointLeft(digit + 1);
//        final double num = Math.rint(bd.doubleValue());
        final double num = sign * (int) (Math.abs(bd.doubleValue()) + 0.50);
        bd = new BigDecimal(num);
        bd = bd.movePointRight(digit + 1);
        return bd.doubleValue();
    }

    /**
     * Round out the number at given digit. For example: value=8715.61, digit=1 ->
     * 8800.0
     * 
     * @param value -
     *            number to be rounded out
     * @param digit -
     *            digit
     * @return the result of rounded off
     */
    public static double roundOutNumber(final double value, final int digit) {
        return roundOutNumber(Double.toString(value), digit);
    }

    /**
     * Round out the number at given digit. For example: value=8715.61, digit=1 ->
     * 8800.0
     * 
     * @param value -
     *            a text string of the number to be rounded off
     * @param digit -
     *            digit
     * @return the result of rounded off
     */
    public static double roundOutNumber(final String value, final int digit) {
        BigDecimal bd = new BigDecimal(value);
        bd = bd.movePointLeft(digit + 1);
        final double num = Math.ceil(bd.doubleValue());
        bd = new BigDecimal(num);
        bd = bd.movePointRight(digit + 1);
        return bd.doubleValue();
    }

    /**
     * 
     * @param x1
     * @param y1
     * @param x2
     * @param y2
     * @return
     */
    public static boolean isOverlapping(final double x1, final double y1,
            final double x2, final double y2) {
        final double value = getOverlap(x1, y1, x2, y2);
        return !(value == 0.0);
    }

    /**
     * 
     * @param x1
     * @param y1
     * @param x2
     * @param y2
     * @return
     */
    public static double getOverlap(final double x1, final double y1,
            final double x2, final double y2) {
        final double min1 = x1 < y1 ? x1 : y1;
        final double max1 = x1 < y1 ? y1 : x1;
        final double min2 = x2 < y2 ? x2 : y2;
        final double max2 = x2 < y2 ? y2 : x2;

        if (!(min1 < max2 || min2 < max1)) {
            return 0.0;
        }

        double value;
        if (min1 < min2) {
            value = max1 - min2;
        } else {
            value = max2 - min1;
        }

        return value;
    }

    /**
     * 
     * @param min
     * @param max
     * @param value
     * @return
     */
    public static boolean contains(final double x1, final double x2,
            final double value) {
        double min;
        double max;
        if (x1 <= x2) {
            min = x1;
            max = x2;
        } else {
            min = x2;
            max = x1;
        }
        return (min <= value && value <= max);
    }

    /**
     * 
     * @param x1
     * @param y1
     * @param x2
     * @param y2
     * @return
     */
    public static boolean contains(final double x1, final double y1,
            final double x2, final double y2) {
        final double min1 = x1 < y1 ? x1 : y1;
        final double max1 = x1 < y1 ? y1 : x1;
        final double min2 = x2 < y2 ? x2 : y2;
        final double max2 = x2 < y2 ? y2 : x2;
        return (min1 <= min2 && max2 <= max1);
    }

    /**
     * 
     * @param flag
     * @param value
     * @param min
     * @param max
     * @param step
     * @param err
     * @return
     */
    public static double stepValue(final boolean flag, final double value,
            final double min, final double max, final double step,
            final double err) {

        // final int indexMax = (int)Math.rint(max/step);

        final int indexNearest = (int) Math.rint(value / step);
        final double valueNearest = indexNearest * step;

        int indexNew;
        if (Math.abs(value - valueNearest) < err) {
            if (flag) {
                indexNew = indexNearest + 1;
            } else {
                indexNew = indexNearest - 1;
            }
        } else {
            if (flag) {
                indexNew = (int) Math.floor(value / step) + 1;
            } else {
                indexNew = (int) Math.floor(value / step);
            }
        }

        final double valueNew = indexNew * step;

        return valueNew;
    }

    /**
     * Returns a rounded number that is the order of given number to given 
     * significant digits with given mode of rounding.
     * 
     * @param value
     *           the number to be rounded
     * @param num
     *           a number to specify the order
     * @param digit
     *           significant digits
     * @param mode
     *           the mode of rounding
     * @return a rounded number
     */
    public static double getNumberInNumberOrder(final double value,
            final double num, final int digit, final int mode) {
        final int order = SGUtilityNumber.getOrder(num) - digit + 1;
        BigDecimal bd = new BigDecimal(value);
        bd = bd.movePointLeft(order);
        bd = bd.setScale(0, mode);
        bd = bd.movePointRight(order);
        return bd.doubleValue();
    }

    /**
     * Returns a rounded number that is the order of given value range to given 
     * significant digits with given mode of rounding.
     * 
     * @param value
     *           the number to be rounded
     * @param min
     *           the minimum value
     * @param max
     *           the maximum value
     * @param digit
     *           significant digits
     * @param mode
     *           the mode of rounding
     * @return a rounded number
     */
    public static double getNumberInRangeOrder(final double value,
            final double min, final double max, final int digit, final int mode) {
        return getNumberInNumberOrder(value, max - min, digit, mode);
    }

    /**
     * Return the minimum value in an array of double numbers.
     * 
     * @param array
     *            an array to be searched
     * @return the minimum value
     */
    public static double min(final double[] array) {
    	boolean valid = false;
        double min = Double.MAX_VALUE;
        for (int ii = 0; ii < array.length; ii++) {
        	if (Double.isNaN(array[ii]) || Double.isInfinite(array[ii])) {
        		continue;
        	}
            if (array[ii] < min) {
                min = array[ii];
                valid = true;
            }
        }
        if (!valid) {
        	return Double.NaN;
        }
        return min;
    }

    /**
     * Return the minimum value in an array of float numbers.
     * 
     * @param array
     *            an array to be searched
     * @return the minimum value
     */
    public static float min(final float[] array) {
    	boolean valid = false;
    	float min = Float.MAX_VALUE;
        for (int ii = 0; ii < array.length; ii++) {
        	if (Float.isNaN(array[ii]) || Float.isInfinite(array[ii])) {
        		continue;
        	}
            if (array[ii] < min) {
                min = array[ii];
                valid = true;
            }
        }
        if (!valid) {
        	return Float.NaN;
        }
        return min;
    }

    /**
     * Return the minimum values in an array of float number tuples.
     * 
     * @param array
     *            an array to be searched
     * @return the minimum value
     */
    public static SGTuple2f min(final SGTuple2f[] array) {
    	float[] xArray = new float[array.length];
    	for (int ii = 0; ii < array.length; ii++) {
    		xArray[ii] = array[ii].x;
    	}
    	float[] yArray = new float[array.length];
    	for (int ii = 0; ii < array.length; ii++) {
    		yArray[ii] = array[ii].y;
    	}
    	final float xMin = min(xArray);
    	final float yMin = min(yArray);
        return new SGTuple2f(xMin, yMin);
    }

    /**
     * Return the maximum value in an array of double numbers.
     * 
     * @param array
     *            an array to be searched
     * @return the maximum value
     */
    public static double max(final double[] array) {
    	boolean valid = false;
        double max = -Double.MAX_VALUE;
        for (int ii = 0; ii < array.length; ii++) {
        	if (Double.isNaN(array[ii]) || Double.isInfinite(array[ii])) {
        		continue;
        	}
            if (array[ii] > max) {
                max = array[ii];
                valid = true;
            }
        }
        if (!valid) {
        	return Double.NaN;
        }
        return max;
    }

    /**
     * Return the maximum value in an array of float numbers.
     * 
     * @param array
     *            an array to be searched
     * @return the maximum value
     */
    public static float max(final float[] array) {
    	boolean valid = false;
    	float max = -Float.MAX_VALUE;
        for (int ii = 0; ii < array.length; ii++) {
        	if (Float.isNaN(array[ii]) || Float.isInfinite(array[ii])) {
        		continue;
        	}
            if (array[ii] > max) {
                max = array[ii];
                valid = true;
            }
        }
        if (!valid) {
        	return Float.NaN;
        }
        return max;
    }
    
    /**
     * Return the maximum values in an array of float number tuples.
     * 
     * @param array
     *            an array to be searched
     * @return the maximum value
     */
    public static SGTuple2f max(final SGTuple2f[] array) {
    	float[] xArray = new float[array.length];
    	for (int ii = 0; ii < array.length; ii++) {
    		xArray[ii] = array[ii].x;
    	}
    	float[] yArray = new float[array.length];
    	for (int ii = 0; ii < array.length; ii++) {
    		yArray[ii] = array[ii].y;
    	}
    	final float xMax = max(xArray);
    	final float yMax = max(yArray);
        return new SGTuple2f(xMax, yMax);
    }

    /**
     * Calculates step values between given range, baseline and step.
     * 
     * @param min
     *           the minimum value
     * @param max
     *           the maximum value
     * @param baseline
     *           the baseline value
     * @param step
     *           the step value
     * @param digit
     *           significant digits
     * @return the set of step values
     */
    public static Set<Number> calcStepValue(final double min, final double max,
    		final double baseline, final double step, final int digit) {
        Set<Number> set = new HashSet<Number>();
        
        if (baseline < min) {
            // when the baseline is smaller than the minimum value
            
            double value = baseline + ((int) ((min - baseline) / step)) * step;
            
            // add one number smaller than the minimum value
            value = SGUtilityNumber.getNumberInRangeOrder(value, min, max, 
            		digit, BigDecimal.ROUND_HALF_UP);
            set.add(Double.valueOf(value));
            
            while (true) {
                double valueNew = value + step;
                valueNew = SGUtilityNumber.getNumberInRangeOrder(valueNew, min, max, 
                		digit, BigDecimal.ROUND_HALF_UP);
                if (valueNew == value) {
                	break;
                }
                value = valueNew;
                if (value >= min && value <= max) {
                    set.add(Double.valueOf(value));
                } else if (value > max) {
                    break;
                }
            }
            
            // add one number larger then the maximum number
            set.add(Double.valueOf(value));
            
        } else if (baseline > max) {
            // when the baseline is larger than the maximum value
            
            double value = baseline - ((int) ((baseline - max) / step)) * step;
            
            // add one number larger then the maximum number
            value = SGUtilityNumber.getNumberInRangeOrder(value, min, max, 
            		digit, BigDecimal.ROUND_HALF_UP);
            set.add(Double.valueOf(value));

            while (true) {
                double valueNew = value - step;
                valueNew = SGUtilityNumber.getNumberInRangeOrder(valueNew, min, max, 
                		digit, BigDecimal.ROUND_HALF_UP);
                if (valueNew == value) {
                	break;
                }
                value = valueNew;
                if (value >= min && value <= max) {
                    set.add(Double.valueOf(value));
                } else if (value < min) {
                    break;
                }
            }

            // add one number smaller than the minimum value
            set.add(Double.valueOf(value));
           
        } else {
            // when the baseline is within the range
            
            set.add(Double.valueOf(baseline));
            
            double value = baseline;
            while (true) {
                double valueNew = value + step;
                valueNew = SGUtilityNumber.getNumberInRangeOrder(valueNew, min, max, 
                		digit, BigDecimal.ROUND_HALF_UP);
                if (valueNew == value) {
                	break;
                }
                value = valueNew;
                if (value >= min && value <= max) {
                    set.add(Double.valueOf(value));
                } else if (value > max) {
                    break;
                }
            }
            
            // add one number larger then the maximum number
            set.add(Double.valueOf(value));
            
            value = baseline;
            while (true) {
                double valueNew = value - step;
                valueNew = SGUtilityNumber.getNumberInRangeOrder(valueNew, min, max, 
                		digit, BigDecimal.ROUND_HALF_UP);
                if (valueNew == value) {
                	break;
                }
                value = valueNew;
                if (value >= min && value <= max) {
                    set.add(Double.valueOf(value));
                } else if (value < min) {
                    break;
                }
            }
            
            // add one number smaller than the minimum value
            set.add(Double.valueOf(value));
        }
        
        return set;
    }

    /**
     * Calculates sorted step values between given range, baseline and step.
     * 
     * @param min
     *           the minimum value
     * @param max
     *           the maximum value
     * @param baseline
     *           the baseline value
     * @param step
     *           the step value
     * @param digit
     *           significant digits
     * @return an array of sorted step values
     */
    public static double[] calcStepValueSorted(final double min, final double max,
    		final double baseline, final double step, final int digit) {
    	Set<Number> set = calcStepValue(min, max, baseline, step, digit);
    	List<Number> list = new ArrayList<Number>(set);
    	double[] array = new double[list.size()];
    	for (int ii = 0; ii < array.length; ii++) {
    		Number num = list.get(ii);
    		array[ii] = num.doubleValue();
    	}
    	Arrays.sort(array);
    	return array;
    }

    /**
     * Transforms a list of double numbers to an array.
     * 
     * @param list
     *           a list of double numbers
     * @return an array of double numbers
     */
    public static double[] toArray(List<Double> list) {
    	double[] array = new double[list.size()];
    	for (int ii = 0; ii < array.length; ii++) {
    		Double d = list.get(ii);
    		array[ii] = d.doubleValue();
    	}
    	return array;
    }

    public static double getNumberInRangeOrder(final double value, final SGAxis axis,
    		final int digit, final int mode) {
        final int type = axis.getScaleType();
        if (type == SGAxis.LINEAR_SCALE) {
            return SGUtilityNumber.getNumberInRangeOrder(value, axis
                    .getMinValue(), axis.getMaxValue(),
                    digit, mode);
        } else if (type == SGAxis.LOG_SCALE) {
            if (value == 0.0) {
            	return 0.0;
            } else {
                final int order = SGUtilityNumber.getOrder(value);
                final double min = SGUtilityNumber.getPowersOfTen(order);
                final double max = min * 10.0;
                return SGUtilityNumber.getNumberInRangeOrder(value, min, max,
                        digit, mode);
            }
        } else {
            throw new IllegalArgumentException("scale type is invalid");
        }
    }
    
    /**
     * Returns an array of strings if they exist.
     * 
     * @param dp
     *          decimal places
     * @param exp
     *          exponent
     * @return string array
     */
    public static String[] getStringArray(final double[] values, final int dp, final int exp) {

        if (dp < 0) {
            throw new IllegalArgumentException("Decimal places must not be negatie: " + dp);
        }
        
        // create NumberFormat object
        NumberFormat nf = SGUtilityNumber.createNumberFormat(dp);

        // create an array of text strings for tick labels
        final int len = values.length;
        String[] sArray = new String[len];
        for (int ii = 0; ii < len; ii++) {
            sArray[ii] = SGUtilityNumber.createExponentialString(values[ii], nf, exp);
        }
        
        return sArray;
    }

    /**
     * Create a pattern and a decimal format.
     * 
     * @param dp
     *          the decimal places
     */
    private static NumberFormat createNumberFormat(final int dp) {
        // create a pattern and a decimal format
        StringBuffer sbPattern = new StringBuffer();
        sbPattern.append('0');
        if (dp > 0) {
            sbPattern.append('.');
            for (int ii = 0; ii < dp; ii++) {
                sbPattern.append('0');
            }
        }
        String pattern = sbPattern.toString();
        NumberFormat nf = new DecimalFormat(pattern);
        return nf;
    }
    
    /**
     * Create a text string with given exponent.
     * 
     * @param value
     *           a value
     * @param nf
     *           the number format
     * @param exp
     *           the exponent
     * @return created text string
     */
    private static String createExponentialString(final double value, 
            final NumberFormat nf, final int exp) {
        if (Double.isNaN(value)) {
            return "";
        }
        BigDecimal bd = new BigDecimal(value);
        bd = bd.movePointLeft(exp);
        StringBuffer sb = new StringBuffer();
        String str = nf.format(bd.doubleValue());
        sb.append(str);
        if (exp != 0) {
            sb.append('E');
            sb.append(exp);
        }
        return sb.toString();
    }

    // only used for debug
    public static void main(String[] args) {
//        double num = 1000.0;
//        int order = getOrder(num);
//        System.out.println(order);
	double value = getNumberInNumberOrder(1.0459, 10.0, 5, BigDecimal.ROUND_HALF_UP);
	System.out.println(value);
    }
}
