package jp.riken.brain.ni.samuraigraph.data;

import java.text.ParseException;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;

/**
 * Data which contains date values.
 */
public abstract class SGDateData extends SGArrayData implements Cloneable {

    /**
     * The index of date strings in an array of string arrays.
     */
    protected Integer mDateIndex = null;
    
    /**
     * The default constructor.
     */
    public SGDateData() {
        super();
    }

    /**
     * Build a data with given columns and indices.
     * 
     * @param dataFile
     *                   the text data file
     * @param dateIndex
     *                   the index for date
     */
    public SGDateData(final SGTextDataFile dataFile, SGDataSourceObserver obs,
            final Integer dateIndex) 
        throws ParseException {
        
        super(dataFile, obs);
        
        // null check
        if (dateIndex == null) {
            throw new IllegalArgumentException("Index for date values must not be null.");
        }
        
        // range check
        if (this.checkColumnIndexRange(dateIndex) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + dateIndex);
        }

        // set the index of date
        this.mDateIndex = dateIndex;
    }
    
    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
        this.mDateIndex = null;
    }

    /**
     * Set values with a given data.
     * 
     * @param data
     *             data to be copied
     * @return
     *             true if succeeded
     */
    public boolean setData(final SGData data) {
        if (!(data instanceof SGDateData)) {
            throw new IllegalArgumentException("!(data instanceof SGDateData)");
        }
        if (super.setData(data) == false) {
            return false;
        }
        SGDateData dData = (SGDateData) data;
        this.mDateIndex = dData.mDateIndex;
        return true;
    }

    public Integer getDateIndex() {
        return mDateIndex;
    }

    public void setDateIndex(Integer index) {
        if (index == null) {
            throw new IllegalArgumentException("index == null");
        }
        if (this.checkColumnIndexRange(index) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        mDateIndex = index;
    }
    
    /**
     * Get properties of this data.
     * @param p
     *          properties to set values
     * @return
     *          true if succeeded
     */
    public boolean getProperties(final SGProperties p) {
        if ((p instanceof DateDataProperties) == false) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        DateDataProperties ep = (DateDataProperties) p;
        ep.mDateIndex = this.mDateIndex;
        return true;
    }
    
    /**
     * Set properties to this data.
     * @param p
     *          properties that have values to set to this data
     * @return
     *          true if succeeded
     */
    public boolean setProperties(final SGProperties p) {
        if ((p instanceof DateDataProperties) == false) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        DateDataProperties ep = (DateDataProperties) p;
//        this.mDateIndex = ep.mDateIndex;
        this.setDateIndex(ep.mDateIndex);
        return true;
    }
    
    /**
     * A class for scalar xy-type data properties.
     *
     */
    public abstract static class DateDataProperties extends ArrayDataProperties {
        
        // indices of data columns
        Integer mDateIndex = null;

        /**
         * The default constructor.
         */
        public DateDataProperties() {
            super();
        }

        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof DateDataProperties) == false) {
                return false;
            }
            DateDataProperties p = (DateDataProperties) obj;
            if (this.mDateIndex != null) {
                if (!this.mDateIndex.equals(p.mDateIndex)) {
                    return false;
                }
            } else {
                if (p.mDateIndex != null) {
                    return false;
                }
            }
            return true;
        }

        /**
         * Returns string representation of this object.
         * @return
         *        string representation of this object
         */
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("[");
            sb.append("DateIndex=");
            sb.append(this.mDateIndex);
            sb.append("]");
            return sb.toString();
        }
        
        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.mDateIndex = null;
        }
    }
    
}
