package jp.riken.brain.ni.samuraigraph.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementGraph;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import ucar.nc2.Dimension;

public class SGDefaultColumnTypeUtility implements SGIDataColumnTypeConstants,
		SGIDataColumnNameConstants {

    /**
	 * A class for the result of default column types.
	 *
	 */
	public static class DefaultColumnTypeResult {
	    private boolean mSucceeded = false;
	    private String[] mColumnTypes = null;
	    public DefaultColumnTypeResult(final String[] columnTypes, final boolean succeeded) {
	        super();
	        if (columnTypes == null) {
	            throw new IllegalArgumentException("columnTypes == null");
	        }
	        this.mColumnTypes = columnTypes;
	        this.mSucceeded = succeeded;
	    }
	
	    public String[] getDefaultColumnTypes() {
	        return (String[]) this.mColumnTypes.clone();
	    }
	    public boolean isSucceeded() {
	        return this.mSucceeded;
	    }
	}

	/**
     * Returns default data column type.
     * 
     * @param dataType
     *            data type
     * @param columnInfoList
     *            a list of column information
     * @param infoMap
     *            information map for data
     * @return default data column types
     */
    public static DefaultColumnTypeResult getDefaultColumnTypes(final String dataType,
            final List<SGDataColumnInfo> columnInfoList, final Map<String, Object> infoMap) {

        if (dataType == null || columnInfoList == null || infoMap == null) {
            throw new IllegalArgumentException("Null input value.");
        }

        SGDataColumnInfo[] columns = new SGDataColumnInfo[columnInfoList.size()];
        for (int ii = 0; ii < columns.length; ii++) {
        	columns[ii] = (SGDataColumnInfo) columnInfoList.get(ii).clone();
        }
        
        // the node name map
		final boolean succeeded = SGDefaultColumnTypeUtility
				.getDefaultColumnTypesSub(dataType, columnInfoList, infoMap, columns);

        // an array of column types
        String[] columnTypes = new String[columns.length];
        for (int ii = 0; ii < columnTypes.length; ii++) {
            columnTypes[ii] = columns[ii].getColumnType();
        }
		
        // returned value
		DefaultColumnTypeResult result = new DefaultColumnTypeResult(
				columnTypes, succeeded);

        return result;
    }
	
	private static boolean getDefaultColumnTypesSub(final String dataType,
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, SGDataColumnInfo[] columns) {

		boolean succeeded = true;
		final int len = columnInfoList.size();
        final NamedNodeMap nodeMap = (NamedNodeMap) infoMap.get(SGIFigureElementGraph.KEY_NODE_MAP);
        final String groupName = (String) infoMap.get(SGIFigureElementGraph.KEY_GROUP_NAME);

		if (SGDataUtility.isArrayData(dataType)) {
			// array data

			List<Integer> numberIndexList = new ArrayList<Integer>();
			List<Integer> textIndexList = new ArrayList<Integer>();
			List<Integer> dateIndexList = new ArrayList<Integer>();
			List<Integer> samplingIndexList = new ArrayList<Integer>();
			for (int ii = 0; ii < len; ii++) {
				SGDataColumnInfo cInfo = columnInfoList.get(ii);
				String valueType = cInfo.getValueType();
				if (VALUE_TYPE_NUMBER.equals(valueType)) {
					numberIndexList.add(Integer.valueOf(ii));
				} else if (VALUE_TYPE_TEXT.equals(valueType)) {
					textIndexList.add(Integer.valueOf(ii));
				} else if (VALUE_TYPE_DATE.equals(valueType)) {
					dateIndexList.add(Integer.valueOf(ii));
				} else if (VALUE_TYPE_SAMPLING_RATE.equals(valueType)) {
					samplingIndexList.add(Integer.valueOf(ii));
				}
			}

			if (SGDataUtility.isSXYTypeData(dataType)
					|| SGDataTypeConstants.SXY_SAMPLING_DATA.equals(dataType)) {
				
				if (nodeMap != null) {
					if (!getDefaultColumnTypesForSXYTypeArrayData(
							columnInfoList, infoMap, nodeMap, groupName, columns)) {
						succeeded = false;
					}
					if (succeeded) {
						if (!SGDataUtility.checkDataColumns(dataType, columns, infoMap)) {
							succeeded = false;
						}
					}
				}
				if (nodeMap == null || !succeeded){
					if (!getDefaultColumnTypesForSXYTypeArrayData(
							columnInfoList, len, infoMap, numberIndexList,
							textIndexList, dateIndexList, samplingIndexList,
							columns)) {
						return false;
					}
				}

			} else if (SGDataTypeConstants.SXYZ_DATA.equals(dataType)) {

				if (nodeMap != null) {
					if (!getDefaultColumnTypesForSXYZTypeArrayData(
							columnInfoList, infoMap, nodeMap, groupName, columns)) {
						succeeded = false;
					}
					if (succeeded) {
						if (!SGDataUtility.checkDataColumns(dataType, columns, infoMap)) {
							succeeded = false;
						}
					}
				}
				if (nodeMap == null || !succeeded){
					if (!getDefaultColumnTypesForSXYZTypeArrayData(
							columnInfoList, len, infoMap, numberIndexList,
							textIndexList, dateIndexList, samplingIndexList,
							columns)) {
						return false;
					}
				}

			} else if (SGDataTypeConstants.VXY_DATA.equals(dataType)) {

				if (nodeMap != null) {
					if (!getDefaultColumnTypesForVXYTypeArrayData(
							columnInfoList, infoMap, nodeMap, groupName, columns)) {
						succeeded = false;
					}
					if (succeeded) {
						if (!SGDataUtility.checkDataColumns(dataType, columns, infoMap)) {
							succeeded = false;
						}
					}
				}
				if (nodeMap == null || !succeeded){
					if (!getDefaultColumnTypesForVXYTypeArrayData(
							columnInfoList, len, infoMap, numberIndexList,
							textIndexList, dateIndexList, samplingIndexList,
							columns)) {
						return false;
					}
				}

			} else {
				throw new Error("Invalid data type: " + dataType);
			}

		} else {
			// netCDF data

			SGNetCDFFile ncFile = (SGNetCDFFile) infoMap.get(KEY_DATA_SOURCE);
			List<SGVariable> varList = ncFile.getVariables();
			final int size = varList.size();

			if (SGDataUtility.isSXYTypeData(dataType)) {
				
				if (nodeMap != null) {
					if (!getDefaultColumnTypesForSXYTypeNetCDFData(
							columnInfoList, infoMap, nodeMap, groupName, columns)) {
						succeeded = false;
					}
					if (succeeded) {
						if (!SGDataUtility.checkDataColumns(dataType, columns, infoMap)) {
							succeeded = false;
						}
					}
				}
				if (nodeMap == null || !succeeded){
					if (!getDefaultColumnTypesForSXYTypeNetCDFData(infoMap,
							varList, size, columns)) {
						return false;
					}
				}

			} else if (SGDataTypeConstants.VXY_NETCDF_DATA.equals(dataType)) {
				
				if (nodeMap != null) {
					if (!getDefaultColumnTypesForVXYTypeNetCDFData(
							columnInfoList, infoMap, nodeMap, groupName, columns)) {
						succeeded = false;
					}
					if (succeeded) {
						if (!SGDataUtility.checkDataColumns(dataType, columns, infoMap)) {
							succeeded = false;
						}
					}
				}
				if (nodeMap == null || !succeeded){
					if (!getDefaultColumnTypesForVXYTypeNetCDFData(infoMap,
							varList, size, columns)) {
						return false;
					}
				}
				
			} else if (SGDataTypeConstants.SXYZ_NETCDF_DATA.equals(dataType)) {
				
				if (nodeMap != null) {
					if (!getDefaultColumnTypesForSXYZTypeNetCDFData(
							columnInfoList, infoMap, nodeMap, groupName, columns)) {
						succeeded = false;
					}
					if (succeeded) {
						if (!SGDataUtility.checkDataColumns(dataType, columns, infoMap)) {
							succeeded = false;
						}
					}
				}
				if (nodeMap == null || !succeeded){
					if (!getDefaultColumnTypesForSXYZTypeNetCDFData(infoMap,
							varList, size, columns)) {
						return false;
					}
				}
				
			} else {
				throw new Error("Invalid data type: " + dataType);
			}
		}

		return true;
	}

	private static boolean getDefaultColumnTypesForSXYTypeArrayData(
			final List<SGDataColumnInfo> columnInfoList, final int len,
			final Map<String, Object> infoMap, List<Integer> numberIndexList,
			List<Integer> textIndexList, List<Integer> dateIndexList,
			List<Integer> samplingIndexList, SGDataColumnInfo[] columns) {
		
		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}

		if (samplingIndexList.size() > 1) {
			return false;
		}

		Double samplingRate = (Double) infoMap.get(KEY_SAMPLING_RATE);
		Boolean multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
		if (multiple == null) {
			return false;
		}

		boolean dateColumnFound = false;
		// when a date column is found first,
		// the date column is used for x values
		for (int ii = 0; ii < len; ii++) {
			SGDataColumnInfo cInfo = (SGDataColumnInfo) columnInfoList.get(ii);
			String valueType = cInfo.getValueType();
			if (VALUE_TYPE_DATE.equals(valueType)) {
				dateColumnFound = true;
				break;
			}
		}

		if (multiple.booleanValue()) {
			if (samplingRate != null) {
				if (numberIndexList.size() >= 1) {
					for (int ii = 0; ii < numberIndexList.size(); ii++) {
						Integer num = (Integer) numberIndexList.get(ii);
						columnTypes[num.intValue()] = Y_VALUE;
					}
				} else {
					return false;
				}
				if (samplingIndexList.size() == 1) {
					Integer num = samplingIndexList.get(0);
					columnTypes[num.intValue()] = X_VALUE;
				}
			} else if (dateColumnFound) {
				if (numberIndexList.size() >= 1) {
					for (int ii = 0; ii < numberIndexList.size(); ii++) {
						Integer num = (Integer) numberIndexList.get(ii);
						columnTypes[num.intValue()] = Y_VALUE;
					}
					Integer date0 = dateIndexList.get(0);
					columnTypes[date0.intValue()] = X_VALUE;
				} else {
					return false;
				}
			} else {
				if (numberIndexList.size() >= 2) {
					Integer num1 = (Integer) numberIndexList.get(0);
					columnTypes[num1.intValue()] = X_VALUE;
					for (int ii = 1; ii < numberIndexList.size(); ii++) {
						Integer num = (Integer) numberIndexList.get(ii);
						columnTypes[num.intValue()] = Y_VALUE;
					}
				} else {
					return false;
				}
			}

		} else {

			final int minNumberSize;
			if (samplingRate != null) {
				minNumberSize = 1;
			} else {
				if (dateColumnFound) {
					minNumberSize = 1;
				} else {
					minNumberSize = 2;
				}
			}
			final int yNumberArrayIndex = minNumberSize - 1;

			// get the column index of y index
			int yNumberColumnIndex = -1;
			int cnt = 0;
			for (int ii = 0; ii < len; ii++) {
				SGDataColumnInfo cInfo = (SGDataColumnInfo) columnInfoList
						.get(ii);
				String valueType = cInfo.getValueType();
				if (VALUE_TYPE_NUMBER.equals(valueType)) {
					if (cnt == yNumberArrayIndex) {
						yNumberColumnIndex = ii;
						break;
					}
					cnt++;
				}
			}
			if (yNumberColumnIndex == -1) {
				return false;
			}

			if (numberIndexList.size() >= minNumberSize) {
				boolean dateAssigned = false;
				if (samplingRate != null) {
					if (samplingIndexList.size() == 1) {
						Integer num = samplingIndexList.get(0);
						columnTypes[num.intValue()] = X_VALUE;
					}
				} else {
					// x values
					if (dateIndexList.size() != 0 && dateColumnFound) {
						Integer num1 = (Integer) dateIndexList.get(0);
						columnTypes[num1.intValue()] = X_VALUE;
						dateAssigned = true;
					} else {
						Integer num1 = (Integer) numberIndexList.get(0);
						columnTypes[num1.intValue()] = X_VALUE;
					}
				}

				// y values
				Integer num2 = (Integer) numberIndexList.get(yNumberArrayIndex);
				columnTypes[num2.intValue()] = Y_VALUE;

				boolean[] isRepeatedTitle = SGDataUtility
						.isEmptyOrRepeatedColumnTitle(columnInfoList
								.toArray(new SGDataColumnInfo[len]));

				// error bars
				if (numberIndexList.size() >= minNumberSize + 2) {
					Integer num3 = (Integer) numberIndexList
							.get(yNumberArrayIndex + 1);
					Integer num4 = (Integer) numberIndexList
							.get(yNumberArrayIndex + 2);
					columnTypes[num3.intValue()] = SGDataUtility
							.appendColumnNoOrTitle(LOWER_ERROR_VALUE,
									yNumberColumnIndex,
									isRepeatedTitle[yNumberColumnIndex],
									columnInfoList.get(yNumberColumnIndex)
											.getTitle());
					columnTypes[num4.intValue()] = SGDataUtility
							.appendColumnNoOrTitle(UPPER_ERROR_VALUE,
									yNumberColumnIndex,
									isRepeatedTitle[yNumberColumnIndex],
									columnInfoList.get(yNumberColumnIndex)
											.getTitle());
				}

				// tick labels
				if (textIndexList.size() > 0) {
					Integer tick = (Integer) textIndexList.get(0);
					columnTypes[tick.intValue()] = SGDataUtility
							.appendColumnNoOrTitle(TICK_LABEL,
									yNumberColumnIndex,
									isRepeatedTitle[yNumberColumnIndex],
									columnInfoList.get(yNumberColumnIndex)
											.getTitle());
				} else if (dateIndexList.size() > 0 && !dateAssigned) {
					Integer date = (Integer) dateIndexList.get(0);
					columnTypes[date.intValue()] = SGDataUtility
							.appendColumnNoOrTitle(TICK_LABEL,
									yNumberColumnIndex,
									isRepeatedTitle[yNumberColumnIndex],
									columnInfoList.get(yNumberColumnIndex)
											.getTitle());
				}

			} else {
				return false;
			}
		}
		
		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}

		return true;
	}

	private static boolean getDefaultColumnTypesForSXYTypeArrayData(
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, final NamedNodeMap nodeMap, 
			final String groupName, final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}
		
		// x and y values
		Integer[] xIndices = getIndices(nodeMap, KEY_X_VALUE_COLUMN_INDICES);
		if (xIndices == null) {
			return false;
		}
		if (xIndices.length == 0) {
			return false;
		}
		if (!checkIndices(xIndices, columnTypes)) {
			return false;
		}
		Integer[] yIndices = getIndices(nodeMap, KEY_Y_VALUE_COLUMN_INDICES);
		if (yIndices == null) {
			return false;
		}
		if (yIndices.length == 0) {
			return false;
		}
		if (!checkIndices(yIndices, columnTypes)) {
			return false;
		}
		if (xIndices.length > 1 && yIndices.length > 1) {
			return false;
		}
		for (int ii = 0; ii < xIndices.length; ii++) {
			columnTypes[xIndices[ii].intValue()] = X_VALUE;
		}
		for (int ii = 0; ii < yIndices.length; ii++) {
			columnTypes[yIndices[ii].intValue()] = Y_VALUE;
		}

		// error values
		getDefaultColumnTypesForSXYTypeArrayDataErrorBar(columnInfoList,
				infoMap, nodeMap, groupName, columnTypes);

		// tick labels
		getDefaultColumnTypesForSXYTypeArrayDataTickLabel(columnInfoList,
				infoMap, nodeMap, groupName, columnTypes);

		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}
		
		return true;
	}
	
	private static boolean getDefaultColumnTypesForSXYTypeArrayDataErrorBar(
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, final NamedNodeMap nodeMap,
			final String groupName, final String[] columnTypes) {
		
		Integer[] leIndices = getIndices(nodeMap,
				KEY_LOWER_ERROR_BAR_COLUMN_INDICES);
		Integer[] ueIndices = getIndices(nodeMap,
				KEY_UPPER_ERROR_BAR_COLUMN_INDICES);
		Integer[] ehIndices = getIndices(nodeMap,
				KEY_ERROR_BAR_HOLDER_COLUMN_INDICES);
		if (leIndices != null && ueIndices != null && ehIndices != null) {
			if (!checkIndices(leIndices, columnTypes)) {
				return false;
			}
			if (!checkIndices(ueIndices, columnTypes)) {
				return false;
			}
			if (!checkIndices(ehIndices, columnTypes)) {
				return false;
			}
			if (ehIndices.length != leIndices.length
					|| ehIndices.length != ueIndices.length) {
				return false;
			}
			for (int ii = 0; ii < ehIndices.length; ii++) {
				final int ehIndex = ehIndices[ii].intValue();
				SGDataColumnInfo ehColInfo = columnInfoList.get(ehIndex);
				final boolean b = isEmptyTitle(columnInfoList, ehIndex)
						|| isRepeatedTitle(columnInfoList, ehIndex);
				final String title = ehColInfo.getTitle();
				if (leIndices[ii] == ueIndices[ii]) {
					columnTypes[leIndices[ii]] = SGDataUtility
							.appendColumnNoOrTitle(LOWER_UPPER_ERROR_VALUE,
									ehIndex, b, title);
				} else {
					columnTypes[leIndices[ii]] = SGDataUtility
							.appendColumnNoOrTitle(LOWER_ERROR_VALUE, ehIndex,
									b, title);
					columnTypes[ueIndices[ii]] = SGDataUtility
							.appendColumnNoOrTitle(UPPER_ERROR_VALUE, ehIndex,
									b, title);
				}
			}
		}
		return true;
	}
	
	private static boolean getDefaultColumnTypesForSXYTypeArrayDataTickLabel(
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, final NamedNodeMap nodeMap,
			final String groupName, final String[] columnTypes) {
		Integer[] tlIndices = getIndices(nodeMap, KEY_TICK_LABEL_COLUMN_INDICES);
		Integer[] thIndices = getIndices(nodeMap,
				KEY_TICK_LABEL_HOLDER_COLUMN_INDICES);
		if (tlIndices != null && thIndices != null) {
			if (!checkIndices(tlIndices, columnTypes)) {
				return false;
			}
			if (!checkIndices(thIndices, columnTypes)) {
				return false;
			}
			if (tlIndices.length != thIndices.length) {
				return false;
			}
			for (int ii = 0; ii < thIndices.length; ii++) {
				final int thIndex = thIndices[ii].intValue();
				SGDataColumnInfo thColInfo = columnInfoList.get(thIndex);
				final boolean b = isEmptyTitle(columnInfoList, thIndex)
						|| isRepeatedTitle(columnInfoList, thIndex);
				final String title = thColInfo.getTitle();
				columnTypes[tlIndices[ii]] = SGDataUtility
						.appendColumnNoOrTitle(TICK_LABEL, thIndex, b, title);
			}
		}
		return true;
	}

	private static boolean getDefaultColumnTypesForVXYTypeArrayData(
			final List<SGDataColumnInfo> columnInfoList, final int len,
			final Map<String, Object> infoMap, List<Integer> numberIndexList,
			List<Integer> textIndexList, List<Integer> dateIndexList,
			List<Integer> samplingIndexList, 
			final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}
		
		if (numberIndexList.size() >= 4) {
			Integer num1 = (Integer) numberIndexList.get(0);
			Integer num2 = (Integer) numberIndexList.get(1);
			Integer num3 = (Integer) numberIndexList.get(2);
			Integer num4 = (Integer) numberIndexList.get(3);
			columnTypes[num1.intValue()] = X_COORDINATE;
			columnTypes[num2.intValue()] = Y_COORDINATE;
			if (SGDataUtility.isPolar(infoMap)) {
				columnTypes[num3.intValue()] = MAGNITUDE;
				columnTypes[num4.intValue()] = ANGLE;
			} else {
				columnTypes[num3.intValue()] = X_COMPONENT;
				columnTypes[num4.intValue()] = Y_COMPONENT;
			}
		} else {
			return false;
		}
		
		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}
		
		return true;
	}

	private static boolean getDefaultColumnTypesForVXYTypeArrayData(
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, final NamedNodeMap nodeMap,
			final String groupName, final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}
		
		Integer[] xIndices = getIndices(nodeMap, KEY_X_COORDINATE_COLUMN_INDEX);
		if (xIndices == null) {
			return false;
		}
		if (xIndices.length != 1) {
			return false;
		}
		if (!checkIndices(xIndices, columnTypes)) {
			return false;
		}
		Integer[] yIndices = getIndices(nodeMap, KEY_Y_COORDINATE_COLUMN_INDEX);
		if (yIndices == null) {
			return false;
		}
		if (yIndices.length != 1) {
			return false;
		}
		if (!checkIndices(yIndices, columnTypes)) {
			return false;
		}
		Integer[] c1Indices = getIndices(nodeMap, KEY_FIRST_COMPONENT_COLUMN_INDEX);
		if (c1Indices == null) {
			return false;
		}
		if (c1Indices.length != 1) {
			return false;
		}
		if (!checkIndices(c1Indices, columnTypes)) {
			return false;
		}
		Integer[] c2Indices = getIndices(nodeMap, KEY_SECOND_COMPONENT_COLUMN_INDEX);
		if (c2Indices == null) {
			return false;
		}
		if (c2Indices.length != 1) {
			return false;
		}
		if (!checkIndices(c2Indices, columnTypes)) {
			return false;
		}
		for (int ii = 0; ii < xIndices.length; ii++) {
			columnTypes[xIndices[ii].intValue()] = X_COORDINATE;
		}
		for (int ii = 0; ii < yIndices.length; ii++) {
			columnTypes[yIndices[ii].intValue()] = Y_COORDINATE;
		}
		String cType1, cType2;
		if (SGDataUtility.isPolar(infoMap)) {
			cType1 = MAGNITUDE;
			cType2 = ANGLE;
		} else {
			cType1 = X_COMPONENT;
			cType2 = Y_COMPONENT;
		}
		for (int ii = 0; ii < c1Indices.length; ii++) {
			columnTypes[c1Indices[ii].intValue()] = cType1;
		}
		for (int ii = 0; ii < c2Indices.length; ii++) {
			columnTypes[c2Indices[ii].intValue()] = cType2;
		}
		
		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}
		
		return true;
	}

	private static boolean getDefaultColumnTypesForSXYZTypeArrayData(
			final List<SGDataColumnInfo> columnInfoList, final int len,
			final Map<String, Object> infoMap, List<Integer> numberIndexList,
			List<Integer> textIndexList, List<Integer> dateIndexList,
			List<Integer> samplingIndexList, 
			final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}
		
		if (numberIndexList.size() >= 3) {
			// x, y and z values
			Integer num1 = (Integer) numberIndexList.get(0);
			Integer num2 = (Integer) numberIndexList.get(1);
			Integer num3 = (Integer) numberIndexList.get(2);
			columnTypes[num1.intValue()] = X_VALUE;
			columnTypes[num2.intValue()] = Y_VALUE;
			columnTypes[num3.intValue()] = Z_VALUE;
		} else {
			return false;
		}
		
		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}
		
		return true;
	}

	private static boolean getDefaultColumnTypesForSXYZTypeArrayData(
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, final NamedNodeMap nodeMap,
			final String groupName, final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}

		// x, y and z values
		Integer[] xIndices = getIndices(nodeMap, KEY_X_VALUE_COLUMN_INDEX);
		if (xIndices == null) {
			return false;
		}
		if (xIndices.length != 1) {
			return false;
		}
		if (!checkIndices(xIndices, columnTypes)) {
			return false;
		}
		Integer[] yIndices = getIndices(nodeMap, KEY_Y_VALUE_COLUMN_INDEX);
		if (yIndices == null) {
			return false;
		}
		if (yIndices.length != 1) {
			return false;
		}
		if (!checkIndices(yIndices, columnTypes)) {
			return false;
		}
		Integer[] zIndices = getIndices(nodeMap, KEY_Z_VALUE_COLUMN_INDEX);
		if (zIndices == null) {
			return false;
		}
		if (zIndices.length != 1) {
			return false;
		}
		if (!checkIndices(zIndices, columnTypes)) {
			return false;
		}
		for (int ii = 0; ii < xIndices.length; ii++) {
			columnTypes[xIndices[ii].intValue()] = X_VALUE;
		}
		for (int ii = 0; ii < yIndices.length; ii++) {
			columnTypes[yIndices[ii].intValue()] = Y_VALUE;
		}
		for (int ii = 0; ii < zIndices.length; ii++) {
			columnTypes[zIndices[ii].intValue()] = Z_VALUE;
		}

		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}

		return true;
	}

	private static boolean isEmptyTitle(
			final List<SGDataColumnInfo> columnInfoList, final int colIndex) {
		SGDataColumnInfo colInfo = columnInfoList.get(colIndex);
		String title = colInfo.getTitle();
		if (title == null || "".equals(title)) {
			return true;
		}
		return false;
	}

	private static boolean isRepeatedTitle(
			final List<SGDataColumnInfo> columnInfoList, final int colIndex) {
		String[] titles = new String[columnInfoList.size()];
		for (int ii = 0; ii < titles.length; ii++) {
			SGDataColumnInfo colInfo = columnInfoList.get(ii);
			titles[ii] = colInfo.getTitle();
		}
		if (null == titles[colIndex]) {
			return true;
		}
		for (int i = 0; i < titles.length; i++) {
			if (colIndex != i && titles[colIndex].equals(titles[i])) {
				return true;
			}
		}
		return false;
	}

	private static Integer[] getIndices(final NamedNodeMap nodeMap,
			final String key) {
		Node nodeIndices = nodeMap.getNamedItem(key);
		if (nodeIndices == null) {
			return null;
		}
		String strIndices = nodeIndices.getNodeValue();
		if (strIndices == null) {
			return null;
		}
		Integer[] indices = SGUtilityText.parseIndices(strIndices);
		return indices;
	}
	
	private static boolean checkIndices(final Integer[] indices, final String[] columnTypes) {
		for (int ii = 0; ii < indices.length; ii++) {
			final int index = indices[ii].intValue();
			if (index < 0 || index >= columnTypes.length) {
				return false;
			}
		}
		return true;
	}

	private static boolean getDefaultColumnTypesForSXYTypeNetCDFData(
			final Map<String, Object> infoMap, final List<SGVariable> varList,
			final int size, 
			final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}

		// a coordinate variable found first is assigned to x variable
		int xIndex = -1;
		SGVariable xVar = null;

		// get unlimited variable
		for (int ii = 0; ii < size; ii++) {
			SGVariable var = varList.get(ii);
			if (var.isUnlimited() && var.isCoordinateVariable()) {
				xIndex = ii;
				xVar = var;
				break;
			}
		}

		// if unlimited variable is not found, search the coordinate variable
		if (xVar == null) {
			for (int ii = 0; ii < size; ii++) {
				SGVariable var = varList.get(ii);
				if (var.isCoordinateVariable()) {
					xIndex = ii;
					xVar = var;
					break;
				}
			}
		}
		if (xIndex == -1) {
			return false;
		}

		Boolean multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
		if (multiple == null) {
			return false;
		}

		Boolean variable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
		if (variable == null) {
			return false;
		}

		// y variable must not be a coordinate variable and it has a dimension
		// that is the same as that of the x variable
		Integer[] yIndices = null;
		Dimension xDim = xVar.getDimension(0);
		List<Integer> yList = new ArrayList<Integer>();
		for (int ii = 0; ii < size; ii++) {
			SGVariable var = varList.get(ii);
			if (!var.isCoordinateVariable()) {
				List<Dimension> dimList = var.getDimensions();
				if (dimList.contains(xDim)) {
					yList.add(ii);
				}
			}
		}
		if (yList.size() == 0) {
			return false;
		} else {
			yIndices = new Integer[yList.size()];
			for (int ii = 0; ii < yList.size(); ii++) {
				yIndices[ii] = yList.get(ii);
			}
		}

		if (variable.booleanValue()) {
			// for multiple variables
			columnTypes[xIndex] = X_VALUE;

			if (multiple.booleanValue()) {
				// multiple y-indices
				for (int yIndex : yIndices) {
					columnTypes[yIndex] = Y_VALUE;
				}
			} else {
				// finds the y-index
				for (int ii = 0; ii < yIndices.length; ii++) {
					final int yIndex = yIndices[ii];
					SGVariable var = varList.get(yIndex);
					if (!(var instanceof SGTextVariable)) {
						columnTypes[yIndex] = Y_VALUE;
						break;
					}
				}
			}

		} else {
			// for multiple dimension indices
			final int yIndex = yIndices[0];
			columnTypes[xIndex] = X_VALUE;
			columnTypes[yIndex] = Y_VALUE;

			int dIndex = -1;
			final SGVariable yVar = varList.get(yIndex);
			List<Dimension> yDimList = yVar.getDimensions();
			for (int ii = 0; ii < varList.size(); ii++) {
				SGVariable var = varList.get(ii);
				if (var.isCoordinateVariable()) {
					Dimension dim = var.getDimension(0);
					if (!xDim.equals(dim) && yDimList.contains(dim)) {
						dIndex = ii;
						break;
					}
				}
			}
			if (dIndex != -1) {
				columnTypes[dIndex] = PICKUP;
			} else {
				return false;
			}
		}

		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}

		return true;
	}
	
	private static boolean getDefaultColumnTypesForSXYTypeNetCDFData(
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, final NamedNodeMap nodeMap,
			final String groupName, final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}

		int index;

		// serial numbers
		String[] serialNumberNames = getNames(nodeMap,
				KEY_SERIAL_NUMBER_VARIABLE_NAME, groupName);
		if (serialNumberNames != null) {
			if (serialNumberNames.length != 1) {
				return false;
			}
			if ((index = setNetCDFColumnType(columnInfoList, columnTypes,
					serialNumberNames[0], SERIAL_NUMBER)) == -1) {
				return false;
			}
		}

		// x and y values
		String[] xNames = getNames(nodeMap, new String[] { KEY_X_VALUE_NAME,
				KEY_X_VALUE_COLUMN_INDEX }, groupName);
		if (xNames == null) {
			xNames = getNames(nodeMap, new String[] { KEY_X_VALUE_NAMES,
					KEY_X_VALUE_COLUMN_INDICES }, groupName);
			if (xNames == null) {
				return false;
			}
		}
		if (xNames.length == 0) {
			return false;
		}
		String[] yNames = getNames(nodeMap, new String[] { KEY_Y_VALUE_NAME,
				KEY_Y_VALUE_COLUMN_INDEX }, groupName);
		if (yNames == null) {
			yNames = getNames(nodeMap, new String[] { KEY_Y_VALUE_NAMES,
					KEY_Y_VALUE_COLUMN_INDICES }, groupName);
			if (yNames == null) {
				return false;
			}
		}
		if (yNames.length == 0) {
			return false;
		}
		if (xNames.length > 1 && yNames.length > 1) {
			return false;
		}
		for (int ii = 0; ii < xNames.length; ii++) {
			index = setNetCDFColumnType(columnInfoList, columnTypes,
					xNames[ii], X_VALUE);
			if (index == -1) {
				return false;
			}
		}
		for (int ii = 0; ii < yNames.length; ii++) {
			index = setNetCDFColumnType(columnInfoList, columnTypes,
					yNames[ii], Y_VALUE);
			if (index == -1) {
				return false;
			}
		}

		// error values
		String[] leNames = getNames(nodeMap,
				new String[] { KEY_LOWER_ERROR_VALUE_NAMES,
						KEY_LOWER_ERROR_BAR_COLUMN_INDICES }, groupName);
		String[] ueNames = getNames(nodeMap,
				new String[] { KEY_UPPER_ERROR_VALUE_NAMES,
						KEY_UPPER_ERROR_BAR_COLUMN_INDICES }, groupName);
		String[] ehNames = getNames(nodeMap,
				new String[] { KEY_ERROR_BAR_HOLDER_NAMES,
						KEY_ERROR_BAR_HOLDER_COLUMN_INDICES }, groupName);
		if (leNames != null && ueNames != null && ehNames != null) {
			if (ehNames.length != leNames.length
					|| ehNames.length != ueNames.length) {
				return false;
			}
			for (int ii = 0; ii < ehNames.length; ii++) {
				index = findNetCDFColumnInfo(columnInfoList, ehNames[ii]);
				if (index == -1) {
					return false;
				}
			}
			for (int ii = 0; ii < ehNames.length; ii++) {
				index = setNetCDFColumnType(columnInfoList, columnTypes,
						leNames[ii], ehNames[ii], LOWER_ERROR_VALUE);
				if (index == -1) {
					return false;
				}
			}
			for (int ii = 0; ii < ehNames.length; ii++) {
				index = setNetCDFColumnType(columnInfoList, columnTypes,
						ueNames[ii], ehNames[ii], UPPER_ERROR_VALUE);
				if (index == -1) {
					return false;
				}
			}
		}

		// tick labels
		String[] tlNames = getNames(nodeMap, new String[] {
				KEY_TICK_LABEL_NAMES, KEY_TICK_LABEL_COLUMN_INDICES },
				groupName);
		String[] thNames = getNames(nodeMap, new String[] {
				KEY_TICK_LABEL_HOLDER_NAMES,
				KEY_TICK_LABEL_HOLDER_COLUMN_INDICES }, groupName);
		if (tlNames != null && thNames != null) {
			if (tlNames.length != thNames.length) {
				return false;
			}
			for (int ii = 0; ii < thNames.length; ii++) {
				index = findNetCDFColumnInfo(columnInfoList, thNames[ii]);
				if (index == -1) {
					return false;
				}
			}
			for (int ii = 0; ii < thNames.length; ii++) {
				index = setNetCDFColumnType(columnInfoList, columnTypes,
						tlNames[ii], thNames[ii], TICK_LABEL);
				if (index == -1) {
					return false;
				}
			}
		}

		// time
		String[] timeNames = getNames(nodeMap, KEY_TIME_VARIABLE_NAME,
				groupName);
		if (timeNames != null) {
			if (timeNames.length != 1) {
				return false;
			}
			index = setNetCDFColumnType(columnInfoList, columnTypes,
					timeNames[0], TIME);
			if (index == -1) {
				return false;
			}
		}

		Boolean variable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
		if (variable != null) {
			if (!variable.booleanValue()) {
				// pick up
				String[] pickupNames = getNames(nodeMap,
						KEY_PICKUP_DIMENSION_NAME, groupName);
				if (pickupNames != null) {
					if (pickupNames.length != 1) {
						return false;
					}
					index = setNetCDFColumnType(columnInfoList, columnTypes,
							pickupNames[0], PICKUP);
					if (index == -1) {
						return false;
					}
				}
			}
		}

		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}

		return true;
	}

	private static boolean getDefaultColumnTypesForVXYTypeNetCDFData(
			final Map<String, Object> infoMap, final List<SGVariable> varList,
			final int size, 
			final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}

		int xIndex = -1;
		int yIndex = -1;
		int tIndex = -1;

		// a coordinate variable found first is assigned to the x variable
		// and found next is assigned to the y variable
		SGVariable xVar = null;
		SGVariable yVar = null;

		int startIndex = 0;
		int comIndex1 = -1;
		int comIndex2 = -1;
		List<Dimension> dimListCom1 = null;
		List<Dimension> dimListCom2 = null;

		for (int ii = 0; ii < varList.size() - 3; ii++) {
			int[] xyIndex = getNetCDFCoordinateVariableXYIndex(varList,
					startIndex);
			if (xyIndex == null) {
				startIndex += 1;
				continue;
			} else {
				xIndex = xyIndex[0];
				yIndex = xyIndex[1];
				xVar = varList.get(xIndex);
				yVar = varList.get(yIndex);
			}

			Dimension xDim = xVar.getDimension(0);
			Dimension yDim = yVar.getDimension(0);

			// variables for two components must have dimensions
			// those of x and y variables
			for (int jj = 0; jj < size; jj++) {
				SGVariable var = varList.get(jj);
				if (var.isCoordinateVariable() == false) {
					List<Dimension> dimList = var.getDimensions();
					if (dimList.contains(xDim) && dimList.contains(yDim)) {
						comIndex1 = jj;
						dimListCom1 = dimList;
						break;
					}
				}
			}
			if (comIndex1 == -1 || comIndex1 == size - 1) {
				startIndex += 1;
				continue;
			}

			for (int jj = comIndex1 + 1; jj < size; jj++) {
				SGVariable var = varList.get(jj);
				if (var.isCoordinateVariable() == false) {
					List<Dimension> dimList = var.getDimensions();
					if (dimList.contains(xDim) && dimList.contains(yDim)) {
						comIndex2 = jj;
						dimListCom2 = dimList;
						break;
					}
				}
			}
			if (comIndex2 != -1) {
				break;
			} else {
				startIndex += 1;
			}
		}

		if (comIndex1 == -1 || comIndex2 == -1) {
			return false;
		} else {
			dimListCom1 = varList.get(comIndex1).getDimensions();
			dimListCom2 = varList.get(comIndex2).getDimensions();
		}

		// unlimited coordinate variable is assigned to time variable
		// if it exists
		for (int ii = 0; ii < size; ii++) {
			if (ii == xIndex || ii == yIndex || ii == comIndex1
					|| ii == comIndex2) {
				continue;
			}
			SGVariable var = varList.get(ii);
			if (var.isUnlimited() && var.isCoordinateVariable()) {
				Dimension dim = var.getDimension(0);
				if (dimListCom1.contains(dim) && dimListCom2.contains(dim)) {
					tIndex = ii;
				}
				break;
			}
		}

		columnTypes[xIndex] = X_COORDINATE;
		columnTypes[yIndex] = Y_COORDINATE;
		if (SGDataUtility.isPolar(infoMap)) {
			columnTypes[comIndex1] = MAGNITUDE;
			columnTypes[comIndex2] = ANGLE;
		} else {
			columnTypes[comIndex1] = X_COMPONENT;
			columnTypes[comIndex2] = Y_COMPONENT;
		}
		if (tIndex != -1) {
			columnTypes[tIndex] = TIME;
		}

		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}

		return true;
	}

	private static boolean getDefaultColumnTypesForVXYTypeNetCDFData(
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, final NamedNodeMap nodeMap,
			final String groupName, final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}

		int index;

		// serial numbers
		String[] serialNumberNames = getNames(nodeMap,
				KEY_SERIAL_NUMBER_VARIABLE_NAME, groupName);
		if (serialNumberNames != null) {
			if (serialNumberNames.length != 1) {
				return false;
			}
			if ((index = setNetCDFColumnType(columnInfoList, columnTypes,
					serialNumberNames[0], SERIAL_NUMBER)) == -1) {
				return false;
			}
		}

		// x-coordinate
		String[] xNames = getNames(nodeMap, new String[] { KEY_X_COORDINATE_VARIABLE_NAME,
				KEY_X_COORDINATE_COLUMN_INDEX }, groupName);
		if (xNames == null) {
			return false;
		}
		if (xNames.length != 1) {
			return false;
		}
		index = setNetCDFColumnType(columnInfoList, columnTypes, xNames[0],
				X_COORDINATE);
		if (index == -1) {
			return false;
		}

		// y-coordinate
		String[] yNames = getNames(nodeMap, new String[] { KEY_Y_COORDINATE_VARIABLE_NAME,
				KEY_Y_COORDINATE_COLUMN_INDEX }, groupName);
		if (yNames == null) {
			return false;
		}
		if (yNames.length != 1) {
			return false;
		}
		index = setNetCDFColumnType(columnInfoList, columnTypes, yNames[0],
				Y_COORDINATE);
		if (index == -1) {
			return false;
		}

		// the first and the second component
		final boolean polar = SGDataUtility.isPolar(infoMap);
		String name1, name2;
		if (polar) {
			name1 = MAGNITUDE;
			name2 = ANGLE;
		} else {
			name1 = X_COMPONENT;
			name2 = Y_COMPONENT;
		}

		String[] c1Names = getNames(nodeMap, new String[] { KEY_FIRST_COMPONENT_VARIABLE_NAME,
				KEY_FIRST_COMPONENT_COLUMN_INDEX }, groupName);
		if (c1Names == null) {
			return false;
		}
		if (c1Names.length != 1) {
			return false;
		}
		index = setNetCDFColumnType(columnInfoList, columnTypes, c1Names[0],
				name1);
		if (index == -1) {
			return false;
		}

		String[] c2Names = getNames(nodeMap, new String[] { KEY_SECOND_COMPONENT_VARIABLE_NAME,
				KEY_SECOND_COMPONENT_COLUMN_INDEX }, groupName);
		if (c2Names == null) {
			return false;
		}
		if (c2Names.length != 1) {
			return false;
		}
		index = setNetCDFColumnType(columnInfoList, columnTypes, c2Names[0],
				name2);
		if (index == -1) {
			return false;
		}

		// time
		String[] timeNames = getNames(nodeMap, KEY_TIME_VARIABLE_NAME, groupName);
		if (timeNames != null) {
			if (timeNames.length != 1) {
				return false;
			}
			index = setNetCDFColumnType(columnInfoList, columnTypes,
					timeNames[0], TIME);
			if (index == -1) {
				return false;
			}
		}

		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}

		return true;
	}

	private static boolean getDefaultColumnTypesForSXYZTypeNetCDFData(
			final Map<String, Object> infoMap, final List<SGVariable> varList,
			final int size, 
			final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}

		int xIndex = -1;
		int yIndex = -1;
		int tIndex = -1;

		// a coordinate variable found first is assigned to the x variable
		// and found next is assigned to the y variable
		SGVariable xVar = null;
		SGVariable yVar = null;

		int startIndex = 0;
		int zIndex = -1;
		List<Dimension> dimListZ = null;

		for (int i = 0; i < varList.size() - 2; i++) {
			int[] xyIndex = getNetCDFCoordinateVariableXYIndex(varList,
					startIndex);
			if (xyIndex == null) {
				startIndex += 1;
				continue;
			} else {
				xIndex = xyIndex[0];
				yIndex = xyIndex[1];
				xVar = varList.get(xIndex);
				yVar = varList.get(yIndex);
			}

			Dimension xDim = xVar.getDimension(0);
			Dimension yDim = yVar.getDimension(0);
			for (int ii = 0; ii < size; ii++) {
				SGVariable var = varList.get(ii);
				if (var.isCoordinateVariable() == false) {
					List<Dimension> dimList = var.getDimensions();
					if (dimList.contains(xDim) && dimList.contains(yDim)) {
						zIndex = ii;
						break;
					}
				}
			}
			if (zIndex != -1) {
				break;
			} else {
				startIndex += 1;
			}
		}

		if (zIndex == -1) {
			return false;
		} else {
			dimListZ = varList.get(zIndex).getDimensions();
		}

		// unlimited coordinate variable is assigned to time variable
		// if it exists
		for (int ii = 0; ii < size; ii++) {
			if (ii == xIndex || ii == yIndex || ii == zIndex) {
				continue;
			}
			SGVariable var = varList.get(ii);
			if (var.isUnlimited() && var.isCoordinateVariable()) {
				Dimension dim = var.getDimension(0);
				if (dimListZ.contains(dim)) {
					tIndex = ii;
				}
				break;
			}
		}

		columnTypes[xIndex] = X_VALUE;
		columnTypes[yIndex] = Y_VALUE;
		columnTypes[zIndex] = Z_VALUE;
		if (tIndex != -1) {
			columnTypes[tIndex] = TIME;
		}

		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}

		return true;
	}

	private static boolean getDefaultColumnTypesForSXYZTypeNetCDFData(
			final List<SGDataColumnInfo> columnInfoList,
			final Map<String, Object> infoMap, final NamedNodeMap nodeMap,
			final String groupName, final SGDataColumnInfo[] columns) {

		String[] columnTypes = new String[columns.length];
		for (int ii = 0; ii < columnTypes.length; ii++) {
			columnTypes[ii] = "";
		}

		int index;

		// serial numbers
		String[] serialNumberNames = getNames(nodeMap,
				KEY_SERIAL_NUMBER_VARIABLE_NAME, groupName);
		if (serialNumberNames != null) {
			if (serialNumberNames.length != 1) {
				return false;
			}
			if ((index = setNetCDFColumnType(columnInfoList, columnTypes,
					serialNumberNames[0], SERIAL_NUMBER)) == -1) {
				return false;
			}
		}

		// x, y and z values
		String[] xNames = getNames(nodeMap, new String[] { KEY_X_VALUE_NAME, 
				KEY_X_VALUE_COLUMN_INDEX }, groupName);
		if (xNames == null) {
			return false;
		}
		if (xNames.length != 1) {
			return false;
		}
		index = setNetCDFColumnType(columnInfoList, columnTypes, xNames[0],
				X_VALUE);
		if (index == -1) {
			return false;
		}

		String[] yNames = getNames(nodeMap, new String[] { KEY_Y_VALUE_NAME, 
				KEY_Y_VALUE_COLUMN_INDEX }, groupName);
		if (yNames == null) {
			return false;
		}
		if (yNames.length != 1) {
			return false;
		}
		index = setNetCDFColumnType(columnInfoList, columnTypes, yNames[0],
				Y_VALUE);
		if (index == -1) {
			return false;
		}

		String[] zNames = getNames(nodeMap, new String[] { KEY_Z_VALUE_NAME, 
				KEY_Z_VALUE_COLUMN_INDEX }, groupName);
		if (zNames == null) {
			return false;
		}
		if (zNames.length != 1) {
			return false;
		}
		index = setNetCDFColumnType(columnInfoList, columnTypes, zNames[0],
				Z_VALUE);
		if (index == -1) {
			return false;
		}

		// time
		String[] timeNames = getNames(nodeMap, KEY_TIME_VARIABLE_NAME, groupName);
		if (timeNames != null) {
			if (timeNames.length != 1) {
				return false;
			}
			index = setNetCDFColumnType(columnInfoList, columnTypes,
					timeNames[0], TIME);
			if (index == -1) {
				return false;
			}
		}

		for (int ii = 0; ii < columns.length; ii++) {
			columns[ii].setColumnType(columnTypes[ii]);
		}

		return true;
	}

	private static int[] getNetCDFCoordinateVariableXYIndex(
			final List<SGVariable> varList, final int startIndex) {
		final int size = varList.size();
		int xIndex = -1;
		int yIndex = -1;

		// a coordinate variable found first is assigned to the x variable
		// and found next is assigned to the y variable
		xIndex = getNetCDFCoordinateVariableIndexWithoutUnlimited(varList,
				startIndex);
		if (xIndex == -1 || xIndex == size - 1) {
			xIndex = getNetCDFCoordinateVariableIndex(varList, startIndex);
		}
		if (xIndex == -1 || xIndex == size - 1) {
			return null;
		}

		yIndex = getNetCDFCoordinateVariableIndexWithoutUnlimited(varList,
				xIndex + 1);
		if (yIndex == -1) {
			yIndex = getNetCDFCoordinateVariableIndex(varList, xIndex + 1);
		}
		if (yIndex == -1) {
			return null;
		}

		return new int[] { xIndex, yIndex };
	}

	private static int getNetCDFCoordinateVariableIndex(
			final List<SGVariable> varList, final int startIndex) {
		final int size = varList.size();
		int index = -1;

		for (int ii = startIndex; ii < size; ii++) {
			SGVariable var = varList.get(ii);
			if (var.isCoordinateVariable()) {
				index = ii;
				break;
			}
		}

		return index;
	}

	private static int setNetCDFColumnType(
			final List<SGDataColumnInfo> columnInfoList,
			final String[] columnTypes, final String varName, final String value) {
		int index = findNetCDFColumnInfo(columnInfoList, varName);
		if (index != -1) {
			if (index < 0 || index >= columnTypes.length) {
				return -1;
			}
			columnTypes[index] = value;
		}
		return index;
	}

	private static int setNetCDFColumnType(
			final List<SGDataColumnInfo> columnInfoList,
			final String[] columnTypes, final String varName,
			final String holderVarName, final String value) {
		int index = findNetCDFColumnInfo(columnInfoList, varName);
		if (index != -1) {
			if (index < 0 || index >= columnTypes.length) {
				return -1;
			}
			columnTypes[index] = SGDataUtility.appendColumnTitle(value,
					holderVarName);
		}
		return index;
	}

	private static int findNetCDFColumnInfo(
			final List<SGDataColumnInfo> columnInfoList, final String varName) {
		for (int ii = 0; ii < columnInfoList.size(); ii++) {
			SGNetCDFDataColumnInfo ncInfo = (SGNetCDFDataColumnInfo) columnInfoList
					.get(ii);
			if (ncInfo.getVariableName().equals(varName)) {
				return ii;
			}
		}
		return -1;
	}

	private static int getNetCDFCoordinateVariableIndexWithoutUnlimited(
			final List<SGVariable> varList, final int startIndex) {
		final int size = varList.size();
		int index = -1;

		for (int ii = startIndex; ii < size; ii++) {
			SGVariable var = varList.get(ii);
			if (var.isCoordinateVariable() && !var.isUnlimited()) {
				// excepting unlimited variables first
				index = ii;
				break;
			}
		}

		return index;
	}

	private static String[] getNames(final NamedNodeMap nodeMap,
			final String[] keys, final String groupName) {
		for (int ii = 0; ii < keys.length; ii++) {
			String[] names = getNames(nodeMap, keys[ii], groupName);
			if (names != null) {
				return names;
			}
		}
		return null;
	}
	
	private static String[] getNames(final NamedNodeMap nodeMap,
			final String key, final String groupName) {
		String strNames = getString(nodeMap, key);
		if (strNames == null) {
			return null;
		}
		String[] names = SGUtilityText.parseStrings(strNames);
		if (groupName != null) {
			for (int ii = 0; ii < names.length; ii++) {
				names[ii] = SGDataUtility.appendGroupName(names[ii], groupName);
			}
		}
		return names;
	}

	private static String getString(final NamedNodeMap nodeMap, final String key) {
		Node nodeNames = nodeMap.getNamedItem(key);
		if (nodeNames == null) {
			return null;
		}
		String strNames = nodeNames.getNodeValue();
		if (strNames == null) {
			return null;
		}
		return strNames;
	}

	/**
	 * Finds and returns the origin map.
	 * If the origin map is not found, returns an empty map.
	 * 
	 * @param nodeMap
	 *           a node map
	 * @return the origin map
	 */
	public static Map<String, Integer> getOriginMap(NamedNodeMap nodeMap) {
		Map<String, Integer> map = new HashMap<String, Integer>();
		if (nodeMap != null) {
			String originMapStr = getString(nodeMap, SGNetCDFData.KEY_ORIGIN_MAP);
			if (originMapStr != null) {
				String[][] originMapArray = SGUtilityText.readStringMaps(originMapStr);
				for (int ii = 0; ii < originMapArray.length; ii++) {
					String key = originMapArray[ii][0];
					String value = originMapArray[ii][1];
					Integer num = SGUtilityText.getInteger(value);
					if (num == null) {
						return new HashMap<String, Integer>();
					}
					map.put(key, num);
				}
			}
		}
		return map;
	}
}
