package jp.riken.brain.ni.samuraigraph.data;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.base.SGIConstants;
import jp.riken.brain.ni.samuraigraph.base.SGIDataSource;
import jp.riken.brain.ni.samuraigraph.base.SGITextDataConstants;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;

import org.w3c.dom.Element;

import ucar.ma2.Array;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.Group;
import ucar.nc2.NetcdfFile;
import ucar.nc2.NetcdfFileWriteable;
import ucar.nc2.Variable;

/**
 * The base class for netCDF data.
 *
 */
public abstract class SGNetCDFData extends SGData 
		implements SGITextDataConstants, SGIDataColumnTypeConstants {

    /**
     * A text string of the key for the fill value.
     */
    public static final String FILL_VALUE = "_FillValue";
    
    /**
     * A text string of the key for the origin map.
     */
    public static final String KEY_ORIGIN_MAP = "Origin";

    /**
     * A netCDF file.
     */
    protected SGNetCDFFile mNCFile = null;

    /**
     * An observer of netCDF files.
     */
    protected SGDataSourceObserver mObserver = null;

    /**
     * The map of origin of coordinate variables.
     * Keys are the name of coordinate variables and values are origin of each variables.
     */
    protected Map<String, Integer> mOriginMap = new HashMap<String, Integer>();

    /**
     * The variable for time values.
     */
    protected SGVariable mTimeVariable = null;

    /**
     * The variable of the serial numbers.
     */
    protected SGVariable mSerialNumberVariable = null;

    /**
     * The default constructor.
     *
     */
    public SGNetCDFData() {
        super();
    }
    
    /**
     * Builds the data object with given NetCDF file.
     * 
     * @param ncfile
     *          NetCDF file
     * @param obs
     *          an observer for netCDF data
     * @param tName
     *          the name of time variable
     * @param serialNumberName
     *          the name of serial number variable
     */
    public SGNetCDFData(SGNetCDFFile ncfile, SGDataSourceObserver obs, String tName,
    		String serialNumberName) {
    	
        super();
        if (ncfile == null || obs == null) {
            throw new IllegalArgumentException("ncfile == null || obs == null");
        }
        this.mNCFile = ncfile;
        this.mObserver = obs;
        
        // add to the observer
        obs.addData(this);
        
        // initialize the origin map
        List<Dimension> dimList = ncfile.getDimensions();
        for (Dimension dim : dimList) { 
            String dimName = dim.getName();
            this.mOriginMap.put(dimName, 0);
        }
        
        // setup time variable
        if (tName != null) {
            SGVariable tVar = ncfile.findVariable(tName);
            if (tVar == null) {
                throw new IllegalArgumentException("Illegal variable name: " + tName);
            }
            if (!tVar.isCoordinateVariable()) {
                throw new IllegalArgumentException("Time variable must be a coordinate variable.");
            }
            this.mTimeVariable = tVar;
        }
        
        // setup serial number variable
        if (serialNumberName != null) {
        	SGVariable snVar = ncfile.findVariable(serialNumberName);
            if (snVar == null) {
                throw new IllegalArgumentException("Illegal variable name: " + serialNumberName);
            }
            if (!snVar.isCoordinateVariable()) {
                throw new IllegalArgumentException("Serial number variable must be a coordinate variable.");
            }
            this.mSerialNumberVariable = snVar;
        }
        
        if (this.mTimeVariable != null && this.mSerialNumberVariable != null) {
        	Dimension tDim = this.mTimeVariable.getDimension(0);
        	Dimension iDim = this.mSerialNumberVariable.getDimension(0);
        	if (tDim.equals(iDim)) {
        		throw new IllegalArgumentException("Dimensions of time and serial number variables must not be the same.");
        	}
        }
    }

    /**
     * Returns the netCDF file.
     * 
     * @return the netCDF file
     */
    public SGNetCDFFile getNetcdfFile() {
        return this.mNCFile;
    }
    
    /**
     * Returns the copy of this data object.
     * 
     * @return
     *         a copy of this data object
     */
    public Object copy() {
        SGNetCDFData data = (SGNetCDFData) super.copy();
        this.mObserver.addData(data);
        data.mOriginMap = new HashMap<String, Integer>(this.mOriginMap);
        data.mVariableCache = new HashMap<String, Object>(this.mVariableCache);
        return data;
    }
    
    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();

        // call a method of netCDF data observer
        this.mObserver.dataDisposed(this);

        // clear attributes
        this.mNCFile = null;

        // clear origin map
        this.mOriginMap.clear();
        this.mOriginMap = null;

        // clear variables
        this.mTimeVariable = null;
        this.mSerialNumberVariable = null;
    }

    /**
     * Set the data.
     * 
     * @param data
     *             data set to this object
     * @return
     *             true if succeeded            
     */
    public boolean setData(final SGData data) {
        if (!(data instanceof SGNetCDFData)) {
            throw new IllegalArgumentException("Given data is not the instance of SGNetCDFData class.");
        }
        this.clearVariableCache();
        SGNetCDFData nData = (SGNetCDFData) data;
        this.mNCFile = nData.mNCFile;
        this.mObserver = nData.mObserver;
        this.mObserver.addData(this);
        this.mOriginMap = new HashMap<String, Integer>(nData.mOriginMap);
        this.mTimeVariable = nData.mTimeVariable;
        this.mSerialNumberVariable = nData.mSerialNumberVariable;
        return true;
    }

    /**
     * Gets the properties of this data.
     * @param p
     *      the properties of this data
     * @return
     *      true if succeeded   
     */
    public boolean getProperties(SGProperties p) {
        if (!(p instanceof NetCDFDataProperties)) {
            return false;
        }
        NetCDFDataProperties np = (NetCDFDataProperties) p;
        if (this.mTimeVariable != null) {
            np.timeName = this.mTimeVariable.getName();
        }
        if (this.mSerialNumberVariable != null) {
        	np.serialNumberName = this.mSerialNumberVariable.getName();
        }
        np.originMap = new HashMap<String, Integer>(this.mOriginMap);
        return true;
    }

    /**
     * Set properties to this data.
     * 
     * @param p
     *          properties to be set
     * @return true if succeeded
     */
    public boolean setProperties(SGProperties p) {
        if (!(p instanceof NetCDFDataProperties)) {
            return false;
        }
        if (!p.equals(this.getProperties())) {
            // clear the cache
            this.clearVariableCache();
        }
        NetCDFDataProperties np = (NetCDFDataProperties) p;
        this.mTimeVariable = this.mNCFile.findVariable(np.timeName);
		this.mSerialNumberVariable = this.mNCFile.findVariable(np.serialNumberName);
		this.mOriginMap = new HashMap<String, Integer>(np.originMap);
        return true;
    }
    
    /**
     * Write properties of this object to the Element.
     * 
     * @param el
     *            the Element object
     * @param type
     *            type of the method to save properties
     * @return true if succeeded
     */
    public boolean writeProperty(Element el, final int type) {
        // write the origins
        StringBuffer sb = new StringBuffer();
        Iterator<String> itr = this.mOriginMap.keySet().iterator();
        while (itr.hasNext()) {
            String key = itr.next();
            Integer value = this.mOriginMap.get(key);
            sb.append(key);
            sb.append('=');
            sb.append(value);
            if (itr.hasNext()) {
                sb.append(',');
            }
        }
        el.setAttribute(KEY_ORIGIN_MAP, sb.toString());
        
        // time variable
        if (this.mTimeVariable != null) {
            el.setAttribute(SGIDataColumnNameConstants.KEY_TIME_VARIABLE_NAME, this.mTimeVariable.getName());
        }
        
        // serial number variable
        if (this.mSerialNumberVariable != null) {
            el.setAttribute(SGIDataColumnNameConstants.KEY_SERIAL_NUMBER_VARIABLE_NAME, this.mSerialNumberVariable.getName());
        }
        
        return true;
    }

    /**
     * Read properties.
     * 
     * @param el
     *           an Element object
     * @return true if succeeded
     */
    public boolean readProperty(Element el) {
        
        String str = null;
        Map<String, Integer> originMap = new HashMap<String, Integer>();
        SGVariable timeVar = null;
        SGVariable serialNumberVar = null;
        
        // origins
        str = el.getAttribute(KEY_ORIGIN_MAP);
        if (str.length() != 0) {
            String[] tokens = str.split(",");
            for (int ii = 0; ii < tokens.length; ii++) {
                String token = tokens[ii].trim();
                String[] keyValue = token.split("=");
                if (keyValue.length != 2) {
                    return false;
                }
                
                // get the name of a variable
                String key = keyValue[0].trim();
                if (this.mNCFile.findVariable(key) == null) {
                    return false;
                }
                
                // get the origin
                String value = keyValue[1].trim();
                Integer origin = SGUtilityText.getInteger(value);
                if (origin == null) {
                    return false;
                }
                
                // put into the map
                originMap.put(key, origin);
            }
            
        } else {
            return false;
        }
        
        // time variable
        str = el.getAttribute(SGIDataColumnNameConstants.KEY_TIME_VARIABLE_NAME);
        if (str.length() != 0) {
            timeVar = this.mNCFile.findVariable(str);
            if (timeVar == null) {
                return false;
            }
        }
        
        // serial number variable
        str = el.getAttribute(SGIDataColumnNameConstants.KEY_SERIAL_NUMBER_VARIABLE_NAME);
        if (str.length() != 0) {
        	serialNumberVar = this.mNCFile.findVariable(str);
            if (serialNumberVar == null) {
                return false;
            }
        }

        // set to attribute
        this.mOriginMap = originMap;
        this.mTimeVariable = timeVar;
        this.mSerialNumberVariable = serialNumberVar;

        return true;
    }
    
    /**
     * Sets the origin of a coordinate variable.
     * 
     * @param name
     *           the name of the coordinate variable
     * @param origin
     *           a value to set to the origin of the coordinate variable
     * @return true if succeeded
     */
    public boolean setOrigin(final String name, final int origin) {
        Dimension dim = this.mNCFile.findDimension(name);
        if (dim == null) {
        	return false;
        }
        final int len = dim.getLength();
        if (origin < 0 || origin >= len) {
        	return false;
        }
        this.mOriginMap.put(dim.getName(), origin);
        return true;
    }
    
    /**
     * Returns the origin of a coordinate variable.
     * 
     * @param name
     *           the name of the coordinate variable
     * @return the origin of the coordinate variable
     */
    public int getOrigin(final String name) {
        Dimension dim = this.mNCFile.findDimension(name);
        if (dim == null) {
            throw new IllegalArgumentException("Dimension is not found: " + name);
        }
        Integer num = this.mOriginMap.get(name);
        if (num == null) {
            throw new Error("Origin is not found in the map for the coordinate variable: " + name);
        }
        return num.intValue();
    }
    
    /**
     * Returns the value of coordinate variable at a given index.
     * 
     * @param name
     *           the name of a coordinate variable
     * @param index
     *           the array index
     * @return the value of coordinate variable at a given index
     * 
     * @throws IOException
     */
    public double getCoordinateValue(final String name, final int index) throws IOException {
    	SGVariable var = this.mNCFile.findVariable(name);
        if (var == null) {
            throw new IllegalArgumentException("Variable is not found: " + name);
        }
        if (var.isCoordinateVariable() == false) {
            throw new IllegalArgumentException("Not a coordinate variable: " + name);
        }
        if (var instanceof SGCoordinateVariable) {
            SGCoordinateVariable cVar = (SGCoordinateVariable) var;
            final int len = cVar.getLength();
            if (index < 0 || index >= len) {
            	throw new IllegalArgumentException("Index out of bounds [0, " + len + " + ]: " + index);
            }
            return index;
        } else {
            Array array = var.getVariable().read();
            final int size = (int) array.getSize();
            if (index < 0 || index >= size) {
                throw new IllegalArgumentException("Index out of bounds [0, " + size + " + ]: " + index);
            }
            return array.getDouble(index);
        }
    }

    /**
     * Returns a text string of the name with units string.
     * 
     * @param var
     *          the variable for an axis
     * @return a text string of the name with units string
     */
    protected String getNameWithUnit(SGVariable var) {
        final String name = var.getNameInPriorityOrder();
        final String unit = var.getUnitsString();
        StringBuffer sb = new StringBuffer();
        sb.append(name);
        if (unit != null) {
            sb.append(" [");
            sb.append(unit);
            sb.append(']');
        }
        return sb.toString();
    }

    /**
     * Returns the fill value of given variable.
     * 
     * @param var
     *            a variable
     * @return the fill value
     */
    protected Number getFillValue(SGVariable var) {
        Attribute attrFillValue = var.findAttribute(FILL_VALUE);
        if (attrFillValue == null) {
            return null;
        }
        Number fillValue = attrFillValue.getNumericValue();
        return fillValue;
    }
    
    
    /**
     * A class of netCDF data properties.
     */
    public static abstract class NetCDFDataProperties extends DataProperties {

        protected String timeName = null;
        
        protected String serialNumberName = null;

        protected Map<String, Integer> originMap = new HashMap<String, Integer>();

        /**
         * The default constructor.
         */
        public NetCDFDataProperties() {
            super();
        }

        /**
         * Dispose this object.
         */
        public void dispose() {
        	super.dispose();
            this.timeName = null;
            this.serialNumberName = null;
            this.originMap.clear();
            this.originMap = null;
        }
        
        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof NetCDFDataProperties) == false) {
                return false;
            }
            NetCDFDataProperties p = (NetCDFDataProperties) obj;
            if (SGUtility.equals(this.timeName, p.timeName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.serialNumberName, p.serialNumberName) == false) {
            	return false;
            }
            if (this.originMap.equals(p.originMap) == false) {
                return false;
            }
            return true;
        }
        
    }

    /**
     * Returns an array of current column types.
     * 
     * @return
     *        an array of current column types
     */
    public abstract String[] getCurrentColumnType();

    /**
     * Sets the column information.
     * 
     * @param colInfo
     *           the column information
     * @return true if succeeded
     */
    public boolean setColumnInfo(SGDataColumnInfo[] colInfo) {
    	
        if (super.setColumnInfo(colInfo) == false) {
        	return false;
        }
        
        // set the origins
        for (int ii = 0; ii < colInfo.length; ii++) {
            SGNetCDFDataColumnInfo nCol = (SGNetCDFDataColumnInfo) colInfo[ii];
        	SGVariable var = nCol.getVariable();
            if (var.isCoordinateVariable()) {
                String name = var.getName();
                this.mOriginMap.put(name, nCol.getOrigin());
            }
        }
        
        // clear the cache
        this.clearVariableCache();
        
        return true;
    }
    
    // clear the cache
    private void clearVariableCache() {
    	this.mVariableCache.clear();
    }

    /**
     * Returns the data type for given type.
     * 
     * @param type
     *           the type of use
     * @return the data type
     */
    public String getDataType(final int type) {
        if (type == SGIConstants.SAVE_PROPERTIES_TO_DATA_SET 
                || type == SGIConstants.SAVE_PROPERTIES_TO_DATA_SET_OLDER) {
            return this.getDataTypeForDataSet();
        } else {
            return this.getDataType();
        }
    }

    /**
     * Returns the data type for the data set.
     * 
     * @return the data type
     */
    protected abstract String getDataTypeForDataSet();

    /**
     * Returns an array of variables.
     * 
     * @return an array of variables
     */
    public abstract SGVariable[] getVariables();

    /**
     * Returns whether the coordinate variable is fixed.
     * 
     * @param name
     *           the name of a coordinate variable
     * @return true if fixed
     */
    public boolean isFixedCoordinateVariable(String name) {
    	SGVariable var = this.mNCFile.findVariable(name);
        if (var == null) {
            throw new IllegalArgumentException("Variable is not found: " + name);
        }
        if (var.isCoordinateVariable() == false) {
            throw new IllegalArgumentException("Not a coordinate variable: " + name);
        }
        SGVariable[] vars = this.getVariables();
        for (int ii = 0; ii < vars.length; ii++) {
            if (name.equals(vars[ii].getName())) {
                return false;
            }
        }
        return true;
    }
    
    /**
     * Read and returns the variable names.
     * 
     * @param el
     *            an Element object
     * @param key
     *            the key of name property
     * @return the indices it exists
     */
    protected String[] readNames(Element el, String key) {
        String str = el.getAttribute(key);
//        String[] array = null;
//        if (str.length() != 0) {
//            final int start = str.indexOf('{');
//            if (start == -1) {
//                return null;
//            }
//            final int end = str.lastIndexOf('}');
//            if (end == -1) {
//                return null;
//            }
//            if (start > end) {
//                return null;
//            }
//            String sub = str.substring(start + 1, end);
//            array = sub.split(",");
//            for (int ii = 0; ii < array.length; ii++) {
//                array[ii] = array[ii].trim();
//            }
//        }
//        return array;
        return SGUtilityText.parseStrings(str);
    }

    /**
     * Returns an array of values of the coordinate variable.
     * 
     * @return an array of values of the coordinate variable
     */
    protected double[] getCoordinateValueArray(SGVariable var) {
    	String varName = var.getName();
    	Object cache = this.getVariableCache(varName);
    	if (cache != null) {
    		return (double[]) cache;
    	}
    	double[] dArray = null;
        if (var instanceof SGCoordinateVariable) {
            SGCoordinateVariable cVar = (SGCoordinateVariable) var;
            final int len = cVar.getLength();
            dArray = new double[len];
            for (int ii = 0; ii < dArray.length; ii++) {
            	dArray[ii] = ii;
            }
//        } else if (var instanceof SGDateVariable) {
//			dArray = ((SGDateVariable) var).getNumberArray();
        } else {
            Array array = null;
            try {
                array = var.getVariable().read();
            } catch (IOException e) {
                return null;
            }
            final int size = (int) array.getSize();
            dArray = new double[size];
            for (int ii = 0; ii < size; ii++) {
                dArray[ii] = array.getDouble(ii);
            }
        }
        this.setVariableCache(varName, dArray);
        return dArray;
    }
    
    /**
     * Returns an array of values of given variable.
     * 
     * @param var
     *           the variable
     * @param cVars
     *           an array of coordinate variables
     * @return an array of values of given variable
     */
    protected double[] getValueArray(SGVariable var, SGVariable[] cVars) {
    	String varName = var.getName();
    	Object cache = this.getVariableCache(varName);
    	if (cache != null) {
    		return (double[]) cache;
    	}
    	return this.getValueArray(var, cVars, this.mOriginMap);
    }

    protected double[] getValueArray(SGVariable var, SGVariable[] cVars, 
    		Map<String, Integer> originMap) {
    	
    	// returns date values
    	if (var instanceof SGDateVariable) {
    		return ((SGDateVariable) var).getNumberArray();
    	}

    	String varName = var.getName();
    	Object cache = this.getVariableCache(varName);
    	if (cache != null) {
    		return (double[]) cache;
    	}

        List<Integer> cIndices = new ArrayList<Integer>();
        for (int ii = 0; ii < cVars.length; ii++) {
            final int cIndex = var.findDimensionIndex(cVars[ii].getName());
            cIndices.add(cIndex);
        }
        
        // get dimensions
        List<Dimension> dimList = var.getDimensions();
        final int dimSize = dimList.size();
        
        // create an array of origins from the map
        int[] origin = new int[dimSize];
        for (Dimension dim : dimList) {
            String name = dim.getName();
            final int dIndex = var.findDimensionIndex(name);
            if (cIndices.contains(dIndex)) {
                origin[dIndex] = 0;
            } else {
                Integer value = originMap.get(name);
                origin[dIndex] = value.intValue();
            }
        }

        // create an array of size
        int[] aSize = new int[dimSize];
        for (int ii = 0; ii < dimSize; ii++) {
            Dimension dim = dimList.get(ii);
            if (cIndices.contains(ii)) {
                aSize[ii] = dim.getLength();
            } else {
                aSize[ii] = 1;
            }
        }
        
        // get the fill value
        Number fillValue = this.getFillValue(var);

        // create an array
        Array data3D = null;
        try {
            data3D = var.read(origin, aSize);
        } catch (InvalidRangeException e) {
            return null;
        } catch (IOException e) {
            return null;
        }
        Array data2D = data3D.reduce();
        
        final int size2D = (int)data2D.getSize();
        Double[] valueArray = this.arrangeValueArray(data2D, var, cVars);
        double[] dArray = new double[size2D];
        for (int ii = 0; ii < size2D; ii++) {
            Double d = valueArray[ii];
            if (fillValue != null && d.doubleValue() == fillValue.doubleValue()) {
                dArray[ii] = Double.NaN;
            } else {
                dArray[ii] = d;
            }
        }
        
        this.setVariableCache(var.getName(), dArray);
        return dArray;
    }

    
    protected abstract Double[] arrangeValueArray(Array data2D, SGVariable var, 
            SGVariable[] cVars);
    
    protected Double[] arrange1Dim(Array data2D) {
        final int size2D = (int)data2D.getSize();
        Double[] dArray = new Double[size2D];
        for (int ii = 0; ii < size2D; ii++) {
            dArray[ii] = data2D.getDouble(ii);
        }
        return dArray;
    }
    
    protected Double[] arrange2Dim(Array data2D, SGVariable var, SGVariable[] cVars) {
        
        // a list of dimensions of coordinate variables
        List<Dimension> cDimList = new ArrayList<Dimension>();
        for (int ii = 0; ii < cVars.length; ii++) {
            cDimList.add(cVars[ii].getDimension(0));
        }
        
        // get dimensions
        List<Dimension> dimList = var.getDimensions();
        final int dimSize = dimList.size();
        List<Dimension> redDimList = new ArrayList<Dimension>();
        for (int ii = 0; ii < dimSize; ii++) {
            Dimension dim = dimList.get(ii);
            if (cDimList.contains(dim)) {
                redDimList.add(dim);
            }
        }
        final int firstLen = redDimList.get(0).getLength();
        final int secondLen = redDimList.get(1).getLength();
        
        Class<?> elType = data2D.getElementType();
        Object obj = data2D.copyToNDJavaArray();
        double[][] dArray;
        if (elType.equals(double.class)) {
            dArray = (double[][]) obj;
        } else {
            dArray = new double[firstLen][];
            if (elType.equals(float.class)) {
                float[][] array = (float[][]) obj;
                for (int ii = 0; ii < firstLen; ii++) {
                    dArray[ii] = new double[secondLen];
                    for (int jj = 0; jj < secondLen; jj++) {
                        dArray[ii][jj] = array[ii][jj];
                    }
                }
            } else if (elType.equals(int.class)) {
                int[][] array = (int[][]) obj;
                for (int ii = 0; ii < firstLen; ii++) {
                    dArray[ii] = new double[secondLen];
                    for (int jj = 0; jj < secondLen; jj++) {
                        dArray[ii][jj] = array[ii][jj];
                    }
                }
            } else if (elType.equals(long.class)) {
                long[][] array = (long[][]) obj;
                for (int ii = 0; ii < firstLen; ii++) {
                    dArray[ii] = new double[secondLen];
                    for (int jj = 0; jj < secondLen; jj++) {
                        dArray[ii][jj] = array[ii][jj];
                    }
                }
            } else if (elType.equals(short.class)) {
                short[][] array = (short[][]) obj;
                for (int ii = 0; ii < firstLen; ii++) {
                    dArray[ii] = new double[secondLen];
                    for (int jj = 0; jj < secondLen; jj++) {
                        dArray[ii][jj] = array[ii][jj];
                    }
                }
            } else if (elType.equals(byte.class)) {
                byte[][] array = (byte[][]) obj;
                for (int ii = 0; ii < firstLen; ii++) {
                    dArray[ii] = new double[secondLen];
                    for (int jj = 0; jj < secondLen; jj++) {
                        dArray[ii][jj] = array[ii][jj];
                    }
                }
            } else {
                throw new Error("Not supported element type: " + elType);
            }
        }
        
        boolean reversed = false;
        if (!cDimList.equals(redDimList)) {
            reversed = true;
        }
        
        final int allLen = firstLen * secondLen;
        Double[] ret = new Double[allLen];
        final int yLen = dArray.length;
        for (int yy = 0; yy < yLen; yy++) {
            final int xLen = dArray[yy].length;
            for (int xx = 0; xx < xLen; xx++) {
                final int index = reversed ? (yy * xLen + xx) : (xx * yLen + yy);
                ret[index] = dArray[yy][xx];
            }
        }

        return ret;
    }
    
    /**
     * 
     * @param file
     * @return
     * @throws IOException
     * @throws InvalidRangeException
     */
    protected boolean saveDataAsNetCDF(final File file)
    throws IOException, InvalidRangeException {
        SGVariable[] vArray = this.getVariables();
        
        NetcdfFile ncRead = this.getNetcdfFile().getNetcdfFile();
        NetcdfFileWriteable ncWrite = null;
        try {
            ncWrite = NetcdfFileWriteable.createNew(file.getAbsolutePath());
            
            List<Attribute> globalAttrList = ncRead.getGlobalAttributes();
            for (Attribute gattr : globalAttrList) {
                ncWrite.addGlobalAttribute(gattr);
            }
            
            final List<Variable> varList = ncRead.getVariables();
            
            Set<Dimension> dimSet = new HashSet<Dimension>();
            Set<String> varNameSet = new HashSet<String>();
            for (int i = 0; i < vArray.length; i++) {
                if (vArray[i] instanceof SGTextVariable) {
                    SGTextVariable textVar = (SGTextVariable)vArray[i];
                    Variable variable = null;
                    for (Variable var : varList) {
                        if (var.getName().equals(textVar.getOriginalName())) {
                            variable = var;
                            break;
                        }
                    }
                    if (variable==null) {
                        return false;
                    }
                    varNameSet.add(variable.getName());
                    List<Dimension> dims = variable.getDimensions();
                    for (int j = 0; j < dims.size(); j++) {
                        Dimension dim = dims.get(j);
                        dimSet.add(dim);
                        for (int k = 0; k < varList.size(); k++) {
                            if (dim.getName().equals(varList.get(i).getName())) {
                                varNameSet.add(varList.get(i).getName());
                                break;
                            }
                        }
                    }
                } else if (vArray[i] instanceof SGCoordinateVariable) {
                    SGCoordinateVariable coordVar = (SGCoordinateVariable)vArray[i];
                    if (coordVar.getVariable()!=null) {
                        Variable variable = vArray[i].getVariable();
                        varNameSet.add(variable.getName());
                        List<Dimension> dims = variable.getDimensions();
                        for (int j = 0; j < dims.size(); j++) {
                            Dimension dim = dims.get(j);
                            dimSet.add(dim);
                            for (int k = 0; k < varList.size(); k++) {
                                if (dim.getName().equals(varList.get(i).getName())) {
                                    varNameSet.add(varList.get(i).getName());
                                    break;
                                }
                            }
                        }
                    }
                    if (coordVar.getDimension()!=null) {
                        Dimension dim = coordVar.getDimension();
                        dimSet.add(dim);
                        for (int k = 0; k < varList.size(); k++) {
                            if (dim.getName().equals(varList.get(i).getName())) {
                                varNameSet.add(varList.get(i).getName());
                                break;
                            }
                        }
                    }
                } else {
                    // SGVariable, SGDateVariable
                    Variable variable = vArray[i].getVariable();
                    varNameSet.add(variable.getName());
                    List<Dimension> dims = variable.getDimensions();
                    for (int j = 0; j < dims.size(); j++) {
                        Dimension dim = dims.get(j);
                        dimSet.add(dim);
                        for (int k = 0; k < varList.size(); k++) {
                            if (dim.getName().equals(varList.get(i).getName())) {
                                varNameSet.add(varList.get(i).getName());
                                break;
                            }
                        }
                    }
                }
            }
            
            for (Dimension dim : dimSet) {
                Dimension d = new Dimension(dim.getName(), dim.getLength());
                ncWrite.addDimension(null, d);
            }
            
            for (int i = 0; i < varList.size(); i++) {
                Variable var = varList.get(i);
                if (varNameSet.contains(var.getName())) {
                    List<Attribute> attrList = var.getAttributes();
                    
                    String vname = var.getName();
                    String[] groupNames = vname.split("/");
                    StringBuffer sb = new StringBuffer();
                    for (int j = 0; j < groupNames.length-1; j++) {
                        sb.append(groupNames[j]);
                        if (j!=groupNames.length-2) sb.append("/");
                    }
                    String groupName = sb.toString();
                    if (groupName.length()==0) {
                        Variable v = new Variable(ncWrite, null, null, var.getName(), var.getDataType(), var.getDimensionsString());
                        for (int j = 0; j < attrList.size(); j++) {
                            v.addAttribute(attrList.get(j));
                        }
                        ncWrite.addVariable(null, v);
                    } else {
                        Group group = ncWrite.findGroup(groupName);
                        if (group==null) {
                            Group g = ncWrite.getRootGroup();
                            for (int j = 0; j < groupNames.length-1; j++) {
                                Group g2 = g.findGroup(groupNames[j]);
                                if (g2==null) {
                                    g2 = new Group(ncWrite, g, groupNames[j]);
                                    g.addGroup(g2);
                                }
                                g = g2;
                            }
                        }
                        group = ncWrite.findGroup(groupName);
                        Variable v = new Variable(ncWrite, null, null, groupNames[groupNames.length-1], var.getDataType(), var.getDimensionsString());
                        for (int j = 0; j < attrList.size(); j++) {
                            v.addAttribute(attrList.get(j));
                        }
                        ncWrite.addVariable(group, v);
                    }
                }
            }
            
            ncWrite.create();
            
            for (int i = 0; i < varList.size(); i++) {
                Variable var = varList.get(i);
                Array array = var.read();
                if (varNameSet.contains(var.getName())) {
                    ncWrite.write(var.getName(), array);
                }
            }
            
        } finally {
            if (ncWrite!=null) {
                ncWrite.close();
            }
        }
        
        return true;
    }

    /**
     * Save values to given file.
     * 
     * @param data
     *             a data
     * @param mode
     *             the mode of saving data
     * @return
     *             true if succeeded
     */
    @Override
    public boolean saveData(final File file, final int mode) {
        if (mode == SGIConstants.SAVE_PROPERTIES_TO_DATA_SET_NETCDF) {
            try {
                return this.saveDataAsNetCDF(file);
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }
        
        boolean status = false;
        SGVariable[] vArray = this.getVariables();
        try {
            FileWriter fw = new FileWriter(file);
            for (int ii = 0; ii < vArray.length; ii++) {
                String title = this.getNameWithUnit(vArray[ii]);
                if (title == null) {
                    title = "";
                }
                String str = SGUtilityText.getCSVString(title);
                if (ii != 0) {
                    fw.write(",");
                }
                fw.write(str);
            }
            fw.write("\n");
            status = this.writeData(fw, mode);
            fw.close();
        } catch (IOException ex) {
            return false;
        }
        return status;
    }
    
    /**
     * Returns an array of values to save to a data set.
     * 
     * @return an array of values to save to a data set
     */
    protected abstract double[][] getDataSetValues();
    
    /**
     * Write this data to a given output stream.
     * 
     * @param w
     *          an output stream
     * @return
     *          true if succeeded
     */
    protected boolean writeData(Writer w, final int mode) throws IOException {
        StringBuffer sb = new StringBuffer();
        double[][] array = this.getDataSetValues();
        for (int ii = 0; ii < array.length; ii++) {
            final int colNum = array[ii].length;

            if (mode == SGIConstants.SAVE_PROPERTIES_TO_DATA_SET_OLDER) {
                // check whether NaN values exist
                boolean isNaN = false;
                for (int jj = 0; jj < colNum; jj++) {
                    Double d = (Double) array[ii][jj];
                    if (d.isNaN()) {
                        isNaN = true;
                        break;
                    }
                }

                // when NaN values exist
                if (isNaN) {
                    sb.append(DATA_COMMENT_PREFIX);
                    sb.append(DATA_COMMENT_HEADER_PREFIX);
                    sb.append(HEADER_NOT_A_COMMENT_LINE);
                    sb.append(DATA_COMMENT_HEADER_SUFFIX);
                }
            }
            
            // export values in a row
            for (int jj = 0; jj < colNum; jj++) {
                sb.append(array[ii][jj]);
                if (jj != colNum - 1) {
                    sb.append(',');
                }
            }
            sb.append('\n');
        }
        w.write(sb.toString());
        return true;
    }

    /**
     * Returns an array of time values.
     * 
     * @return an array of time values
     */
    public double[] getTimeValueArray() {
        if (this.mTimeVariable == null) {
            return null;
        } else {
            return this.getCoordinateValueArray(this.mTimeVariable);
        }
    }
    
    /**
     * Returns whether time variable is available.
     * 
     * @return true if available
     */
    public boolean isTimeVariableAvailable() {
        return (this.mTimeVariable != null);
    }

    /**
     * Returns the current index of time value.
     * 
     * @return the current index of time value
     */
    public int getCurrentTimeValueIndex() {
        if (this.mTimeVariable == null) {
            return -1;
        }
        String key = this.mTimeVariable.getName();
        Integer num = this.mOriginMap.get(key);
        return num;
    }

    /**
     * Returns the time variable.
     * 
     * @return the time variable
     */
    public SGVariable getTimeVariable() {
        return this.mTimeVariable;
    }
    
    /**
     * Returns the current time value.
     * 
     * @return the current time value
     */
    public Number getCurrentTimeValue() {
        if (this.mTimeVariable == null) {
            return -1;
        }
        final int index = this.getCurrentTimeValueIndex();
        final double[] values = this.getTimeValueArray();
        return Double.valueOf(values[index]);
    }

    /**
     * Sets the current index of time value.
     * 
     * @param index
     *           a value to set to the index of time values
     */
    public void setCurrentTimeValueIndex(final int index) {
        if (this.mTimeVariable == null) {
            return;
        }
        Dimension tDim = this.mTimeVariable.getDimension(0);
        final int len = tDim.getLength();
        if (index < 0 || index >= len) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        String key = this.mTimeVariable.getName();
        this.mOriginMap.put(key, index);
        
        // clear the cache
        this.clearVariableCache();
    }

    /**
     * Returns the name of given variable if it is not null.
     * 
     * @param var
     *            a variable
     * @return the name of given variable or null
     */
    protected String getName(SGVariable var) {
    	return (var != null) ? var.getName() : null;
    }

    /**
     * Returns the name of given dimension if it is not null.
     * 
     * @param dim
     *            a dimension
     * @return the name of given dimension or null
     */
    protected String getName(Dimension dim) {
    	return (dim != null) ? dim.getName() : null;
    }

    /**
     * Finds and returns a variable of given name.
     * 
     * @param name
     *            the name of variable
     * @return a variable of given name
     */
    protected SGVariable findVariable(String name) {
    	return (name != null) ? this.mNCFile.findVariable(name) : null;
    }

    /**
     * Finds and returns a dimension of given name.
     * 
     * @param name
     *            the name of dimension
     * @return a variable of given name
     */
    protected Dimension findDimension(String name) {
    	return (name != null) ? this.mNCFile.findDimension(name) : null;
    }

    // Finds and returns a variable.
    protected SGVariable getVariable(SGNetCDFFile ncfile, String name) {
    	SGVariable var = null;
        if (name != null) {
        	var = ncfile.findVariable(name);
        	if (var == null) {
                throw new IllegalArgumentException("Invalid variable name: " + name);
        	}
        }
        return var;
    }
    
    // Checks whether a variable is not a coordinate variable and have given
    // dimensions.
    protected void checkNotCoordinateVariable(SGVariable var, List<Dimension> cDimList) {
		if (var != null) {
			if (var.isCoordinateVariable()) {
				throw new IllegalArgumentException("The variable "
						+ var.getName() + " must not be coordinate values.");
			}
			for (int ii = 0; ii < cDimList.size(); ii++) {
				Dimension cDim = cDimList.get(ii);
				if (!var.getDimensions().contains(cDim)) {
					throw new IllegalArgumentException("The variable "
							+ var.getName() + " does not have a dimension "
							+ cDim.getName());
				}
			}
		}
	}

    /**
     * Returns a map which has data information.
     * Overrode to set the data type and netCDF file.
     * 
     * @return a map which has data information
     */
    public Map<String, Object> getInfoMap() {
    	Map<String, Object> map = super.getInfoMap();
        map.put(KEY_DATA_TYPE, this.getDataType());
    	map.put(KEY_DATA_SOURCE, this.mNCFile);
    	return map;
    }

    /**
     * Returns the data source.
     * 
     * @return the data source
     */
    public SGIDataSource getDataSource() {
    	return this.getNetcdfFile();
    }

    /**
     * Returns the variable of serial numbers.
     * 
     * @return the variable of serial numbers
     */
    public SGVariable getSerialNumberVariable() {
    	return this.mSerialNumberVariable;
    }
    
    /**
     * Returns whether the variable of serial numbers is available.
     * 
     * @return true if the variable of serial numbers is available
     */
    public boolean isSerialNumberAvailable() {
    	return (this.mSerialNumberVariable != null);
    }
    
	/**
	 * The cache for variables.
	 */
	protected Map<String, Object> mVariableCache = new HashMap<String, Object>();

	/**
	 * Returns the cache of a variable of given name.
	 * 
	 * @param varName
	 *           the name of a variable
	 * 
	 * @return the cache of a variable of given name
	 */
	public Object getVariableCache(String varName) {
		Object obj = null;
		if (this.useFileCache(varName)) {
			obj = this.mNCFile.getVariableCache(varName);
		} else {
			String key = this.getCacheKey(varName).toString();
			obj = this.mVariableCache.get(key);
		}
		return obj;
	}

	/**
	 * Sets the cache of a variable of given name.
	 * 
	 * @param varName
	 *           the name of a variable
	 * @param cache
	 *           the cache of a variable of given name
	 */
	public void setVariableCache(String varName, Object cache) {
		if (this.useFileCache(varName)) {
			this.mNCFile.setVariableCache(varName, cache);
		} else {
			String key = this.getCacheKey(varName).toString();
			this.mVariableCache.put(key, cache);
		}
	}
	
	protected boolean useFileCache(String varName) {
		SGVariable var = this.findVariable(varName);
		if (var == null) {
			throw new IllegalArgumentException("var == null :" + varName);
		}
		return (var.isCoordinateVariable() || (var instanceof SGCoordinateVariable) || (var instanceof SGTextVariable));
	}

	protected Integer getCacheKey(String varName) {
		StringBuffer sb = new StringBuffer();
		sb.append(varName);
		sb.append(this.mOriginMap.toString());
		return Integer.valueOf(sb.toString().hashCode());
	}

	/**
	 * Sets the cache map.
	 * 
	 * @param map
	 *           the cache map
	 */
	protected void setCacheMap(Map<String, Object> map) {
		this.mVariableCache = map;
	}

}
