package jp.riken.brain.ni.samuraigraph.data;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGIConstants;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityNumber;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;

import org.w3c.dom.Element;

/**
 * Scalar type XY data. This object has of x and y values. If given, this object
 * can keep error bar values and text strings.
 * 
 */
public class SGSXYData extends SGArrayData implements SGISXYTypeData,
        SGIDataColumnTypeConstants, SGIDataColumnNameConstants, SGIConstants {

    /**
     * The column index for x-values.
     */
    protected Integer mXIndex = null;
    
    /**
     * The column index for y-values.
     */
    protected Integer mYIndex = null;
    
    /**
     * The column index for lower error values.
     */
    protected Integer mLowerErrorIndex = null;
    
    /**
     * The column index for upper error values.
     */
    protected Integer mUpperErrorIndex = null;
    
    /**
     * The column index for values that holds error values.
     */
    protected Integer mErrorBarHolderIndex = null;
    
    /**
     * The column index for tick labels.
     */
    protected Integer mTickLabelIndex = null;

    /**
     * The column index for values that holds tick labels.
     */
    protected Integer mTickLabelHolderIndex = null;

    /**
     * The decimal places for the tick labels.
     */
    protected int mDecimalPlaces = 0;

    /**
     * The exponent for the tick labels.
     */
    protected int mExponent = 0;
    
    
    /**
     * Builds a data object.
     * 
     */
    public SGSXYData() {
        super();
    }

    /**
     * Build a data with given columns and indices.
     * 
     * @param dataFile
     *                   the text data file
     * @param obs
     *                   an observer for array data
     * @param xIndex
     *                   the column index for x-values
     * @param yIndex
     *                   the column index for y-values
     */
    public SGSXYData(
            final SGTextDataFile dataFile, 
    		final SGDataSourceObserver obs,
            final Integer xIndex, final Integer yIndex) {
        this(dataFile, obs, xIndex, yIndex, null, null, null, null, null);
    }
    
    /**
     * Builds a data object with given columns and indices.
     * 
     * @param dataFile
     *                   the text data file
     * @param obs
     *                   an observer for array data
     * @param xIndex
     *                   the column index for x-values
     * @param yIndex
     *                   the column index for y-values
     * @param lowerErrorIndex
     *                   the column index for lower error values
     * @param upperErrorIndex
     *                   the column index for upper error values
     * @param errorBarHolderIndex
     *                   the column index for values that holds error values
     * @param tickLabelIndex
     *                   the column index for strings
     * @param tickLabelHolderIndex
     *                   the column index for values that holds tick labels
     */
    public SGSXYData(
            final SGTextDataFile dataFile,
            final SGDataSourceObserver obs,
            final Integer xIndex, final Integer yIndex,
            final Integer lowerErrorIndex,
            final Integer upperErrorIndex,
            final Integer errorBarHolderIndex,
            final Integer tickLabelIndex,
            final Integer tickLabelHolderIndex) {
        
        super(dataFile, obs);
        
        // null check
        if (xIndex == null || yIndex == null) {
            throw new IllegalArgumentException("Indices for x and y values must not be null.");
        }
        if (!(lowerErrorIndex == null && upperErrorIndex == null) && !(lowerErrorIndex != null && upperErrorIndex != null)) {
            throw new IllegalArgumentException(
                    "A pair of indices for the lower and upper error bars as (null, null) or (not null, not null) are permitted.");
        }
        if (lowerErrorIndex != null && errorBarHolderIndex == null) {
			throw new IllegalArgumentException("Error bar holder index is null.");
		}
		if (tickLabelIndex != null && tickLabelHolderIndex == null) {
			throw new IllegalArgumentException("Tick label holder index is null.");
		}
        
        // x values
        if (this.checkColumnIndexRange(xIndex) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + xIndex);
        }

        // y values
        if (this.checkColumnIndexRange(yIndex) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + yIndex);
        }
        
        // error values
        if (lowerErrorIndex != null) {
            if (this.checkColumnIndexRange(lowerErrorIndex) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + lowerErrorIndex);
            }
        }
        if (upperErrorIndex != null) {
            if (this.checkColumnIndexRange(upperErrorIndex) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + upperErrorIndex);
            }
        }
        if (errorBarHolderIndex != null) {
            if (this.checkColumnIndexRange(errorBarHolderIndex) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + errorBarHolderIndex);
            }
            if (!errorBarHolderIndex.equals(xIndex) && !errorBarHolderIndex.equals(yIndex)) {
                throw new IllegalArgumentException("Error holder value index must be equal to x value index or y value index.");
            }
        }

        // strings
        if (tickLabelIndex != null) {
            if (this.checkColumnIndexRange(tickLabelIndex) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + tickLabelIndex);
            }
        }
        if (tickLabelHolderIndex != null) {
        	if (this.checkColumnIndexRange(tickLabelHolderIndex) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + tickLabelHolderIndex);
        	}
        }
        
        // set indices
        this.mXIndex = xIndex;
        this.mYIndex = yIndex;
        this.mLowerErrorIndex = lowerErrorIndex;
        this.mUpperErrorIndex = upperErrorIndex;
        this.mErrorBarHolderIndex = errorBarHolderIndex;
        this.mTickLabelIndex = tickLabelIndex;
        this.mTickLabelHolderIndex = tickLabelHolderIndex;
        
        if (tickLabelIndex != null) {
            // initialize the decimal places and exponent when the column index for tick labels 
            // are given and the column is of the number column
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
            SGDataColumn col = columns[tickLabelIndex.intValue()];
            if (col instanceof SGNumberDataColumn) {
                SGNumberDataColumn nCol = (SGNumberDataColumn) col;
                
                // set the maximum order to exponent
                final double[] values = nCol.getNumberArray();
                int maxOrder = - Integer.MAX_VALUE;
                for (int ii = 0; ii < values.length; ii++) {
                    if (values[ii] == 0.0) {
                        continue;
                    }
                    if (Double.isNaN(values[ii])) {
                        continue;
                    }
                    final int order = SGUtilityNumber.getOrder(Math.abs(values[ii]));
                    if (order > maxOrder) {
                        maxOrder = order;
                    }
                }
                this.mExponent = maxOrder;
                
                // set the default value to decimal places
                this.mDecimalPlaces = 2;
            }
            
        } else {
        	// if tick labels are not set yet
        	this.setupTickLabel(xIndex, yIndex);
        }
    }
    
    private void setupTickLabel(Integer xIndex, Integer yIndex) {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGDataColumn xColumn = columns[xIndex.intValue()];
        SGDataColumn yColumn = columns[yIndex.intValue()];
    	if (xColumn instanceof SGDateDataColumn) {
    		this.mTickLabelIndex = xIndex;
    		this.mTickLabelHolderIndex = yIndex;
    	} else if (yColumn instanceof SGDateDataColumn) {
    		this.mTickLabelIndex = yIndex;
    		this.mTickLabelHolderIndex = xIndex;
    	}
    }
    
    /**
     * Builds a data object with given data.
     * 
     * @param data
     *             a data object that are set to this object
     */
    public SGSXYData(final SGData data) {
        super();
        this.setData(data);
    }

    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
        this.mXIndex = null;
        this.mYIndex = null;
        this.mLowerErrorIndex = null;
        this.mUpperErrorIndex = null;
        this.mErrorBarHolderIndex = null;
        this.mTickLabelIndex = null;
        this.mTickLabelHolderIndex = null;
    }

    /**
     * Sets the data.
     * 
     * @param data
     *             data set to this object
     * @return
     *             true if succeeded            
     */
    public boolean setData(final SGData data) {
        
        if (!(data instanceof SGSXYData)) {
            throw new IllegalArgumentException("!(data instanceof SGSXYData)");
        }
        
        if (super.setData(data) == false) {
            return false;
        }

        SGSXYData dataSXY = (SGSXYData) data;
        this.mXIndex = dataSXY.mXIndex;
        this.mYIndex = dataSXY.mYIndex;
        this.mLowerErrorIndex = dataSXY.mLowerErrorIndex;
        this.mUpperErrorIndex = dataSXY.mUpperErrorIndex;
        this.mErrorBarHolderIndex = dataSXY.mErrorBarHolderIndex;
        this.mTickLabelIndex = dataSXY.mTickLabelIndex;
        this.mTickLabelHolderIndex = dataSXY.mTickLabelHolderIndex;
        
        return true;
    }

    /**
     * Returns a number of data points.
     * 
     * @return
     *         a number of data points
     */
    public int getPointsNumber() {
        return this.getXArray().length;
    }

    /**
     * Returns the type of data.
     * 
     * @return
     *         the type of data
     */
    public String getDataType() {
        return SGDataTypeConstants.SXY_DATA;
    }

    /**
     * Returns the bounds of x-values.
     * 
     * @return the bounds of x-values
     */
    public SGValueRange getBoundsX() {
        return SGDataUtility.getBoundsX(this);
    }

    /**
     * Returns the bounds of y-values.
     * 
     * @return the bounds of y-values
     */
    public SGValueRange getBoundsY() {
        return SGDataUtility.getBoundsY(this);
    }
    
    /**
     * Returns the x value at a given index.
     * 
     * @param n
     *          the index of x value
     * @return
     *          x-value at a given index
     */
    public Double getXValue(final int n) {
    	final int len = this.getPointsNumber();
    	if (n < 0 || n >= len) {
    	    throw new IllegalArgumentException("given index is out of bounds: " + n);
    	}
    	return Double.valueOf(this.getXArray()[n]);
    }

    /**
     * Returns the y value at a given index.
     * 
     * @param n
     *          the index of y value
     * @return
     *          y-value at a given index
     */
    public Double getYValue(final int n) {
    	final int len = this.getPointsNumber();
    	if (n < 0 || n >= len) {
    	    throw new IllegalArgumentException("given index is out of bounds: " + n);
    	}
    	return Double.valueOf(this.getYArray()[n]);
    }

    /**
     * Returns the lower error value at a given index if it exists.
     * 
     * @param n
     *          the index of error value
     * @return
     *          lower error value at a given index or null it this data has no error bars
     */
    public Double getLowerErrorValue(final int n) {
    	final int len = this.getPointsNumber();
    	if (n < 0 || n >= len) {
    	    throw new IllegalArgumentException("given index is out of bounds: " + n);
    	}
        if (this.isErrorBarAvailable()) {
            return Double.valueOf(- Math.abs(this.getLowerErrorArray()[n]));
        }
        return null;
    }

    /**
     * Returns the upper error value at a given index if it exists.
     * 
     * @param n
     *          the index of error value
     * @return
     *          upper error value with the given index or null it this data has no error bars
     */
    public Double getUpperErrorValue(final int n) {
    	final int len = this.getPointsNumber();
    	if (n < 0 || n >= len) {
    	    throw new IllegalArgumentException("given index is out of bounds: " + n);
    	}
        if (this.isErrorBarAvailable()) {
            return Double.valueOf(Math.abs(this.getUpperErrorArray()[n]));
        }
        return null;
    }

    /**
     * Returns the string at a given index if it exists.
     * 
     * @param n
     *          the index of string
     * @return
     *          a string with the given index or null it this data has no strings
     */
    public String getString(final int n) {
        final int len = this.getPointsNumber();
        if (n < 0 || n >= len) {
            throw new IllegalArgumentException("given index is out of bounds: " + n);
        }
        if (this.isTickLabelAvailable()) {
            String[] array = this.getTextArray();
            if (array == null) {
                return null;
            } else {
                return array[n];
            }
        }
        return null;
    }

    /**
     * Returns a copy of x-value array.
     * 
     * @return
     *         x-value array
     */
    public double[] getXValueArray() {
//        return SGUtility.copyDoubleArray(this.getXArray());
    	return this.getXArray();
    }

    /**
     * Returns a copy of y-value array.
     * 
     * @return
     *         y-value array
     */
    public double[] getYValueArray() {
//        return SGUtility.copyDoubleArray(this.getYArray());
    	return this.getYArray();
    }

    /**
     * Returns a copy of lower-error-value array if they exist.
     * 
     * @return
     *         an array of lower-error-values or null if this data has no error bars
     */
    public double[] getLowerErrorValueArray() {
        if (this.isErrorBarAvailable()) {
            double[] eArray = SGUtility.copyDoubleArray(this.getLowerErrorArray());
            for (int ii = 0; ii < eArray.length; ii++) {
                eArray[ii] = - Math.abs(eArray[ii]);
            }
            return eArray;
        }
        return null;
    }

    /**
     * Returns a copy of upper-error-value array if they exist.
     * 
     * @return
     *         an array of upper-error-values or null if this data has no error bars
     */
    public double[] getUpperErrorValueArray() {
        if (this.isErrorBarAvailable()) {
            double[] eArray = SGUtility.copyDoubleArray(this.getUpperErrorArray());
            for (int ii = 0; ii < eArray.length; ii++) {
                eArray[ii] = Math.abs(eArray[ii]);
            }
            return eArray;
        }
        return null;
    }

    /**
     * Returns an array of strings if they exist.
     * Decimal places and exponent are effective when the column for tick labels is of the number column.
     * 
     * @param dp
     *          decimal places
     * @param exp
     *          exponent
     * @return string array
     */
    public String[] getStringArray() {
        if (this.mTickLabelIndex != null) {
            SGDataColumn[] columns = this.mDataFile.mDataColumns;
            SGDataColumn col = columns[this.mTickLabelIndex.intValue()];
            if (col instanceof SGNumberDataColumn) {
                SGNumberDataColumn numCol = (SGNumberDataColumn) col;
                return numCol.getStringArray(this.mDecimalPlaces, this.mExponent);
            } else {
                return col.getStringArray();
            }
        } else {
            return null;
        }
    }

    /**
     * Returns whether error bars are available.
     * 
     * @return true if error bars are available
     */
    public boolean isErrorBarAvailable() {
        return (this.mUpperErrorIndex != null && this.mLowerErrorIndex != null);
    }

    /**
     * Returns whether tick labels are available.
     * 
     * @return true if tick labels are available
     */
    public boolean isTickLabelAvailable() {
        return (this.mTickLabelIndex != null);
    }

    /**
     * Returns whether a given data object has the same tick label resources.
     * 
     * @param data
     *           a data to compare
     * @return true if a given data object has the same tick label resources
     */
    public boolean hasEqualTickLabelResource(SGISXYTypeData data) {
    	if (!(data instanceof SGSXYData)) {
    		return false;
    	}
    	SGSXYData dataSXY = (SGSXYData) data;
    	if (this.mTickLabelIndex == null) {
    		return (dataSXY.mTickLabelIndex == null);
    	} else {
        	return this.mTickLabelIndex.equals(dataSXY.mTickLabelIndex);
    	}
    }
    
    /**
     * Returns whether a data type column is used.
     * 
     * @return true if a data type column is used
     */
    public boolean isDateColumnUsed() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
		if (columns[this.mXIndex] instanceof SGDateDataColumn) {
			return true;
		}
		if (columns[this.mYIndex] instanceof SGDateDataColumn) {
			return true;
		}
		return false;
    }

    /**
     * Returns an array of x-values.
     * @return
     *         an array of x-values
     */
    protected double[] getXArray() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
    	SGINumberDataColumn col = (SGINumberDataColumn) columns[this.mXIndex.intValue()];
        return col.getNumberArray();
    }
    
    /**
     * Returns an array of y-values.
     * @return
     *         an array of y-values
     */
    protected double[] getYArray() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGINumberDataColumn col = (SGINumberDataColumn) columns[this.mYIndex.intValue()];
        return col.getNumberArray();
    }
    
    /**
     * Returns an array of lower error values.
     * @return
     *         an array of lower error values
     */
    protected double[] getLowerErrorArray() {
    	if (this.mLowerErrorIndex == null) {
    	    return null;
    	} else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
    		SGINumberDataColumn col = (SGINumberDataColumn) columns[this.mLowerErrorIndex.intValue()];
            return col.getNumberArray();
    	}
    }
    
    /**
     * Returns an array of upper error values.
     * @return
     *         an array of upper error values
     */
    protected double[] getUpperErrorArray() {
    	if (this.mUpperErrorIndex == null) {
    	    return null;
    	} else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
    		SGINumberDataColumn col = (SGINumberDataColumn) columns[this.mUpperErrorIndex.intValue()];
            return col.getNumberArray();
    	}
    }
    
    /**
     * Returns an array of strings.
     * @return
     *         an array of strings
     */
    protected String[] getTextArray() {
        if (this.mTickLabelIndex == null) {
            return null;
        } else {
            SGDataColumn[] columns = this.mDataFile.mDataColumns;
            SGDataColumn col = columns[this.mTickLabelIndex.intValue()];
            return col.getStringArray();
        }
    }
    
    /**
     * Writes data column indices in the attributes.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected boolean writeAttributeColumnIndices(Element el) {
//        el.setAttribute(KEY_X_VALUE_COLUMN_INDEX, this.mXIndex.toString());
//        el.setAttribute(KEY_Y_VALUE_COLUMN_INDEX, this.mYIndex.toString());
//        if (this.isErrorBarAvailable()) {
//            el.setAttribute(KEY_LOWER_ERROR_BAR_COLUMN_INDEX, this.mLowerErrorIndex.toString());
//            el.setAttribute(KEY_UPPER_ERROR_BAR_COLUMN_INDEX, this.mUpperErrorIndex.toString());
//        	el.setAttribute(KEY_ERROR_BAR_HOLDER_COLUMN_INDEX, this.mErrorBarHolderIndex.toString());
//        }
//        if (this.isTickLabelAvailable()) {
//            el.setAttribute(KEY_TICK_LABEL_COLUMN_INDEX, this.mTickLabelIndex.toString());
//        	el.setAttribute(KEY_TICK_LABEL_HOLDER_COLUMN_INDEX, this.mTickLabelHolderIndex.toString());
//        }
        return true;
    }

    /**
     * Writes data column indices as sequential numbers.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected boolean writeSequentialColumnIndices(Element el) {
//    	int tlIndex = 0;
//        el.setAttribute(KEY_X_VALUE_COLUMN_INDEX, Integer.toString(0));
//        el.setAttribute(KEY_Y_VALUE_COLUMN_INDEX, Integer.toString(1));
//        tlIndex += 2;
//        if (this.isErrorBarAvailable()) {
//            el.setAttribute(KEY_LOWER_ERROR_BAR_COLUMN_INDEX, Integer.toString(2));
//            el.setAttribute(KEY_UPPER_ERROR_BAR_COLUMN_INDEX, Integer.toString(3));
//            if (this.mYIndex.equals(this.mErrorBarHolderIndex)) {
//                el.setAttribute(KEY_ERROR_BAR_HOLDER_COLUMN_INDEX, this.mYIndex.toString());
//            } else if (this.mXIndex.equals(this.mErrorBarHolderIndex)) {
//                el.setAttribute(KEY_ERROR_BAR_HOLDER_COLUMN_INDEX, this.mXIndex.toString());
//            }
//            tlIndex += 3;
//        }
//        if (this.isTickLabelAvailable()) {
//            el.setAttribute(KEY_TICK_LABEL_COLUMN_INDEX, Integer.toString(tlIndex));
//            el.setAttribute(KEY_TICK_LABEL_HOLDER_COLUMN_INDEX, Integer.toString(tlIndex + 1));
//        }
        return true;
    }
    
    @Override
    protected boolean writeSequentialColumnName(Element el) {
//        int tlIndex = 0;
//        el.setAttribute(KEY_X_VALUE_COLUMN_INDEX, this.getSequentialColumnName(0));
//        el.setAttribute(KEY_Y_VALUE_COLUMN_INDEX, this.getSequentialColumnName(1));
//        tlIndex += 2;
//        if (this.isErrorBarAvailable()) {
//            el.setAttribute(KEY_LOWER_ERROR_BAR_COLUMN_INDEX, this.getSequentialColumnName(2));
//            el.setAttribute(KEY_UPPER_ERROR_BAR_COLUMN_INDEX, this.getSequentialColumnName(3));
//            if (this.mYIndex.equals(this.mErrorBarHolderIndex)) {
//                el.setAttribute(KEY_ERROR_BAR_HOLDER_COLUMN_INDEX, this.getSequentialColumnName(this.mYIndex.intValue()));
//            } else if (this.mXIndex.equals(this.mErrorBarHolderIndex)) {
//                el.setAttribute(KEY_ERROR_BAR_HOLDER_COLUMN_INDEX, this.getSequentialColumnName(this.mXIndex.intValue()));
//            }
//            tlIndex += 3;
//        }
//        if (this.isTickLabelAvailable()) {
//            el.setAttribute(KEY_TICK_LABEL_COLUMN_INDEX, this.getSequentialColumnName(tlIndex));
//            el.setAttribute(KEY_TICK_LABEL_HOLDER_COLUMN_INDEX, this.getSequentialColumnName(tlIndex + 1));
//        }
//        
//        // serial numbers
//        el.setAttribute(KEY_SERIAL_NUMBER_COLUMN_INDEX, SERIAL_NUMBER);
//        
        return true;
    }

    /**
     * Read properties from a given Element and set to this data.
     * @param el
     *          an Element
     * @return
     *          true if succeeded
     */
    public boolean readProperty(Element el) {
//        Integer index = null;
//        Integer xIndex = null;
//        Integer yIndex = null;
//        Integer lIndex = null;
//        Integer uIndex = null;
//        Integer ehIndex = null;
//        Integer tIndex = null;
//        Integer thIndex = null;
//        if ((index = this.readIndex(el, KEY_X_VALUE_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            xIndex = index;
//        } else {
//            return false;
//        }
//        if ((index = this.readIndex(el, KEY_Y_VALUE_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            yIndex = index;
//        } else {
//            return false;
//        }
//        if ((index = this.readIndex(el, KEY_LOWER_ERROR_BAR_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            lIndex = index;
//        }
//        if ((index = this.readIndex(el, KEY_UPPER_ERROR_BAR_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            uIndex = index;
//        }
//        if ((index = this.readIndex(el, KEY_ERROR_BAR_HOLDER_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            ehIndex = index;
//        } else {
//        	ehIndex = yIndex;
//        }
//        if (!(lIndex == null && uIndex == null) && !(lIndex != null && uIndex != null)) {
//            return false;
//        }
//        if ((index = this.readIndex(el, KEY_TICK_LABEL_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            tIndex = index;
//        }
//        if ((index = this.readIndex(el, KEY_TICK_LABEL_HOLDER_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            thIndex = index;
//        }
//        
//        // set to attributes
//        this.mXIndex = xIndex;
//        this.mYIndex = yIndex;
//        this.mLowerErrorIndex = lIndex;
//        this.mUpperErrorIndex = uIndex;
//        this.mErrorBarHolderIndex = ehIndex;
//        this.mTickLabelIndex = tIndex;
//        this.mTickLabelHolderIndex = thIndex;

        return true;
    }
    
    /**
     * Returns the title of X values.
     * @return
     *         the title of X values
     */
    public String getTitleX() {
        if (this.mXIndex == null) {
            return "";
        } else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mXIndex.intValue()].getTitle();
        }
    }
    
    /**
     * Returns the title of Y values.
     * @return
     *         the title of Y values
     */
    public String getTitleY() {
        if (this.mYIndex == null) {
            return "";
        } else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mYIndex.intValue()].getTitle();
        }
    }

    public Integer getXIndex() {
        return this.mXIndex;
    }

    public Integer getYIndex() {
        return this.mYIndex;
    }
    
    public Integer getLowerErrorIndex() {
        return this.mLowerErrorIndex;
    }
    
    public Integer getUpperErrorIndex() {
        return this.mUpperErrorIndex;
    }

    public Integer getErrorBarHolderIndex() {
        return this.mErrorBarHolderIndex;
    }

    public Integer getTickLabelIndex() {
        return this.mTickLabelIndex;
    }
    
    public Integer getTickLabelHolderIndex() {
    	return this.mTickLabelHolderIndex;
    }

    public void setXIndex(Integer index) {
        if (index == null) {
            throw new IllegalArgumentException("index == null");
        }
        if (this.checkColumnIndexRange(index) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        this.mXIndex = index;
    }

    public void setYIndex(Integer index) {
        if (index == null) {
            throw new IllegalArgumentException("index == null");
        }
        if (this.checkColumnIndexRange(index) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        this.mYIndex = index;
    }
    
    public void setLowerErrorIndex(Integer index) {
        if (index != null) {
            if (this.checkColumnIndexRange(index) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + index);
            }
        }
        this.mLowerErrorIndex = index;
    }

    public void setUpperErrorIndex(Integer index) {
        if (index != null) {
            if (this.checkColumnIndexRange(index) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + index);
            }
        }
        this.mUpperErrorIndex = index;
    }

    public void setErrorBarHolderIndex(Integer index) {
        if (index != null) {
            if (this.checkColumnIndexRange(index) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + index);
            }
        }
        this.mErrorBarHolderIndex = index;
    }

    public void setTickLabelIndex(Integer index) {
        if (index != null) {
            if (this.checkColumnIndexRange(index) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + index);
            }
        }
        this.mTickLabelIndex = index;
    }

    public void setTickLabelHolderIndex(Integer index) {
        if (index != null) {
            if (this.checkColumnIndexRange(index) == false) {
                throw new IllegalArgumentException("Index out of bounds: " + index);
            }
        }
        this.mTickLabelHolderIndex = index;
    }

    /**
     * Returns properties of this data.
     * @return
     *         properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new SXYDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Get properties of this data.
     * @param p
     *          properties to set values
     * @return
     *          true if succeeded
     */
    public boolean getProperties(final SGProperties p) {
        if ((p instanceof SXYDataProperties) == false) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        SXYDataProperties ep = (SXYDataProperties) p;
        ep.mXIndex = this.getXIndex();
        ep.mYIndex = this.getYIndex();
        ep.mLowerErrorIndex = this.getLowerErrorIndex();
        ep.mUpperErrorIndex = this.getUpperErrorIndex();
        ep.mErrorBarHolderIndex = this.getErrorBarHolderIndex();
        ep.mTickLabelIndex = this.getTickLabelIndex();
        ep.mTickLabelHolderIndex = this.getTickLabelHolderIndex();
        return true;
    }
    
    /**
     * Set properties to this data.
     * @param p
     *          properties that have values to set to this data
     * @return
     *          true if succeeded
     */
    public boolean setProperties(final SGProperties p) {
        if ((p instanceof SXYDataProperties) == false) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        SXYDataProperties ep = (SXYDataProperties) p;
        this.mXIndex = ep.mXIndex;
        this.mYIndex = ep.mYIndex;
        this.mLowerErrorIndex = ep.mLowerErrorIndex;
        this.mUpperErrorIndex = ep.mUpperErrorIndex;
        this.mErrorBarHolderIndex = ep.mErrorBarHolderIndex;
        this.mTickLabelIndex = ep.mTickLabelIndex;
        this.mTickLabelHolderIndex = ep.mTickLabelHolderIndex;
        return true;
    }

    /**
     * Sets the type of data columns.
     * 
     * @param column
     *           an array of column types
     * @return true if succeeded
     */
    public boolean setColumnType(String[] columns) {
        if (columns == null) {
            throw new IllegalArgumentException("columns == null");
        }
        Integer x = null;
        Integer y = null;
        Integer le = null;
        Integer ue = null;
        Integer tl = null;
        for (int ii = 0; ii < columns.length; ii++) {
            if (SGUtilityText.isEqualString(X_VALUE, columns[ii])) {
                x = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(Y_VALUE, columns[ii])) {
                y = Integer.valueOf(ii);
            } else if (SGUtilityText.startsWith(columns[ii], LOWER_ERROR_VALUE)) {
                le = Integer.valueOf(ii);
            } else if (SGUtilityText.startsWith(columns[ii], UPPER_ERROR_VALUE)) {
                ue = Integer.valueOf(ii);
            } else if (SGUtilityText.startsWith(columns[ii], LOWER_UPPER_ERROR_VALUE)) {
                le = Integer.valueOf(ii);
                ue = Integer.valueOf(ii);
            } else if (SGUtilityText.startsWith(columns[ii], TICK_LABEL)) {   
                tl = Integer.valueOf(ii);
            } else if ("".equals(columns[ii])) {
                continue;
            } else {
            	return false;
            }
        }
        
        // check the necessary columns
        if (x == null || y == null) {
        	return false;
        }
        
        // check the error bar
        if (!(le == null && ue == null) && !(le != null && ue != null)) {
        	return false;
        }
        
        String[] columnTitles = this.mDataFile.getTitles();
        
        Integer eh = null;
        if (le != null) {
            String strle = columns[le.intValue()];
            String strue = columns[ue.intValue()];
            Integer ehle = SGDataUtility.getAppendedColumnIndex(strle, columnTitles);
            Integer ehue = SGDataUtility.getAppendedColumnIndex(strue, columnTitles);
            if (ehle == null || ehue==null || ehle.equals(ehue)==false) {
                return false;
            }
            eh = ehle;
        }
        
        Integer th = null;
        if (tl != null) {
            String str = columns[tl.intValue()];
            th = SGDataUtility.getAppendedColumnIndex(str, columnTitles);
            if (th == null) {
                return false;
            }
        }
        
        // set to the attributes
        this.mXIndex = x;
        this.mYIndex = y;
        this.mLowerErrorIndex = le;
        this.mUpperErrorIndex = ue;
        this.mErrorBarHolderIndex = eh;
        this.mTickLabelIndex = tl;
        this.mTickLabelHolderIndex = th;
        
    	// if tick labels are not set yet
        if (tl == null && th == null) {
        	this.setupTickLabel(x, y);
        }

        return true;
    }
    
    
    /**
     * A class for scalar xy-type data properties.
     *
     */
    public static class SXYDataProperties extends ArrayDataProperties {
        
        // indices of data columns
        Integer mXIndex = null;

        Integer mYIndex = null;
        
        Integer mLowerErrorIndex = null;
        
        Integer mUpperErrorIndex = null;
        
        Integer mErrorBarHolderIndex = null;
        
        Integer mTickLabelIndex = null;

        Integer mTickLabelHolderIndex = null;

        /**
         * The default constructor.
         */
        public SXYDataProperties() {
            super();
        }

        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof SXYDataProperties) == false) {
                return false;
            }

            SXYDataProperties p = (SXYDataProperties) obj;
            
            if (!SGUtility.equals(this.mXIndex, p.mXIndex)) {
            	return false;
            }
            if (!SGUtility.equals(this.mYIndex, p.mYIndex)) {
            	return false;
            }
            if (!SGUtility.equals(this.mLowerErrorIndex, p.mLowerErrorIndex)) {
            	return false;
            }
            if (!SGUtility.equals(this.mUpperErrorIndex, p.mUpperErrorIndex)) {
            	return false;
            }
            if (!SGUtility.equals(this.mErrorBarHolderIndex, p.mErrorBarHolderIndex)) {
            	return false;
            }
            if (!SGUtility.equals(this.mTickLabelIndex, p.mTickLabelIndex)) {
            	return false;
            }
            if (!SGUtility.equals(this.mTickLabelHolderIndex, p.mTickLabelHolderIndex)) {
            	return false;
            }

            return true;
        }

        /**
         * Returns string representation of this object.
         * @return
         *        string representation of this object
         */
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("[");
            sb.append("XIndex=");
            sb.append(this.mXIndex);
            sb.append(", YIndex=");
            sb.append(this.mYIndex);
            sb.append(", LowerErrorIndex=");
            sb.append(this.mLowerErrorIndex);
            sb.append(", UpperErrorIndex=");
            sb.append(this.mUpperErrorIndex);
            sb.append(", ErrorBarHolderIndex=");
            sb.append(this.mErrorBarHolderIndex);
            sb.append(", TickLabelIndex=");
            sb.append(this.mTickLabelIndex);
            sb.append(", TickLabelHolderIndex=");
            sb.append(this.mTickLabelHolderIndex);
            sb.append("]");
            return sb.toString();
        }
        
        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.mXIndex = null;
            this.mYIndex = null;
            this.mLowerErrorIndex = null;
            this.mUpperErrorIndex = null;
            this.mErrorBarHolderIndex = null;
            this.mTickLabelIndex = null;
            this.mTickLabelHolderIndex = null;
        }
    }

    
    /**
     * Returns an array of current column types.
     * @return
     *        an array of current column types
     */
    public String[] getCurrentColumnType() {
        final int colNum = this.getColNum();
        String[] array = new String[colNum];
        if (this.mXIndex != null) {
            array[this.mXIndex.intValue()] = X_VALUE;
        }
        if (this.mYIndex != null) {
            array[this.mYIndex.intValue()] = Y_VALUE;
        }
        if (this.mLowerErrorIndex != null && this.mUpperErrorIndex != null 
                && this.mErrorBarHolderIndex != null) {
            int index = this.mErrorBarHolderIndex.intValue();
            if (this.mLowerErrorIndex.equals(this.mUpperErrorIndex)) {
                array[this.mLowerErrorIndex.intValue()] =
                    SGDataUtility.appendColumnNoOrTitle(
                        LOWER_UPPER_ERROR_VALUE, index,
                        this.mDataFile.isEmptyTitle(index) || this.mDataFile.isRepeatedTitle(index),
                        this.getTitle(index));
            } else {
                array[this.mLowerErrorIndex.intValue()] =
                    SGDataUtility.appendColumnNoOrTitle(
                        LOWER_ERROR_VALUE, index,
                        this.mDataFile.isEmptyTitle(index) || this.mDataFile.isRepeatedTitle(index),
                        this.getTitle(index));
                array[this.mUpperErrorIndex.intValue()] =
                    SGDataUtility.appendColumnNoOrTitle(
                        UPPER_ERROR_VALUE, index,
                        this.mDataFile.isEmptyTitle(index) || this.mDataFile.isRepeatedTitle(index),
                        this.getTitle(index));
            }
        }
        if (this.mTickLabelIndex != null && this.mTickLabelHolderIndex != null
                && !this.hasDateTickLabel()) {
            int index = this.mTickLabelHolderIndex.intValue();
            array[this.mTickLabelIndex.intValue()] =
                SGDataUtility.appendColumnNoOrTitle(
                    TICK_LABEL, index,
                    this.mDataFile.isEmptyTitle(index) || this.mDataFile.isRepeatedTitle(index),
                    this.getTitle(index));
        }
        return array;
    }
    
    private boolean hasDateTickLabel() {
		if (this.isTickLabelAvailable()) {
			if (this.mTickLabelIndex.equals(this.mXIndex)
					|| this.mTickLabelIndex.equals(this.mYIndex)) {
				return true;
			} else {
				return false;
			}
		} else {
			return false;
		}
	}
    
    /**
	 * Returns an array of data columns to save to a data set.
	 * 
	 * @return an array of data columns to save to a data set
	 */
    protected SGDataColumn[] getDataSetColumns() {
        SGDataColumn[] colArray;
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        if (this.isErrorBarAvailable()) {
            if (this.isTickLabelAvailable()) {
                colArray = new SGDataColumn[7];
                colArray[5] = columns[this.mTickLabelIndex.intValue()];
                colArray[6] = columns[this.mTickLabelHolderIndex.intValue()];
            } else {
                colArray = new SGDataColumn[5];
            }
            colArray[2] = columns[this.mLowerErrorIndex.intValue()];
            colArray[3] = columns[this.mUpperErrorIndex.intValue()];
            colArray[4] = columns[this.mErrorBarHolderIndex.intValue()];
        } else {
            if (this.isTickLabelAvailable()) {
                colArray = new SGDataColumn[4];
                colArray[2] = columns[this.mTickLabelIndex.intValue()];                
                colArray[3] = columns[this.mTickLabelHolderIndex.intValue()];                
            } else {
                colArray = new SGDataColumn[2];
            }
        }
        colArray[0] = columns[this.mXIndex.intValue()];
        colArray[1] = columns[this.mYIndex.intValue()];
        return colArray;
    }
    
    @Override
    public SGDataColumn[] getUsedDataColumnsClone() {
        SGDataColumn[] columns = this.mDataFile.mDataColumns;
        String[] columnTypes = this.getCurrentColumnType();
        List<SGDataColumn> collist = new ArrayList<SGDataColumn>();
        
        int index = this.mXIndex.intValue();
        SGDataColumn col = (SGDataColumn)columns[index].clone();
        col.setColumnType(columnTypes[index]);
        collist.add(col);
        
        index = this.mYIndex.intValue();
        col = (SGDataColumn)columns[index].clone();
        col.setColumnType(columnTypes[index]);
        collist.add(col);
        
        if (this.isErrorBarAvailable()) {
            if (this.mLowerErrorIndex.equals(this.mUpperErrorIndex)) {
                index = this.mLowerErrorIndex.intValue();
                col = (SGDataColumn)columns[index].clone();
                col.setColumnType(columnTypes[index]);
                collist.add(col);
            } else {
                index = this.mLowerErrorIndex.intValue();
                col = (SGDataColumn)columns[index].clone();
                col.setColumnType(columnTypes[index]);
                collist.add(col);
                index = this.mUpperErrorIndex.intValue();
                col = (SGDataColumn)columns[index].clone();
                col.setColumnType(columnTypes[index]);
                collist.add(col);
                
            }
        }
        if (this.isTickLabelAvailable()) {
            index = this.mTickLabelIndex.intValue();
            col = (SGDataColumn)columns[index].clone();
            col.setColumnType(columnTypes[index]);
            collist.add(col);
        }
        
        return collist.toArray(new SGDataColumn[collist.size()]);
    }
    
    /**
     * Returns an array of values to save to a data set.
     * 
     * @return an array of values to save to a data set
     */
    protected Object[][] getDataSetValues() {
        
        final boolean bError = this.isErrorBarAvailable();
        final boolean bTick = this.isTickLabelAvailable();
        final int dataNum = this.getPointsNumber();
        SGDataColumn[] colArray = this.getDataSetColumns();
        final int colNum = colArray.length;
        
        // get string array from the number column
        SGNumberDataColumn numCol = null;
        String[] numStrArray = null;
        if (bTick) {
            SGDataColumn col = null;
            if (bError) {
                col = colArray[4];
            } else {
                col = colArray[2];
            }
            if (col instanceof SGNumberDataColumn) {
                numCol = (SGNumberDataColumn) col;
                numStrArray = numCol.getStringArray(this.mDecimalPlaces, this.mExponent);
            }
        }
        
        Object[][] array = new Object[dataNum][];
        for (int ii = 0; ii < dataNum; ii++) {
            array[ii] = new Object[colNum];
            for (int jj = 0; jj < colNum; jj++) {
                SGDataColumn col = colArray[jj];
                Object value = col.getExportedValue(ii);;
                if (bError) {
                    if (jj == 2) {
                        // lower error value
                        Double d = (Double) value;
                        value = Double.valueOf(- Math.abs(d.doubleValue()));
                    } else if (jj == 3) {
                        // upper error value
                        Double d = (Double) value;
                        value = Double.valueOf(Math.abs(d.doubleValue()));
                    }
                }
                if (bTick) {
                    // tick label
                    if ((bError && jj == 4) || (!bError && jj == 2)) {
                        // tick labels are from the number column
                        if (numStrArray != null) {
                            value = numStrArray[ii];
                        }
                    }
                }
                array[ii][jj] = value;
            }
        }
        return array;
    }

    /**
     * Sets the decimal places for the tick labels.
     * 
     * @param dp
     *          a value to set to the decimal places
     */
    public void setDecimalPlaces(final int dp) {
        if (dp < 0) {
            throw new IllegalArgumentException("Decimal places must not be negative: " + dp);
        }
        this.mDecimalPlaces = dp;
    }

    /**
     * Sets the exponent for the tick labels.
     * 
     * @param exp
     *           a value to set to the exponent
     */
    public void setExponent(final int exp) {
        this.mExponent = exp;
    }

    /**
     * Returns the decimal places for the tick labels.
     * 
     * @return the decimal places for the tick labels
     */
    public int getDecimalPlaces() {
        return this.mDecimalPlaces;
    }

    /**
     * Returns the exponent for tick labels.
     * 
     * @return the exponent for tick labels
     */
    public int getExponent() {
        return this.mExponent;
    }

    /**
	 * Returns whether the error bars are vertical. If this data does not have
	 * error values, returns null.
	 * 
	 * @return true if error bars are vertical, false if they are horizontal and
	 *         null if this data do not have error values
	 */
    public Boolean isErrorBarVertical() {
    	if (this.isErrorBarAvailable()) {
    		return this.mYIndex.equals(this.mErrorBarHolderIndex);
    	} else {
    		return null;
    	}
    }
    
    /**
	 * Returns whether the tick labels align horizontally. If this data does not have
	 * tick labels, returns null.
	 * 
	 * @return true if tick labels align horizontally, false if they do not so and
	 *         null if this data do not have tick labels
	 */
    public Boolean isTickLabelHorizontal() {
    	if (this.isTickLabelAvailable()) {
    		return this.mYIndex.equals(this.mTickLabelHolderIndex);
    	} else if (this.isDateColumnUsed()) {
    		SGDataColumn[] columns = this.mDataFile.mDataColumns;
    		if (columns[this.mXIndex] instanceof SGDateDataColumn) {
    			return Boolean.TRUE;
    		} else if (columns[this.mYIndex] instanceof SGDateDataColumn) {
    			return Boolean.FALSE;
    		}
    	}
		return null;
    }

    /**
     * Transforms this data to a multiple type data.
     * 
     * @return a transformed data
     */
    public SGISXYTypeMultipleData toMultiple() {
        final boolean be = this.isErrorBarAvailable();
        final boolean bt = this.isTickLabelAvailable();
        Integer[] x = new Integer[] { this.mXIndex };
        Integer[] y = new Integer[] { this.mYIndex };
        Integer[] le = be ? new Integer[] { this.mLowerErrorIndex } : null;
        Integer[] ue = be ? new Integer[] { this.mUpperErrorIndex } : null;
        Integer[] eh = be ? new Integer[] { this.mErrorBarHolderIndex } : null;
        Integer[] t = bt ? new Integer[] { this.mTickLabelIndex } : null;
        Integer[] th = bt ? new Integer[] { this.mTickLabelHolderIndex } : null;
        SGSXYMultipleData data = new SGSXYMultipleData(this.mDataFile,
                this.mObserver, x, y, le, ue, eh, t, th);
        return data;
    }

	public SGDataColumn getXColumn() {
		return this.getColumn(this.mXIndex);
	}

	public SGDataColumn getYColumn() {
		return this.getColumn(this.mYIndex);
	}

	public SGDataColumn getLowerErrorColumn() {
		return this.getColumn(this.mLowerErrorIndex);
	}
	
	public SGDataColumn getUpperErrorColumn() {
		return this.getColumn(this.mUpperErrorIndex);
	}

	public SGDataColumn getErrorBarHolderColumn() {
		return this.getColumn(this.mErrorBarHolderIndex);
	}

	public SGDataColumn getTickLableColumn() {
		return this.getColumn(this.mTickLabelIndex);
	}

	public SGDataColumn getTickLabelHolderColumn() {
		return this.getColumn(this.mTickLabelHolderIndex);
	}

    /**
     * Returns a map which has data information.
     * 
     * @return a map which has data information
     */
    public Map<String, Object> getInfoMap() {
    	Map<String, Object> infoMap = super.getInfoMap();
    	infoMap.put(KEY_MULTIPLE, Boolean.FALSE);
    	return infoMap;
    }

}
