package jp.riken.brain.ni.samuraigraph.data;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementGraph;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;

import org.w3c.dom.Element;

/**
 * Data with multiple scalar type XY data.
 * 
 */
public class SGSXYMultipleData extends SGArrayData implements
        SGISXYTypeMultipleData, SGIDataColumnTypeConstants, SGIDataColumnNameConstants {

    /**
     * An array of column indices for x-values.
     */
    protected Integer[] mXIndices = null;
    
    /**
     * An array of column indices for y-values.
     */
    protected Integer[] mYIndices = null;
    
    /**
     * An array of column indices for lower error values.
     */
    protected Integer[] mLowerErrorIndices = null;
    
    /**
     * An array of column indices for upper error values.
     */
    protected Integer[] mUpperErrorIndices = null;

    /**
     * An array of column indices for error bar holders.
     */
    protected Integer[] mErrorBarHolderIndices = null;

    /**
     * An array of column indices for tick labels.
     */
    protected Integer[] mTickLabelIndices = null;

    /**
     * An array of column indices for tick label holders.
     */
    protected Integer[] mTickLabelHolderIndices = null;

    /**
     * The decimal places for the tick labels.
     */
    protected int mDecimalPlaces = 0;

    /**
     * The exponent for the tick labels.
     */
    protected int mExponent = 0;

    /**
     * The default constructor.
     */
    public SGSXYMultipleData() {
        super();
    }

    /**
     * Build a data with given columns.
     * 
     * @param dataFile
     *           the text data file
     * @param obs
     *           an observer for array data
     */
    protected SGSXYMultipleData(final SGTextDataFile dataFile, SGDataSourceObserver obs) {
        super(dataFile, obs);
    }
    
    /**
     * Build a data with given columns and indices.
     * 
     * @param dataFile
     *                   the text data file
     * @param obs
     *                   an observer for array data
     * @param xIndices
     *                   an array of the indices for x-values
     * @param yIndices
     *                   an array of the indices for y-values
     */
    public SGSXYMultipleData(final SGTextDataFile dataFile, 
            final SGDataSourceObserver obs,
            final Integer[] xIndices, final Integer[] yIndices) {
        
        super(dataFile, obs);

        // check indices
        if (!this.isValidXYValueIndices(xIndices, yIndices)) {
            throw new IllegalArgumentException("Column indices for x or y values are invalid.");
        }
        
        // set indices
        this.mXIndices = (Integer[]) xIndices.clone();
        this.mYIndices = (Integer[]) yIndices.clone();
    }
    
    /**
     * Build a data with given columns and indices.
     * 
     * @param dataFile
     *                   the text data file
     * @param obs
     *                   an observer for array data
     * @param xIndices
     *                   an array of the indices for x-values
     * @param yIndices
     *                   an array of the indices for y-values
     * @param lowerErrorIndices
     *                   an array of the indices for lower error values
     * @param upperErrorIndices
     *                   an array of the indices for upper error values
     * @param errorBarHolderIndices
     *                   an array of the indices of columns those hold error bars
     * @param tickLabelIndices
     *                   an array of the indices for tick labels
     * @param tickLabelHolderIndices
     *                   an array of the indices of columns those hold tick labels
     */
    public SGSXYMultipleData(final SGTextDataFile dataFile,
    		final SGDataSourceObserver obs,
            final Integer[] xIndices, final Integer[] yIndices,
            final Integer[] lowerErrorIndices, final Integer[] upperErrorIndices,
            final Integer[] errorBarHolderIndices, final Integer[] tickLabelIndices,
            final Integer[] tickLabelHolderIndices) {
    	this(dataFile, obs, xIndices, yIndices);
    	
    	// check the input for error bars
    	Integer[] multipleIndices = (xIndices.length == 1) ? yIndices : xIndices;
    	if (!this.isValidErrorBarIndices(multipleIndices, lowerErrorIndices, upperErrorIndices,
    			errorBarHolderIndices)) {
            throw new IllegalArgumentException("Column indices for the error bars are invalid.");
    	}
    	if (!this.isValidTickLabelIndices(multipleIndices, tickLabelIndices,
    			tickLabelHolderIndices)) {
            throw new IllegalArgumentException("Column indices for the tick labels are invalid.");
    	}
    	
        // set indices
    	if (lowerErrorIndices != null) {
    		if (lowerErrorIndices.length != 0) {
                this.mLowerErrorIndices = (Integer[]) lowerErrorIndices.clone();
    		}
    	}
    	if (upperErrorIndices != null) {
    		if (upperErrorIndices.length != 0) {
                this.mUpperErrorIndices = (Integer[]) upperErrorIndices.clone();
    		}
    	}
    	if (errorBarHolderIndices != null) {
    		if (errorBarHolderIndices.length != 0) {
        		this.mErrorBarHolderIndices = (Integer[]) errorBarHolderIndices.clone();
    		}
    	}
    	if (tickLabelIndices != null) {
    		if (tickLabelIndices.length != 0) {
                this.mTickLabelIndices = (Integer[]) tickLabelIndices.clone();
    		}
    	}
    	if (tickLabelHolderIndices != null) {
    		if (tickLabelHolderIndices.length != 0) {
        		this.mTickLabelHolderIndices = (Integer[]) tickLabelHolderIndices.clone();
    		}
    	}
    	
    	// set tick label indices for date type columns
        if (tickLabelIndices == null || tickLabelIndices.length == 0) {
        	this.updateDateTickLabelColumns();
        }
    }
    
    private void updateDateTickLabelColumns() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
    	SGDataColumn xColumn = columns[this.mXIndices[0].intValue()];
    	SGDataColumn yColumn = columns[this.mYIndices[0].intValue()];
    	if (xColumn instanceof SGDateDataColumn) {
    		this.mTickLabelHolderIndices = this.mYIndices;
    		this.mTickLabelIndices = new Integer[this.mYIndices.length];
    		for (int ii = 0; ii < this.mTickLabelIndices.length; ii++) {
    			this.mTickLabelIndices[ii] = this.mXIndices[0];
    		}
    	} else if (yColumn instanceof SGDateDataColumn) {
    		this.mTickLabelHolderIndices = this.mXIndices;
    		this.mTickLabelIndices = new Integer[this.mXIndices.length];
    		for (int ii = 0; ii < this.mTickLabelIndices.length; ii++) {
    			this.mTickLabelIndices[ii] = this.mYIndices[0];
    		}
    	}
    }
    
    protected boolean isValidIndices(final Integer[] indices) {
    	if (indices == null) {
    		return false;
    	}
        for (int ii = 0; ii < indices.length; ii++) {
            if (indices[ii] == null) {
            	return false;
            }
            if (this.checkColumnIndexRange(indices[ii]) == false) {
            	return false;
            }
        }
        return true;
    }

    protected boolean isValidXYValueIndices(final Integer[] xIndices, 
    		final Integer[] yIndices) {
        if (xIndices == null || yIndices == null) {
        	return false;
        }
        if (xIndices.length == 0 || yIndices.length == 0) {
        	return false;
        }
        if (xIndices.length != 1 && yIndices.length != 1) {
        	return false;
        }
        if (!this.isValidIndices(xIndices)) {
        	return false;
        }
        if (!this.isValidIndices(yIndices)) {
        	return false;
        }
        return true;
    }

    protected boolean isValidErrorBarIndices(
    		final Integer[] multipleIndices,
    		final Integer[] lowerErrorIndices, 
    		final Integer[] upperErrorIndices,
    		final Integer[] errorBarHolderIndices) {
    	
        if (lowerErrorIndices != null && upperErrorIndices != null) {
        	if (!this.isValidIndices(lowerErrorIndices)) {
        		return false;
        	}
        	if (!this.isValidIndices(upperErrorIndices)) {
        		return false;
        	}
        	if (lowerErrorIndices.length != upperErrorIndices.length) {
        		return false;
        	}
        	if (errorBarHolderIndices != null) {
            	if (!this.isValidIndices(errorBarHolderIndices)) {
            		return false;
            	}
                if (errorBarHolderIndices.length != lowerErrorIndices.length) {
                	return false;
                }
        	}
        }
        return true;
    }

    protected boolean isValidTickLabelIndices(
    		final Integer[] multipleIndices,
    		final Integer[] tickLabelIndices, 
    		final Integer[] tickLabelHolderIndices) {
    	
        if (tickLabelIndices != null) {
        	if (!this.isValidIndices(tickLabelIndices)) {
        		return false;
        	}
        	if (tickLabelHolderIndices != null) {
            	if (!this.isValidIndices(tickLabelHolderIndices)) {
            		return false;
            	}
                if (tickLabelHolderIndices.length != tickLabelIndices.length) {
                	return false;
                }
        	}
        }
        return true;
    }

    protected boolean hasMultipleYValues() {
    	if (this.hasPairOfXYValues()) {
    		if (this.isErrorBarAvailable()) {
    			return SGUtility.contains(this.mErrorBarHolderIndices, this.mYIndices);
    		} else if (this.isTickLabelAvailable()) {
    			return SGUtility.contains(this.mTickLabelHolderIndices, this.mYIndices);
    		} else {
    			return true;
    		}
    	} else {
        	return (this.mXIndices.length == 1);
    	}
    }

    protected boolean hasPairOfXYValues() {
    	return (this.mXIndices.length == 1 && this.mYIndices.length == 1);
    }
    
    protected double[][] getMultipleValues() {
    	return (this.hasMultipleYValues()) ? this.getYArray() : this.getXArray();
    }

    /**
     * Returns a number of data points.
     * 
     * @return
     *         a number of data points
     */
    public int getPointsNumber() {
        return this.getXArray()[0].length;
    }

    /**
     * Returns the type of data.
     * 
     * @return
     *         the type of data
     */
    public String getDataType() {
        return SGDataTypeConstants.SXY_MULTIPLE_DATA;
    }

    /**
	 * Returns the m-th x value at the n-th index.
	 * 
	 * @param m
	 *            the index of data
	 * @param n
	 *            the index of the data point
	 * @return the n-th x-value of the m-th data
	 */
	public Double getXValue(final int m, final int n) {
		final int len = this.getPointsNumber();
		if (n < 0 || n >= len) {
			throw new IllegalArgumentException("given index is out of bounds: "
					+ n);
		}
		if (m < 0 || m >= this.mXIndices.length) {
			throw new IllegalArgumentException("given index is out of bounds: "
					+ m);
		}
		return Double.valueOf(this.getXArray(m)[n]);
	}

	/**
	 * Returns the m-th y value at the n-th index.
	 * 
	 * @param m
	 *            the index of data
	 * @param n
	 *            the index of the data point
	 * @return the n-th y-value of the m-th data
	 */
	public Double getYValue(final int m, final int n) {
		final int len = this.getPointsNumber();
		if (n < 0 || n >= len) {
			throw new IllegalArgumentException("given index is out of bounds: "
					+ n);
		}
		if (m < 0 || m >= this.mYIndices.length) {
			throw new IllegalArgumentException("given index is out of bounds: "
					+ m);
		}
		return Double.valueOf(this.getYArray(m)[n]);
	}

    /**
     * Returns an array of SXYData objects.
     * 
     * @return an array of SXYData objects
     */
    public SGISXYTypeData[] getSXYDataArray() {
        final boolean by = this.hasMultipleYValues();
        final boolean be = this.isErrorBarAvailable();
        final boolean bt = this.isTickLabelAvailable();
        final int dataNum = by ? this.getYNumber() : this.getXNumber();
        final SGISXYTypeData[] dataArray = new SGISXYTypeData[dataNum];
        Integer xIndex = null;
        Integer yIndex = null;
        Integer[] lArray = null;
        Integer[] uArray = null;
        Integer[] ehArray = null;
        Integer[] tArray = null;
        Integer[] thArray = null;
        if (by) {
            xIndex = this.mXIndices[0];
        } else {
            yIndex = this.mYIndices[0];
        }
        if (be) {
            lArray = new Integer[dataNum];
            uArray = new Integer[dataNum];
            ehArray = new Integer[dataNum];
            for (int ii = 0; ii < dataArray.length; ii++) {
                lArray[ii] = null;
                uArray[ii] = null;
                ehArray[ii] = null;
                for (int jj = 0; jj < this.mErrorBarHolderIndices.length; jj++) {
                    final boolean b = by ? this.mErrorBarHolderIndices[jj].equals(this.mYIndices[ii])
                            : this.mErrorBarHolderIndices[jj].equals(this.mXIndices[ii]);
                    if (b) {
                        lArray[ii] = this.mLowerErrorIndices[jj];
                        uArray[ii] = this.mUpperErrorIndices[jj];
                        ehArray[ii] = this.mErrorBarHolderIndices[jj];
                        break;
                    }
                }
            }
        }
        if (bt) {
            tArray = new Integer[dataNum];
            thArray = new Integer[dataNum];
            for (int ii = 0; ii < dataArray.length; ii++) {
                tArray[ii] = null;
                thArray[ii] = null;
                for (int jj = 0; jj < this.mTickLabelHolderIndices.length; jj++) {
                    final boolean b = by ? this.mTickLabelHolderIndices[jj].equals(this.mYIndices[ii])
                            : this.mTickLabelHolderIndices[jj].equals(this.mXIndices[ii]);
                    if (b) {
                        tArray[ii] = this.mTickLabelIndices[jj];
                        thArray[ii] = this.mTickLabelHolderIndices[jj];
                        break;
                    }
                }
            }
        }
        for (int ii = 0; ii < dataArray.length; ii++) {
            Integer x = by ? xIndex : this.mXIndices[ii];
            Integer y = by ? this.mYIndices[ii] : yIndex;
            Integer le = be ? lArray[ii] : null;
            Integer ue = be ? uArray[ii] : null;
            Integer eh = be ? ehArray[ii] : null;
            Integer t = bt ? tArray[ii] : null;
            Integer th = bt ? thArray[ii] : null;
            dataArray[ii] = new SGSXYData(
                    this.mDataFile, this.mObserver, x, y,
                    le, ue, eh, t, th);
        }

        // set common values to exponent and decimal places
        for (int ii = 0; ii < dataArray.length; ii++) {
            dataArray[ii].setDecimalPlaces(this.mDecimalPlaces);
            dataArray[ii].setExponent(this.mExponent);
        }
        
        return dataArray;
    }

    /**
     * Returns an array of multiple SXYData objects.
     * 
     * @return an array of multiple SXYData objects
     */
    @Override
    public SGISXYTypeMultipleData[] getSXYTypeMultipleDataArray() {
        SGISXYTypeData[] array = this.getSXYDataArray();
        SGISXYTypeMultipleData[] ret = new SGISXYTypeMultipleData[array.length];
        for (int ii = 0; ii < array.length; ii++) {
            SGSXYData data = (SGSXYData) array[ii];
            ret[ii] = data.toMultiple();
            
            // disposes data
            data.dispose();
        }
        return ret;
    }

    /**
	 * Returns a copy of x-value array.
	 * 
	 * @return x-value array
	 */
    public double[][] getXValueArray() {
    	return this.getXArray();
    }

    /**
	 * Returns a copy of y-value array.
	 * 
	 * @return y-value array
	 */
    public double[][] getYValueArray() {
    	return this.getYArray();
    }

    /**
	 * Returns the bounds of x-values.
	 * 
	 * @return the bounds of x-values
	 */
    public SGValueRange getBoundsX() {
        return SGDataUtility.getBoundsX(this);
    }

    /**
     * Returns the bounds of y-values.
     * 
     * @return the bounds of y-values
     */
    public SGValueRange getBoundsY() {
        return SGDataUtility.getBoundsY(this);
    }

    /**
     * Set values with a given data.
     * 
     * @param data
     *             data to be copied
     * @return true if succeeded
     */
    public boolean setData(final SGData data) {        
        if (!(data instanceof SGSXYMultipleData)) {
            return false;
        }
        if (super.setData(data) == false) {
            return false;
        }
        SGSXYMultipleData dataMult = (SGSXYMultipleData) data;
        if (dataMult.mXIndices != null) {
            this.mXIndices = SGUtility.copyIntegerArray(dataMult.mXIndices);
        }
        if (dataMult.mYIndices != null) {
            this.mYIndices = SGUtility.copyIntegerArray(dataMult.mYIndices);
        }
        if (dataMult.mLowerErrorIndices != null) {
            this.mLowerErrorIndices = SGUtility.copyIntegerArray(dataMult.mLowerErrorIndices);
        }
        if (dataMult.mUpperErrorIndices != null) {
            this.mUpperErrorIndices = SGUtility.copyIntegerArray(dataMult.mUpperErrorIndices);
        }
        if (dataMult.mErrorBarHolderIndices != null) {
            this.mErrorBarHolderIndices = SGUtility.copyIntegerArray(dataMult.mErrorBarHolderIndices);
        }
        if (dataMult.mTickLabelIndices != null) {
            this.mTickLabelIndices = SGUtility.copyIntegerArray(dataMult.mTickLabelIndices);
        }
        if (dataMult.mTickLabelHolderIndices != null) {
            this.mTickLabelHolderIndices = SGUtility.copyIntegerArray(dataMult.mTickLabelHolderIndices);
        }
        
        return true;
    }

    /**
     * Returns the copy of this data object.
     * 
     * @return copy of this data object
     */
    public Object copy() {
        SGSXYMultipleData data = (SGSXYMultipleData) super.copy();
        if (this.mXIndices != null) {
            data.mXIndices = SGUtility.copyIntegerArray(this.mXIndices);
        }
        if (this.mYIndices != null) {
            data.mYIndices = SGUtility.copyIntegerArray(this.mYIndices);
        }
        if (this.mLowerErrorIndices != null) {
            data.mLowerErrorIndices = SGUtility.copyIntegerArray(this.mLowerErrorIndices);
        }
        if (this.mUpperErrorIndices != null) {
            data.mUpperErrorIndices = SGUtility.copyIntegerArray(this.mUpperErrorIndices);
        }
        if (this.mErrorBarHolderIndices != null) {
            data.mErrorBarHolderIndices = SGUtility.copyIntegerArray(this.mErrorBarHolderIndices);
        }
        if (this.mTickLabelIndices != null) {
            data.mTickLabelIndices = SGUtility.copyIntegerArray(this.mTickLabelIndices);
        }
        if (this.mTickLabelHolderIndices != null) {
            data.mTickLabelHolderIndices = SGUtility.copyIntegerArray(this.mTickLabelHolderIndices);
        }
        return data;
    }

    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
        this.mXIndices = null;
        this.mYIndices = null;
        this.mLowerErrorIndices = null;
        this.mUpperErrorIndices = null;
        this.mErrorBarHolderIndices = null;
        this.mTickLabelIndices = null;
        this.mTickLabelHolderIndices = null;
    }

//    /**
//     * Write this data to a given output stream.
//     * @param w
//     *          an output stream
//     * @return
//     *          true if succeeded
//     */
//    protected boolean writeData(Writer fw) throws IOException {
//        final int num = this.getPointsNumber();
//        final double[] xArray = this.getXArray();
//        final double[][] yArray = this.getYArray();
//        for (int ii = 0; ii < num; ii++) {
//            StringBuffer sb = new StringBuffer();
//            sb.append(xArray[ii]);
//            for (int jj = 0; jj < yArray.length; jj++) {
//        	sb.append(',');
//        	sb.append(yArray[jj][ii]);
//            }
//            sb.append('\n');
//            fw.write(sb.toString());
//        }
//        return true;
//    }

    /**
     * Returns an array of x-values.
     * 
     * @return
     *         an array of x-values
     */
    protected double[][] getXArray() {
    	return this.getValueArray(this.mXIndices);
    }

    /**
     * Returns an array of y-values.
     * 
     * @return
     *         an array of y-values
     */
    protected double[][] getYArray() {
    	return this.getValueArray(this.mYIndices);
    }
        
    protected double[][] getValueArray(Integer[] indices) {
        SGDataColumn[] columns = this.mDataFile.mDataColumns;
    	double[][] array = new double[indices.length][];
    	for (int ii = 0; ii < array.length; ii++) {
    		SGINumberDataColumn col = (SGINumberDataColumn) columns[indices[ii].intValue()];
            array[ii] = col.getNumberArray();
    	}
    	return array;
    }
    
    /**
     * Returns an array of the m-th x-values.
     * @param m
     *         the index of x-values
     * @return
     *         an array of the m-th x-values
     */
    protected double[] getXArray(final int m) {
    	if (m < 0 || m >= this.mXIndices.length) {
    	    throw new IllegalArgumentException(m + ": m < 0 || m >= this.mXIndices.length");
    	}
        SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGINumberDataColumn col = (SGINumberDataColumn) columns[this.mXIndices[m].intValue()];
        return col.getNumberArray();
    }

    /**
     * Returns an array of the m-th y-values.
     * @param m
     *         the index of y-values
     * @return
     *         an array of the m-th y-values
     */
    protected double[] getYArray(final int m) {
    	if (m < 0 || m >= this.mYIndices.length) {
    	    throw new IllegalArgumentException(m + ": m < 0 || m >= this.mYIndices.length");
    	}
        SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGINumberDataColumn col = (SGINumberDataColumn) columns[this.mYIndices[m].intValue()];
        return col.getNumberArray();
    }
    
    /**
     * Returns the number of arrays of x-values.
     * 
     * @return
     *         the number of arrays of x-values
     */
    public int getXNumber() {
        return this.mXIndices.length;
    }

    /**
     * Returns the number of arrays of y-values.
     * 
     * @return
     *         the number of arrays of y-values
     */
    public int getYNumber() {
        return this.mYIndices.length;
    }

    /**
     * Writes data column indices in the attributes.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected boolean writeAttributeColumnIndices(Element el) {
//        if (this.hasMultipleYValues()) {
////            el.setAttribute(KEY_X_VALUE_COLUMN_INDICES, this.mXIndices[0].toString());
//            if (this.writeAttributeColumnIndicesSub(el, KEY_Y_VALUE_COLUMN_INDICES,
//                    this.mYIndices) == false) {
//                return false;
//            }
//        } else {
//            el.setAttribute(KEY_Y_VALUE_COLUMN_INDICES, this.mYIndices[0].toString());
//            if (this.writeAttributeColumnIndicesSub(el, KEY_X_VALUE_COLUMN_INDICES,
//                    this.mXIndices) == false) {
//                return false;
//            }
//        }
        if (this.writeAttributeColumnIndicesSub(el, KEY_X_VALUE_COLUMN_INDICES,
                this.mXIndices) == false) {
            return false;
        }
        if (this.writeAttributeColumnIndicesSub(el, KEY_Y_VALUE_COLUMN_INDICES,
                this.mYIndices) == false) {
            return false;
        }
		if (this.mLowerErrorIndices != null && this.mUpperErrorIndices != null
				&& this.mErrorBarHolderIndices != null) {
			if (this.writeAttributeColumnIndicesSub(el,
					KEY_LOWER_ERROR_BAR_COLUMN_INDICES, this.mLowerErrorIndices) == false) {
				return false;
			}
			if (this.writeAttributeColumnIndicesSub(el,
					KEY_UPPER_ERROR_BAR_COLUMN_INDICES, this.mUpperErrorIndices) == false) {
				return false;
			}
			if (this.writeAttributeColumnIndicesSub(el,
					KEY_ERROR_BAR_HOLDER_COLUMN_INDICES,
					this.mErrorBarHolderIndices) == false) {
				return false;
			}
		}
		if (this.mTickLabelIndices != null
				&& this.mTickLabelHolderIndices != null) {
			SGDataColumn[] columns = this.mDataFile.getDataColumns();
			List<Integer> tlIndicesList = new ArrayList<Integer>();
			List<Integer> thIndicesList = new ArrayList<Integer>();
			for (int ii = 0; ii < this.mTickLabelIndices.length; ii++) {
				if (!(columns[this.mTickLabelIndices[ii].intValue()] instanceof SGDateDataColumn)) {
					tlIndicesList.add(this.mTickLabelIndices[ii]);
					thIndicesList.add(this.mTickLabelHolderIndices[ii]);
				}
			}
			if (tlIndicesList.size() != 0) {
				Integer[] tlIndices = new Integer[tlIndicesList.size()];
				tlIndicesList.toArray(tlIndices);
				Integer[] thIndices = new Integer[thIndicesList.size()];
				thIndicesList.toArray(thIndices);
				if (this.writeAttributeColumnIndicesSub(el,
						KEY_TICK_LABEL_COLUMN_INDICES, tlIndices) == false) {
					return false;
				}
				if (this.writeAttributeColumnIndicesSub(el,
						KEY_TICK_LABEL_HOLDER_COLUMN_INDICES, thIndices) == false) {
					return false;
				}
			}
		}
        return true;
    }

    private boolean writeAttributeColumnIndicesSub(Element el, String key,
			Integer[] indices) {
		int[] array = new int[indices.length];
		for (int ii = 0; ii < array.length; ii++) {
			array[ii] = indices[ii].intValue();
		}
		return this.writeAttributeColumnIndicesSub(el, key, array);
	}

	private boolean writeAttributeColumnIndicesSub(Element el, String key,
			int[] indices) {
		StringBuffer sb = new StringBuffer();
		sb.append('{');
		for (int ii = 0; ii < indices.length; ii++) {
			sb.append(indices[ii]);
			if (ii != indices.length - 1) {
				sb.append(',');
			}
		}
		sb.append('}');
		el.setAttribute(key, sb.toString());
		return true;
	}

    /**
	 * Writes data column indices as sequential numbers.
	 * 
	 * @param el
	 *            the Element object
	 * @return true if succeeded
	 */
    protected boolean writeSequentialColumnIndices(Element el) {
    	int[] xIndices, yIndices, hIndices;
    	if (this.hasMultipleYValues()) {
    		xIndices = new int[] { 0 };
    		yIndices = new int[this.mYIndices.length];
    		for (int ii = 0; ii < yIndices.length; ii++) {
    			yIndices[ii] = ii + 1;
    		}
    		hIndices = yIndices;
    	} else {
    		xIndices = new int[this.mXIndices.length];
    		for (int ii = 0; ii < xIndices.length; ii++) {
    			xIndices[ii] = ii;
    		}
    		yIndices = new int[] { xIndices.length };
    		hIndices = xIndices;
    	}
    	
    	// write indices
        if (!this.writeAttributeColumnIndicesSub(el, KEY_X_VALUE_COLUMN_INDICES, xIndices)) {
        	return false;
        }
        if (!this.writeAttributeColumnIndicesSub(el, KEY_Y_VALUE_COLUMN_INDICES, yIndices)) {
        	return false;
        }
        
        int offset = xIndices.length + yIndices.length;
        if (this.isErrorBarAvailable()) {
        	// error bars
        	
        	int[] lIndices = new int[this.mLowerErrorIndices.length];
    		for (int ii = 0; ii < lIndices.length; ii++) {
    			lIndices[ii] = ii + offset;
    		}
            if (!this.writeAttributeColumnIndicesSub(el, KEY_LOWER_ERROR_BAR_COLUMN_INDICES, 
            		lIndices)) {
            	return false;
            }
    		offset += lIndices.length;
    		
        	int[] uIndices = new int[this.mUpperErrorIndices.length];
    		for (int ii = 0; ii < uIndices.length; ii++) {
    			uIndices[ii] = ii + offset;
    		}
            if (!this.writeAttributeColumnIndicesSub(el, KEY_UPPER_ERROR_BAR_COLUMN_INDICES, 
            		uIndices)) {
            	return false;
            }
    		offset += uIndices.length;

        	int[] ehIndices = new int[this.mErrorBarHolderIndices.length];
    		for (int ii = 0; ii < ehIndices.length; ii++) {
    			ehIndices[ii] = hIndices[ii];
    		}
            if (!this.writeAttributeColumnIndicesSub(el, KEY_ERROR_BAR_HOLDER_COLUMN_INDICES, 
            		ehIndices)) {
            	return false;
            }
        }
        
        if (this.isTickLabelAvailable() && !this.isTickLabelFromDataColumn()) {
        	// tick labels
        	
        	int[] tIndices = new int[this.mTickLabelIndices.length];
    		for (int ii = 0; ii < tIndices.length; ii++) {
    			tIndices[ii] = ii + offset;
    		}
            if (!this.writeAttributeColumnIndicesSub(el, KEY_TICK_LABEL_COLUMN_INDICES, 
            		tIndices)) {
            	return false;
            }

        	int[] thIndices = new int[this.mTickLabelHolderIndices.length];
    		for (int ii = 0; ii < thIndices.length; ii++) {
    			thIndices[ii] = hIndices[ii];
    		}
            if (!this.writeAttributeColumnIndicesSub(el, KEY_TICK_LABEL_HOLDER_COLUMN_INDICES, 
            		thIndices)) {
            	return false;
            }
        }
    	
        return true;
    }
    
    private boolean writeAttributeColumnNameSub(Element el, String key, int[] indices) {
        StringBuffer sb = new StringBuffer();
        sb.append('{');
        for (int ii = 0; ii < indices.length; ii++) {
            sb.append(this.getSequentialColumnName(indices[ii]));
            if (ii != indices.length - 1) {
                sb.append(',');
            }
        }
        sb.append('}');
        el.setAttribute(key, sb.toString());
        return true;
    }
    
    @Override
    protected boolean writeSequentialColumnName(Element el) {
        int[] xIndices, yIndices, hIndices;
        if (this.hasMultipleYValues()) {
            xIndices = new int[] { 0 };
            yIndices = new int[this.mYIndices.length];
            for (int ii = 0; ii < yIndices.length; ii++) {
                yIndices[ii] = ii + 1;
            }
            hIndices = yIndices;
        } else {
            xIndices = new int[this.mXIndices.length];
            for (int ii = 0; ii < xIndices.length; ii++) {
                xIndices[ii] = ii;
            }
            yIndices = new int[] { xIndices.length };
            hIndices = xIndices;
        }
        
        // write indices
        if (!this.writeAttributeColumnNameSub(el, KEY_X_VALUE_COLUMN_INDICES, xIndices)) {
            return false;
        }
        if (!this.writeAttributeColumnNameSub(el, KEY_Y_VALUE_COLUMN_INDICES, yIndices)) {
            return false;
        }
        
        int offset = xIndices.length + yIndices.length;
        if (this.isErrorBarAvailable()) {
            // error bars
            
            int[] lIndices = new int[this.mLowerErrorIndices.length];
            for (int ii = 0; ii < lIndices.length; ii++) {
                lIndices[ii] = ii + offset;
            }
            if (!this.writeAttributeColumnNameSub(el, KEY_LOWER_ERROR_BAR_COLUMN_INDICES, lIndices)) {
                return false;
            }
            offset += lIndices.length;
            
            int[] uIndices = new int[this.mUpperErrorIndices.length];
            for (int ii = 0; ii < uIndices.length; ii++) {
                uIndices[ii] = ii + offset;
            }
            if (!this.writeAttributeColumnNameSub(el, KEY_UPPER_ERROR_BAR_COLUMN_INDICES, uIndices)) {
                return false;
            }
            offset += uIndices.length;

            int[] ehIndices = new int[this.mErrorBarHolderIndices.length];
            for (int ii = 0; ii < ehIndices.length; ii++) {
                ehIndices[ii] = hIndices[ii];
            }
            if (!this.writeAttributeColumnNameSub(el, KEY_ERROR_BAR_HOLDER_COLUMN_INDICES, ehIndices)) {
                return false;
            }
        }
        
        if (this.isTickLabelAvailable() && !this.isTickLabelFromDataColumn()) {
            // tick labels
            
            int[] tIndices = new int[this.mTickLabelIndices.length];
            for (int ii = 0; ii < tIndices.length; ii++) {
                tIndices[ii] = ii + offset;
            }
            if (!this.writeAttributeColumnNameSub(el, KEY_TICK_LABEL_COLUMN_INDICES, tIndices)) {
                return false;
            }

            int[] thIndices = new int[this.mTickLabelHolderIndices.length];
            for (int ii = 0; ii < thIndices.length; ii++) {
                thIndices[ii] = hIndices[ii];
            }
            if (!this.writeAttributeColumnNameSub(el, KEY_TICK_LABEL_HOLDER_COLUMN_INDICES, thIndices)) {
                return false;
            }
        }

        // serial numbers
        el.setAttribute(KEY_SERIAL_NUMBER_VARIABLE_NAME, SERIAL_NUMBER);
        
        return true;
    }
    
    /**
     * Write sampling rate to the Element.
     * 
     * @param el the Element object
     * @return true
     */
    public boolean writeSamplingRateToProperty(Element el) {
        SGDataColumn[] columns = this.getDataSetColumnsClone();
        for (int i = 0; i < columns.length; i++) {
            if (SGIDataColumnTypeConstants.VALUE_TYPE_SAMPLING_RATE.equals(columns[i].getValueType())) {
                SGSamplingDataColumn scol = (SGSamplingDataColumn)columns[i];
                el.setAttribute(SGIFigureElementGraph.KEY_SAMPLING_RATE, Double.toString(scol.getSamplingRate()));
                break;
            }
        }
        return true;
    }
    
    /**
     * Read properties from a given Element and set to this data.
     * @param el
     *          an Element
     * @return
     *          true if succeeded
     */
    public boolean readProperty(Element el) {
//        Integer index = null;
        Integer[] array = null;
        Integer[] xIndices = null;
        Integer[] yIndices = null;
        if ((array = this.readIndices(el, KEY_X_VALUE_COLUMN_INDICES)) != null) {
        	if (array.length != 0) {
                xIndices = new Integer[array.length];
                for (int ii = 0; ii < array.length; ii++) {
                    if (this.checkColumnIndexRange(array[ii]) == false) {
                        return false;
                    }
                    xIndices[ii] = array[ii];
                }
        	}
        } else {
            return false;
        }
//        if ((index = this.readIndex(el, KEY_X_VALUE_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            xIndices = new Integer[] {index};
//        } else if ((array = this.readIndices(el, KEY_X_VALUE_COLUMN_INDICES)) != null) {
//        	if (array.length != 0) {
//                xIndices = new Integer[array.length];
//                for (int ii = 0; ii < array.length; ii++) {
//                    if (this.checkColumnIndexRange(array[ii]) == false) {
//                        return false;
//                    }
//                    xIndices[ii] = array[ii];
//                }
//        	}
//        } else {
//            return false;
//        }
        
	    if ((array = this.readIndices(el, KEY_Y_VALUE_COLUMN_INDICES)) != null) {
	    	if (array.length != 0) {
	            yIndices = new Integer[array.length];
	            for (int ii = 0; ii < array.length; ii++) {
	                if (this.checkColumnIndexRange(array[ii]) == false) {
	                    return false;
	                }
	                yIndices[ii] = array[ii];
	            }
	    	}
	    } else {
	        return false;
	    }

//        if ((index = this.readIndex(el, KEY_Y_VALUE_COLUMN_INDEX)) != null) {
//            if (this.checkColumnIndexRange(index) == false) {
//                return false;
//            }
//            yIndices = new Integer[] {index};
//        } else if ((array = this.readIndices(el, KEY_Y_VALUE_COLUMN_INDICES)) != null) {
//        	if (array.length != 0) {
//                yIndices = new Integer[array.length];
//                for (int ii = 0; ii < array.length; ii++) {
//                    if (this.checkColumnIndexRange(array[ii]) == false) {
//                        return false;
//                    }
//                    yIndices[ii] = array[ii];
//                }
//        	}
//        } else {
//            return false;
//        }
        
        if (!this.isValidXYValueIndices(xIndices, yIndices)) {
        	return false;
        }

        Integer[] lIndices = null;
        Integer[] uIndices = null;
        Integer[] ehIndices = null;
        Integer[] tIndices = null;
        Integer[] thIndices = null;
	    if ((array = this.readIndices(el, KEY_LOWER_ERROR_BAR_COLUMN_INDICES)) != null) {
	    	if (array.length != 0) {
	            lIndices = new Integer[array.length];
	            for (int ii = 0; ii < array.length; ii++) {
	                if (this.checkColumnIndexRange(array[ii]) == false) {
	                    return false;
	                }
	                lIndices[ii] = array[ii];
	            }
	    	}
	    }
	    if ((array = this.readIndices(el, KEY_UPPER_ERROR_BAR_COLUMN_INDICES)) != null) {
	    	if (array.length != 0) {
	            uIndices = new Integer[array.length];
	            for (int ii = 0; ii < array.length; ii++) {
	                if (this.checkColumnIndexRange(array[ii]) == false) {
	                    return false;
	                }
	                uIndices[ii] = array[ii];
	            }
	    	}
	    }
	    if ((array = this.readIndices(el, KEY_ERROR_BAR_HOLDER_COLUMN_INDICES)) != null) {
	    	if (array.length != 0) {
	            ehIndices = new Integer[array.length];
	            for (int ii = 0; ii < array.length; ii++) {
	                if (this.checkColumnIndexRange(array[ii]) == false) {
	                    return false;
	                }
	                ehIndices[ii] = array[ii];
	            }
	    	}
	    }
	    if ((array = this.readIndices(el, KEY_TICK_LABEL_COLUMN_INDICES)) != null) {
	    	if (array.length != 0) {
	    		tIndices = new Integer[array.length];
	            for (int ii = 0; ii < array.length; ii++) {
	                if (this.checkColumnIndexRange(array[ii]) == false) {
	                    return false;
	                }
	                tIndices[ii] = array[ii];
	            }
	    	}
	    }
	    if ((array = this.readIndices(el, KEY_TICK_LABEL_HOLDER_COLUMN_INDICES)) != null) {
	    	if (array.length != 0) {
	            thIndices = new Integer[array.length];
	            for (int ii = 0; ii < array.length; ii++) {
	                if (this.checkColumnIndexRange(array[ii]) == false) {
	                    return false;
	                }
	                thIndices[ii] = array[ii];
	            }
	    	}
	    }

    	Integer[] multipleIndices = (xIndices.length == 1) ? yIndices : xIndices;
    	if (!this.isValidErrorBarIndices(multipleIndices, lIndices, uIndices, ehIndices)) {
    		return false;
    	}
    	if (!this.isValidTickLabelIndices(multipleIndices, tIndices, thIndices)) {
    		return false;
    	}

        // set to the attributes
        this.mXIndices = xIndices;
        this.mYIndices = yIndices;
        this.mLowerErrorIndices = lIndices;
        this.mUpperErrorIndices = uIndices;
        this.mErrorBarHolderIndices = ehIndices;
        this.mTickLabelIndices = tIndices;
        this.mTickLabelHolderIndices = thIndices;
        
        return true;
    }
    
    /**
     * Returns the title of X values.
     * 
     * @return
     *         the title of X values
     */
    public String getTitleX() {
    	if (this.hasMultipleYValues() || this.hasPairOfXYValues()) {
            SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mXIndices[0].intValue()].getTitle();
    	} else {
    		return "";
    	}
    }
    
    /**
     * Returns the title of Y values.
     * 
     * @return
     *         the title of Y values
     */
    public String getTitleY() {
    	if (!this.hasMultipleYValues() || this.hasPairOfXYValues()) {
            SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mYIndices[0].intValue()].getTitle();
    	} else {
    		return "";
    	}
    }

    public Integer[] getXIndices() {
        if (this.mXIndices == null) {
            return null;
        } else {
            return (Integer[]) mXIndices.clone();
        }
    }

    public void setXIndices(Integer[] indices) {
        if (indices == null) {
            this.mXIndices = null;
        } else {
        	if (!this.isValidIndices(indices)) {
                throw new IllegalArgumentException("Column indices for x values are invalid.");
        	}
            this.mXIndices = (Integer[]) indices.clone();
        }
    }

    public Integer[] getYIndices() {
        if (this.mYIndices == null) {
            return null;
        } else {
            return (Integer[]) mYIndices.clone();
        }
    }

    public void setYIndices(Integer[] indices) {
        if (indices == null) {
            this.mYIndices = null;
        } else {
        	if (!this.isValidIndices(indices)) {
                throw new IllegalArgumentException("Column indices for y values are invalid.");
        	}
            this.mYIndices = (Integer[]) indices.clone();
        }
    }

    public Integer[] getLowerErrorIndices() {
        if (this.mLowerErrorIndices == null) {
            return null;
        } else {
            return (Integer[]) this.mLowerErrorIndices.clone();
        }
    }

    public void setLowerErrorIndices(Integer[] indices) {
        if (indices == null) {
        	this.mLowerErrorIndices = null;
        } else {
            for (int ii = 0; ii < indices.length; ii++) {
                final Integer index = indices[ii];
                if (this.checkColumnIndexRange(index) == false) {
                    throw new IllegalArgumentException("Index out of bounds: " + index);
                }
            }
            this.mLowerErrorIndices = (Integer[]) indices.clone();
        }
    }

    public Integer[] getUpperErrorIndices() {
        if (this.mUpperErrorIndices == null) {
            return null;
        } else {
            return (Integer[]) this.mUpperErrorIndices.clone();
        }
    }

    public void setUpperErrorIndices(Integer[] indices) {
        if (indices == null) {
        	this.mUpperErrorIndices = null;
        } else {
            for (int ii = 0; ii < indices.length; ii++) {
                final Integer index = indices[ii];
                if (this.checkColumnIndexRange(index) == false) {
                    throw new IllegalArgumentException("Index out of bounds: " + index);
                }
            }
            this.mUpperErrorIndices = (Integer[]) indices.clone();
        }
    }

    public Integer[] getErrorBarHolderIndices() {
        if (this.mErrorBarHolderIndices == null) {
            return null;
        } else {
            return (Integer[]) this.mErrorBarHolderIndices.clone();
        }
    }

    public void setErrorBarHolderIndices(Integer[] indices) {
        if (indices == null) {
        	this.mErrorBarHolderIndices = null;
        } else {
            for (int ii = 0; ii < indices.length; ii++) {
                final Integer index = indices[ii];
                if (this.checkColumnIndexRange(index) == false) {
                    throw new IllegalArgumentException("Index out of bounds: " + index);
                }
            }
            this.mErrorBarHolderIndices = (Integer[]) indices.clone();
        }
    }

    public Integer[] getTickLabelIndices() {
        if (this.mTickLabelIndices == null) {
            return null;
        } else {
            return (Integer[]) this.mTickLabelIndices.clone();
        }
    }

    public void setTickLabelIndices(Integer[] indices) {
        if (indices == null) {
        	this.mTickLabelIndices = null;
        } else {
            for (int ii = 0; ii < indices.length; ii++) {
                final Integer index = indices[ii];
                if (this.checkColumnIndexRange(index) == false) {
                    throw new IllegalArgumentException("Index out of bounds: " + index);
                }
            }
            this.mTickLabelIndices = (Integer[]) indices.clone();
        }
    }

    public Integer[] getTickLabelHolderIndices() {
        if (this.mTickLabelHolderIndices == null) {
            return null;
        } else {
            return (Integer[]) this.mTickLabelHolderIndices.clone();
        }
    }

    public void setTickLabelHolderIndices(Integer[] indices) {
        if (indices == null) {
        	this.mTickLabelHolderIndices = null;
        } else {
            for (int ii = 0; ii < indices.length; ii++) {
                final Integer index = indices[ii];
                if (this.checkColumnIndexRange(index) == false) {
                    throw new IllegalArgumentException("Index out of bounds: " + index);
                }
            }
            this.mTickLabelHolderIndices = (Integer[]) indices.clone();
        }
    }

    /**
     * Returns properties of this data.
     * @return
     *         properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new SXYMultipleDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Get properties of this data.
     * @param p
     *          properties to set values
     * @return
     *          true if succeeded
     */
    public boolean getProperties(final SGProperties p) {
        if ((p instanceof SXYMultipleDataProperties) == false) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        SXYMultipleDataProperties ep = (SXYMultipleDataProperties) p;
        ep.mXIndices = this.getXIndices();
        ep.mYIndices = this.getYIndices();
        ep.mLowerErrorIndices = this.getLowerErrorIndices();
        ep.mUpperErrorIndices = this.getUpperErrorIndices();
        ep.mErrorBarHolderIndices = this.getErrorBarHolderIndices();
        ep.mTickLabelIndices = this.getTickLabelIndices();
        ep.mTickLabelHolderIndices = this.getTickLabelHolderIndices();
        return true;
    }
    
    /**
     * Set properties to this data.
     * @param p
     *          properties that have values to set to this data
     * @return
     *          true if succeeded
     */
    public boolean setProperties(final SGProperties p) {
        if ((p instanceof SXYMultipleDataProperties) == false) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        SXYMultipleDataProperties ep = (SXYMultipleDataProperties) p;
        this.setXIndices(ep.mXIndices);
        this.setYIndices(ep.mYIndices);
        this.setLowerErrorIndices(ep.mLowerErrorIndices);
        this.setUpperErrorIndices(ep.mUpperErrorIndices);
        this.setErrorBarHolderIndices(ep.mErrorBarHolderIndices);
        this.setTickLabelIndices(ep.mTickLabelIndices);
        this.setTickLabelHolderIndices(ep.mTickLabelHolderIndices);
        
        return true;
    }

    /**
     * Sets the type of data columns.
     * 
     * @param column
     *           an array of column types
     * @return true if succeeded
     */
    public boolean setColumnType(String[] columns) {
        if (columns == null) {
            throw new IllegalArgumentException("columns == null");
        }
        List<Integer> xList = new ArrayList<Integer>();
        List<Integer> yList = new ArrayList<Integer>();
        List<Integer> lList = new ArrayList<Integer>();
        List<Integer> uList = new ArrayList<Integer>();
        List<Integer> tList = new ArrayList<Integer>();
        for (int ii = 0; ii < columns.length; ii++) {
            if (SGUtilityText.isEqualString(X_VALUE, columns[ii])) {
                xList.add(Integer.valueOf(ii));
            } else if (SGUtilityText.isEqualString(Y_VALUE, columns[ii])) {
                yList.add(Integer.valueOf(ii));
            } else if (SGUtilityText.startsWith(columns[ii], LOWER_ERROR_VALUE)) {
                lList.add(Integer.valueOf(ii));
            } else if (SGUtilityText.startsWith(columns[ii], UPPER_ERROR_VALUE)) {
                uList.add(Integer.valueOf(ii));
            } else if (SGUtilityText.startsWith(columns[ii], LOWER_UPPER_ERROR_VALUE)) {
                lList.add(Integer.valueOf(ii));
                uList.add(Integer.valueOf(ii));
            } else if (SGUtilityText.startsWith(columns[ii], TICK_LABEL)) {
                tList.add(Integer.valueOf(ii));
            } else if ("".equals(columns[ii])) {
                continue;
            } else {
            	return false;
            }
        }
        
        // check values
        if (xList.size() == 0 || yList.size() == 0) {
        	return false;
        }
        if (xList.size() != 1 && yList.size() != 1) {
        	return false;
        }
        if (lList.size() != uList.size()) {
        	return false;
        }
        
        String[] columnTitles = this.mDataFile.getTitles();

        Integer[] xIndices = xList.toArray(new Integer[xList.size()]);
        Integer[] yIndices = yList.toArray(new Integer[yList.size()]);
        Integer[] lIndices = null;
        if (lList.size() != 0) {
            lIndices = lList.toArray(new Integer[lList.size()]);
        }
        Integer[] uIndices = null;
        if (uList.size() != 0) {
            uIndices = uList.toArray(new Integer[uList.size()]);
        }
        Integer[] ehIndices = null;
        if (lList.size() != 0) {
            ehIndices = new Integer[lList.size()];
            
            List<Integer> ehueList = new ArrayList<Integer>();
            for (int ii = 0; ii < ehIndices.length; ii++) {
                String str = columns[uIndices[ii].intValue()];
                Integer eh = SGDataUtility.getAppendedColumnIndex(str, columnTitles);
                if (eh==null) {
                    return false;
                }
                ehueList.add(eh);
            }
            
            for (int ii = 0; ii < ehIndices.length; ii++) {
                String str = columns[lIndices[ii].intValue()];
                Integer eh = SGDataUtility.getAppendedColumnIndex(str, columnTitles);
                if (eh == null) {
                    return false;
                }
                if (ehueList.contains(eh)==false) {
                    return false;
                }
                ehIndices[ii] = eh;
            }
        }
        Integer[] tIndices = null;
        if (tList.size() != 0) {
            tIndices = tList.toArray(new Integer[tList.size()]);
        }
        Integer[] thIndices = null;
        if (tList.size() != 0) {
            thIndices = new Integer[tList.size()];
            for (int ii = 0; ii < thIndices.length; ii++) {
                String str = columns[tIndices[ii].intValue()];
                Integer th = SGDataUtility.getAppendedColumnIndex(str, columnTitles);
                if (th == null) {
                    return false;
                }
                thIndices[ii] = th;
            }
        }
        
        // set to the attributes
        this.mXIndices = xIndices;
        this.mYIndices = yIndices;
        this.mLowerErrorIndices = lIndices;
        this.mUpperErrorIndices = uIndices;
        this.mErrorBarHolderIndices = ehIndices;
        this.mTickLabelIndices = tIndices;
        this.mTickLabelHolderIndices = thIndices;
        
        if (tIndices == null) {
        	this.updateDateTickLabelColumns();
        }

        return true;
    }
    
    /**
     * A class for multiple scalar xy-type data properties.
     *
     */
    public static class SXYMultipleDataProperties extends ArrayDataProperties {
        
        // indices of data columns
        Integer[] mXIndices = null;
    
        Integer[] mYIndices = null;
        
        Integer[] mLowerErrorIndices = null;

        Integer[] mUpperErrorIndices = null;
        
        Integer[] mErrorBarHolderIndices = null;

        Integer[] mTickLabelIndices = null;

        Integer[] mTickLabelHolderIndices = null;
        
        /**
         * The default constructor.
         */
        public SXYMultipleDataProperties() {
            super();
        }
        
        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof SXYMultipleDataProperties) == false) {
                return false;
            }
            SXYMultipleDataProperties p = (SXYMultipleDataProperties) obj;
            if (!SGUtility.equals(this.mXIndices, p.mXIndices)) {
            	return false;
            }
            if (!SGUtility.equals(this.mYIndices, p.mYIndices)) {
            	return false;
            }
            if (!SGUtility.equals(this.mLowerErrorIndices, p.mLowerErrorIndices)) {
            	return false;
            }
            if (!SGUtility.equals(this.mUpperErrorIndices, p.mUpperErrorIndices)) {
            	return false;
            }
            if (!SGUtility.equals(this.mErrorBarHolderIndices, p.mErrorBarHolderIndices)) {
            	return false;
            }
            if (!SGUtility.equals(this.mTickLabelIndices, p.mTickLabelIndices)) {
            	return false;
            }
            if (!SGUtility.equals(this.mTickLabelHolderIndices, p.mTickLabelHolderIndices)) {
            	return false;
            }
            return true;
        }
        
        /**
         * Returns string representation of this object.
         * @return
         *        string representation of this object
         */
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("[");
            sb.append("XIndices={");
            printIndices(sb, this.mXIndices);
            sb.append("}, YIndices={");
            printIndices(sb, this.mYIndices);
            sb.append("}, LowerErrorIndices={");
            printIndices(sb, this.mLowerErrorIndices);
            sb.append("}, UpperErrorIndices={");
            printIndices(sb, this.mUpperErrorIndices);
            sb.append("}, ErrorBarHolderIndices={");
            printIndices(sb, this.mErrorBarHolderIndices);
            sb.append("}, TickLabelIndices={");
            printIndices(sb, this.mTickLabelIndices);
            sb.append("}, TickLabelHolderIndices={");
            printIndices(sb, this.mTickLabelHolderIndices);
            sb.append("}]");
            return sb.toString();
        }
        
        private void printIndices(StringBuffer sb, Integer[] array) {
            if (array != null) {
                for (int ii = 0; ii < array.length; ii++) {
                    sb.append(array[ii]);
                    if (ii != array.length - 1) {
                        sb.append(", ");
                    }
                }
            }
        }

        /**
         * Copy this object.
         * @return a copied object
         */
        public Object copy() {
            Object obj = super.copy();
            SXYMultipleDataProperties p = (SXYMultipleDataProperties) obj;
            if (this.mXIndices != null) {
                p.mXIndices = (Integer[]) this.mXIndices.clone();
            }
            if (this.mYIndices != null) {
                p.mYIndices = (Integer[]) this.mYIndices.clone();
            }
            if (this.mLowerErrorIndices != null) {
                p.mLowerErrorIndices = (Integer[]) this.mLowerErrorIndices.clone();
            }
            if (this.mUpperErrorIndices != null) {
                p.mUpperErrorIndices = (Integer[]) this.mUpperErrorIndices.clone();
            }
            if (this.mErrorBarHolderIndices != null) {
                p.mErrorBarHolderIndices = (Integer[]) this.mErrorBarHolderIndices.clone();
            }
            if (this.mTickLabelIndices != null) {
            	p.mTickLabelIndices = (Integer[]) this.mTickLabelIndices.clone();
            }
            if (this.mTickLabelHolderIndices != null) {
            	p.mTickLabelHolderIndices = (Integer[]) this.mTickLabelHolderIndices.clone();
            }
            return p;
        }
        
        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.mXIndices = null;
            this.mYIndices = null;
            this.mLowerErrorIndices = null;
            this.mUpperErrorIndices = null;
            this.mErrorBarHolderIndices = null;
            this.mTickLabelIndices = null;
            this.mTickLabelHolderIndices = null;
        }
    }
    
    /**
     * Returns an array of current column types.
     * @return
     *        an array of current column types
     */
    public String[] getCurrentColumnType() {
        final int colNum = this.getColNum();
        String[] array = new String[colNum];
        for (int ii = 0; ii < this.mXIndices.length; ii++) {
            array[this.mXIndices[ii].intValue()] = X_VALUE;
        }
        for (int ii = 0; ii < this.mYIndices.length; ii++) {
            array[this.mYIndices[ii].intValue()] = Y_VALUE;
        }
        if (this.isErrorBarAvailable()) {
            for (int ii = 0; ii < this.mLowerErrorIndices.length; ii++) {
                int index = this.mErrorBarHolderIndices[ii].intValue();
                if (this.mLowerErrorIndices[ii].equals(this.mUpperErrorIndices[ii])) {
                    array[this.mLowerErrorIndices[ii].intValue()] =
                        SGDataUtility.appendColumnNoOrTitle(
                                LOWER_UPPER_ERROR_VALUE, index,
                                this.mDataFile.isEmptyTitle(index) || this.mDataFile.isRepeatedTitle(index),
                                this.getTitle(index));
                } else {
                    array[this.mLowerErrorIndices[ii].intValue()] =
                        SGDataUtility.appendColumnNoOrTitle(
                                LOWER_ERROR_VALUE, index,
                                this.mDataFile.isEmptyTitle(index) || this.mDataFile.isRepeatedTitle(index),
                                this.getTitle(index));
                    array[this.mUpperErrorIndices[ii].intValue()] =
                        SGDataUtility.appendColumnNoOrTitle(
                                UPPER_ERROR_VALUE, index,
                                this.mDataFile.isEmptyTitle(index) || this.mDataFile.isRepeatedTitle(index),
                                this.getTitle(index));
                }
            }
        }
        if (this.isTickLabelAvailable() && !this.isTickLabelFromDataColumn()) {
            for (int ii = 0; ii < this.mTickLabelIndices.length; ii++) {
                int index = this.mTickLabelHolderIndices[ii].intValue();
                array[this.mTickLabelIndices[ii].intValue()] =
                    SGDataUtility.appendColumnNoOrTitle(
                            TICK_LABEL, index,
                            this.mDataFile.isEmptyTitle(index) || this.mDataFile.isRepeatedTitle(index),
                            this.getTitle(index));
            }
        }
        return array;
    }
    
    private boolean isTickLabelFromDataColumn() {
    	if (this.isTickLabelAvailable()) {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        	for (int ii = 0; ii < this.mTickLabelIndices.length; ii++) {
        		SGDataColumn col = columns[this.mTickLabelIndices[ii]];
        		if (col instanceof SGDateDataColumn) {
        			return true;
        		}
        	}
    	}
    	return false;
    }
    
    /**
	 * Returns the maximum number of SXY data objects that can be obtained from
	 * this data object.
	 * 
	 * @return the maximum number of SXY data objects that can be obtained from
	 *         this data object
	 */
    public int getSXYDataNumMax() {
//        return this.getColNum(SGNumberDataColumn.class) - 1;
        int num = this.getColNum(SGNumberDataColumn.class) - 1;
        final int sNum = this.getColNum(SGSamplingDataColumn.class);
        if (sNum != 0) {
        	num++;
        }
        return num;
    }
    
    /**
     * Returns whether a data type column is used.
     * 
     * @return true if a data type column is used
     */
    public boolean isDateColumnUsed() {
    	if (this.getDateColumnIndex(this.mXIndices) != -1) {
    		return true;
    	}
    	if (this.getDateColumnIndex(this.mYIndices) != -1) {
    		return true;
    	}
    	return false;
    }

    private int getDateColumnIndex(Integer[] indices) {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
    	for (int ii = 0; ii < indices.length; ii++) {
    		SGDataColumn col = columns[indices[ii]];
    		if (col instanceof SGDateDataColumn) {
    			return ii;
    		}
    	}
    	return -1;
    }

    /**
     * Returns an array of data columns to save to a data set.
     * 
     * @return an array of data columns to save to a data set
     */
    protected SGDataColumn[] getDataSetColumns() {
    	final boolean be = this.isErrorBarAvailable();
    	final boolean bt = this.isTickLabelAvailable() && !this.isTickLabelFromDataColumn();
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
    	int len = this.mXIndices.length + this.mYIndices.length;
    	if (be) {
    		len += this.mLowerErrorIndices.length + this.mUpperErrorIndices.length;
    	}
    	if (bt) {
    		len += this.mTickLabelIndices.length;
    	}
    	int offset = 0;
        SGDataColumn[] colArray = new SGDataColumn[len];
		for (int ii = 0; ii < this.mXIndices.length; ii++) {
			colArray[ii] = columns[this.mXIndices[ii].intValue()];
		}
		offset += this.mXIndices.length;
		for (int ii = 0; ii < this.mYIndices.length; ii++) {
			colArray[ii + offset] = columns[this.mYIndices[ii].intValue()];
		}
		offset += this.mYIndices.length;
		if (be) {
			for (int ii = 0; ii < this.mLowerErrorIndices.length; ii++) {
				colArray[ii + offset] = columns[this.mLowerErrorIndices[ii].intValue()];
			}
			offset += this.mLowerErrorIndices.length;
			for (int ii = 0; ii < this.mUpperErrorIndices.length; ii++) {
				colArray[ii + offset] = columns[this.mUpperErrorIndices[ii].intValue()];
			}
			offset += this.mUpperErrorIndices.length;
		}
		if (bt) {
			for (int ii = 0; ii < this.mTickLabelIndices.length; ii++) {
				colArray[ii + offset] = columns[this.mTickLabelIndices[ii].intValue()];
			}
		}
		return colArray;
    }
    
    @Override
    public SGDataColumn[] getUsedDataColumnsClone() {
        SGDataColumn[] columns = this.mDataFile.mDataColumns;
        String[] columnTypes = this.getCurrentColumnType();
        List<SGDataColumn> collist = new ArrayList<SGDataColumn>();
        
        for (int i = 0; i < this.mXIndices.length; i++) {
            int index = this.mXIndices[i].intValue();
            SGDataColumn col = (SGDataColumn)columns[index].clone();
            col.setColumnType(columnTypes[index]);
            collist.add(col);
        }
        
        for (int i = 0; i < this.mYIndices.length; i++) {
            int index = this.mYIndices[i].intValue();
            SGDataColumn col = (SGDataColumn)columns[index].clone();
            col.setColumnType(columnTypes[index]);
            collist.add(col);
        }
        
        if (this.isErrorBarAvailable()) {
            for (int i = 0; i < this.mLowerErrorIndices.length; i++) {
                int index1 = this.mLowerErrorIndices[i].intValue();
                int index2 = this.mUpperErrorIndices[i].intValue();
                if (index1==index2) {
                    SGDataColumn col = (SGDataColumn)columns[index1].clone();
                    col.setColumnType(columnTypes[index1]);
                    collist.add(col);
                } else {
                    SGDataColumn col = (SGDataColumn)columns[index1].clone();
                    col.setColumnType(columnTypes[index1]);
                    collist.add(col);
                    col = (SGDataColumn)columns[index2].clone();
                    col.setColumnType(columnTypes[index2]);
                    collist.add(col);
                }
            }
        }
        if (this.isTickLabelAvailable()) {
            for (int i = 0; i < this.mTickLabelIndices.length; i++) {
                int index = this.mTickLabelIndices[i].intValue();
                SGDataColumn col = (SGDataColumn)columns[index].clone();
                col.setColumnType(columnTypes[index]);
                collist.add(col);
            }
        }
        
        return collist.toArray(new SGDataColumn[collist.size()]);
    }

    /**
     * Returns whether error bars are available.
     * 
     * @return true if error bars are available
     */
    public boolean isErrorBarAvailable() {
        return (this.mLowerErrorIndices != null && this.mUpperErrorIndices != null);
    }

    /**
     * Returns whether tick labels are available.
     * 
     * @return true if tick labels are available
     */
    public boolean isTickLabelAvailable() {
		return (this.mTickLabelIndices != null);
    }

    /**
     * Returns the decimal places for the tick labels.
     * 
     * @return the decimal places for the tick labels
     */
    public int getDecimalPlaces() {
        return this.mDecimalPlaces;
    }

    /**
     * Returns the exponent for tick labels.
     * 
     * @return the exponent for tick labels
     */
    public int getExponent() {
        return this.mExponent;
    }

    /**
     * Sets the decimal places for the tick labels.
     * 
     * @param dp
     *          a value to set to the decimal places
     */
    public void setDecimalPlaces(final int dp) {
        if (dp < 0) {
            throw new IllegalArgumentException("Decimal places must not be negative: " + dp);
        }
        this.mDecimalPlaces = dp;
    }

    /**
     * Sets the exponent for the tick labels.
     * 
     * @param exp
     *           a value to set to the exponent
     */
    public void setExponent(final int exp) {
    	this.mExponent = exp;
    }

    /**
	 * Returns whether the error bars are vertical. If this data does not have
	 * error values, returns null.
	 * 
	 * @return true if error bars are vertical, false if they are horizontal and
	 *         null if this data do not have error values
	 */
	public Boolean isErrorBarVertical() {
    	if (this.isErrorBarAvailable()) {
    		if (SGUtility.contains(this.mYIndices, this.mErrorBarHolderIndices[0])) {
    			return Boolean.TRUE;
    		}
    		if (SGUtility.contains(this.mXIndices, this.mErrorBarHolderIndices[0])) {
    			return Boolean.FALSE;
    		}
    	}
		return null;
	}

    /**
	 * Returns whether the tick labels align horizontally. If this data does not have
	 * tick labels, returns null.
	 * 
	 * @return true if tick labels align horizontally, false if they do not so and
	 *         null if this data do not have tick labels
	 */
	public Boolean isTickLabelHorizontal() {
    	if (this.isTickLabelAvailable()) {
    		if (SGUtility.contains(this.mYIndices, this.mTickLabelHolderIndices[0])) {
    			return Boolean.TRUE;
    		}
    		if (SGUtility.contains(this.mXIndices, this.mTickLabelHolderIndices[0])) {
    			return Boolean.FALSE;
    		}
    	} else {
    		if (this.getDateColumnIndex(this.mXIndices) != -1) {
        		return Boolean.TRUE;
    		}
    		if (this.getDateColumnIndex(this.mYIndices) != -1) {
        		return Boolean.FALSE;
    		}
    	}
		return null;
	}

	public SGDataColumn[] getXColumns() {
		return this.getColumns(this.mXIndices);
	}

	public SGDataColumn[] getYColumns() {
		return this.getColumns(this.mYIndices);
	}

	public SGDataColumn[] getLowerErrorColumns() {
		return this.getColumns(this.mLowerErrorIndices);
	}
	
	public SGDataColumn[] getUpperErrorColumns() {
		return this.getColumns(this.mUpperErrorIndices);
	}

	public SGDataColumn[] getErrorBarHolderColumns() {
		return this.getColumns(this.mErrorBarHolderIndices);
	}

	public SGDataColumn[] getTickLableColumns() {
		return this.getColumns(this.mTickLabelIndices);
	}

	public SGDataColumn[] getTickLabelHolderColumns() {
		return this.getColumns(this.mTickLabelHolderIndices);
	}

	public SGDataColumn[] getMultipleColumns() {
        if (this.hasMultipleYValues()) {
        	return this.getYColumns();
        } else {
        	return this.getXColumns();
        }
	}
	
    /**
     * Returns a map which has data information.
     * 
     * @return a map which has data information
     */
    public Map<String, Object> getInfoMap() {
    	Map<String, Object> infoMap = super.getInfoMap();
    	infoMap.put(KEY_MULTIPLE, Boolean.TRUE);
    	return infoMap;
    }

    @Override
    public boolean isSplitEnabled() {
        final boolean by = this.hasMultipleYValues();
        final int dataNum = by ? this.getYNumber() : this.getXNumber();

        if (dataNum>1) {
            return true;
        }
        return false;
    }
    
    public static SGISXYTypeMultipleData merge(final List<SGData> dataList) {
        if (dataList.size()==0) {
            return null;
        }
        for (SGData data : dataList) {
            if ((data instanceof SGSXYMultipleData)==false) {
                return null;
            }
        }
        
        SGSXYMultipleData data0 = (SGSXYMultipleData)dataList.get(0);
        SGTextDataFile dataFile = data0.getDataFile();
        SGDataSourceObserver obs = data0.mObserver;
        
        Set<Integer> xset = new TreeSet<Integer>();
        Set<Integer> yset = new TreeSet<Integer>();
        Set<Integer> leset = new TreeSet<Integer>();
        Set<Integer> ueset = new TreeSet<Integer>();
        Set<Integer> ehset = new TreeSet<Integer>();
        Set<Integer> tset = new TreeSet<Integer>();
        Set<Integer> thset = new TreeSet<Integer>();
        for (SGData data : dataList) {
            SGSXYMultipleData dataMult = (SGSXYMultipleData)data;
            Integer[] x = dataMult.getXIndices();
            Integer[] y = dataMult.getYIndices();
            Integer[] le = dataMult.getLowerErrorIndices();
            Integer[] ue = dataMult.getUpperErrorIndices();
            Integer[] eh = dataMult.getErrorBarHolderIndices();
            Integer[] t = dataMult.getTickLabelIndices();
            Integer[] th = dataMult.getTickLabelHolderIndices();
            if (null!=x) {
                for (int i = 0; i < x.length; i++) {
                    xset.add(x[i]);
                }
            }
            if (null!=y) {
                for (int i = 0; i < y.length; i++) {
                    yset.add(y[i]);
                }
            }
            if (null!=le) {
                for (int i = 0; i < le.length; i++) {
                    leset.add(le[i]);
                }
            }
            if (null!=ue) {
                for (int i = 0; i < ue.length; i++) {
                    ueset.add(ue[i]);
                }
            }
            if (null!=eh) {
                for (int i = 0; i < eh.length; i++) {
                    ehset.add(eh[i]);
                }
            }
            if (null!=t) {
                for (int i = 0; i < t.length; i++) {
                    tset.add(t[i]);
                }
            }
            if (null!=th) {
                for (int i = 0; i < th.length; i++) {
                    thset.add(th[i]);
                }
            }
        }
        
        Integer[] x = new Integer[xset.size()];
        Integer[] y = new Integer[yset.size()];
        Integer[] le = new Integer[leset.size()];
        Integer[] ue = new Integer[ueset.size()];
        Integer[] eh = new Integer[ehset.size()];
        Integer[] t = new Integer[tset.size()];
        Integer[] th = new Integer[thset.size()];
        xset.toArray(x);
        yset.toArray(y);
        leset.toArray(le);
        ueset.toArray(ue);
        ehset.toArray(eh);
        tset.toArray(t);
        thset.toArray(th);
        
        SGSXYMultipleData data = new SGSXYMultipleData(
                dataFile, obs, x, y, le, ue, eh, t, th);
        data.setDecimalPlaces(data0.getDecimalPlaces());
        data.setExponent(data0.getExponent());
        return data;
    }
    
}
