package jp.riken.brain.ni.samuraigraph.data;

import java.util.List;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;
import ucar.ma2.Array;
import ucar.nc2.Dimension;

/**
 * The class of multiple scalar XY type data with netCDF data.
 *
 */
public abstract class SGSXYMultipleNetCDFData extends SGNetCDFData implements
        SGISXYTypeMultipleData, SGIDataColumnTypeConstants, SGIDataColumnNameConstants {

    /**
     * The index start of dimension.
     */
    protected Integer mDimensionStart = null;

    /**
     * The index end of dimension.
     */
    protected Integer mDimensionEnd = null;

    /**
     * The index step of dimension.
     */
    protected Integer mDimensionStep = null;

    /**
     * The decimal places for the tick labels.
     */
    protected int mDecimalPlaces = 0;

    /**
     * The exponent for the tick labels.
     */
    protected int mExponent = 0;
    
    /**
     * The default constructor.
     *
     */
    public SGSXYMultipleNetCDFData() {
        super();
    }

    /**
     * Builds the data object with given NetCDF file.
     * 
     * @param ncfile
     *          NetCDF file
     * @param obs
     *          an observer of netCDF data
     * @param timeName
     *          the name of time variable
     * @param serialNumberName
     *          the name of serial number variable
     */
    public SGSXYMultipleNetCDFData(SGNetCDFFile ncfile, SGDataSourceObserver obs, 
    		String timeName, String serialNumberName) {
    	super(ncfile, obs, timeName, serialNumberName);
    }
    
    /**
     * Returns the number of points.
     * 
     * @return the number of points
     */
    public int getPointsNumber() {
    	SGVariable var = this.getCoordinateVariable();
        Dimension dim = var.getDimension(0);
        return dim.getLength();
    }

    /**
     * Returns the bounds of x-values.
     * 
     * @return the bounds of x-values
     */
    public SGValueRange getBoundsX() {
        return SGDataUtility.getBoundsX(this);
    }

    /**
     * Returns the bounds of y-values.
     * 
     * @return the bounds of y-values
     */
    public SGValueRange getBoundsY() {
        return SGDataUtility.getBoundsY(this);
    }


    protected SGVariable[] getVariables(final String[] names, 
    		SGNetCDFFile ncfile, List<Dimension> cDimList) {
        SGVariable[] vars = null;
    	if (names != null) {
            vars = new SGVariable[names.length];
            for (int ii = 0; ii < vars.length; ii++) {
                SGVariable var = ncfile.findVariable(names[ii]);
                this.checkNotCoordinateVariable(var, cDimList);
                vars[ii] = var;
            }
    	}
    	return vars;
    }

    /**
     * Returns whether this multiple data has multiple arrays for y-values.
     * 
     * @return true if this multiple data has multiple arrays for y-values
     */
    protected abstract boolean hasMultipleYValues();

    /**
     * Returns a variable for "not" multiple values.
     * 
     * @return a variable for "not" multiple values
     */
    public abstract SGVariable getSingleVariable();
    
    protected Double[] arrangeValueArray(Array data2D, SGVariable var, SGVariable[] cVars) {
        return this.arrange1Dim(data2D);
    }

    protected int find(SGVariable var, SGVariable[] vars) {
    	if (vars == null) {
    		return -1;
    	}
        for (int ii = 0; ii < vars.length; ii++) {
            if (vars[ii].equals(var)) {
                return ii;
            }
        }
        return -1;
    }
    
    @Override
    public boolean setData(SGData data) {
        if (!(data instanceof SGSXYMultipleNetCDFData)) {
            throw new IllegalArgumentException("Given data is not the instance of SGSXYMultipleNetCDFData class.");
        }
        if (super.setData(data) == false) {
            return false;
        }
        SGSXYMultipleNetCDFData nData = (SGSXYMultipleNetCDFData) data;
        this.mDimensionStart = nData.mDimensionStart;
        this.mDimensionEnd = nData.mDimensionEnd;
        this.mDimensionStep = nData.mDimensionStep;
        return true;
    }


    /**
     * Sets the decimal places for the tick labels.
     * 
     * @param dp
     *          a value to set to the decimal places
     */
	public void setDecimalPlaces(int dp) {
        if (dp < 0) {
            throw new IllegalArgumentException("Decimal places must not be negative: " + dp);
        }
        this.mDecimalPlaces = dp;
	}

    /**
     * Sets the exponent for the tick labels.
     * 
     * @param exp
     *           a value to set to the exponent
     */
	public void setExponent(int exp) {
        this.mExponent = exp;
	}

    /**
     * Returns the decimal places for the tick labels.
     * 
     * @return the decimal places for the tick labels
     */
	public int getDecimalPlaces() {
        return this.mDecimalPlaces;
	}

    /**
     * Returns the exponent for tick labels.
     * 
     * @return the exponent for tick labels
     */
	public int getExponent() {
        return this.mExponent;
	}

    /**
	 * Returns the m-th x value at the n-th index.
	 * 
	 * @param m
	 *            the index of data
	 * @param n
	 *            the index of the data point
	 * @return the n-th x-value of the m-th data
	 */
	public Double getXValue(final int m, final int n) {
        return Double.valueOf(this.getXValueArray()[m][n]);
    }

    /**
     * Returns the m-th y value at the n-th index.
     * 
     * @param m
     *          the index of data
     * @param n
     *          the index of the data point
     * @return
     *          the n-th y-value of the m-th data
     */
    public Double getYValue(final int m, final int n) {
        return Double.valueOf(this.getYValueArray()[m][n]);
    }

    protected SGVariable getCoordinateVariable() {
    	if (this.isSerialNumberAvailable()) {
    		return this.mSerialNumberVariable;
    	} else {
    		return this.getSingleVariable();
    	}
    }
    
    /**
     * Returns an array of values of the single variable.
     * 
     * @return an array of values of the single variable
     */
    protected double[][] getSingleVariableValueArray() {
    	if (this.isSerialNumberAvailable()) {
			return new double[][] { 
					this.getValueArray(this.getSingleVariable(),
					new SGVariable[] { this.mSerialNumberVariable }) };
		} else {
			return new double[][] { this.getCoordinateValueArray(this
					.getSingleVariable()) };
		}
    }
    
    /**
	 * Returns an array of values of multiple variables.
	 * 
	 * @return an array of values of multiple variables
	 */
    protected abstract double[][] getMultipleVariableValueArray();

    /**
     * Returns an array of X-values.
     * 
     * @return an array of X-values
     */
    public double[][] getXValueArray() {
    	if (this.hasMultipleYValues()) {
    		return this.getSingleVariableValueArray();
    	} else {
    		return this.getMultipleVariableValueArray();
    	}
    }

    /**
     * Returns an array of Y-values.
     * 
     * @return an array of Y-values
     */
    public double[][] getYValueArray() {
    	if (this.hasMultipleYValues()) {
    		return this.getMultipleVariableValueArray();
    	} else {
    		return this.getSingleVariableValueArray();
    	}
    }
    
    /**
     * Returns the start index of multiple dimensions.
     * 
     * @return the start index of multiple dimensions
     */
    public int getDimensionStart() {
        if (null!=this.mDimensionStart) {
            return this.mDimensionStart.intValue();
        } else {
            return -1;
        }
    }
    
    /**
     * Returns the end index of multiple dimensions.
     * 
     * @return the end index of multiple dimensions
     */
    public int getDimensionEnd() {
        if (null!=this.mDimensionEnd) {
            return this.mDimensionEnd.intValue();
        } else {
            return -1;
        }
    }

    /**
     * Returns the step of multiple dimensions.
     * 
     * @return the step of multiple dimensions
     */
    public int getDimensionStep() {
        if (null!=this.mDimensionStep) {
            return this.mDimensionStep.intValue();
        } else {
            return -1;
        }
    }

    /**
     * Sets the start index of multiple dimensions.
     * 
     * @param start
     *           the start index of multiple dimensions
     * @return true if succeeded
     */
    public boolean setDimensionStart(final int start) {
        this.mDimensionStart = Integer.valueOf(start);
        return true;
    }

    /**
     * Sets the end index of multiple dimensions.
     * 
     * @param end
     *           the end index of multiple dimensions
     * @return true if succeeded
     */
    public boolean setDimensionEnd(final int end) {
        this.mDimensionEnd = Integer.valueOf(end);
        return true;
    }

    /**
     * Sets the step of multiple dimensions.
     * 
     * @param step
     *           the step of multiple dimensions
     * @return true if succeeded
     */
    public boolean setDimensionStep(final int step) {
        this.mDimensionStep = Integer.valueOf(step);
        return true;
    }
    
    /**
     * Return whether the XVariable is a coordinate variable.
     * 
     * @return whether the XVariable is a coordinate variable.
     */
    public abstract boolean isXVariableCoordinate();

    /**
     * Sets the column information.
     * 
     * @param colInfo
     *           the column information
     * @return true if succeeded
     */
    public boolean setColumnInfo(SGDataColumnInfo[] colInfo) {
        if (super.setColumnInfo(colInfo) == false) {
        	return false;
        }
    	return true;
    }
    
    /**
     * Sets the current index of time value.
     * 
     * @param index
     *           a value to set to the index of time values
     */
    public void setCurrentTimeValueIndex(final int index) {
    	super.setCurrentTimeValueIndex(index);
    }
    
    /**
     * Set properties to this data.
     * 
     * @param p
     *          properties to be set
     * @return true if succeeded
     */
    public boolean setProperties(SGProperties p) {
        if (super.setProperties(p) == false) {
        	return false;
        }
        return true;
    }

    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
        
        // clear the cache
        this.mVariableCache.clear();
    }
}
