package jp.riken.brain.ni.samuraigraph.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;

import org.w3c.dom.Element;

import ucar.nc2.Dimension;

/**
 * Scalar type XY data with multiple variables.
 * @author kuromaru
 *
 */
public class SGSXYMultipleVariableNetCDFData extends SGSXYMultipleNetCDFData {

    /**
     * The variables for x-values.
     */
    protected SGVariable[] mXVariables = null;
    
    /**
     * The variables for y-values.
     */
    protected SGVariable[] mYVariables = null;    
    
    /**
     * The variables for lower error values.
     */
    protected SGVariable[] mLowerErrorVariables = null;

    /**
     * The variables for upper error values.
     */
    protected SGVariable[] mUpperErrorVariables = null;
    
    /**
     * The variables for values that holds error values.
     */
    protected SGVariable[] mErrorBarHolderVariables = null;
    
    /**
     * The variables for tick labels.
     */
    protected SGVariable[] mTickLabelVariables = null;

    /**
     * The variables for values that holds tick labels.
     */
    protected SGVariable[] mTickLabelHolderVariables = null;

    /**
     * Whether single variable is date.
     */
    private boolean mIsSingleVariableDateFlag = false;

    /**
     * The default constructor.
     *
     */
    public SGSXYMultipleVariableNetCDFData() {
        super();
    }

    /**
     * Builds a data object with given netCDF data.
     * 
     * @param ncfile
     *            netCDF data file
     * @param obs
     *            netCDF data observer
     * @param xNames
     *            the names of x variables
     * @param yNames
     *            the names of y variables
     * @param timeName
     *            the name of time variable
     * @param serialNumberName
     *          the name of serial number variable
     */
    public SGSXYMultipleVariableNetCDFData(final SGNetCDFFile ncfile,
            final SGDataSourceObserver obs, final String[] xNames,
            final String[] yNames, final String timeName, final String serialNumberName) {
    	
        super(ncfile, obs, timeName, serialNumberName);
        
        if (xNames == null || yNames == null) {
            throw new IllegalArgumentException("xNames == null || yNames == null");
        }
        final int xNum = xNames.length;
        if (xNum == 0) {
            throw new IllegalArgumentException("The number of variables for x-values is equal to zero.");
        }
        final int yNum = yNames.length;
        if (yNum == 0) {
            throw new IllegalArgumentException("The number of variables for y-values is equal to zero.");
        }
        if (!this.isValidXYNames(ncfile, xNames, yNames)) {
            throw new IllegalArgumentException("Either one of the two arrays must have only one element.");
        }

        // get variables
        SGVariable[] xVars = new SGVariable[xNum];
        for (int ii = 0; ii < xNum; ii++) {
            xVars[ii] = ncfile.findVariable(xNames[ii]);
            if (xVars[ii] == null) {
                throw new IllegalArgumentException("Illegal variable name: " + xNames[ii]);
            }
        }
        SGVariable[] yVars = new SGVariable[yNum];
        for (int ii = 0; ii < yNum; ii++) {
            yVars[ii] = ncfile.findVariable(yNames[ii]);
            if (yVars[ii] == null) {
                throw new IllegalArgumentException("Illegal variable name: " + yNames[ii]);
            }            
        }

        //
        // check variables
        //
        
    	// x and y variables
    	SGVariable varSingle = null;
    	SGVariable[] varMultiple = null;
    	if (xVars.length == 1 && yVars.length == 1) {
    		SGVariable xVar = xVars[0];
    		SGVariable yVar = yVars[0];
    		if (xVar.isCoordinateVariable()) {
    			varSingle = xVar;
    			varMultiple = yVars;
    		} else {
    			varSingle = yVar;
    			varMultiple = xVars;
    		}
    	} else if (xVars.length == 1) {
    		varSingle = xVars[0];
    		varMultiple = yVars;
    	} else if (yVars.length == 1) {
    		varSingle = yVars[0];
    		varMultiple = xVars;
    	} else {
            throw new IllegalArgumentException("Either of x and y variables must be size one array.");
    	}
    	
        List<Dimension> cDimList = new ArrayList<Dimension>();
        
        // time variable
    	if (this.mTimeVariable != null) {
    		Dimension tDim = this.mTimeVariable.getDimension(0);
    		cDimList.add(tDim);
    	}
    	
        if (this.mSerialNumberVariable != null) {
        	// x and y variables
        	if (varSingle.isCoordinateVariable()) {
                throw new IllegalArgumentException("The single variable must not be a coordinate variable.");
        	}
            for (int ii = 0; ii < varMultiple.length; ii++) {
            	if (varMultiple[ii].isCoordinateVariable()) {
                    throw new IllegalArgumentException("The multiple variables must not be coordinate variables.");
            	}
            }
        	
        	// index variable
            Dimension iDim = this.mSerialNumberVariable.getDimension(0);
            cDimList.add(iDim);
            
            // single variable
        	this.checkNotCoordinateVariable(varSingle, cDimList);
        } else {
        	// x and y variables
        	if (!varSingle.isCoordinateVariable()) {
                throw new IllegalArgumentException("The single variable must be a coordinate variable.");
        	}
            cDimList.add(varSingle.getDimension(0));
        }

        // multiple variables
        for (int ii = 0; ii < varMultiple.length; ii++) {
        	this.checkNotCoordinateVariable(varMultiple[ii], cDimList);
        }
        
        // set to attributes
        this.mXVariables = xVars;
        this.mYVariables = yVars;
        
        this.updateIsSingleVariableDateFlag();
    }

    /**
     * Builds a data object with given netCDF data.
     * 
     * @param ncfile
     *            netCDF data file
     * @param obs
     *            netCDF data observer
     * @param xNames
     *            the names of x variable
     * @param yNames
     *            the names of y variables
     * @param lNames
     *            the names of lower error variables
     * @param uNames
     *            the names of upper error variables
     * @param ehNames
     *            the names of error bar holder variables
     * @param tNames
     *            the names of tick label variables
     * @param thNames
     *            the names of tick label holder variables
     * @param timeName
     *            the name of time variables
     * @param index Name
     *            the name of serialNumber variables
     */
    public SGSXYMultipleVariableNetCDFData(SGNetCDFFile ncfile, 
    		SGDataSourceObserver obs,
            final String[] xNames, final String[] yNames, final String[] lNames, 
            final String[] uNames, final String[] ehNames, final String[] tNames, 
            final String[] thNames, final String timeName, final String serialNumberName) {
    	
        this(ncfile, obs, xNames, yNames, timeName, serialNumberName);

    	// check the input for error bars
    	String[] multipleNames = (xNames.length == 1) ? yNames : xNames;
    	if (!this.isValidErrorBarNames(ncfile, multipleNames, lNames, uNames, ehNames)) {
            throw new IllegalArgumentException("Column indices for the error bars are invalid.");
    	}
    	if (!this.isValidTickLabelNames(ncfile, multipleNames, tNames, thNames)) {
            throw new IllegalArgumentException("Column indices for the tick labels are invalid.");
    	}

        //
        // check variables
        //
        SGVariable cVar = (this.mXVariables.length == 1) ? this.mXVariables[0] : this.mYVariables[0];
        List<Dimension> cDimList = new ArrayList<Dimension>();
        cDimList.add(cVar.getDimension(0));
    	if (this.mTimeVariable != null) {
    		cDimList.add(this.mTimeVariable.getDimension(0));
    	}
    	if (this.mSerialNumberVariable != null) {
    		cDimList.add(this.mSerialNumberVariable.getDimension(0));
    	}

        // get variables with the check of consistency
    	final boolean be = (lNames != null && lNames.length != 0);
    	final boolean bt = (tNames != null && tNames.length != 0);
    	SGVariable[] lVars = be ? this.getVariables(lNames, ncfile, cDimList): null;
    	SGVariable[] uVars = be ? this.getVariables(uNames, ncfile, cDimList): null;
    	SGVariable[] ehVars = be ? this.getVariables(ehNames, ncfile, cDimList): null;
    	SGVariable[] tVars = bt ? this.getVariables(tNames, ncfile, cDimList): null;
    	SGVariable[] thVars = bt ? this.getVariables(thNames, ncfile, cDimList): null;
    	
        // set to attributes
    	this.mLowerErrorVariables = lVars;
    	this.mUpperErrorVariables = uVars;
    	this.mErrorBarHolderVariables = ehVars;
    	this.mTickLabelVariables = tVars;
    	this.mTickLabelHolderVariables = thVars;
    }

    private boolean isValidXYNames(
    		final SGNetCDFFile ncfile,
    		final String[] xNames, 
    		final String[] yNames) {
        if (xNames.length == 0 || yNames.length == 0) {
        	return false;
        }
        if (xNames.length != 1 && yNames.length != 1) {
        	return false;
        }
        if (this.isValidNames(ncfile, xNames) == false) {
        	return false;
        }
        if (this.isValidNames(ncfile, yNames) == false) {
        	return false;
        }
        return true;
    }
    
    private boolean isValidErrorBarNames(
    		final SGNetCDFFile ncfile,
    		final String[] multipleNames,
    		final String[] lowerErrorNames, 
    		final String[] upperErrorNames,
    		final String[] errorBarHolderNames) {
    	
        if (lowerErrorNames != null && upperErrorNames != null) {
        	if (!this.isValidNames(ncfile, lowerErrorNames)) {
        		return false;
        	}
        	if (!this.isValidNames(ncfile, upperErrorNames)) {
        		return false;
        	}
        	if (lowerErrorNames.length != upperErrorNames.length) {
        		return false;
        	}
        	if (errorBarHolderNames != null) {
            	if (!this.isValidNames(ncfile, errorBarHolderNames)) {
            		return false;
            	}
                if (errorBarHolderNames.length != lowerErrorNames.length) {
                	return false;
                }
        	}
        }
        return true;
    }

    private boolean isValidTickLabelNames(
    		final SGNetCDFFile ncfile,
    		final String[] multipleNames,
    		final String[] tickLabelNames, 
    		final String[] tickLabelHolderNames) {
    	
        if (tickLabelNames != null) {
        	if (!this.isValidNames(ncfile, tickLabelNames)) {
        		return false;
        	}
        	if (tickLabelHolderNames != null) {
            	if (!this.isValidNames(ncfile, tickLabelHolderNames)) {
            		return false;
            	}
                if (tickLabelHolderNames.length != tickLabelNames.length) {
                	return false;
                }
        	}
        }
        return true;
    }

    private boolean isValidNames(
    		final SGNetCDFFile ncfile,
    		final String[] names) {
    	if (names == null) {
    		return false;
    	}
        for (int ii = 0; ii < names.length; ii++) {
            if (names[ii] == null) {
            	return false;
            }
            if (ncfile.findVariable(names[ii]) == null) {
            	return false;
            }
        }
        return true;
    }
    
    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
        this.mXVariables = null;
        this.mYVariables = null;
        this.mLowerErrorVariables = null;
        this.mUpperErrorVariables = null;
        this.mTickLabelVariables = null;
        this.mTickLabelHolderVariables = null;
    }
    
    /**
     * Returns whether this multiple data has multiple arrays for y-values.
     * 
     * @return true if this multiple data has multiple arrays for y-values
     */
    protected boolean hasMultipleYValues() {
    	// if the array of x indices has only one elements,
    	// this data has "multiple" y values
    	return (this.mXVariables.length == 1);
    }

    /**
     * Returns a variable for multiple values.
     * 
     * @return a variable for multiple values
     */
    public SGVariable[] getMultipleVariables() {
        if (this.hasMultipleYValues()) {
            return this.mYVariables;
        } else {
            return this.mXVariables;
        }
    }

    /**
     * Returns a variable for "not" multiple values.
     * 
     * @return a variable for "not" multiple values
     */
    public SGVariable getSingleVariable() {
        if (this.hasMultipleYValues()) {
            return this.mXVariables[0];
        } else {
            return this.mYVariables[0];
        }
    }

    /**
     * Returns the data type for the data set.
     * 
     * @return the data type
     */
    protected String getDataTypeForDataSet() {
        return SGDataTypeConstants.SXY_MULTIPLE_DATA;
    }

    /**
     * Returns an array of current column types.
     * 
     * @return an array of current column types
     */
    @Override
    public String[] getCurrentColumnType() {
        List<SGVariable> varList = this.mNCFile.getVariables();
        final int varNum = varList.size();
        String[] array = new String[varNum];
        for (int ii = 0; ii < varNum; ii++) {
            SGVariable var = varList.get(ii);
            final int lIndex = this.find(var, this.mLowerErrorVariables);
            final int uIndex = this.find(var, this.mUpperErrorVariables);
            final int tIndex = this.find(var, this.mTickLabelVariables);
            if (this.find(var, this.mXVariables) != -1) {
                array[ii] = X_VALUE;
            } else if (this.find(var, this.mYVariables) != -1) {
                array[ii] = Y_VALUE;
            } else if ((lIndex != -1) || (uIndex != -1)) {
                String name = null;
                int index = -1;
                if (lIndex == uIndex) {
                    name = LOWER_UPPER_ERROR_VALUE;
                    index = lIndex;
                } else if (lIndex != -1) {
                    name = LOWER_ERROR_VALUE;
                    index = lIndex;
                } else if (uIndex != -1) {
                    name = UPPER_ERROR_VALUE;
                    index = uIndex;
                }
                if (name == null) {
                    return null;
                }
                final int hIndex = this.mNCFile.getVariableIndex(this.mErrorBarHolderVariables[index].getName());
                if (hIndex == -1) {
                    return null;
                }
                array[ii] = SGDataUtility.appendColumnTitle(name, this.mErrorBarHolderVariables[index].getName());
            } else if (tIndex != -1) {
                final int hIndex = this.mNCFile.getVariableIndex(this.mTickLabelHolderVariables[tIndex].getName());
                if (hIndex == -1) {
                    return null;
                }
                array[ii] = SGDataUtility.appendColumnTitle(TICK_LABEL, this.mTickLabelHolderVariables[tIndex].getName());
            } else if (var.equals(this.mTimeVariable)) {
                array[ii] = TIME;
            } else if (var.equals(this.mSerialNumberVariable)) {
				array[ii] = SERIAL_NUMBER;
            } else {
                array[ii] = "";
            }
        }        
        return array;
    }

    /**
     * Returns an array of all variables.
     * 
     * @return an array of all variables
     */
    public SGVariable[] getVariables() {
        List<SGVariable> varList = new ArrayList<SGVariable>();
        for (int ii = 0; ii < this.mXVariables.length; ii++) {
            varList.add(this.mXVariables[ii]);
        }
        for (int ii = 0; ii < this.mYVariables.length; ii++) {
            varList.add(this.mYVariables[ii]);
        }
        if (this.mLowerErrorVariables != null) {
            for (int ii = 0; ii < this.mLowerErrorVariables.length; ii++) {
            	varList.add(this.mLowerErrorVariables[ii]);
            }
        }
        if (this.mUpperErrorVariables != null) {
            for (int ii = 0; ii < this.mUpperErrorVariables.length; ii++) {
            	varList.add(this.mUpperErrorVariables[ii]);
            }
        }
        if (this.mTickLabelVariables != null) {
            for (int ii = 0; ii < this.mTickLabelVariables.length; ii++) {
            	varList.add(this.mTickLabelVariables[ii]);
            }
        }
        if (this.mTimeVariable != null) {
        	varList.add(this.mTimeVariable);
        }
        if (this.mSerialNumberVariable != null) {
        	varList.add(this.mSerialNumberVariable);
        }
        SGVariable[] varArray = new SGVariable[varList.size()];
        return (SGVariable[]) varList.toArray(varArray);
    }

    public SGVariable[] getXVariables() {
    	return this.getVariables(this.mXVariables);
    }

    public SGVariable[] getYVariables() {
    	return this.getVariables(this.mYVariables);
    }

    public SGVariable[] getLowerErrorVariables() {
    	return this.getVariables(this.mLowerErrorVariables);
    }

    public SGVariable[] getUpperErrorVariables() {
    	return this.getVariables(this.mUpperErrorVariables);
    }

    public SGVariable[] getErrorHolderVariables() {
    	return this.getVariables(this.mErrorBarHolderVariables);
    }

    public SGVariable[] getTickLabelVariables() {
    	return this.getVariables(this.mTickLabelVariables);
    }

    public SGVariable[] getTickLabelHolderVariables() {
    	return this.getVariables(this.mTickLabelHolderVariables);
    }

    private SGVariable[] getVariables(SGVariable[] vars) {
    	if (vars != null) {
        	return (SGVariable[]) vars.clone();
    	} else {
    		return null;
    	}
    }

    /**
     * Sets the type of data columns.
     * 
     * @param column
     *              an array of column types
     * @return true if succeeded
     */
    public boolean setColumnType(String[] columns) {
    	
        List<SGVariable> xVarList = new ArrayList<SGVariable>();
        List<SGVariable> yVarList = new ArrayList<SGVariable>();
        List<SGVariable> lVarList = new ArrayList<SGVariable>();
        List<SGVariable> uVarList = new ArrayList<SGVariable>();
        List<SGVariable> tVarList = new ArrayList<SGVariable>();
    	SGVariable timeVar = null;
    	SGVariable serialNumberVar = null;
    	List<String> lColList = new ArrayList<String>();
    	List<String> uColList = new ArrayList<String>();
    	List<String> tColList = new ArrayList<String>();

        List<SGVariable> varList = this.mNCFile.getVariables();
        for (int ii = 0; ii < columns.length; ii++) {
            SGVariable var = varList.get(ii);
            if (SGUtilityText.isEqualString(X_VALUE, columns[ii])) {
                xVarList.add(var);
            } else if (SGUtilityText.isEqualString(Y_VALUE, columns[ii])) {
                yVarList.add(var);
            } else if (SGUtilityText.startsWith(columns[ii], LOWER_ERROR_VALUE)) {
                lVarList.add(var);
                lColList.add(columns[ii]);
            } else if (SGUtilityText.startsWith(columns[ii], UPPER_ERROR_VALUE)) {
                uVarList.add(var);
                uColList.add(columns[ii]);
            } else if (SGUtilityText.startsWith(columns[ii], LOWER_UPPER_ERROR_VALUE)) {
                lVarList.add(var);
                uVarList.add(var);
                lColList.add(columns[ii]);
                uColList.add(columns[ii]);
            } else if (SGUtilityText.startsWith(columns[ii], TICK_LABEL)) {
                tVarList.add(var);
                tColList.add(columns[ii]);
            } else if (SGUtilityText.isEqualString(TIME, columns[ii])) {
                timeVar = var;
            } else if (SGUtilityText.isEqualString(SERIAL_NUMBER, columns[ii])) {
            	serialNumberVar = var;
            } else if ("".equals(columns[ii])) {
            	continue;
            } else {
            	return false;
            }
        }
        if (xVarList.size() == 0 || yVarList.size() == 0) {
        	return false;
        }
        if (xVarList.size() != 1 && yVarList.size() != 1) {
        	return false;
        }
        if (lVarList.size() != uVarList.size()) {
            return false;
        }
        
        List<SGVariable> ehVarList = new ArrayList<SGVariable>();
        List<Integer> ehueList = new ArrayList<Integer>();
        for (int ii = 0; ii < uVarList.size(); ii++) {
            String col = uColList.get(ii);
            Integer eh = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(col, this.mNCFile);
            if (eh == null) {
                return false;
            }
            ehueList.add(eh);
        }
        for (int ii = 0; ii < lVarList.size(); ii++) {
            String col = lColList.get(ii);
            Integer eh = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(col, this.mNCFile);
            if (eh == null) {
                return false;
            }
            if (ehueList.contains(eh)==false) {
                return false;
            }
            SGVariable ehVar = varList.get(eh.intValue());
            ehVarList.add(ehVar);
        }

        List<SGVariable> thVarList = new ArrayList<SGVariable>();
        for (int ii = 0; ii < tVarList.size(); ii++) {
            String col = tColList.get(ii);
            Integer th = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(col, this.mNCFile);
            if (th == null) {
                return false;
            }
            SGVariable thVar = varList.get(th.intValue());
            thVarList.add(thVar);
        }

        SGVariable[] xVars = xVarList.toArray(new SGVariable[xVarList.size()]);
        SGVariable[] yVars = yVarList.toArray(new SGVariable[yVarList.size()]);
        SGVariable[] lVars = null;
        SGVariable[] uVars = null;
        SGVariable[] ehVars = null;
        if (lVarList.size() != 0) {
            lVars = lVarList.toArray(new SGVariable[lVarList.size()]);
            uVars = uVarList.toArray(new SGVariable[uVarList.size()]);
            ehVars = ehVarList.toArray(new SGVariable[ehVarList.size()]);
        }
        SGVariable[] tVars = null;
        SGVariable[] thVars = null;
        if (tVarList.size() != 0) {
            tVars = tVarList.toArray(new SGVariable[tVarList.size()]);
            thVars = thVarList.toArray(new SGVariable[thVarList.size()]);
        }
        
        // set to the attributes
        this.mXVariables = xVars;
        this.mYVariables = yVars;
        this.mLowerErrorVariables = lVars;
        this.mUpperErrorVariables = uVars;
        this.mErrorBarHolderVariables = ehVars;
        this.mTickLabelVariables = tVars;
        this.mTickLabelHolderVariables = thVars;
        this.mTimeVariable = timeVar;
        this.mSerialNumberVariable = serialNumberVar;
        
        this.updateIsSingleVariableDateFlag();
        
        return true;
    }

    /**
     * Sets the data.
     * 
     * @param data
     *             data set to this object
     * @return
     *             true if succeeded            
     */
    @Override
    public boolean setData(SGData data) {
        if (!(data instanceof SGSXYMultipleVariableNetCDFData)) {
            throw new IllegalArgumentException("Given data is not the instance of SGSXYMultipleNetCDFData class.");
        }
        if (super.setData(data) == false) {
            return false;
        }
        SGSXYMultipleVariableNetCDFData nData = (SGSXYMultipleVariableNetCDFData) data;
        this.mXVariables = (SGVariable[]) nData.mXVariables.clone();
        this.mYVariables = (SGVariable[]) nData.mYVariables.clone();
        this.updateIsSingleVariableDateFlag();
        if (nData.isErrorBarAvailable()) {
            this.mLowerErrorVariables = (SGVariable[]) nData.mLowerErrorVariables.clone();
            this.mUpperErrorVariables = (SGVariable[]) nData.mUpperErrorVariables.clone();
        	this.mErrorBarHolderVariables = (SGVariable[]) nData.mErrorBarHolderVariables.clone();
        }
        if (nData.mTickLabelVariables!=null) {
        	this.mTickLabelVariables = (SGVariable[]) nData.mTickLabelVariables.clone();
        	this.mTickLabelHolderVariables = (SGVariable[]) nData.mTickLabelHolderVariables.clone();
        }
        return true;
    }
    
    /**
     * Set properties to this data.
     * 
     * @param p
     *          properties to be set
     * @return true if succeeded
     */
    public boolean setProperties(SGProperties p) {
        if (!(p instanceof SXYMultipleVariableNetCDFDataProperties)) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        SXYMultipleVariableNetCDFDataProperties sp = (SXYMultipleVariableNetCDFDataProperties) p;
        SGVariable[] xVars = new SGVariable[sp.xNames.length];
        for (int ii = 0; ii < xVars.length; ii++) {
            xVars[ii] = this.mNCFile.findVariable(sp.xNames[ii]);
        }
        SGVariable[] yVars = new SGVariable[sp.yNames.length];
        for (int ii = 0; ii < yVars.length; ii++) {
            yVars[ii] = this.mNCFile.findVariable(sp.yNames[ii]);
        }
        SGVariable[] lVars = null;
        if (sp.lNames != null) {
            lVars = new SGVariable[sp.lNames.length];
            for (int ii = 0; ii < lVars.length; ii++) {
                lVars[ii] = this.mNCFile.findVariable(sp.lNames[ii]);
            }
        }
        SGVariable[] uVars = null;
        if (sp.uNames != null) {
            uVars = new SGVariable[sp.uNames.length];
            for (int ii = 0; ii < uVars.length; ii++) {
                uVars[ii] = this.mNCFile.findVariable(sp.uNames[ii]);
            }
        }
        SGVariable[] ehVars = null;
        if (sp.ehNames != null) {
            ehVars = new SGVariable[sp.ehNames.length];
            for (int ii = 0; ii < ehVars.length; ii++) {
                ehVars[ii] = this.mNCFile.findVariable(sp.ehNames[ii]);
            }
        }
        SGVariable[] tVars = null;
        if (sp.tNames != null) {
            tVars = new SGVariable[sp.tNames.length];
            for (int ii = 0; ii < tVars.length; ii++) {
                tVars[ii] = this.mNCFile.findVariable(sp.tNames[ii]);
            }
        }
        SGVariable[] thVars = null;
        if (sp.thNames != null) {
        	thVars = new SGVariable[sp.thNames.length];
            for (int ii = 0; ii < thVars.length; ii++) {
                thVars[ii] = this.mNCFile.findVariable(sp.thNames[ii]);
            }
        }
        
        // set to the attributes
        this.mXVariables = xVars;
        this.mYVariables = yVars;
        this.mLowerErrorVariables = lVars;
        this.mUpperErrorVariables = uVars;
        this.mErrorBarHolderVariables = ehVars;
        this.mTickLabelVariables = tVars;
        this.mTickLabelHolderVariables = thVars;
        
        this.updateIsSingleVariableDateFlag();
        
        return true;
    }
    
    /**
     * Returns the properties of this data.
     * 
     * @return the properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new SXYMultipleVariableNetCDFDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Gets the properties of this data.
     * @param p
     *      the properties of this data
     * @return
     *      true if succeeded   
     */
    public boolean getProperties(SGProperties p) {
        if (!(p instanceof SXYMultipleVariableNetCDFDataProperties)) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        SXYMultipleVariableNetCDFDataProperties sp = (SXYMultipleVariableNetCDFDataProperties) p;
        String[] xNames = new String[this.mXVariables.length];
        for (int ii = 0; ii < xNames.length; ii++) {
            xNames[ii] = this.mXVariables[ii].getName();
        }
        String[] yNames = new String[this.mYVariables.length];
        for (int ii = 0; ii < yNames.length; ii++) {
            yNames[ii] = this.mYVariables[ii].getName();
        }
        String[] lNames = null;
        String[] uNames = null;
        String[] ehNames = null;
        if (this.isErrorBarAvailable()) {
        	lNames = new String[this.mLowerErrorVariables.length];
        	for (int ii = 0; ii < lNames.length; ii++) {
        		lNames[ii] = this.mLowerErrorVariables[ii].getName();
        	}
        	uNames = new String[this.mUpperErrorVariables.length];
        	for (int ii = 0; ii < uNames.length; ii++) {
        		uNames[ii] = this.mUpperErrorVariables[ii].getName();
        	}
        	ehNames = new String[this.mErrorBarHolderVariables.length];
        	for (int ii = 0; ii < ehNames.length; ii++) {
        		ehNames[ii] = this.mErrorBarHolderVariables[ii].getName();
        	}
        }
        String[] tNames = null;
        String[] thNames = null;
        if (this.mTickLabelVariables!=null) {
        	tNames = new String[this.mTickLabelVariables.length];
        	for (int ii = 0; ii < tNames.length; ii++) {
        		tNames[ii] = this.mTickLabelVariables[ii].getName();
        	}
        	thNames = new String[this.mTickLabelHolderVariables.length];
        	for (int ii = 0; ii < thNames.length; ii++) {
        		thNames[ii] = this.mTickLabelHolderVariables[ii].getName();
        	}
        }
        sp.xNames = xNames;
        sp.yNames = yNames;
        sp.lNames = lNames;
        sp.uNames = uNames;
        sp.ehNames = ehNames;
        sp.tNames = tNames;
        sp.thNames = thNames;
        return true;
    }
    
    /**
     * Read properties.
     * 
     * @param el
     *           an Element object
     * @return true if succeeded
     */
    public boolean readProperty(Element el) {
        
        if (super.readProperty(el) == false) {
            return false;
        }
        
        String str = null;
        String[] names = null;
        SGVariable[] xVars = null;
        SGVariable[] yVars = null;
        SGVariable[] lVars = null;
        SGVariable[] uVars = null;
        SGVariable[] ehVars = null;
        SGVariable[] tVars = null;
        SGVariable[] thVars = null;
        
        // x variables
//        str = el.getAttribute(KEY_X_VALUE_NAME);
//        if (str.length() != 0) {
//            SGVariable var = this.mNCFile.findVariable(str);
//            if (var == null) {
//                return false;
//            }
//            xVars = new SGVariable[] { var };
//        } else if ((names = this.readNames(el, KEY_X_VALUE_NAMES)) != null) {
//        	xVars = this.readVariables(names);
//        	if (xVars == null) {
//        		return false;
//        	}
//        } else {
//            return false;
//        }
        str = el.getAttribute(KEY_X_VALUE_NAMES);
        if (str.length() != 0) {
            SGVariable var = this.mNCFile.findVariable(str);
            if (var == null) {
                return false;
            }
        	xVars = this.readVariables(names);
        } else {
            return false;
        }
        
        // y variables
//        names = this.readNames(el, KEY_Y_VALUE_NAMES);
//        if (names == null) {
//            return false;
//        }
//    	yVars = this.readVariables(names);
//    	if (yVars == null) {
//    		return false;
//    	}
        str = el.getAttribute(KEY_Y_VALUE_NAMES);
        if (str.length() != 0) {
            SGVariable var = this.mNCFile.findVariable(str);
            if (var == null) {
                return false;
            }
        	yVars = this.readVariables(names);
        } else {
            return false;
        }
        
        // lower errors
        names = this.readNames(el, KEY_LOWER_ERROR_VALUE_NAMES);
        if (names != null) {
        	lVars = this.readVariables(names);
        	if (lVars == null) {
        		return false;
        	}
        }

        // upper errors
        names = this.readNames(el, KEY_UPPER_ERROR_VALUE_NAMES);
        if (names != null) {
        	uVars = this.readVariables(names);
        	if (uVars == null) {
        		return false;
        	}
        }

        // error bar holder
        names = this.readNames(el, KEY_ERROR_BAR_HOLDER_NAMES);
        if (names != null) {
        	ehVars = this.readVariables(names);
        	if (ehVars == null) {
        		return false;
        	}
        }

        if (!(lVars == null && uVars == null && ehVars == null) 
        		&& !(lVars != null && uVars != null && ehVars != null)) {
            return false;
        }

        // tick labels
        names = this.readNames(el, KEY_TICK_LABEL_NAMES);
        if (names != null) {
        	tVars = this.readVariables(names);
        	if (tVars == null) {
        		return false;
        	}
        }

        // tick label holder
        names = this.readNames(el, KEY_TICK_LABEL_HOLDER_NAMES);
        if (names != null) {
        	thVars = this.readVariables(names);
        	if (thVars == null) {
        		return false;
        	}
        }

        if (!(tVars == null && thVars == null) 
        		&& !(tVars != null && thVars != null)) {
            return false;
        }

        // set to attribute
        this.mXVariables = xVars;
        this.mYVariables = yVars;
        this.mLowerErrorVariables = lVars;
        this.mUpperErrorVariables = uVars;
        this.mErrorBarHolderVariables = ehVars;
        this.mTickLabelVariables = tVars;
        this.mTickLabelHolderVariables = thVars;

        this.updateIsSingleVariableDateFlag();
        
        return true;
    }
    
    protected SGVariable[] readVariables(final String[] names) {
        SGVariable[] vars = new SGVariable[names.length];
        for (int ii = 0; ii < names.length; ii++) {
            SGVariable var = this.mNCFile.findVariable(names[ii]);
            if (var == null) {
                return null;
            }
            vars[ii] = var;
        }
        return vars;
    }

//    /**
//     * Save values to given file.
//     * 
//     * @param data
//     *             a data
//     * @return
//     *             true if succeeded
//     */
//    public boolean saveData(final File file) {
//        boolean status = false;
//        Variable[] vArray = this.getVariables();
//        try {
//            FileWriter fw = new FileWriter(file);
//            for (int ii = 0; ii < vArray.length; ii++) {
//                String title = this.getNameWithUnit(vArray[ii]);
//                if (title == null) {
//                    title = "";
//                }
//                String str = SGUtilityText.getCSVString(title);
//                if (ii != 0) {
//                    fw.write(",");
//                }
//                fw.write(str);
//            }
//            fw.write("\n");
//            status = this.writeData(fw);
//            fw.close();
//        } catch (IOException ex) {
//            return false;
//        }
//        return status;
//    }
    
//    /**
//     * Write this data to a given output stream.
//     * 
//     * @param w
//     *          an output stream
//     * @return
//     *          true if succeeded
//     */
//    protected boolean writeData(Writer w) throws IOException {
//        double[] xArray = this.getXValueArray();
//        double[][] yArray = this.getYValueArray();
//        StringBuffer sb = new StringBuffer();
//        for (int ii = 0; ii < xArray.length; ii++) {
//            sb.append(xArray[ii]);
//            sb.append(',');
//            for (int jj = 0; jj < yArray.length; jj++) {
//                sb.append(yArray[jj][ii]);
//                if (jj != yArray.length - 1) {
//                    sb.append(',');
//                }
//            }
//            sb.append('\n');
//        }
//        w.write(sb.toString());
//        return true;
//    }

    /**
     * Returns an array of values to save to a data set.
     * 
     * @return an array of values to save to a data set
     */
    protected double[][] getDataSetValues() {
        double[][] xArray = this.getXValueArray();
        double[][] yArray = this.getYValueArray();
        final int pNum = xArray.length;
        double[][] values = new double[pNum][];
        for (int ii = 0; ii < pNum; ii++) {
            values[ii] = new double[xArray.length + yArray.length];
            for (int jj = 0; jj < xArray.length; jj++) {
                values[ii][jj] = xArray[jj][ii];
            }
            for (int jj = 0; jj < yArray.length; jj++) {
                values[ii][jj + xArray.length] = yArray[jj][ii];
            }
        }
//      TODO fix after the specification of data set is determined 
        return values;
    }
    
    /**
     * Return string that variable names are binded into curly bracket comma separatedly.
     * 
     * @param variables
     * @return
     */
    protected String bindVariableNamesInBracket(SGVariable[] variables) {
        StringBuffer sb = new StringBuffer("{");
        for (int ii = 0; ii < variables.length; ii++) {
            sb.append(variables[ii].getOriginalName());
            if (ii!=variables.length-1) {
                sb.append(",");
            }
        }
        sb.append("}");
        return sb.toString();
    }
    
    /**
     * Write properties of this object to the Element.
     * 
     * @param el
     *            the Element object
     * @param type
     *            type of the method to save properties
     * @return true if succeeded
     */
    @Override
    public boolean writeProperty(Element el, final int type) {
        if (super.writeProperty(el, type) == false) {
            return false;
        }
        
        String value = null;
        
        value = this.bindVariableNamesInBracket(this.mXVariables);
        el.setAttribute(KEY_X_VALUE_NAMES, value);
        value = this.bindVariableNamesInBracket(this.mYVariables);
        el.setAttribute(KEY_Y_VALUE_NAMES, value);
        if (this.isErrorBarAvailable()) {
            value = this.bindVariableNamesInBracket(this.mLowerErrorVariables);
            el.setAttribute(KEY_LOWER_ERROR_VALUE_NAMES, value);
            value = this.bindVariableNamesInBracket(this.mUpperErrorVariables);
            el.setAttribute(KEY_UPPER_ERROR_VALUE_NAMES, value);
            value = this.bindVariableNamesInBracket(this.mErrorBarHolderVariables);
            el.setAttribute(KEY_ERROR_BAR_HOLDER_NAMES, value);
        }
        if (this.mTickLabelVariables!=null) {
            value = this.bindVariableNamesInBracket(this.mTickLabelVariables);
            el.setAttribute(KEY_TICK_LABEL_NAMES, value);
            value = this.bindVariableNamesInBracket(this.mTickLabelHolderVariables);
            el.setAttribute(KEY_TICK_LABEL_HOLDER_NAMES, value);
        }
        return true;
    }
    
    /**
     * Returns a text string of data type.
     * 
     * @return a text string of data type
     */
    public String getDataType() {
        return SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DATA;
    }

    /**
     * Returns an array of SXYData objects.
     * 
     * @return an array of SXYData objects
     */
    public SGISXYTypeData[] getSXYDataArray() {
    	SGVariable varSingle = this.getSingleVariable();
    	SGVariable[] varMultiple = this.getMultipleVariables();
        final int len = varMultiple.length;

    	String nameSingle = varSingle.getName();
        String[] namesMultiple = new String[len];
        for (int ii = 0; ii < len; ii++) {
            namesMultiple[ii] = varMultiple[ii].getName();
        }
    	final String timeName = (this.mTimeVariable != null) ? this.mTimeVariable.getName() : null;
    	final String indexName = (this.mSerialNumberVariable != null) ? this.mSerialNumberVariable.getName() : null;
    	
        List<SGVariable> varList = this.mNCFile.getVariables();
        final boolean by = this.hasMultipleYValues();
        final boolean be = this.isErrorBarAvailable();
        final boolean bt = this.mTickLabelVariables!=null;
        
        final SGISXYTypeData[] dataArray = new SGISXYTypeData[len];
        for (int ii = 0; ii < dataArray.length; ii++) {
            String xName = by ? nameSingle : namesMultiple[ii];
            String yName = by ? namesMultiple[ii] : nameSingle;
            String lName = null;
            String uName = null;
            String ehName = null;
            if (be) {
                for (int jj = 0; jj < this.mErrorBarHolderVariables.length; jj++) {
                    String name = this.mErrorBarHolderVariables[jj].getName();
                    if (xName.equals(name) || yName.equals(name)) {
                        ehName = name;
                        lName = this.mLowerErrorVariables[jj].getName();
                        uName = this.mUpperErrorVariables[jj].getName();
                        break;
                    }
                }
            }
            String tName = null;
            String thName = null;
            if (bt) {
                for (int jj = 0; jj < this.mTickLabelHolderVariables.length; jj++) {
                    String name = this.mTickLabelHolderVariables[jj].getName();
                    if (xName.equals(name) || yName.equals(name)) {
                        thName = name;
                        tName = this.mTickLabelVariables[jj].getName();
                        break;
                    }
                }
            }
            SGNetCDFData data;
            if (this.mNCFile.findVariable(xName) instanceof SGDateVariable ||
                    this.mNCFile.findVariable(yName) instanceof SGDateVariable) {
                data = new SGSXYNetCDFDateData(this.mNCFile, this.mObserver, 
                        xName, yName, lName, uName, ehName, tName, thName, timeName,
                        indexName);
            } else {
                data = new SGSXYNetCDFData(this.mNCFile, this.mObserver, 
                        xName, yName, lName, uName, ehName, tName, thName, timeName,
                        indexName);
            }
            for (SGVariable var : varList) {
                if (var.isCoordinateVariable()) {
                    String name = var.getName();
                    if (this.mOriginMap.containsKey(name)) {
                        final int origin = this.mOriginMap.get(name);
                        data.setOrigin(name, origin);
                    }
                }
            }
            
            // sets the sub cache
            SGSXYNetCDFData d = (SGSXYNetCDFData) data;
            d.setCacheMap(this.mVariableCache);
            
            dataArray[ii] = (SGISXYTypeData) data;
        }
        return dataArray;
    }

    /**
     * Returns the maximum number of SXY data objects that can be obtained from
     * this data object.
     * 
     * @return the maximum number of SXY data objects that can be obtained from
     *         this data object
     */
    public int getSXYDataNumMax() {
        int max = 0;
        List<SGVariable> varList = this.mNCFile.getVariables();
        for (SGVariable cVar : varList) {
            if (cVar.isCoordinateVariable()) {
                Dimension cDim = cVar.getDimension(0);
                int cnt = 0;
                for (SGVariable var : varList) {
                    List<Dimension> dimList = var.getDimensions();
                    if (dimList.contains(cDim)) {
                        cnt++;
                    }
                }
                if (cnt > max) {
                    max = cnt;
                }
            }
        }
        return max;
    }

    /**
	 * Returns the m-th x value at the n-th index.
	 * 
	 * @param m
	 *            the index of data
	 * @param n
	 *            the index of the data point
	 * @return the n-th x-value of the m-th data
	 */
	public Double getXValue(final int m, final int n) {
        final int len = this.getPointsNumber();
        if (n < 0 || n >= len) {
            throw new IllegalArgumentException("given index is out of bounds: " + n);
        }
        if (m < 0 || m >= this.mXVariables.length) {
            throw new IllegalArgumentException("given index is out of bounds: " + m);
        }
        return super.getXValue(m, n);
    }

    /**
     * Returns the m-th y value at the n-th index.
     * 
     * @param m
     *          the index of data
     * @param n
     *          the index of the data point
     * @return
     *          the n-th y-value of the m-th data
     */
    public Double getYValue(final int m, final int n) {
        final int len = this.getPointsNumber();
        if (n < 0 || n >= len) {
            throw new IllegalArgumentException("given index is out of bounds: " + n);
        }
        if (m < 0 || m >= this.mYVariables.length) {
            throw new IllegalArgumentException("given index is out of bounds: " + m);
        }
        return super.getYValue(m, n);
    }
    
    /**
     * Returns an array of values of multiple variables.
     * 
     * @return an array of values of multiple variables
     */
    protected double[][] getMultipleVariableValueArray() {
        SGVariable[] cVars = new SGVariable[] { this.getCoordinateVariable() };
        SGVariable[] vars = this.getMultipleVariables();
        double[][] array = new double[vars.length][];
        for (int ii = 0; ii < array.length; ii++) {
            if (vars[ii].isCoordinateVariable()==false) {
                array[ii] = this.getValueArray(vars[ii], cVars);
            } else {
                array[ii] = this.getCoordinateValueArray(vars[ii]);
            }
        }
        return array;
    }

    /**
     * Returns the title for the X-axis.
     * 
     * @return the title for the X-axis
     */
    public String getTitleX() {
        if (this.mXVariables.length == 1 && this.mYVariables.length == 1) {
            return this.getNameWithUnit(this.mXVariables[0]);
        } else {
            if (this.hasMultipleYValues()) {
                return this.getNameWithUnit(this.mXVariables[0]);
            } else {
                return "";
            }
        }
    }

    /**
     * Returns the title for the Y-axis.
     * 
     * @return the title for the Y-axis
     */
    public String getTitleY() {
        if (this.mXVariables.length == 1 && this.mYVariables.length == 1) {
            return this.getNameWithUnit(this.mYVariables[0]);
        } else {
            if (this.hasMultipleYValues()) {
                return "";
            } else {
                return this.getNameWithUnit(this.mYVariables[0]);
            }
        }
    }
    
    /**
     * A class of multiple scalar XY netCDF data properties.
     */
    public static class SXYMultipleVariableNetCDFDataProperties extends NetCDFDataProperties {

        protected String[] xNames = null;
        
        protected String[] yNames = null;
        
        protected String[] lNames = null;
        
        protected String[] uNames = null;
        
        protected String[] ehNames = null;
        
        protected String[] tNames = null;
        
        protected String[] thNames = null;
        
        /**
         * The default constructor.
         */
        public SXYMultipleVariableNetCDFDataProperties() {
            super();
        }

        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.xNames = null;
            this.yNames = null;
            this.lNames = null;
            this.uNames = null;
            this.ehNames = null;
            this.tNames = null;
            this.thNames = null;
        }
        
        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof SXYMultipleVariableNetCDFDataProperties) == false) {
                return false;
            }
            if (super.equals(obj) == false) {
                return false;
            }
            SXYMultipleVariableNetCDFDataProperties p = (SXYMultipleVariableNetCDFDataProperties) obj;
            if (SGUtility.equals(this.xNames, p.xNames) == false) {
                return false;
            }
            if (SGUtility.equals(this.yNames, p.yNames) == false) {
                return false;
            }
            if (SGUtility.equals(this.lNames, p.lNames) == false) {
            	return false;
            }
            if (SGUtility.equals(this.uNames, p.uNames) == false) {
            	return false;
            }
            if (SGUtility.equals(this.ehNames, p.ehNames) == false) {
            	return false;
            }
            if (SGUtility.equals(this.tNames, p.tNames) == false) {
            	return false;
            }
            if (SGUtility.equals(this.thNames, p.thNames) == false) {
            	return false;
            }
            return true;
        }
    }
    
    /**
     * Returns whether error bars are available.
     * 
     * @return true if error bars are available
     */
    public boolean isErrorBarAvailable() {
		return (this.mLowerErrorVariables != null && this.mUpperErrorVariables != null);
    }
    
    /**
     * Returns whether tick labels are available.
     * 
     * @return true if tick labels are available
     */
    public boolean isTickLabelAvailable() {
        if (this.mIsSingleVariableDateFlag) {
            return true;
        }
		return (this.mTickLabelVariables != null);
    }

    /**
     * Returns whether the error bars are vertical. If this data does not have
     * error values, returns null.
     * 
     * @return true if error bars are vertical, false if they are horizontal and
     *         null if this data do not have error values
     */
    public Boolean isErrorBarVertical() {
        if (this.isErrorBarAvailable()) {
            return SGUtility.contains(this.mErrorBarHolderVariables, this.mYVariables);
        } else {
            return null;
        }
    }

    /**
     * Returns whether the tick labels align horizontally. If this data does not have
     * tick labels, returns null.
     * 
     * @return true if tick labels align horizontally, false if they do not so and
     *         null if this data do not have tick labels
     */
    public Boolean isTickLabelHorizontal() {
        if (this.mIsSingleVariableDateFlag) {
            if (this.hasMultipleYValues() && this.mXVariables[0] instanceof SGDateVariable) {
                return true;
            } else {
                return false;
            }
        }
        if (this.isTickLabelAvailable()) {
            return SGUtility.contains(this.mTickLabelHolderVariables, this.mYVariables);
        } else {
            return null;
        }
    }

    /**
     * Returns a map which has data information.
     * 
     * @return a map which has data information
     */
    public Map<String, Object> getInfoMap() {
        Map<String, Object> infoMap = super.getInfoMap();
        infoMap.put(KEY_MULTIPLE, Boolean.TRUE);
        infoMap.put(KEY_MULTIPLE_VARIABLE, Boolean.TRUE);
        if (null!=this.mDimensionStart) {
            infoMap.put(KEY_MULTIPLE_DIMENSION_START, this.mDimensionStart);
        }
        if (null!=this.mDimensionEnd) {
            infoMap.put(KEY_MULTIPLE_DIMENSION_END, this.mDimensionEnd);
        }
        if (null!=this.mDimensionStep) {
            infoMap.put(KEY_MULTIPLE_DIMENSION_STEP, this.mDimensionStep);
        }
        return infoMap;
    }

    /**
     * Returns an array of multiple SXYData objects.
     * 
     * @return an array of multiple SXYData objects
     */
    public SGISXYTypeMultipleData[] getSXYTypeMultipleDataArray() {
        SGISXYTypeData[] array = this.getSXYDataArray();
        SGISXYTypeMultipleData[] ret = new SGISXYTypeMultipleData[array.length];
        for (int ii = 0; ii < array.length; ii++) {
            SGSXYNetCDFData data = (SGSXYNetCDFData) array[ii];
            ret[ii] = data.toMultiple();
            
            // disposes data
            data.dispose();
        }
        return ret;
    }

    @Override
    public boolean isSplitEnabled() {
        SGVariable[] varMultiple = this.getMultipleVariables();
        final int len = varMultiple.length;
        if (len > 1) {
            return true;
        }
        return false;
    }
    
    public static SGISXYTypeMultipleData merge(final List<SGData> dataList) {
        if (dataList.size()==0) {
            return null;
        }
        for (SGData data : dataList) {
            if ((data instanceof SGSXYMultipleVariableNetCDFData)==false) {
                return null;
            }
        }
        
        SGSXYMultipleVariableNetCDFData dataLast = (SGSXYMultipleVariableNetCDFData)dataList.get(dataList.size()-1);
        SGNetCDFFile ncFile = dataLast.getNetcdfFile();
        SGDataSourceObserver obs = dataLast.mObserver;
        String time = dataLast.isTimeVariableAvailable() ? dataLast.getTimeVariable().getName() : null;
        String index = dataLast.isSerialNumberAvailable() ? dataLast.getSerialNumberVariable().getName() : null;
        
        Set<String> xset = new TreeSet<String>();
        Set<String> yset = new TreeSet<String>();
        Set<String> leset = new TreeSet<String>();
        Set<String> ueset = new TreeSet<String>();
        Set<String> ehset = new TreeSet<String>();
        Set<String> tset = new TreeSet<String>();
        Set<String> thset = new TreeSet<String>();
        for (SGData data : dataList) {
            SGSXYMultipleVariableNetCDFData dataMult = (SGSXYMultipleVariableNetCDFData)data;
            SGVariable[] xvar = dataMult.getXVariables();
            SGVariable[] yvar = dataMult.getYVariables();
            SGVariable[] levar = dataMult.getLowerErrorVariables();
            SGVariable[] uevar = dataMult.getUpperErrorVariables();
            SGVariable[] ehvar = dataMult.getErrorHolderVariables();
            SGVariable[] tvar = dataMult.getTickLabelVariables();
            SGVariable[] thvar = dataMult.getTickLabelHolderVariables();
            if (null!=xvar) {
                for (int i = 0; i < xvar.length; i++) {
                    xset.add(xvar[i].getName());
                }
            }
            if (null!=yvar) {
                for (int i = 0; i < yvar.length; i++) {
                    yset.add(yvar[i].getName());
                }
            }
            if (null!=levar) {
                for (int i = 0; i < levar.length; i++) {
                    leset.add(levar[i].getName());
                }
            }
            if (null!=uevar) {
                for (int i = 0; i < uevar.length; i++) {
                    ueset.add(uevar[i].getName());
                }
            }
            if (null!=ehvar) {
                for (int i = 0; i < ehvar.length; i++) {
                    ehset.add(ehvar[i].getName());
                }
            }
            if (null!=tvar) {
                for (int i = 0; i < tvar.length; i++) {
                    tset.add(tvar[i].getName());
                }
            }
            if (null!=thvar) {
                for (int i = 0; i < thvar.length; i++) {
                    thset.add(thvar[i].getName());
                }
            }
        }
        
        String[] x = new String[xset.size()];
        String[] y = new String[yset.size()];
        String[] le = new String[leset.size()];
        String[] ue = new String[ueset.size()];
        String[] eh = new String[ehset.size()];
        String[] t = new String[tset.size()];
        String[] th = new String[thset.size()];
        xset.toArray(x);
        yset.toArray(y);
        leset.toArray(le);
        ueset.toArray(ue);
        ehset.toArray(eh);
        tset.toArray(t);
        thset.toArray(th);
        
        SGSXYMultipleVariableNetCDFData data = new SGSXYMultipleVariableNetCDFData(
                ncFile, obs,
                x, y, le, ue, eh, t, th,
                time, index);
        data.mOriginMap = new HashMap<String, Integer>(dataLast.mOriginMap);
        data.setDecimalPlaces(dataLast.getDecimalPlaces());
        data.setExponent(dataLast.getExponent());
        
        return data;
    }
    
    private void updateIsSingleVariableDateFlag() {
        if (this.mXVariables.length==1 && this.mXVariables[0] instanceof SGDateVariable) {
            this.mIsSingleVariableDateFlag = true;
        } else if (this.mYVariables.length==1 && this.mYVariables[0] instanceof SGDateVariable) {
            this.mIsSingleVariableDateFlag = true;
        } else {
            this.mIsSingleVariableDateFlag = false;
        }
    }
    
    @Override
    public boolean isXVariableCoordinate() {
        if (this.getXVariables().length==1) {
            return this.getXVariables()[0].isCoordinateVariable();
        } else {
            return false;
        }
    }
    
}
