package jp.riken.brain.ni.samuraigraph.data;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2d;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;

import org.w3c.dom.Element;

/**
 * Scalar type XYZ data. This object has of x, y and z values. 
 *
 */
public class SGSXYZData extends SGArrayData implements SGISXYZTypeData,
        SGIDataColumnTypeConstants, SGIDataColumnNameConstants {

    /**
     * The column index for x-values.
     */
    protected Integer mXIndex = null;
    
    /**
     * The column index for y-values.
     */
    protected Integer mYIndex = null;

    /**
     * The column index for z-values.
     */
    protected Integer mZIndex = null;
    
    /**
     * Builds a data object.
     * 
     */
    public SGSXYZData() {
        super();
    }

    /**
     * Builds a data object with given columns and indices.
     * 
     * @param dataFile
     *                   the text data file
     * @param obs
     *                   an observer for array data
     * @param xIndex
     *                   the column index for x-values
     * @param yIndex
     *                   the column index for y-values
     * @param zIndex
     *                   the column index for z-values
     */
    public SGSXYZData(final SGTextDataFile dataFile,
    		final SGDataSourceObserver obs,
            final Integer xIndex, final Integer yIndex,
            final Integer zIndex) {
        
        super(dataFile, obs);
        
        // null check
        if (xIndex == null || yIndex == null || zIndex == null) {
            throw new IllegalArgumentException("Indices for values must not be null.");
        }
        
        // x-values
        if (this.checkColumnIndexRange(xIndex) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + xIndex);
        }
        
        // y-values
        if (this.checkColumnIndexRange(yIndex) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + yIndex);
        }

        // z-values
        if (this.checkColumnIndexRange(zIndex) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + zIndex);
        }
        
        // set indices
        this.mXIndex = xIndex;
        this.mYIndex = yIndex;
        this.mZIndex = zIndex;
    }

    /**
     * Builds a data object with given data.
     * 
     * @param data
     *             a data object that are set to this object
     */
    public SGSXYZData(final SGData data) {
        super();
        this.setData(data);
    }

    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
    	super.dispose();
    	this.mXIndex = null;
    	this.mYIndex = null;
    	this.mZIndex = null;
    }
    
    /**
     * Sets the data.
     * 
     * @param data
     *             data set to this object
     * @return
     *             true if succeeded            
     */
    public boolean setData(final SGData data) {
        
        if (!(data instanceof SGSXYZData)) {
            throw new IllegalArgumentException("!(data instanceof SGSXYZData)");
        }
        
        if (super.setData(data) == false) {
            return false;
        }

        SGSXYZData dataSXYZ = (SGSXYZData) data;
        this.mXIndex = dataSXYZ.mXIndex;
        this.mYIndex = dataSXYZ.mYIndex;
        this.mZIndex = dataSXYZ.mZIndex;
        
        return true;
    }
    
    /**
     * Returns a copy of x-value array.
     * 
     * @return
     *         x-value array
     */
    public double[] getXValueArray() {
//        return SGUtility.copyDoubleArray(this.getXArray());
    	return this.getXArray();
    }

    /**
     * Returns a copy of y-value array.
     * 
     * @return
     *         y-value array
     */
    public double[] getYValueArray() {
//        return SGUtility.copyDoubleArray(this.getYArray());
    	return this.getYArray();
    }

    /**
     * Returns a copy of z-value array.
     * 
     * @return
     *         z-value array
     */
    public double[] getZValueArray() {
//        return SGUtility.copyDoubleArray(this.getZArray());
    	return this.getZArray();
    }

    /**
     * Returns an array of xy-values.
     * 
     * @return an array of xy-values
     */
    public SGTuple2d[] getXYValueArray() {
        double[] xArray = this.getXArray();
        double[] yArray = this.getYArray();
        SGTuple2d[] cArray = new SGTuple2d[xArray.length];
        for (int ii = 0; ii < xArray.length; ii++) {
            cArray[ii] = new SGTuple2d(xArray[ii], yArray[ii]);
        }
        return cArray;
    }
    
    /**
     * Returns an array of x-values.
     * @return
     *         an array of x-values
     */
    protected double[] getXArray() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGNumberDataColumn col = (SGNumberDataColumn) columns[this.mXIndex.intValue()];
        return col.getNumberArray();
    }

    /**
     * Returns an array of y-values.
     * @return
     *         an array of y-values
     */
    protected double[] getYArray() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGNumberDataColumn col = (SGNumberDataColumn) columns[this.mYIndex.intValue()];
        return col.getNumberArray();
    }

    /**
     * Returns an array of z-values.
     * @return
     *         an array of z-values
     */
    protected double[] getZArray() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGNumberDataColumn col = (SGNumberDataColumn) columns[this.mZIndex.intValue()];
        return col.getNumberArray();
    }

    /**
     * Returns a number of data points.
     * 
     * @return
     *         a number of data points
     */
    public int getPointsNumber() {
        return this.getXYValueArray().length;
    }

    /**
     * Returns the type of data.
     * 
     * @return
     *         the type of data
     */
    public String getDataType() {
        return SGDataTypeConstants.SXYZ_DATA;
    }

//    /**
//     * Returns the bounds of x-values.
//     * 
//     * @return the bounds of x-values
//     */
//    public SGTuple2d getBoundsX() {
//        return SGDataUtility.getBoundsX(this);
//    }
//
//    /**
//     * Returns the bounds of y-values.
//     * 
//     * @return the bounds of y-values
//     */
//    public SGTuple2d getBoundsY() {
//        return SGDataUtility.getBoundsY(this);
//    }
    
    
    /**
     * Returns an array of current column types.
     * @return
     *        an array of current column types
     */
    public String[] getCurrentColumnType() {
        final int colNum = this.getColNum();
        String[] array = new String[colNum];
        if (this.mXIndex != null) {
            array[this.mXIndex.intValue()] = X_VALUE;
        }
        if (this.mYIndex != null) {
            array[this.mYIndex.intValue()] = Y_VALUE;
        }
        if (this.mZIndex != null) {
            array[this.mZIndex.intValue()] = Z_VALUE;
        }
        return array;
    }

    /**
     * Returns an array of data columns to save to a data set.
     * 
     * @return an array of data columns to save to a data set
     */
    protected SGDataColumn[] getDataSetColumns() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGDataColumn[] colArray = new SGDataColumn[3];
        colArray[0] = columns[this.mXIndex.intValue()];
        colArray[1] = columns[this.mYIndex.intValue()];
        colArray[2] = columns[this.mZIndex.intValue()];
        
        colArray[0].setColumnType(X_VALUE);
        colArray[1].setColumnType(Y_VALUE);
        colArray[2].setColumnType(Z_VALUE);
        return colArray;
    }
    
    @Override
    public SGDataColumn[] getUsedDataColumnsClone() {
        return this.getDataSetColumnsClone();
    }

    /**
     * Sets the type of data columns.
     * 
     * @param column
     *           an array of column types
     * @return true if succeeded
     */
    public boolean setColumnType(String[] columns) {
        if (columns == null) {
            throw new IllegalArgumentException("columns == null");
        }
        Integer x = null;
        Integer y = null;
        Integer z = null;
        for (int ii = 0; ii < columns.length; ii++) {
            if (SGUtilityText.isEqualString(X_VALUE, columns[ii])) {
                x = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(Y_VALUE, columns[ii])) {
                y = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(Z_VALUE, columns[ii])) {
                z = Integer.valueOf(ii);
            } else if ("".equals(columns[ii])) {
                continue;
            } else {
            	return false;
            }
        }
        
        // check the necessary columns
        if (x == null || y == null || z == null) {
        	return false;
        }
        
        // set to the attributes
        this.mXIndex = x;
        this.mYIndex = y;
        this.mZIndex = z;

        return true;
    }

    /**
     * Returns properties of this data.
     * @return
     *         properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new SXYZDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Get properties of this data.
     * @param p
     *          properties to set values
     * @return
     *          true if succeeded
     */
    public boolean getProperties(final SGProperties p) {
        if ((p instanceof SXYZDataProperties) == false) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        SXYZDataProperties ep = (SXYZDataProperties) p;
        ep.mXIndex = this.mXIndex;
        ep.mYIndex = this.mYIndex;
        ep.mZIndex = this.mZIndex;
        return true;
    }

    /**
     * Set properties to this data.
     * @param p
     *          properties that have values to set to this data
     * @return
     *          true if succeeded
     */
    public boolean setProperties(final SGProperties p) {
        if ((p instanceof SXYZDataProperties) == false) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        SXYZDataProperties ep = (SXYZDataProperties) p;
        this.mXIndex = ep.mXIndex;
        this.mYIndex = ep.mYIndex;
        this.mZIndex = ep.mZIndex;
        return true;
    }
    
    /**
     * Writes data column indices in the attributes.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected boolean writeAttributeColumnIndices(Element el) {
        el.setAttribute(KEY_X_VALUE_COLUMN_INDEX, this.mXIndex.toString());
        el.setAttribute(KEY_Y_VALUE_COLUMN_INDEX, this.mYIndex.toString());
        el.setAttribute(KEY_Z_VALUE_COLUMN_INDEX, this.mZIndex.toString());
        return true;
    }

    /**
     * Writes data column indices as sequential numbers.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected boolean writeSequentialColumnIndices(Element el) {
        el.setAttribute(KEY_X_VALUE_COLUMN_INDEX, Integer.toString(0));
        el.setAttribute(KEY_Y_VALUE_COLUMN_INDEX, Integer.toString(1));
        el.setAttribute(KEY_Z_VALUE_COLUMN_INDEX, Integer.toString(2));
        return true;
    }
    
    @Override
    protected boolean writeSequentialColumnName(Element el) {
        el.setAttribute(KEY_X_VALUE_COLUMN_INDEX, this.getSequentialColumnName(0));
        el.setAttribute(KEY_Y_VALUE_COLUMN_INDEX, this.getSequentialColumnName(1));
        el.setAttribute(KEY_Z_VALUE_COLUMN_INDEX, this.getSequentialColumnName(2));
        
        // serial numbers
        el.setAttribute(KEY_SERIAL_NUMBER_VARIABLE_NAME, SERIAL_NUMBER);
        
        return true;
    }

    /**
     * Read properties from a given Element and set to this data.
     * @param el
     *          an Element
     * @return
     *          true if succeeded
     */
    public boolean readProperty(Element el) {
        Integer index = null;
        Integer xIndex = null;
        Integer yIndex = null;
        Integer zIndex = null;
        if ((index = this.readIndex(el, KEY_X_VALUE_COLUMN_INDEX)) != null) {
            if (this.checkColumnIndexRange(index) == false) {
                return false;
            }
            xIndex = index;
        } else {
            return false;
        }
        if ((index = this.readIndex(el, KEY_Y_VALUE_COLUMN_INDEX)) != null) {
            if (this.checkColumnIndexRange(index) == false) {
                return false;
            }
            yIndex = index;
        } else {
            return false;
        }
        if ((index = this.readIndex(el, KEY_Z_VALUE_COLUMN_INDEX)) != null) {
            if (this.checkColumnIndexRange(index) == false) {
                return false;
            }
            zIndex = index;
        } else {
            return false;
        }
        
        // set to attributes
        this.mXIndex = xIndex;
        this.mYIndex = yIndex;
        this.mZIndex = zIndex;

        return true;
    }

    /**
     * A class for scalar xyz-type data properties.
     *
     */
    public static class SXYZDataProperties extends ArrayDataProperties {
        
        // indices of data columns
        Integer mXIndex = null;

        Integer mYIndex = null;
        
        Integer mZIndex = null;
        
        /**
         * The default constructor.
         */
        public SXYZDataProperties() {
            super();
        }

        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof SXYZDataProperties) == false) {
                return false;
            }

            SXYZDataProperties p = (SXYZDataProperties) obj;
            if (this.mXIndex != null) {
                if (!this.mXIndex.equals(p.mXIndex)) {
                    return false;
                }
            } else {
                if (p.mXIndex != null) {
                    return false;
                }
            }
            if (this.mYIndex != null) {
                if (!this.mYIndex.equals(p.mYIndex)) {
                    return false;
                }
            } else {
                if (p.mYIndex != null) {
                    return false;
                }
            }
            if (this.mZIndex != null) {
                if (!this.mZIndex.equals(p.mZIndex)) {
                    return false;
                }
            } else {
                if (p.mZIndex != null) {
                    return false;
                }
            }
            return true;
        }

        /**
         * Returns string representation of this object.
         * @return
         *        string representation of this object
         */
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("[");
            sb.append("XIndex=");
            sb.append(this.mXIndex);
            sb.append(", YIndex=");
            sb.append(this.mYIndex);
            sb.append(", ZIndex=");
            sb.append(this.mZIndex);
            sb.append("]");
            return sb.toString();
        }
        
        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.mXIndex = null;
            this.mYIndex = null;
            this.mZIndex = null;
        }
    }

    /**
     * Returns the bounds of x-values.
     * 
     * @return the bounds of x-values
     */
    public SGValueRange getBoundsX() {
        return SGDataUtility.getBoundsX(this);
    }

    /**
     * Returns the bounds of y-values.
     * 
     * @return the bounds of y-values
     */
    public SGValueRange getBoundsY() {
        return SGDataUtility.getBoundsY(this);
    }

    /**
     * Returns the bounds of z-values.
     * 
     * @return the bounds of z-values
     */
    public SGValueRange getBoundsZ() {
        return SGDataUtility.getBoundsZ(this);
    }

    /**
     * Returns the title of X values.
     * 
     * @return
     *         the title of X values
     */
    public String getTitleX() {
        if (this.mXIndex == null) {
            return "";
        } else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mXIndex.intValue()].getTitle();
        }
    }

    /**
     * Returns the title of Y values.
     * 
     * @return
     *         the title of Y values
     */
    public String getTitleY() {
        if (this.mYIndex == null) {
            return "";
        } else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mYIndex.intValue()].getTitle();
        }
    }

    /**
     * Returns the title of Z values.
     * 
     * @return
     *         the title of Z values
     */
    public String getTitleZ() {
        if (this.mZIndex == null) {
            return "";
        } else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mZIndex.intValue()].getTitle();
        }
    }

	public SGDataColumn getXColumn() {
		return this.getColumn(this.mXIndex);
	}

	public SGDataColumn getYColumn() {
		return this.getColumn(this.mYIndex);
	}

	public SGDataColumn getZColumn() {
		return this.getColumn(this.mZIndex);
	}

}
