package jp.riken.brain.ni.samuraigraph.data;

import java.util.ArrayList;
import java.util.List;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2d;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;

import org.w3c.dom.Element;

import ucar.ma2.Array;
import ucar.nc2.Dimension;

public class SGSXYZNetCDFData extends SGNetCDFData implements 
    SGISXYZTypeData, SGIDataColumnTypeConstants, SGIDataColumnNameConstants {

    /**
     * The variable for x-values.
     */
    protected SGVariable mXVariable = null;
    
    /**
     * The variable for y-values.
     */
    protected SGVariable mYVariable = null;

    /**
     * The variable for z-values.
     */
    protected SGVariable mZVariable = null;
    
    /**
     * The default constructor.
     *
     */
    public SGSXYZNetCDFData() {
        super();
    }

    /**
     * Builds a data object with given netCDF data.
     * 
     * @param ncfile
     *            netCDF data file
     * @param obs
     *            netCDF data observer
     * @param xName
     *            the name of x-values
     * @param yName
     *            the name of y-values
     * @param zName
     *            the name of z-values
     * @param timeName
     *            the name of time values
     * @param serialNumberName
     *          the name of serial number variable
     */
    public SGSXYZNetCDFData(final SGNetCDFFile ncfile, SGDataSourceObserver obs,
            final String xName, final String yName, final String zName, 
            final String timeName, final String serialNumberName) {

    	super(ncfile, obs, timeName, serialNumberName);
        
        if (xName == null || yName == null || zName == null) {
            throw new IllegalArgumentException("xName == null || yName == null || zName == null");
        }
        
        // get variables
        SGVariable xVar = ncfile.findVariable(xName);
        if (xVar == null) {
            throw new IllegalArgumentException("Illegal variable name: " + xName);
        }
        SGVariable yVar = ncfile.findVariable(yName);
        if (yVar == null) {
            throw new IllegalArgumentException("Illegal variable name: " + yName);
        }
        SGVariable zVar = ncfile.findVariable(zName);
        if (zVar == null) {
            throw new IllegalArgumentException("Illegal variable name: " + zName);
        }

        //
        // check variables
        //
        List<Dimension> cDimList = new ArrayList<Dimension>();
        
        // time variable
    	if (this.mTimeVariable != null) {
    		cDimList.add(this.mTimeVariable.getDimension(0));
    	}
        
        if (this.isSerialNumberAvailable()) {
        	// x and y variables
            if (xVar.isCoordinateVariable() || yVar.isCoordinateVariable()
					|| zVar.isCoordinateVariable()) {
				throw new IllegalArgumentException(
						"x, y and z variables must not be coordinate variables.");
			}
            
        	// index variable
            Dimension iDim = this.mSerialNumberVariable.getDimension(0);
            cDimList.add(iDim);
            
            // x and y variables
            this.checkNotCoordinateVariable(xVar, cDimList);
            this.checkNotCoordinateVariable(yVar, cDimList);
        } else {
            if (!xVar.isCoordinateVariable() || !yVar.isCoordinateVariable()) {
                throw new IllegalArgumentException("x and y variables must be coordinate variable.");
            }
            if (zVar.isCoordinateVariable()) {
                throw new IllegalArgumentException("z variable must not be a coordinate variable.");
            }
            cDimList.add(xVar.getDimension(0));
            cDimList.add(yVar.getDimension(0));
        }

        // z variable
        this.checkNotCoordinateVariable(zVar, cDimList);
        
        // set to attributes
        this.mXVariable = xVar;
        this.mYVariable = yVar;
        this.mZVariable = zVar;
    }

    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
        this.mXVariable = null;
        this.mYVariable = null;
        this.mZVariable = null;
        
        // clear the cache
        this.mVariableCache.clear();
    }
    
    /**
     * Returns an array of current column types.
     * 
     * @return an array of current column types
     */
    public String[] getCurrentColumnType() {
        List<SGVariable> varList = this.mNCFile.getVariables();
        final int varNum = varList.size();
        String[] array = new String[varNum];
        for (int ii = 0; ii < varNum; ii++) {
        	SGVariable var = varList.get(ii);
            if (var.equals(this.mXVariable)) {
                array[ii] = X_VALUE;
            } else if (var.equals(this.mYVariable)) {
                array[ii] = Y_VALUE;
            } else if (var.equals(this.mZVariable)) {
                array[ii] = Z_VALUE;
            } else if (var.equals(this.mTimeVariable)) {
                array[ii] = TIME;
			} else if (var.equals(this.mSerialNumberVariable)) {
				array[ii] = SERIAL_NUMBER;
            } else {
                array[ii] = "";
            }
        }        
        return array;
    }

    /**
     * Returns the data type for the data set.
     * 
     * @return the data type
     */
    protected String getDataTypeForDataSet() {
        return SGDataTypeConstants.SXYZ_DATA;
    }

    /**
     * Returns an array of all variables.
     * 
     * @return an array of all variables
     */
    public SGVariable[] getVariables() {
    	List<SGVariable> varList = new ArrayList<SGVariable>();
    	varList.add(this.mXVariable);
    	varList.add(this.mYVariable);
    	varList.add(this.mZVariable);
        if (this.mTimeVariable != null) {
        	varList.add(this.mTimeVariable);
        }
    	if (this.mSerialNumberVariable != null) {
    		varList.add(this.mSerialNumberVariable);
    	}
        SGVariable[] varArray = new SGVariable[varList.size()];
        return (SGVariable[]) varList.toArray(varArray);
    }

    public SGVariable getXVariable() {
    	return this.mXVariable;
    }

    public SGVariable getYVariable() {
    	return this.mYVariable;
    }

    public SGVariable getZVariable() {
    	return this.mZVariable;
    }

    /**
     * Sets the type of data columns.
     * 
     * @param column
     *              an array of column types
     * @return true if succeeded
     */
    public boolean setColumnType(String[] columns) {
        
    	SGVariable xVar = null;
    	SGVariable yVar = null;
    	SGVariable zVar = null;
    	SGVariable timeVar = null;
    	SGVariable serialNumberVar = null;

        // set variables
        List<SGVariable> varList = this.mNCFile.getVariables();
        for (int ii = 0; ii < columns.length; ii++) {
        	SGVariable var = varList.get(ii);
            if (SGUtilityText.isEqualString(X_VALUE, columns[ii])) {
                xVar = var;
            } else if (SGUtilityText.isEqualString(Y_VALUE, columns[ii])) {
                yVar = var;
            } else if (SGUtilityText.isEqualString(Z_VALUE, columns[ii])) {
                zVar = var;
            } else if (SGUtilityText.isEqualString(TIME, columns[ii])) {
                timeVar = var;
            } else if (SGUtilityText.isEqualString(SERIAL_NUMBER, columns[ii])) {
                serialNumberVar = var;
            } else if ("".equals(columns[ii])) {
            	continue;
            } else {
            	return false;
            }
        }
        if (xVar == null || yVar == null || zVar == null) {
        	return false;
        }
        
        // set variables
        this.mXVariable = xVar;
        this.mYVariable = yVar;
        this.mZVariable = zVar;
        this.mTimeVariable = timeVar;
        this.mSerialNumberVariable = serialNumberVar;

        return true;
    }

    /**
     * Sets the data.
     * 
     * @param data
     *             data set to this object
     * @return
     *             true if succeeded            
     */
    public boolean setData(SGData data) {
        if (!(data instanceof SGSXYZNetCDFData)) {
            throw new IllegalArgumentException("Given data is not the instance of SGSXYZNetCDFData class.");
        }
        if (super.setData(data) == false) {
            return false;
        }
        SGSXYZNetCDFData nData = (SGSXYZNetCDFData) data;
        this.mXVariable = nData.mXVariable;
        this.mYVariable = nData.mYVariable;
        this.mZVariable = nData.mZVariable;
        return true;
    }
    
    /**
     * Returns the properties of this data.
     * 
     * @return the properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new SXYZNetCDFDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Gets the properties of this data.
     * @param p
     *      the properties of this data
     * @return
     *      true if succeeded   
     */
    public boolean getProperties(SGProperties p) {
        if (!(p instanceof SXYZNetCDFDataProperties)) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        SXYZNetCDFDataProperties sp = (SXYZNetCDFDataProperties) p;
        sp.xName = this.mXVariable.getName();
        sp.yName = this.mYVariable.getName();
        sp.zName = this.mZVariable.getName();
        return true;
    }
    
    /**
     * Set properties to this data.
     * 
     * @param p
     *          properties to be set
     * @return true if succeeded
     */
    public boolean setProperties(SGProperties p) {
        if (!(p instanceof SXYZNetCDFDataProperties)) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        SXYZNetCDFDataProperties sp = (SXYZNetCDFDataProperties) p;
        this.mXVariable = this.mNCFile.findVariable(sp.xName);
        this.mYVariable = this.mNCFile.findVariable(sp.yName);
        this.mZVariable = this.mNCFile.findVariable(sp.zName);
        return true;
    }
    
    /**
     * Write properties of this object to the Element.
     * 
     * @param el
     *            the Element object
     * @param type
     *            type of the method to save properties
     * @return true if succeeded
     */
    public boolean writeProperty(Element el, final int type) {
        if (super.writeProperty(el, type) == false) {
            return false;
        }
        el.setAttribute(KEY_X_VALUE_NAME, this.mXVariable.getName());
        el.setAttribute(KEY_Y_VALUE_NAME, this.mYVariable.getName());
        el.setAttribute(KEY_Z_VALUE_NAME, this.mZVariable.getName());
        return true;
    }

    /**
     * Read properties.
     * 
     * @param el
     *           an Element object
     * @return true if succeeded
     */
    public boolean readProperty(Element el) {
        
        if (super.readProperty(el) == false) {
            return false;
        }
        
        String str = null;
        SGVariable xVar = null;
        SGVariable yVar = null;
        SGVariable zVar = null;
        
        // x variable
        str = el.getAttribute(KEY_X_VALUE_NAME);
        if (str.length() != 0) {
            xVar = this.mNCFile.findVariable(str);
            if (xVar == null) {
                return false;
            }
        } else {
            return false;
        }
        
        // y variable
        str = el.getAttribute(KEY_Y_VALUE_NAME);
        if (str.length() != 0) {
            yVar = this.mNCFile.findVariable(str);
            if (yVar == null) {
                return false;
            }
        } else {
            return false;
        }

        // z variable
        str = el.getAttribute(KEY_Z_VALUE_NAME);
        if (str.length() != 0) {
            zVar = this.mNCFile.findVariable(str);
            if (zVar == null) {
                return false;
            }
        } else {
            return false;
        }
        
        // set to attribute
        this.mXVariable = xVar;
        this.mYVariable = yVar;
        this.mZVariable = zVar;

        return true;
    }
    
    /**
     * Returns a text string of data type.
     * 
     * @return a text string of data type
     */
    public String getDataType() {
        return SGDataTypeConstants.SXYZ_NETCDF_DATA;
    }

    /**
     * Returns an array of values to save to a data set.
     * 
     * @return an array of values to save to a data set
     */
    @Override
    public double[][] getDataSetValues() {
        final SGTuple2d[] xyArray = this.getXYValueArray();
        final double[] zArray = this.getZValueArray();
        final int pNum = xyArray.length;
        final int cNum = 3;
        final double[][] values = new double[pNum][];
        for (int ii = 0; ii < pNum; ii++) {
            values[ii] = new double[cNum];
            values[ii][0] = xyArray[ii].x;
            values[ii][1] = xyArray[ii].y;
            values[ii][2] = zArray[ii];
        }
        return values;
    }
    
    /**
     * A class of scalar XY netCDF data properties.
     */
    public static class SXYZNetCDFDataProperties extends NetCDFDataProperties {

        protected String xName = null;
        
        protected String yName = null;
        
        protected String zName = null;
        
        /**
         * The default constructor.
         */
        public SXYZNetCDFDataProperties() {
            super();
        }

        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.xName = null;
            this.yName = null;
            this.zName = null;
        }
        
        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof SXYZNetCDFDataProperties) == false) {
                return false;
            }
            if (super.equals(obj) == false) {
                return false;
            }
            SXYZNetCDFDataProperties p = (SXYZNetCDFDataProperties) obj;
            if (SGUtility.equals(this.xName, p.xName) == false) {
                return false;
            }
            if (SGUtility.equals(this.yName, p.yName) == false) {
                return false;
            }
            if (SGUtility.equals(this.zName, p.zName) == false) {
                return false;
            }
            return true;
        }
    }

    /**
     * Returns the number of points.
     * 
     * @return the number of points
     */
    public int getPointsNumber() {
    	final int num;
        if (this.isSerialNumberAvailable()) {
        	Dimension dim = this.mSerialNumberVariable.getDimension(0);
        	num = dim.getLength();
        } else {
        	Dimension dimX = this.mXVariable.getDimension(0);
        	Dimension dimY = this.mYVariable.getDimension(0);
        	num = dimX.getLength() * dimY.getLength();
        }
        return num;
    }

    /**
     * Returns an array of X-values.
     * 
     * @return an array of X-values
     */
    public double[] getXValueArray() {
        if (this.isSerialNumberAvailable()) {
            return this.getValueArray(this.mXVariable, new SGVariable[]{ this.mSerialNumberVariable });
        } else {
            return this.getCoordinateValueArray(this.mXVariable);
        }
    }

    /**
     * Returns an array of Y-values.
     * 
     * @return an array of Y-values
     */
    public double[] getYValueArray() {
        if (this.isSerialNumberAvailable()) {
            return this.getValueArray(this.mYVariable, new SGVariable[]{ this.mSerialNumberVariable });
        } else {
            return this.getCoordinateValueArray(this.mYVariable);
        }
    }

    /**
     * Returns an array of Z-values.
     * 
     * @return an array of Z-values
     */
    public double[] getZValueArray() {
        if (this.isSerialNumberAvailable()) {
			return this.getValueArray(this.mZVariable,
					new SGVariable[] { this.mSerialNumberVariable });
		} else {
			return this.getValueArray(this.mZVariable, new SGVariable[] {
					this.mXVariable, this.mYVariable });
		}
    }

    /**
	 * Returns an array of xy-values.
	 * 
	 * @return an array of xy-values
	 */
    public SGTuple2d[] getXYValueArray() {
        final double[] xArray = this.getXValueArray();
        final double[] yArray = this.getYValueArray();
        if (this.isSerialNumberAvailable()) {
        	final int len = xArray.length;
            SGTuple2d[] cArray = new SGTuple2d[len];
            for (int ii = 0; ii < len; ii++) {
            	cArray[ii] = new SGTuple2d(xArray[ii], yArray[ii]);
            }
            return cArray;
        } else {
            SGTuple2d[] cArray = new SGTuple2d[xArray.length * yArray.length];
            for (int yy = 0; yy < yArray.length; yy++) {
                for (int xx = 0; xx < xArray.length; xx++) {
                    final int index = yy * xArray.length + xx;
                    cArray[index] = new SGTuple2d(xArray[xx], yArray[yy]);
                }
            }
            return cArray;
        }
    }
    
    /**
     * Returns the bounds of x-values.
     * 
     * @return the bounds of x-values
     */
    public SGValueRange getBoundsX() {
        return SGDataUtility.getBoundsX(this);
    }

    /**
     * Returns the bounds of y-values.
     * 
     * @return the bounds of y-values
     */
    public SGValueRange getBoundsY() {
        return SGDataUtility.getBoundsY(this);
    }

    /**
     * Returns the bounds of z-values.
     * 
     * @return the bounds of z-values
     */
    public SGValueRange getBoundsZ() {
        return SGDataUtility.getBoundsZ(this);
    }

    /**
     * Returns the title for the X-axis.
     * 
     * @return the title for the X-axis
     */
    public String getTitleX() {
        return this.getNameWithUnit(this.mXVariable);
    }

    /**
     * Returns the title for the Y-axis.
     * 
     * @return the title for the Y-axis
     */
    public String getTitleY() {
        return this.getNameWithUnit(this.mYVariable);
    }

    /**
     * Returns the title for the Y-axis.
     * 
     * @return the title for the Y-axis
     */
    public String getTitleZ() {
        return this.getNameWithUnit(this.mZVariable);
    }

    protected Double[] arrangeValueArray(Array data2D, SGVariable var, SGVariable[] cVars) {
    	if (this.isSerialNumberAvailable()) {
    		return this.arrange1Dim(data2D);
    	} else {
            return this.arrange2Dim(data2D, var, cVars);
    	}
    }

}
