package jp.riken.brain.ni.samuraigraph.data;

import java.util.HashMap;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2d;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;

import org.w3c.dom.Element;

/**
 * Two dimensional vector data.
 * 
 */
public class SGVXYData extends SGArrayData implements SGIVXYTypeData, 
        SGIDataColumnTypeConstants, SGIDataColumnNameConstants {

    /**
     * The column index for x-coordinate values.
     */
    protected Integer mXCoordinateIndex = null;
    
    /**
     * The column index for y-coordinate values.
     */
    protected Integer mYCoordinateIndex = null;
    
    /**
     * The column index for the first component values.
     */
    protected Integer mComponentIndex1 = null;
    
    /**
     * The column index for the second component values.
     */
    protected Integer mComponentIndex2 = null;

    /**
     * A flag for polar mode.
     */
    private boolean mPolarFlag = false;

    /**
     * The default constructor.
     */
    public SGVXYData() {
        super();
    }

    /**
     * Build a data with given columns and indices.
     * 
     * @param dataFile
     *                   the text data file
     * @param obs
     *                   an observer for array data
     * @param xCoordinateIndex
     *                   the index for x-coordinate values
     * @param yCoordinateIndex
     *                   the index for y-coordinate values
     * @param componentIndex1
     *                   the index for the first component values
     * @param componentIndex2
     *                   the index for the second component values
     * @param isPolar
     *                   true if two arrays are given in polar coordinate
     */
    public SGVXYData(final SGTextDataFile dataFile,
    		final SGDataSourceObserver obs,
            final Integer xCoordinateIndex, final Integer yCoordinateIndex,
            final Integer componentIndex1, final Integer componentIndex2, 
            final boolean isPolar) {
        
        super(dataFile, obs);
        
        // null check
        if (xCoordinateIndex == null || yCoordinateIndex == null || componentIndex1 == null || componentIndex2 == null) {
            throw new IllegalArgumentException("Indices must not be null.");
        }
        
        // x coordinate values
        if (this.checkColumnIndexRange(xCoordinateIndex) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + xCoordinateIndex);
        }

        // y coordinate values
        if (this.checkColumnIndexRange(yCoordinateIndex) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + yCoordinateIndex);
        }
        
        // the first component values
        if (this.checkColumnIndexRange(componentIndex1) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + componentIndex1);
        }
        
        // the second component values
        if (this.checkColumnIndexRange(componentIndex2) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + componentIndex2);
        }
        
        // set to attributes
        this.mXCoordinateIndex = xCoordinateIndex;
        this.mYCoordinateIndex = yCoordinateIndex;
        this.mComponentIndex1 = componentIndex1;
        this.mComponentIndex2 = componentIndex2;
        this.mPolarFlag = isPolar;
    }

    /**
     * 
     * @return
     */
    public SGVXYData(final SGData data) {
        super();
        this.setData(data);
    }

    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
    	super.dispose();
    	this.mXCoordinateIndex = null;
    	this.mYCoordinateIndex = null;
    	this.mComponentIndex1 = null;
    	this.mComponentIndex2 = null;
    }

    /**
     * Set values with a given data.
     * 
     * @param data
     *             data to be copied
     * @return
     *             true if succeeded
     */
    public boolean setData(final SGData data) {
        
        if (!(data instanceof SGVXYData)) {
            return false;
        }
        
        if (super.setData(data) == false) {
            return false;
        }
        
        SGVXYData dataVXY = (SGVXYData) data;
        this.mXCoordinateIndex = dataVXY.mXCoordinateIndex;
        this.mYCoordinateIndex = dataVXY.mYCoordinateIndex;
        this.mComponentIndex1 = dataVXY.mComponentIndex1;
        this.mComponentIndex2 = dataVXY.mComponentIndex2;
        this.mPolarFlag = dataVXY.isPolar();
        
        return true;
    }

    /**
     * Returns the number of points.
     * 
     * @return the number of points
     */
    public int getPointsNumber() {
        return this.getXArray().length;
    }

    /**
     * Returns the name of data type.
     * 
     * @return the name of data type
     */
    public String getDataType() {
        return SGDataTypeConstants.VXY_DATA;
    }

    /**
     * Returns the bounds of x-values.
     * 
     * @return the bounds of x-values
     */
    public SGValueRange getBoundsX() {
        return SGDataUtility.getBoundsX(this);
    }

    /**
     * Returns the bounds of y-values.
     * 
     * @return the bounds of y-values
     */
    public SGValueRange getBoundsY() {
        return SGDataUtility.getBoundsY(this);
    }

    /**
     * 
     * @return
     */
    public double[] getXCoordinateArray() {
//        return SGUtility.copyDoubleArray(this.getXArray());
    	return this.getXArray();
    }

    /**
     * 
     * @return
     */
    public double[] getYCoordinateArray() {
//        return SGUtility.copyDoubleArray(this.getYArray());
    	return this.getYArray();
    }
    
    /**
     * Returns an array of coordinates.
     * 
     * @return an array of coordinates
     */
    public SGTuple2d[] getCoordinateArray() {
        double[] xArray = this.getXArray();
        double[] yArray = this.getYArray();
        SGTuple2d[] cArray = new SGTuple2d[xArray.length];
        for (int ii = 0; ii < xArray.length; ii++) {
            cArray[ii] = new SGTuple2d(xArray[ii], yArray[ii]);
        }
        return cArray;
    }

    /**
     * Returns an array of x component of vectors.
     * 
     * @return an array of x component of vectors 
     */
    public double[] getXComponentArray() {
        if (this.isPolar()) {
            final double[] magArray = this.getComponentArray1();
            final double[] angleArray = this.getComponentArray2();
            return SGVXYDataUtility.getXComponentArray(magArray, angleArray);
        } else {
//            return SGUtility.copyDoubleArray(this.getComponentArray1());
        	return this.getComponentArray1();
        }
    }

    /**
     * Returns an array of y component of vectors.
     * 
     * @return an array of y component of vectors 
     */
    public double[] getYComponentArray() {
        if (this.isPolar()) {
            final double[] magArray = this.getComponentArray1();
            final double[] angleArray = this.getComponentArray2();
            return SGVXYDataUtility.getYComponentArray(magArray, angleArray);
        } else {
//            return SGUtility.copyDoubleArray(this.getComponentArray2());
        	return this.getComponentArray2();
        }
    }

    /**
     * Returns an array of magnitude of vectors.
     * 
     * @return an array of magnitude of vectors 
     */
    public double[] getMagnitudeArray() {
        if (this.isPolar()) {
//            return SGUtility.copyDoubleArray(this.getComponentArray1());
        	return this.getComponentArray1();
        } else {
            final double[] xComArray = this.getComponentArray1();
            final double[] yComArray = this.getComponentArray2();
            return SGVXYDataUtility.getMagnitudeArray(xComArray, yComArray);
        }
    }

    /**
     * Returns an array of angle of vectors.
     * 
     * @return an array of angle of vectors 
     */
    public double[] getAngleArray() {
        if (this.isPolar()) {
//            return SGUtility.copyDoubleArray(this.getComponentArray2());
        	return this.getComponentArray2();
        } else {
            final double[] xComArray = this.getComponentArray1();
            final double[] yComArray = this.getComponentArray2();
            return SGVXYDataUtility.getAngleArray(xComArray, yComArray);
        }
    }

    /**
     * Returns an array of x-values.
     * @return
     *         an array of x-values
     */
    protected double[] getXArray() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGNumberDataColumn col = (SGNumberDataColumn) columns[this.mXCoordinateIndex.intValue()];
        return col.getNumberArray();
    }
    
    /**
     * Returns an array of y-values.
     * @return
     *         an array of y-values
     */
    protected double[] getYArray() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGNumberDataColumn col = (SGNumberDataColumn) columns[this.mYCoordinateIndex.intValue()];
        return col.getNumberArray();
    }
    
    /**
     * Returns an array of the first components.
     * @return
     *         an array of the first components
     */
    protected double[] getComponentArray1() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGNumberDataColumn col = (SGNumberDataColumn) columns[this.mComponentIndex1.intValue()];
        return col.getNumberArray();
    }
    
    /**
     * Returns an array of the second components.
     * @return
     *         an array of the second components
     */
    protected double[] getComponentArray2() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGNumberDataColumn col = (SGNumberDataColumn) columns[this.mComponentIndex2.intValue()];
        return col.getNumberArray();
    }
    
    /**
     * Returns whether the data is given in the polar coordinate.
     * 
     * @return true if the data is given in the polar coordinate
     * 
     */
    public boolean isPolar() {
        return this.mPolarFlag;
    }

    /**
     * Writes data column indices in the attributes.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected boolean writeAttributeColumnIndices(Element el) {
        el.setAttribute(KEY_X_COORDINATE_COLUMN_INDEX, this.mXCoordinateIndex.toString());
        el.setAttribute(KEY_Y_COORDINATE_COLUMN_INDEX, this.mYCoordinateIndex.toString());
        el.setAttribute(KEY_FIRST_COMPONENT_COLUMN_INDEX, this.mComponentIndex1.toString());
        el.setAttribute(KEY_SECOND_COMPONENT_COLUMN_INDEX, this.mComponentIndex2.toString());
        return true;
    }

    /**
     * Writes data column indices as sequential numbers.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected boolean writeSequentialColumnIndices(Element el) {
        el.setAttribute(KEY_X_COORDINATE_COLUMN_INDEX, Integer.toString(0));
        el.setAttribute(KEY_Y_COORDINATE_COLUMN_INDEX, Integer.toString(1));
        el.setAttribute(KEY_FIRST_COMPONENT_COLUMN_INDEX, Integer.toString(2));
        el.setAttribute(KEY_SECOND_COMPONENT_COLUMN_INDEX, Integer.toString(3));
        return true;
    }
    
    @Override
    protected boolean writeSequentialColumnName(Element el) {
        el.setAttribute(KEY_X_COORDINATE_COLUMN_INDEX, this.getSequentialColumnName(0));
        el.setAttribute(KEY_Y_COORDINATE_COLUMN_INDEX, this.getSequentialColumnName(1));
        el.setAttribute(KEY_FIRST_COMPONENT_COLUMN_INDEX, this.getSequentialColumnName(2));
        el.setAttribute(KEY_SECOND_COMPONENT_COLUMN_INDEX, this.getSequentialColumnName(3));
        
        // serial numbers
        el.setAttribute(KEY_SERIAL_NUMBER_VARIABLE_NAME, SERIAL_NUMBER);
        
        return true;
    }

    /**
     * Read properties from a given Element and set to this data.
     * @param el
     *          an Element
     * @return
     *          true if succeeded
     */
    public boolean readProperty(Element el) {
        Integer index = null;
        if ((index = this.readIndex(el, KEY_X_COORDINATE_COLUMN_INDEX)) != null) {
            if (this.checkColumnIndexRange(index) == false) {
                return false;
            }
            this.mXCoordinateIndex = index;
        } else {
            return false;
        }
        if ((index = this.readIndex(el, KEY_Y_COORDINATE_COLUMN_INDEX)) != null) {
            if (this.checkColumnIndexRange(index) == false) {
                return false;
            }
            this.mYCoordinateIndex = index;
        } else {
            return false;
        }
        if ((index = this.readIndex(el, KEY_FIRST_COMPONENT_COLUMN_INDEX)) != null) {
            if (this.checkColumnIndexRange(index) == false) {
                return false;
            }
            this.mComponentIndex1 = index;
        } else {
            return false;
        }
        if ((index = this.readIndex(el, KEY_SECOND_COMPONENT_COLUMN_INDEX)) != null) {
            if (this.checkColumnIndexRange(index) == false) {
                return false;
            }
            this.mComponentIndex2 = index;
        } else {
            return false;
        }
        return true;
    }
    
    public Map<String, Object> getInfoMap() {
        Map<String, Object> map = new HashMap<String, Object>();
        map.put(SGIDataColumnTypeConstants.KEY_POLAR_SELECTED, 
                Boolean.valueOf(this.isPolar()));
        return map;
    }

    /**
     * Returns the title of X values.
     * @return
     *         the title of X values
     */
    public String getTitleX() {
        if (this.mXCoordinateIndex == null) {
            return "";
        } else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mXCoordinateIndex.intValue()].getTitle();
        }
    }
    
    /**
     * Returns the title of Y values.
     * @return
     *         the title of Y values
     */
    public String getTitleY() {
        if (this.mYCoordinateIndex == null) {
            return "";
        } else {
        	SGDataColumn[] columns = this.mDataFile.mDataColumns;
            return columns[this.mYCoordinateIndex.intValue()].getTitle();
        }
    }

    public Integer getXCoordinateIndex() {
        return mXCoordinateIndex;
    }

    public Integer getYCoordinateIndex() {
        return mYCoordinateIndex;
    }
    
    public Integer getComponentIndex1() {
        return mComponentIndex1;
    }

    public Integer getComponentIndex2() {
        return mComponentIndex2;
    }
    
    public void setXCoordinateIndex(Integer index) {
        if (this.checkColumnIndexRange(index) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        mXCoordinateIndex = index;
    }

    public void setYCoordinateIndex(Integer index) {
        if (this.checkColumnIndexRange(index) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        mYCoordinateIndex = index;
    }

    public void setComponentIndex1(Integer index) {
        if (this.checkColumnIndexRange(index) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        mComponentIndex1 = index;
    }

    public void setComponentIndex2(Integer index) {
        if (this.checkColumnIndexRange(index) == false) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        mComponentIndex2 = index;
    }
    
    /**
     * Returns properties of this data.
     * @return
     *         properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new VXYDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Get properties of this data.
     * @param p
     *          properties to set values
     * @return
     *          true if succeeded
     */
    public boolean getProperties(final SGProperties p) {
        if ((p instanceof VXYDataProperties) == false) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        VXYDataProperties ep = (VXYDataProperties) p;
        ep.mXCoordinateIndex = this.getXCoordinateIndex();
        ep.mYCoordinateIndex = this.getYCoordinateIndex();
        ep.mComponentIndex1 = this.getComponentIndex1();
        ep.mComponentIndex2 = this.getComponentIndex2();
        return true;
    }
    
    /**
     * Set properties to this data.
     * @param p
     *          properties that have values to set to this data
     * @return
     *          true if succeeded
     */
    public boolean setProperties(final SGProperties p) {
        if ((p instanceof VXYDataProperties) == false) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        VXYDataProperties ep = (VXYDataProperties) p;
        this.mXCoordinateIndex = ep.mXCoordinateIndex;
        this.mYCoordinateIndex = ep.mYCoordinateIndex;
        this.mComponentIndex1 = ep.mComponentIndex1;
        this.mComponentIndex2 = ep.mComponentIndex2;
        return true;
    }

    /**
     * Sets the type of data columns.
     * 
     * @param column
     *           an array of column types
     * @return true if succeeded
     */
    public boolean setColumnType(String[] columns) {
        if (columns == null) {
            throw new IllegalArgumentException("columns == null");
        }
        
//        if (this.isValidColumn(columns) == false) {
//        	return false;
//        }

        Integer xIndex = null;
        Integer yIndex = null;
        Integer fIndex = null;
        Integer sIndex = null;
        for (int ii = 0; ii < columns.length; ii++) {
            if (SGUtilityText.isEqualString(X_COORDINATE, columns[ii])) {
                xIndex = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(Y_COORDINATE, columns[ii])) {
                yIndex = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(MAGNITUDE, columns[ii]) 
            		|| SGUtilityText.isEqualString(X_COMPONENT, columns[ii])) {
                fIndex = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(ANGLE, columns[ii]) 
            		|| SGUtilityText.isEqualString(Y_COMPONENT, columns[ii])) {
                sIndex = Integer.valueOf(ii);
            } else if ("".equals(columns[ii])) {
                continue;
            } else {
            	return false;
            }
        }
        
        // set to the attributes
        this.mXCoordinateIndex = xIndex;
        this.mYCoordinateIndex = yIndex;
        this.mComponentIndex1 = fIndex;
        this.mComponentIndex2 = sIndex;
        
        return true;
    }

//    /**
//     * Returns whether all given data columns are valid.
//     * 
//     * @param columns
//     *           data columns
//     * @return true if all given data columns are valid
//     */
//    public boolean isValidColumn(String[] columns) {
//    	
//        Integer xOld = this.mXCoordinateIndex;
//        Integer yOld = this.mYCoordinateIndex;
//        Integer fOld = this.mComponentIndex1;
//        Integer sOld = this.mComponentIndex2;
//        
//        Integer xIndex = null;
//        Integer yIndex = null;
//        Integer fIndex = null;
//        Integer sIndex = null;
//        
//        for (int ii = 0; ii < columns.length; ii++) {
//            if (SGUtilityText.isEqualString(X_COORDINATE, columns[ii])) {
//                xIndex = Integer.valueOf(ii);
//            } else if (SGUtilityText.isEqualString(Y_COORDINATE, columns[ii])) {
//                yIndex = Integer.valueOf(ii);
//            } else if (SGUtilityText.isEqualString(MAGNITUDE, columns[ii]) 
//            		|| SGUtilityText.isEqualString(X_COMPONENT, columns[ii])) {
//                fIndex = Integer.valueOf(ii);
//            } else if (SGUtilityText.isEqualString(ANGLE, columns[ii]) 
//            		|| SGUtilityText.isEqualString(Y_COMPONENT, columns[ii])) {
//                sIndex = Integer.valueOf(ii);
//            } else if ("".equals(columns[ii])) {
//                continue;
//            } else {
//            	return false;
//            }
//        }
//        
//        // check the necessary columns
//        if (xIndex == null || yIndex == null || fIndex == null || sIndex == null) {
//        	return false;
//        }
//
//        // set to attributes
//        this.mXCoordinateIndex = xIndex;
//        this.mYCoordinateIndex = yIndex;
//        this.mComponentIndex1 = fIndex;
//        this.mComponentIndex2 = sIndex;
//        
////        try {
////            // check magnitude
////            double[] magArray = this.getMagnitudeArray();
////            for (int ii = 0; ii < magArray.length; ii++) {
////            	if (magArray[ii] < 0.0) {
////            		return false;
////            	}
////            }
////        	
////        } finally {
////            // recover
////    		this.mXCoordinateIndex = xOld;
////    		this.mYCoordinateIndex = yOld;
////    		this.mComponentIndex1 = fOld;
////    		this.mComponentIndex2 = sOld;
////        }
//
//        // recover
//		this.mXCoordinateIndex = xOld;
//		this.mYCoordinateIndex = yOld;
//		this.mComponentIndex1 = fOld;
//		this.mComponentIndex2 = sOld;
//
//    	return true;
//    }
    
    /**
     * A class of properties of vector xy-type data.
     */
    public static class VXYDataProperties extends ArrayDataProperties {

        // indices of data columns
        Integer mXCoordinateIndex = null;

        Integer mYCoordinateIndex = null;
        
        Integer mComponentIndex1 = null;
        
        Integer mComponentIndex2 = null;

        /**
         * The default constructor.
         */
        public VXYDataProperties() {
            super();
        }

        /**
         * Returns whether this object is equal to given object.
         * 
         * @param obj
         *            an object to be compared
         * @return true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof VXYDataProperties) == false) {
                return false;
            }
            VXYDataProperties p = (VXYDataProperties) obj;
            if (this.mXCoordinateIndex != null) {
                if (!this.mXCoordinateIndex.equals(p.mXCoordinateIndex)) {
                    return false;
                }
            } else {
                if (p.mXCoordinateIndex != null) {
                    return false;
                }
            }
            if (this.mYCoordinateIndex != null) {
                if (!this.mYCoordinateIndex.equals(p.mYCoordinateIndex)) {
                    return false;
                }
            } else {
                if (p.mYCoordinateIndex != null) {
                    return false;
                }
            }
            if (this.mComponentIndex1 != null) {
                if (!this.mComponentIndex1.equals(p.mComponentIndex1)) {
                    return false;
                }
            } else {
                if (p.mComponentIndex1 != null) {
                    return false;
                }
            }
            if (this.mComponentIndex2 != null) {
                if (!this.mComponentIndex2.equals(p.mComponentIndex2)) {
                    return false;
                }
            } else {
                if (p.mComponentIndex2 != null) {
                    return false;
                }
            }

            return true;
        }

        /**
         * Returns string representation of this object.
         * 
         * @return string representation of this object
         */
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("[");
            sb.append("XIndex=");
            sb.append(this.mXCoordinateIndex);
            sb.append(", YIndex=");
            sb.append(this.mYCoordinateIndex);
            sb.append(", ComponentIndex1=");
            sb.append(this.mComponentIndex1);
            sb.append(", ComponentIndex2=");
            sb.append(this.mComponentIndex2);
            sb.append("]");
            return sb.toString();
        }

        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.mXCoordinateIndex = null;
            this.mYCoordinateIndex = null;
            this.mComponentIndex1 = null;
            this.mComponentIndex2 = null;
        }
    }

    /**
     * Returns an array of current column types.
     * @return
     *        an array of current column types
     */
    public String[] getCurrentColumnType() {
        String type1, type2;
        if (this.isPolar()) {
            type1 = MAGNITUDE;
            type2 = ANGLE;
        } else {
            type1 = X_COMPONENT;
            type2 = Y_COMPONENT;
        }
        final int colNum = this.getColNum();
        String[] array = new String[colNum];
        if (this.mXCoordinateIndex != null) {
            array[this.mXCoordinateIndex.intValue()] = X_COORDINATE;
        }
        if (this.mYCoordinateIndex != null) {
            array[this.mYCoordinateIndex.intValue()] = Y_COORDINATE;
        }
        if (this.mComponentIndex1 != null) {
            array[this.mComponentIndex1.intValue()] = type1;
        }
        if (this.mComponentIndex2 != null) {
            array[this.mComponentIndex2.intValue()] = type2;
        }
        return array;
    }
    
    /**
     * Returns an array of data columns to save to a data set.
     * 
     * @return an array of data columns to save to a data set
     */
    protected SGDataColumn[] getDataSetColumns() {
    	SGDataColumn[] columns = this.mDataFile.mDataColumns;
        SGDataColumn[] colArray = new SGDataColumn[4];
        colArray[0] = columns[this.mXCoordinateIndex.intValue()];
        colArray[1] = columns[this.mYCoordinateIndex.intValue()];
        colArray[2] = columns[this.mComponentIndex1.intValue()];
        colArray[3] = columns[this.mComponentIndex2.intValue()];
        
        colArray[0].setColumnType(X_COORDINATE);
        colArray[1].setColumnType(Y_COORDINATE);
        if (this.isPolar()) {
            colArray[2].setColumnType(MAGNITUDE);
            colArray[3].setColumnType(ANGLE);
        } else {
            colArray[2].setColumnType(X_COMPONENT);
            colArray[3].setColumnType(Y_COMPONENT);
        }
        return colArray;
    }
    
    @Override
    public SGDataColumn[] getUsedDataColumnsClone() {
        return this.getDataSetColumnsClone();
    }
    
}
