package jp.riken.brain.ni.samuraigraph.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2d;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;

import org.w3c.dom.Element;

import ucar.ma2.Array;
import ucar.nc2.Dimension;

public class SGVXYNetCDFData extends SGNetCDFData implements SGIVXYTypeData,
        SGIDataColumnTypeConstants, SGIDataColumnNameConstants {

    /**
     * The varialble for the x coordinate.
     */
    protected SGVariable mXLocationVariable = null;
    
    /**
     * The varialble for the y coordinate.
     */
    protected SGVariable mYLocationVariable = null;
    
    /**
     * The variable for the first component
     */
    protected SGVariable mFirstComponentVariable = null;

    /**
     * The variable for the second component
     */
    protected SGVariable mSecondComponentVariable = null;
    
    /**
     * A flag for polar mode.
     */
    private boolean mPolarFlag = false;
    
    /**
     * The default constructor.
     *
     */
    public SGVXYNetCDFData() {
        super();
    }

    /**
     * Builds a data object with given netCDF data.
     * 
     * @param ncfile
     *            netCDF data file
     * @param obs
     *            netCDF data observer
     * @param xName
     *            the name of x coordinate variable
     * @param yName
     *            the name of y coordinate variable
     * @param fName
     *            the name of the first component variable
     * @param sName
     *            the name of the second component variable
     * @param isPolar
     *            true for polar coordinate
     * @param timeName
     *            the name of time variable
     * @param serialNumberName
     *          the name of serial number variable
     */
    public SGVXYNetCDFData(SGNetCDFFile ncfile, SGDataSourceObserver obs,
            final String xName, final String yName, final String fName, 
            final String sName, final boolean isPolar, final String timeName,
            final String serialNumberName) {
    	
        super(ncfile, obs, timeName, serialNumberName);
        
        if (xName == null || yName == null || fName == null || sName == null) {
            throw new IllegalArgumentException("xName == null || yName == null || fName == null || sName == null");
        }
        
        // get variables
        SGVariable xVar = ncfile.findVariable(xName);
        if (xVar == null) {
            throw new IllegalArgumentException("Illegal variable name: " + xName);
        }
        SGVariable yVar = ncfile.findVariable(yName);
        if (yVar == null) {
            throw new IllegalArgumentException("Illegal variable name: " + yName);
        }
        SGVariable fVar = ncfile.findVariable(fName);
        if (fVar == null) {
            throw new IllegalArgumentException("Illegal variable name: " + fName);
        }
        SGVariable sVar = ncfile.findVariable(sName);
        if (sVar == null) {
            throw new IllegalArgumentException("Illegal variable name: " + sName);
        }

        //
        // check variables
        //
        List<Dimension> cDimList = new ArrayList<Dimension>();

        // time variable
    	if (this.mTimeVariable != null) {
    		cDimList.add(this.mTimeVariable.getDimension(0));
    	}
    	
        if (fVar.isCoordinateVariable() || sVar.isCoordinateVariable()) {
            throw new IllegalArgumentException("Both of two components must not be coordinate variables.");
        }
        
        if (this.isSerialNumberAvailable()) {
        	// x and y variables
            if (xVar.isCoordinateVariable() || yVar.isCoordinateVariable()) {
                throw new IllegalArgumentException("Both of x and y must not be coordinate variables.");
            }
        	
        	// index variable
            Dimension iDim = this.mSerialNumberVariable.getDimension(0);
            cDimList.add(iDim);
            
        	// x and y variables
            this.checkNotCoordinateVariable(xVar, cDimList);
            this.checkNotCoordinateVariable(yVar, cDimList);
        } else {
        	// x and y variables
            if (!xVar.isCoordinateVariable() || !yVar.isCoordinateVariable()) {
                throw new IllegalArgumentException("Both of x and y must be coordinate variables.");
            }
            cDimList.add(xVar.getDimension(0));
            cDimList.add(yVar.getDimension(0));
        }

        // two components
        this.checkNotCoordinateVariable(fVar, cDimList);
        this.checkNotCoordinateVariable(sVar, cDimList);
        
        // set to attributes
        this.mXLocationVariable = xVar;
        this.mYLocationVariable = yVar;
        this.mFirstComponentVariable = fVar;
        this.mSecondComponentVariable = sVar;
        this.mPolarFlag = isPolar;
    }
    
    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
        this.mXLocationVariable = null;
        this.mYLocationVariable = null;
        this.mFirstComponentVariable = null;
        this.mSecondComponentVariable = null;
        
        // clear the cache
        this.mVariableCache.clear();
    }

    /**
     * Returns whether the data is given in the polar coordinate.
     * 
     * @return true if the data is given in the polar coordinate
     * 
     */
    public boolean isPolar() {
        return this.mPolarFlag;
    }

    /**
     * Returns a map of information.
     * 
     * @return a map of information
     */
    public Map<String, Object> getInfoMap() {
        Map<String, Object> map = new HashMap<String, Object>();
        map.put(SGIDataColumnTypeConstants.KEY_POLAR_SELECTED, 
                Boolean.valueOf(this.isPolar()));
        return map;
    }

    /**
     * Returns a text string that denotes the first component.
     * 
     * @return a text string that denotes the first component
     */
    private String getFirstComponentName() {
        if (this.isPolar()) {
            return MAGNITUDE;
        } else {
            return X_COMPONENT;
        }
    }

    /**
     * Returns a text string that denotes the second component.
     * 
     * @return a text string that denotes the second component
     */
    private String getSecondComponentName() {
        if (this.isPolar()) {
            return ANGLE;
        } else {
            return Y_COMPONENT;
        }
    }
    
    /**
     * Returns a text string of data type.
     * 
     * @return a text string of data type
     */
    public String getDataType() {
        return SGDataTypeConstants.VXY_NETCDF_DATA;
    }
    
    /**
     * Returns the data type for the data set.
     * 
     * @return the data type
     */
    protected String getDataTypeForDataSet() {
        return SGDataTypeConstants.VXY_DATA;
    }

    /**
     * Returns the number of points.
     * 
     * @return the number of points
     */
    public int getPointsNumber() {
    	final int num;
    	if (this.isSerialNumberAvailable()) {
        	Dimension dim = this.mSerialNumberVariable.getDimension(0);
        	num = dim.getLength();
    	} else {
            Dimension dimX = this.mXLocationVariable.getDimension(0);
            Dimension dimY = this.mYLocationVariable.getDimension(0);
        	num = dimX.getLength() * dimY.getLength();
    	}
    	return num;
    }
    
    /**
     * Returns an array of all variables.
     * 
     * @return an array of all variables
     */
    public SGVariable[] getVariables() {
    	List<SGVariable> varList = new ArrayList<SGVariable>();
    	varList.add(this.mXLocationVariable);
    	varList.add(this.mYLocationVariable);
    	varList.add(this.mFirstComponentVariable);
    	varList.add(this.mSecondComponentVariable);
        if (this.mTimeVariable != null) {
        	varList.add(this.mTimeVariable);
        }
    	if (this.mSerialNumberVariable != null) {
    		varList.add(this.mSerialNumberVariable);
    	}
        SGVariable[] varArray = new SGVariable[varList.size()];
        return (SGVariable[]) varList.toArray(varArray);
    }

    /**
     * Returns an array of current column types.
     * @return
     *        an array of current column types
     */
    public String[] getCurrentColumnType() {
        String com1 = this.getFirstComponentName();
        String com2 = this.getSecondComponentName();
        List<SGVariable> varList = this.mNCFile.getVariables();
        final int varNum = varList.size();
        String[] array = new String[varNum];
        for (int ii = 0; ii < varNum; ii++) {
        	SGVariable var = varList.get(ii);
            if (var.equals(this.mXLocationVariable)) {
                array[ii] = X_COORDINATE;
            } else if (var.equals(this.mYLocationVariable)) {
                array[ii] = Y_COORDINATE;
            } else if (var.equals(this.mFirstComponentVariable)) {
                array[ii] = com1;
            } else if (var.equals(this.mSecondComponentVariable)) {
                array[ii] = com2;
            } else if (var.equals(this.mTimeVariable)) {
                array[ii] = TIME;
			} else if (var.equals(this.mSerialNumberVariable)) {
				array[ii] = SERIAL_NUMBER;
            } else {
                array[ii] = "";
            }
        }        
        return array;
    }
    
    /**
     * Sets the type of data columns.
     * 
     * @param column
     *              an array of column types
     * @return true if succeeded
     */
    public boolean setColumnType(String[] columns) {

    	SGVariable xVar = null;
    	SGVariable yVar = null;
    	SGVariable fVar = null;
    	SGVariable sVar = null;
    	SGVariable timeVar = null;
    	SGVariable serialNumberVar = null;

        String fCom = this.getFirstComponentName();
        String sCom = this.getSecondComponentName();
        List<SGVariable> varList = this.mNCFile.getVariables();
        for (int ii = 0; ii < columns.length; ii++) {
        	SGVariable var = varList.get(ii);
            if (SGUtilityText.isEqualString(X_COORDINATE, columns[ii])) {
                xVar = var;
            } else if (SGUtilityText.isEqualString(Y_COORDINATE, columns[ii])) {
                yVar = var;
            } else if (SGUtilityText.isEqualString(fCom, columns[ii])) {
                fVar = var;
            } else if (SGUtilityText.isEqualString(sCom, columns[ii])) {
                sVar = var;
            } else if (SGUtilityText.isEqualString(TIME, columns[ii])) {
                timeVar = var;
            } else if (SGUtilityText.isEqualString(SERIAL_NUMBER, columns[ii])) {
                serialNumberVar = var;
            } else if ("".equals(columns[ii])) {
            	continue;
            } else {
            	return false;
            }
        }

        if (xVar == null || yVar == null || fVar == null || sVar == null) {
        	return false;
        }
        
        // set variables
        this.mXLocationVariable = xVar;
        this.mYLocationVariable = yVar;
        this.mFirstComponentVariable = fVar;
        this.mSecondComponentVariable = sVar;
        this.mTimeVariable = timeVar;
        this.mSerialNumberVariable = serialNumberVar;
        
        return true;
    }

    /**
     * Returns whether all given data columns are valid.
     * 
     * @param columns
     *           data columns
     * @return true if all given data columns are valid
     */
    public boolean isValidColumn(String[] columns) {
    	
    	SGVariable xVar = null;
    	SGVariable yVar = null;
    	SGVariable fVar = null;
    	SGVariable sVar = null;
    	SGVariable timeVar = null;
    	SGVariable indexVar = null;

        String fCom = this.getFirstComponentName();
        String sCom = this.getSecondComponentName();
        List<SGVariable> varList = this.mNCFile.getVariables();
        for (int ii = 0; ii < columns.length; ii++) {
        	SGVariable var = varList.get(ii);
            if (SGUtilityText.isEqualString(X_COORDINATE, columns[ii])) {
                xVar = var;
            } else if (SGUtilityText.isEqualString(Y_COORDINATE, columns[ii])) {
                yVar = var;
            } else if (SGUtilityText.isEqualString(fCom, columns[ii])) {
                fVar = var;
            } else if (SGUtilityText.isEqualString(sCom, columns[ii])) {
                sVar = var;
            } else if (SGUtilityText.isEqualString(TIME, columns[ii])) {
                timeVar = var;
//            } else if (SGUtilityText.isEqualString(INDEX, columns[ii])) {
//                indexVar = var;
            } else if ("".equals(columns[ii])) {
            	continue;
            } else {
            	return false;
            }
        }

        if (xVar == null || yVar == null || fVar == null || sVar == null) {
        	return false;
        }
        
        // set to attributes
        this.mXLocationVariable = xVar;
        this.mYLocationVariable = yVar;
        this.mFirstComponentVariable = fVar;
        this.mSecondComponentVariable = sVar;
        this.mTimeVariable = timeVar;
        this.mSerialNumberVariable = indexVar;
        
    	return true;
    }

    /**
     * Sets the data.
     * 
     * @param data
     *             data set to this object
     * @return
     *             true if succeeded            
     */
    public boolean setData(SGData data) {
        if (!(data instanceof SGVXYNetCDFData)) {
            throw new IllegalArgumentException(
            		"Given data is not the instance of SGVXYNetCDFData class.");
        }
        if (super.setData(data) == false) {
            return false;
        }
        SGVXYNetCDFData nData = (SGVXYNetCDFData) data;
        this.mXLocationVariable = nData.mXLocationVariable;
        this.mYLocationVariable = nData.mYLocationVariable;
        this.mFirstComponentVariable = nData.mFirstComponentVariable;
        this.mSecondComponentVariable = nData.mSecondComponentVariable;
        this.mPolarFlag = nData.mPolarFlag;
        return true;
    }

    /**
     * Set properties to this data.
     * 
     * @param p
     *          properties to be set
     * @return true if succeeded
     */
    public boolean setProperties(SGProperties p) {
        if (!(p instanceof VXYNetCDFDataProperties)) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        VXYNetCDFDataProperties np = (VXYNetCDFDataProperties) p;
        this.mXLocationVariable = this.mNCFile.findVariable(np.xName);
        this.mYLocationVariable = this.mNCFile.findVariable(np.yName);
        this.mFirstComponentVariable = this.mNCFile.findTopVariable(np.fName);
        this.mSecondComponentVariable = this.mNCFile.findTopVariable(np.sName);
        return true;
    }
    
    /**
     * Returns the properties of this data.
     * 
     * @return the properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new VXYNetCDFDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Gets the properties of this data.
     * @param p
     *      the properties of this data
     * @return
     *      true if succeeded   
     */
    public boolean getProperties(SGProperties p) {
        if (!(p instanceof VXYNetCDFDataProperties)) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        VXYNetCDFDataProperties sp = (VXYNetCDFDataProperties) p;
        sp.xName = this.mXLocationVariable.getName();
        sp.yName = this.mYLocationVariable.getName();
        sp.fName = this.mFirstComponentVariable.getName();
        sp.sName = this.mSecondComponentVariable.getName();
        return true;
    }

//    /**
//     * Write this data to a given output stream.
//     * 
//     * @param w
//     *          an output stream
//     * @return
//     *          true if succeeded
//     */
//    protected boolean writeData(Writer w) throws IOException {
//        SGTuple2d[] cArray = this.getCoordinateArray();
//        double[] fArray = this.getFirstComponentValueArray();
//        double[] sArray = this.getSecondComponentValueArray();
//        StringBuffer sb = new StringBuffer();
//        for (int ii = 0; ii < cArray.length; ii++) {
//            sb.append(cArray[ii].x);
//            sb.append(',');
//            sb.append(cArray[ii].y);
//            sb.append(',');
//            sb.append(fArray[ii]);
//            sb.append(',');
//            sb.append(sArray[ii]);                
//            sb.append('\n');
//        }
//        w.write(sb.toString());
//        return true;
//    }

    /**
     * Returns an array of values to save to a data set.
     * 
     * @return an array of values to save to a data set
     */
    @Override
    public double[][] getDataSetValues() {
        SGTuple2d[] cArray = this.getCoordinateArray();
        double[] fArray = this.getFirstComponentValueArray();
        double[] sArray = this.getSecondComponentValueArray();
        final int pNum = cArray.length;
        double[][] values = new double[pNum][];
        for (int ii = 0; ii < pNum; ii++) {
            values[ii] = new double[4];
            values[ii][0] = cArray[ii].x;
            values[ii][1] = cArray[ii].y;
            values[ii][2] = fArray[ii];
            values[ii][3] = sArray[ii];
        }
        return values;
    }
    
    /**
     * Write properties of this object to the Element.
     * 
     * @param el
     *            the Element object
     * @param type
     *            type of the method to save properties
     * @return true if succeeded
     */
    public boolean writeProperty(Element el, final int type) {
        if (super.writeProperty(el, type) == false) {
            return false;
        }
        el.setAttribute(KEY_X_COORDINATE_VARIABLE_NAME, this.mXLocationVariable.getName());
        el.setAttribute(KEY_Y_COORDINATE_VARIABLE_NAME, this.mYLocationVariable.getName());
        el.setAttribute(KEY_FIRST_COMPONENT_VARIABLE_NAME, this.mFirstComponentVariable.getName());
        el.setAttribute(KEY_SECOND_COMPONENT_VARIABLE_NAME, this.mSecondComponentVariable.getName());
        return true;
    }
    
    /**
     * Returns an array of coordinate variables.
     * 
     * @return an array of coordinate variables
     */
    private SGVariable[] getCoordinateVariables() {
    	if (this.isSerialNumberAvailable()) {
    		return new SGVariable[] { this.mSerialNumberVariable };
    	} else {
            return new SGVariable[] { this.mXLocationVariable, this.mYLocationVariable };
    	}
    }
    
    /**
     * Returns an array of values for the first component.
     * 
     * @return an array of values for the first component
     */
    private double[] getFirstComponentValueArray() {
    	if (this.isSerialNumberAvailable()) {
            return this.getValueArray(this.mFirstComponentVariable, 
            		new SGVariable[]{ this.mSerialNumberVariable });
    	} else {
            return this.getValueArray(this.mFirstComponentVariable, 
                    this.getCoordinateVariables());
    	}
    }
    
    /**
     * Returns an array of values for the second component.
     * 
     * @return an array of values for the second component
     */
    private double[] getSecondComponentValueArray() {
    	if (this.isSerialNumberAvailable()) {
            return this.getValueArray(this.mSecondComponentVariable, 
            		new SGVariable[]{ this.mSerialNumberVariable });
    	} else {
            return this.getValueArray(this.mSecondComponentVariable, 
                    this.getCoordinateVariables());
    	}
    }

    /**
     * Returns a copy of x array.
     * 
     * @return x-value array
     */
    public double[] getXCoordinateArray() {
    	if (this.isSerialNumberAvailable()) {
            return this.getValueArray(this.mXLocationVariable, 
            		new SGVariable[]{ this.mSerialNumberVariable });
    	} else {
            return this.getCoordinateValueArray(this.mXLocationVariable);
    	}
    }

    /**
     * Returns a copy of y array.
     * 
     * @return y-value array
     */
    public double[] getYCoordinateArray() {
    	if (this.isSerialNumberAvailable()) {
            return this.getValueArray(this.mYLocationVariable, 
            		new SGVariable[]{ this.mSerialNumberVariable });
    	} else {
            return this.getCoordinateValueArray(this.mYLocationVariable);
    	}
    }

    /**
     * Returns an array of values for the x coordinates.
     * 
     * @return an array of values for the x coordinates
     */
    public SGTuple2d[] getCoordinateArray() {
        double[] xArray = this.getXCoordinateArray();
        double[] yArray = this.getYCoordinateArray();
        if (this.isSerialNumberAvailable()) {
        	final int len = xArray.length;
            SGTuple2d[] cArray = new SGTuple2d[len];
            for (int ii = 0; ii < len; ii++) {
            	cArray[ii] = new SGTuple2d(xArray[ii], yArray[ii]);
            }
            return cArray;
        } else {
            SGTuple2d[] cArray = new SGTuple2d[xArray.length * yArray.length];
            for (int yy = 0; yy < yArray.length; yy++) {
                for (int xx = 0; xx < xArray.length; xx++) {
                    final int index = yy * xArray.length + xx;
                    cArray[index] = new SGTuple2d(xArray[xx], yArray[yy]);
                }
            }
            return cArray;
        }
    }

    /**
     * Returns an array of x component of vectors.
     * 
     * @return an array of x component of vectors 
     */
    public double[] getXComponentArray() {
        if (this.isPolar()) {
            final double[] magArray = this.getFirstComponentValueArray();
            final double[] angleArray = this.getSecondComponentValueArray();
            return SGVXYDataUtility.getXComponentArray(magArray, angleArray);
        } else {
            return this.getFirstComponentValueArray();
        }
    }
    
    /**
     * Returns an array of y component of vectors.
     * 
     * @return an array of y component of vectors 
     */
    public double[] getYComponentArray() {
        if (this.isPolar()) {
            final double[] magArray = this.getFirstComponentValueArray();
            final double[] angleArray = this.getSecondComponentValueArray();
            return SGVXYDataUtility.getYComponentArray(magArray, angleArray);
        } else {
            return this.getSecondComponentValueArray();
        }
    }

    /**
     * Returns an array of magnitude of vectors.
     * 
     * @return an array of magnitude of vectors 
     */
    public double[] getMagnitudeArray() {
        if (this.isPolar()) {
            return this.getFirstComponentValueArray();
        } else {
            final double[] xComArray = this.getFirstComponentValueArray();
            final double[] yComArray = this.getSecondComponentValueArray();
            return SGVXYDataUtility.getMagnitudeArray(xComArray, yComArray);
        }
    }
    
    /**
     * Returns an array of angle of vectors.
     * 
     * @return an array of angle of vectors 
     */
    public double[] getAngleArray() {
        if (this.isPolar()) {
            return this.getSecondComponentValueArray();
        } else {
            final double[] xComArray = this.getFirstComponentValueArray();
            final double[] yComArray = this.getSecondComponentValueArray();
            return SGVXYDataUtility.getAngleArray(xComArray, yComArray);
        }
    }

    /**
     * Returns the bounds of x-values.
     * 
     * @return the bounds of x-values
     */
    public SGValueRange getBoundsX() {
        return SGDataUtility.getBoundsX(this);
    }

    /**
     * Returns the bounds of y-values.
     * 
     * @return the bounds of y-values
     */
    public SGValueRange getBoundsY() {
        return SGDataUtility.getBoundsY(this);
    }

    /**
     * Returns the title for the X-axis.
     * 
     * @return the title for the X-axis
     */
    public String getTitleX() {
        return this.getNameWithUnit(this.mXLocationVariable);
    }

    /**
     * Returns the title for the Y-axis.
     * 
     * @return the title for the Y-axis
     */
    public String getTitleY() {
        return this.getNameWithUnit(this.mYLocationVariable);
    }

    /**
     * A class of vector XY netCDF data properties.
     */
    public static class VXYNetCDFDataProperties extends NetCDFDataProperties {

        protected String xName = null;
        
        protected String yName = null;
        
        protected String fName = null;
        
        protected String sName = null;
        
        /**
         * The default constructor.
         */
        public VXYNetCDFDataProperties() {
            super();
        }

        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.xName = null;
            this.yName = null;
            this.fName = null;
            this.sName = null;
        }
        
        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof VXYNetCDFDataProperties) == false) {
                return false;
            }
            if (super.equals(obj) == false) {
                return false;
            }
            VXYNetCDFDataProperties p = (VXYNetCDFDataProperties) obj;
            if (SGUtility.equals(this.xName, p.xName) == false) {
                return false;
            }
            if (SGUtility.equals(this.yName, p.yName) == false) {
                return false;
            }
            if (SGUtility.equals(this.fName, p.fName) == false) {
                return false;
            }
            if (SGUtility.equals(this.sName, p.sName) == false) {
                return false;
            }
            return true;
        }
    }
    
    protected Double[] arrangeValueArray(Array data2D, SGVariable var, SGVariable[] cVars) {
    	if (this.isSerialNumberAvailable()) {
    		return this.arrange1Dim(data2D);
    	} else {
            return this.arrange2Dim(data2D, var, cVars);
    	}
    }

}
