package jp.riken.brain.ni.samuraigraph.figure;

import java.awt.Color;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import jp.riken.brain.ni.samuraigraph.base.SGColorBarModel;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;

/**
 * A class that manages color bar styles.
 * 
 * @author kuromaru
 *
 */
public class SGColorBarStyleManager {

    /*
     * The name of color bar.
     */
    public static final String COLOR_BAR_NAME_COLORS = "Colors";

    public static final String COLOR_BAR_NAME_GRADIENT_COLORS = "Gradient Colors";

    public static final String COLOR_BAR_NAME_BLACK_AND_WHITE = "Black and White";

    public static final String COLOR_BAR_NAME_GRAY_SCALE = "Gray Scale";

    public static final String COLOR_BAR_NAME_REPEATED_GRAY_SCALE = "Gray Scale Saw";
    
    public static final String COLOR_BAR_NAME_COLORS_SCALE_WHITE = "Colors Scale White";
    
    public static final String COLOR_BAR_NAME_COLORS_SCALE_BLACK = "Colors Scale Black";
    
    public static final String COLOR_BAR_NAME_REPEATED_GRADIENT_COLORS = "Gradient Colors Saw";

    /**
     * The default color bar names.
     */
    public static final String[] DEFAULT_COLOR_BAR_NAMES = {
            COLOR_BAR_NAME_COLORS,
            COLOR_BAR_NAME_GRADIENT_COLORS,
            COLOR_BAR_NAME_BLACK_AND_WHITE,
            COLOR_BAR_NAME_GRAY_SCALE,
            COLOR_BAR_NAME_REPEATED_GRAY_SCALE,
            COLOR_BAR_NAME_COLORS_SCALE_WHITE,
            COLOR_BAR_NAME_COLORS_SCALE_BLACK,
            COLOR_BAR_NAME_REPEATED_GRADIENT_COLORS
    };

    /**
     * The default color bar models.
     */
    private static final SGColorBarModel[] DEFAULT_COLOR_BAR_MODELS;

    static {
        // create default color bar models
        Color[] base1 = { Color.BLUE, Color.CYAN, new Color(0, 255, 0), Color.YELLOW, Color.RED };
        Color[] base2 = { Color.BLACK, Color.WHITE };
        Color[][] base3 = {
                { Color.BLUE, Color.WHITE}, { Color.CYAN, Color.WHITE },
                { new Color(0, 255, 0), Color.WHITE }, { Color.YELLOW, Color.WHITE },
                { Color.RED, Color.WHITE }
        };
        Color[][] base4 = {
                { Color.BLUE, Color.BLACK}, { Color.CYAN, Color.BLACK },
                { new Color(0, 255, 0), Color.BLACK }, { Color.YELLOW, Color.BLACK },
                { Color.RED, Color.BLACK }
        };
        DEFAULT_COLOR_BAR_MODELS = new SGColorBarModel[DEFAULT_COLOR_BAR_NAMES.length];
        DEFAULT_COLOR_BAR_MODELS[0] = new DiscreteColorBarModel(base1);
        DEFAULT_COLOR_BAR_MODELS[1] = new HueColorBarModel();
        DEFAULT_COLOR_BAR_MODELS[2] = new DiscreteColorBarModel(base2);
        DEFAULT_COLOR_BAR_MODELS[3] = new GradationalColorBarModel(base2);
        DEFAULT_COLOR_BAR_MODELS[4] = new RepeatedGradationalColorBarModel(base2, 3);
        DEFAULT_COLOR_BAR_MODELS[5] = new MultipleGradationalColorBarModel(base3);
        DEFAULT_COLOR_BAR_MODELS[6] = new MultipleGradationalColorBarModel(base4);
        DEFAULT_COLOR_BAR_MODELS[7] = new RepeatedHueColorBarModel(3);
    }

    /**
     * A map of color bar models.
     */
    private Map<String, SGColorBarModel> mColorBarModelMap = new HashMap<String, SGColorBarModel>();

    /**
     * A flag whether the order of colors is revered.
     */
    private boolean mReversedOrderFlag = false;
    
    /**
     * The default constructor.
     *
     */
    public SGColorBarStyleManager() {
        super();

        // set to the map
        for (int ii = 0; ii < DEFAULT_COLOR_BAR_NAMES.length; ii++) {
            this.mColorBarModelMap.put(DEFAULT_COLOR_BAR_NAMES[ii],
                    (SGColorBarModel) DEFAULT_COLOR_BAR_MODELS[ii].clone());
        }
    }
    
    /**
     * Returns the color bar model.
     * 
     * @param name
     *          the name of color bar style
     * @return a color bar model of the given style if it exists
     */
    public SGColorBarModel getColorBarModel(final String name) {
    	if (name == null) {
    		throw new IllegalArgumentException("name == null");
    	}
    	Set<String> keys = this.mColorBarModelMap.keySet();
    	for (String key : keys) {
    		if (SGUtilityText.isEqualString(key, name)) {
    	        return this.mColorBarModelMap.get(key);
    		}
    	}
    	return null;
    }

    /**
     * Adds a color bar model. If a color bar model of the same name already
     * exists, it is removed.
     * 
     * @param name
     *            the name of color bar style
     * @param model
     *            the color bar model
     * @return true if succeeded, false when a given name is equal to that of
     *         the one of the default color bar model
     */
    public boolean addColorBarModel(final String name,
            final SGColorBarModel model) {

        // check input values
        if (name == null || model == null) {
            throw new IllegalArgumentException("name == null || model == null");
        }
        for (int ii = 0; ii < DEFAULT_COLOR_BAR_NAMES.length; ii++) {
            if (DEFAULT_COLOR_BAR_NAMES[ii].equals(name)) {
                return false;
            }
        }

        // set to an attribute
        this.mColorBarModelMap.put(name, model);

        return true;
    }
    
    /**
     * Returns the name of the color bar model.
     * 
     * @param model
     *           the color bar model
     * @return the name of a given color bar model if it exists
     */
    public String getColorBarName(final SGColorBarModel model) {
        String name = null;
        Set<String> keys = this.mColorBarModelMap.keySet();
        for (String key : keys) {
            SGColorBarModel m = this.mColorBarModelMap.get(key);
            if (m.equals(model)) {
                name = key;
                break;
            }
        }
        return name;
    }

    /**
     * Returns whether the order of colors is reversed.
     * 
     * @return whether the order of colors is reversed
     */
    public boolean isReversedOrder() {
        return this.mReversedOrderFlag;
    }
    
    /**
     * Sets whether the order of colors is reversed.
     * 
     * @param b
     *          true to reverse the order of colors
     */
    public void setReversedOrder(final boolean b) {
        this.mReversedOrderFlag = b;
        Iterator<SGColorBarModel> itr = this.mColorBarModelMap.values().iterator();
        while (itr.hasNext()) {
            itr.next().setReversedOrder(b);
        }
    }

    /**
     * A color bar model that gives discretely varying colors.
     * 
     * @author kuromaru
     *
     */
    public static class DiscreteColorBarModel extends SGColorBarModel {
        public DiscreteColorBarModel(final Color[] colors) {
            super(colors);
        }
        
        public Color eval(final double value) {
            if (value <= 0.0) {
                return this.getColor(0);
            } else if (value >= 1.0) {
                return this.getColor(this.mColors.length - 1);
            } else {
                final double indexValue = value * (this.mColors.length - 1) + 0.50;
                final int index = (int) (indexValue);
                return this.getColor(index);
            }
        }
    }

    /**
     * A color bar model that gives gradationaly varying colors.
     * 
     * @author kuromaru
     *
     */
    public static class GradationalColorBarModel extends SGColorBarModel {
        public GradationalColorBarModel(final Color[] colors) {
            super(colors);
        }
        
        public Color eval(final double value) {
            if (value <= 0.0) {
                return this.getColor(0);
            } else if (value >= 1.0) {
                return this.getColor(this.mColors.length - 1);
            } else {
                final double indexValue = value * (this.mColors.length - 1) + 0.50;
                final int index = (int) (indexValue);
                final double ratioSingle = indexValue - index;
                if (index == 0 && ratioSingle < 0.50) {
                    return this.getColor(0);
                } else if (index == this.mColors.length - 1 && ratioSingle >= 0.50) {
                    return this.getColor(this.mColors.length - 1);
                } else {
                    // linear interpolation
                    final int index1, index2;
                    final double r;
                    if (ratioSingle < 0.50) {
                        index1 = index - 1;
                        index2 = index;
                        r = 0.50 + ratioSingle;
                    } else {
                        index1 = index;
                        index2 = index + 1;
                        r = ratioSingle - 0.50;
                    }
                    final Color col1 = this.getColor(index1);
                    final Color col2 = this.getColor(index2);
                    final double m1 = 1.0 - r;
                    final double m2 = r;
                    final double red = m1 * col1.getRed() + m2 * col2.getRed();
                    final double green = m1 * col1.getGreen() + m2 * col2.getGreen();
                    final double blue = m1 * col1.getBlue() + m2 * col2.getBlue();
                    final double alpha = m1 * col1.getAlpha() + m2 * col2.getAlpha();
                    Color col = new Color((int) red, (int) green,
                            (int) blue, (int) alpha);
                    return col;
                }
            }
        }
    }
    
    /**
     * A color bar model that gives gradationaly varying colors as hue changes
     * from blue to red.
     * 
     * @author kuromaru
     *
     */
    public static class HueColorBarModel extends SGColorBarModel {
        public HueColorBarModel() {
            super();
        }

        public Color eval(final double value) {
            final double valueNew;
            if (value <= 0.0) {
                valueNew = 0.0;
            } else if (value >= 1.0) {
                valueNew = 1.0;
            } else {
                valueNew = value;
            }
            final double v = this.mReversedOrderFlag ? valueNew : (1.0 - valueNew);
            final float r = (float) ((2.0 / 3.0) * v);
            Color cl = new Color(Color.HSBtoRGB(r, 1.0f, 1.0f));
            return cl;
        }
    }
    
    /**
     * A color bar model that gives repeated varying colors.
     * 
     * @author kuromaru
     *
     */
    public abstract static class RepeatedColorBarModel extends SGColorBarModel {

        /**
         * The number of repetetion.
         * 
         */
        protected int mRepeateNum;
        
        /**
         * A color bar model.
         */
        protected SGColorBarModel mModel;
        
        /**
         * Builds this color bar model.
         * 
         * @param colors
         *           an array of colors
         * @param repeatNum
         *           the number of repetion
         */
        public RepeatedColorBarModel(final Color[] colors, final int repeatNum) {
            super();
            this.mRepeateNum = repeatNum;
            this.mModel = this.createInstance(colors);
        }
        
        public Color eval(final double value) {
            final int digit = (int) (value);
            final float valueNew = (float) (value - digit);
            final int digitNew = (int) (valueNew * mRepeateNum);
            final float valueNew2 = (float) (valueNew * mRepeateNum - digitNew);
            return this.mModel.eval(valueNew2);
        }
        
        public void setReversedOrder(final boolean b) {
            super.setReversedOrder(b);
            this.mModel.setReversedOrder(b);
        }
        
        protected abstract SGColorBarModel createInstance(final Color[] colors);
    }
    
    
    /**
     * A color bar model that gives repeated gradationally varying colors.
     * 
     * @author kuromaru
     *
     */
    public static class RepeatedGradationalColorBarModel extends RepeatedColorBarModel {
        public RepeatedGradationalColorBarModel(final Color[] colors, final int repeatNum) {
            super(colors, repeatNum);
        }
        
        protected SGColorBarModel createInstance(final Color[] colors) {
            return new GradationalColorBarModel(colors);
        }
    }
    
    /**
     * A color bar model that contains multiple color bar models.
     * 
     * @author minemoto
     *
     */
    public abstract static class MultipleColorBarModel extends SGColorBarModel {
        /**
         * Color bar models.
         */
        protected SGColorBarModel[] mModel;
        
        public MultipleColorBarModel(final Color[] colors, final int repeatNum) {
            this.mModel = new SGColorBarModel[repeatNum];
            for (int i = 0; i < repeatNum; i++) {
                this.mModel[i] = this.createInstance(colors);
            }
        }
        
        public MultipleColorBarModel(final Color[][] colors) {
            super();
            this.mModel = new SGColorBarModel[colors.length];
            for (int i = 0; i < colors.length; i++) {
                this.mModel[i] = this.createInstance(colors[i]);
            }
        }
        
        public Color eval(final double value) {
            if (value <= 0.0) {
                if (!isReversedOrder()) {
                    return this.mModel[0].eval(0);
                } else {
                    return this.mModel[this.mModel.length-1].eval(0.0);
                }
            } else if (value >= 1.0) {
                if (!isReversedOrder()) {
                    return this.mModel[this.mModel.length-1].eval(1.0);
                } else {
                    return this.mModel[0].eval(1.0);
                }
            } else {
                final int digit = (int) (value);
                final double valueNew = (value - digit);
                final int digitNew = (int) (valueNew * this.mModel.length);
                final double valueNew2 = (valueNew * this.mModel.length - digitNew);
                if (!isReversedOrder()) {
                    return this.mModel[digitNew].eval(valueNew2);
                } else {
                    return this.mModel[this.mModel.length-1-digitNew].eval(valueNew2);
                }
            }
        }
        
        public void setReversedOrder(final boolean b) {
            super.setReversedOrder(b);
            for (int i = 0; i < this.mModel.length; i++) {
                this.mModel[i].setReversedOrder(b);
            }
        }
        
        protected abstract SGColorBarModel createInstance(final Color[] colors);
    }
    
    /**
     * A color bar model that contains multiple color bar models
     * which give gradationally varying colors.
     * 
     * @author minemoto
     *
     */
    public static class MultipleGradationalColorBarModel extends MultipleColorBarModel {
        public MultipleGradationalColorBarModel(final Color[][] colors) {
            super(colors);
        }
        
        protected SGColorBarModel createInstance(final Color[] colors) {
            return new GradationalColorBarModel(colors);
        }
    }
    
    /**
     * A color bar model that gives repeated gradationaly varying colors 
     * as hue changes from blue to red.
     * 
     * @author minemoto
     *
     */
    public static class RepeatedHueColorBarModel extends RepeatedColorBarModel {
        public RepeatedHueColorBarModel(final int repeatNum) {
            super(null, repeatNum);
        }
        protected SGColorBarModel createInstance(final Color[] colors) {
            return new HueColorBarModel();
        }
    }

    /**
     * A color bar model that gives repeated discretely varying colors.
     * 
     * @author kuromaru
     *
     */
    public static class RepeatedDiscreteColorBarModel extends RepeatedColorBarModel {
        public RepeatedDiscreteColorBarModel(final Color[] colors, final int repeatNum) {
            super(colors, repeatNum);
        }
        
        protected SGColorBarModel createInstance(final Color[] colors) {
            return new DiscreteColorBarModel(colors);
        }
    }

    public static void main(String[] args) {
        HueColorBarModel m = new HueColorBarModel();
        Color cl = m.eval(0.50);
        System.out.println(cl);
    }

}
