package jp.riken.brain.ni.samuraigraph.figure.java2d;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import jp.riken.brain.ni.samuraigraph.base.SGDrawingElement;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2f;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityNumber;
import jp.riken.brain.ni.samuraigraph.data.SGIVXYTypeData;
import jp.riken.brain.ni.samuraigraph.figure.SGIVXYDataConstants;

/**
 * Class of a group of arrows.
 * 
 * @author okumura
 */
public class SGElementGroupArrowInGraph extends SGElementGroupArrowForData implements
        SGIElementGroupInGraph, SGIVXYDataConstants {

    /**
     * The graph.
     */
    protected SGFigureElementGraph mGraph = null;

    /**
     * The group set.
     */
    protected SGElementGroupSetInGraph mGroupSet = null;

    /**
     * A flag whether this group is focused.
     */
    private boolean mFocusedFlag = false;

    /**
     * Sets the group set that this element group set belongs to
     * @param gs
     *          a group set that this element group set belongs to
     * @return true if succeeded
     */
    public boolean setElementGroupSet(SGElementGroupSetInGraph gs) {
        this.mGroupSet = gs;
        return true;
    }

    /**
     * Sets the flag whether this group is focused.
     * 
     * @param b
     *         a value to set the the flag whether this group is focused
     */
    public void setFocused(final boolean b) {
        this.mFocusedFlag = b;
    }

    /**
     * Returns whether this group is focused
     * 
     * @return a flag whether this group is focused
     */
    public boolean isFocused() {
        return this.mFocusedFlag;
    }

    /**
     * Builds the element group.
     * 
     * @param graph
     *           the graph that this element group set belongs to
     */
    protected SGElementGroupArrowInGraph(SGFigureElementGraph graph) {
        super();
        this.mGraph = graph;
        
        // initialize
        this.setVisible(DEFAULT_ARROW_VISIBLE);
    }

    /**
     * The number of anchors for one direction.
     */
    public static final int NUMBER_OF_ANCHORS = 8;
    
    public static final Color FOCUS_COLOR = new Color(63, 63, 63, 63);

    /**
     * Paint this object.
     */
    public boolean paintElement(final Graphics2D g2d, final Rectangle2D rect) {
        if (super.paintElement(g2d, rect) == false) {
        	return false;
        }

        if (this.mGridMode) {
            if (this.isFocused()) {
                final float minX = (float) this.mBounds.getMinX();
                final float maxX = (float) this.mBounds.getMaxX();
                final float minY = (float) this.mBounds.getMinY();
                final float maxY = (float) this.mBounds.getMaxY();
                final float width = maxX - minX;
                final float height = maxY - minY;

                g2d.setPaint(FOCUS_COLOR);
                Rectangle2D bounds = new Rectangle2D.Double(minX, minY, width, height);
                g2d.fill(bounds);

                // draw anchors
                final int num = NUMBER_OF_ANCHORS - 1;
                final float diffX = (maxX - minX) / num;
                final float diffY = (maxY - minY) / num;
                for (int ii = 0; ii <= num; ii++) {
                    final float y = minY + ii * diffY;
                    for (int jj = 0; jj <= num; jj++) {
                        final float x = minX + jj * diffX;
                        SGUtilityForFigureElementJava2D.drawAnchorAsFocusedObject(x, y, g2d);
                    }
                }
            }

        } else {
            if (this.isFocused()
                    && this.mGraph.isSymbolsVisibleAroundFocusedObjects()) {
                SGDrawingElement[] array = this.mDrawingElementArray;
                for (int ii = 0; ii < array.length; ii++) {
                    SGDrawingElementArrow2D arrow = (SGDrawingElementArrow2D) array[ii];
                    emphasisArrow(arrow, g2d);
                }

            }
        }

        return true;
    }

    /**
     * 
     * @param g2d
     * @param symbol
     * @return
     */
    private boolean emphasisArrow(final SGDrawingElementArrow2D arrow,
            final Graphics2D g2d) {
        SGTuple2f start = arrow.getStart();
        SGTuple2f end = arrow.getEnd();
        SGUtilityForFigureElementJava2D.drawAnchorAsFocusedObject(
        	new Point2D.Float(start.x, start.y), g2d);
        SGUtilityForFigureElementJava2D.drawAnchorAsFocusedObject(
                new Point2D.Float(end.x, end.y), g2d);
        return true;
    }

    /**
     * Returns whether this group set contains the given point.
     * 
     * @param x
     *          the x coordinate
     * @param y
     *          the y coordinate
     * @return
     *          true if this element group contains the given point
     */
    public boolean contains(final int x, final int y) {
        Rectangle2D gRect = this.mGraph.getGraphRect();
		if (!gRect.contains(x, y)) {
		    return false;
		}
    	if (this.mGridMode) {
            return this.mBounds.contains(x, y);
    	} else {
    		// if a given point is out of the graph rectangle, returns false;
    		return super.contains(x, y);
    	}
    }
    
    /**
     * 
     * @return
     */
    protected SGDrawingElement createDrawingElementInstance(final int index) {
        return new ArrowInGroup(this, index);
    }

    @Override
    public SGTuple2f getEndLocation(int index) {
        if (this.mGroupSet instanceof SGElementGroupSetInGraphVXY) {
            SGElementGroupSetInGraphVXY gs = (SGElementGroupSetInGraphVXY) this.mGroupSet;
            return gs.getEndLocation(index);
        } else {
            throw new Error("Not supported");
        }
    }

    @Override
    public SGTuple2f getStartLocation(int index) {
        if (this.mGroupSet instanceof SGElementGroupSetInGraphVXY) {
            SGElementGroupSetInGraphVXY gs = (SGElementGroupSetInGraphVXY) this.mGroupSet;
            return gs.getStartLocation(index);
        } else {
            throw new Error("Not supported");
        }
    }

	public boolean initDrawingElement(float[] xArray, float[] yArray) {
        final int num = xArray.length * yArray.length;
        if (this.initDrawingElement(num) == false) {
            return false;
        }
        return true;
	}

    /**
     * Sets the location of drawing elements.
     * 
     * @param xCoordinateArray
     *           an array of x-coordinates
     * @param yCoordinateArray
     *           an array of y-coordinates
     * @return true if succeeded
     */
    public boolean setLocation(float[] xCoordinateArray, float[] yCoordinateArray) {
        if (xCoordinateArray == null || yCoordinateArray == null) {
            throw new IllegalArgumentException(
                    "xCoordinateArray == null || yCoordinateArray == null");
        }

        final int xNum = xCoordinateArray.length;
        final int yNum = yCoordinateArray.length;
        
        // update the drawing elements
        final int totalNum = xNum * yNum;
        if (this.mDrawingElementArray.length != totalNum) {
        	this.initDrawingElement(totalNum);
        }

        SGIVXYTypeData dataVXY = (SGIVXYTypeData) this.mGroupSet.getData();
        SGElementGroupSetInGraphVXY groupSet = (SGElementGroupSetInGraphVXY) this.mGroupSet;
        SGTuple2f[] endArray = groupSet.mEndPointsArray;
        if (SGArrowUtility.calcEndLocation(dataVXY, this.mGraph, this.mGroupSet, 
        		groupSet.getMagnitudePerCM(), groupSet.isDirectionInvariant(),
        		xCoordinateArray, yCoordinateArray, null, endArray) == false) {
        	return false;
        }
        
        float xMin = 0.0f;
        float xMax = 0.0f;
        float yMin = 0.0f;
        float yMax = 0.0f;
        if (xNum == 1) {
        	xMin = xCoordinateArray[0];
        	xMax = xCoordinateArray[0];
        } else {
        	xMin = SGUtilityNumber.min(xCoordinateArray);
        	xMax = SGUtilityNumber.max(xCoordinateArray);
        }
        if (yNum == 1) {
        	yMin = yCoordinateArray[0];
        	yMax = yCoordinateArray[0];
        } else {
        	yMin = SGUtilityNumber.min(yCoordinateArray);
        	yMax = SGUtilityNumber.max(yCoordinateArray);
        }

        SGTuple2f endMin = SGUtilityNumber.min(endArray);
        if (endMin.x < xMin) {
        	xMin = endMin.x;
        }
        if (endMin.y < yMin) {
        	yMin = endMin.y;
        }
        SGTuple2f endMax = SGUtilityNumber.max(endArray);
        if (endMax.x > xMax) {
        	xMax = endMax.x;
        }
        if (endMax.y > yMax) {
        	yMax = endMax.y;
        }

        xMin -= SINGLE_VALUE_MARGIN;
        xMax += SINGLE_VALUE_MARGIN;
        yMin -= SINGLE_VALUE_MARGIN;
        yMax += SINGLE_VALUE_MARGIN;

        this.mBounds.setRect(xMin, yMin, xMax - xMin, yMax - yMin);

        return true;
    }

}
