package jp.riken.brain.ni.samuraigraph.figure.java2d;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import jp.riken.brain.ni.samuraigraph.base.SGAxis;
import jp.riken.brain.ni.samuraigraph.base.SGDrawingElement;
import jp.riken.brain.ni.samuraigraph.base.SGFillPaint;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2f;
import jp.riken.brain.ni.samuraigraph.figure.SGISXYZDataConstants;

import org.w3c.dom.Element;

/**
 * The color map in the graph object.
 *
 */
public class SGElementGroupColorMapInGraph extends SGElementGroupColorMapForData
    implements SGIElementGroupInGraph, SGISXYZDataConstants {

    /**
     * Maximum number of anchors.
     */
    public static final int MAX_NUMBER_OF_ANCHORS = 20;
    
    private SGFigureElementGraph mGraph = null;

    protected SGElementGroupSetInGraph mGroupSet = null;
    
    public boolean setElementGroupSet(SGElementGroupSetInGraph gs) {
        this.mGroupSet = gs;
        return true;
    }

    private boolean mFocusedFlag = false;

    public void setFocused(final boolean b) {
        this.mFocusedFlag = b;
    }

    public boolean isFocused() {
        return this.mFocusedFlag;
    }
    
    public SGElementGroupColorMapInGraph(SGFigureElementGraph graph) {
        super();
        this.mGraph = graph;
        
        // initialize
        this.setVisible(DEFAULT_COLOR_MAP_VISIBLE);
    }

    /**
     * Sets the location of each rectangle.
     * 
     * @param pointArray
     *          an array of location
     */
    public boolean setLocation(SGTuple2f[] pointArray) {
        if (this.mDrawingElementArray == null) {
            return true;
        }

        if (pointArray.length != this.mDrawingElementArray.length) {
        	this.initDrawingElement(pointArray);
        }

        final SGElementGroupSetInGraphSXYZ gs = (SGElementGroupSetInGraphSXYZ) this.mGroupSet;
        final SGAxis xAxis = gs.getXAxis();
        final SGAxis yAxis = gs.getYAxis();
        final double xMin = xAxis.getMinValue();
        final double xMax = xAxis.getMaxValue();
        final double yMin = yAxis.getMinValue();
        final double yMax = yAxis.getMaxValue();
        final float w = this.mGraph.getGraphRectWidth()
                * (float) (this.mWidthValue / (xMax - xMin));
        final float h = this.mGraph.getGraphRectHeight()
                * (float) (this.mHeightValue / (yMax - yMin));
        this.setRectangleWidth(w);
        this.setRectangleHeight(h);

        SGDrawingElement[] array = this.mDrawingElementArray;
        for (int ii = 0; ii < array.length; ii++) {
            // set the bounds of rectangle
            final ColorMapRectangle rect = (ColorMapRectangle) array[ii];
            final float x = pointArray[ii].x;
            final float y = pointArray[ii].y;
            if (Float.isInfinite(x) || Float.isNaN(x) 
                    || Float.isInfinite(y) || Float.isNaN(y)) {
                rect.setVisible(false);
                continue;
            }
            rect.setBounds(x, y, w, h);
            
            // update the color
            final double zValue = rect.getZValue();
            if (Double.isNaN(zValue) || Double.isInfinite(zValue)) {
                rect.setVisible(false);
                continue;
            } else {
            	if (this.mColorBarModel.getScaleType() == SGAxis.LOG_SCALE && zValue <= 0.0) {
            		rect.setVisible(false);
            		continue;
            	} else {
                    rect.setVisible(true);
                    final Color cl = this.mColorBarModel.evaluate(zValue);
                    rect.setInnerPaint(new SGFillPaint(cl));
            	}
            }
        }
        
        return true;
    }

    /**
     * The number of anchors for one direction.
     */
    public static final int NUMBER_OF_ANCHORS = 8;
    
    public static final Color FOCUS_COLOR = new Color(63, 63, 63, 63);

    /**
     * Paint the objects.
     */
    public boolean paintElement(Graphics2D g2d, Rectangle2D clipRect) {
        super.paintElement(g2d, clipRect);

        if (this.mGridMode) {
            if (this.isFocused()) {
                final float minX = (float) this.mBounds.getMinX();
                final float maxX = (float) this.mBounds.getMaxX();
                final float minY = (float) this.mBounds.getMinY();
                final float maxY = (float) this.mBounds.getMaxY();
                final float width = maxX - minX;
                final float height = maxY - minY;

                g2d.setPaint(FOCUS_COLOR);
                Rectangle2D rect = new Rectangle2D.Double(minX, minY, width, height);
                g2d.fill(rect);

                // draw anchors
                final int num = NUMBER_OF_ANCHORS - 1;
                final float diffX = (maxX - minX) / num;
                final float diffY = (maxY - minY) / num;
                for (int ii = 0; ii <= num; ii++) {
                    final float y = minY + ii * diffY;
                    for (int jj = 0; jj <= num; jj++) {
                        final float x = minX + jj * diffX;
                        SGUtilityForFigureElementJava2D.drawAnchorAsFocusedObject(x, y, g2d);
                    }
                }
            }

        } else {
            SGDrawingElement[] array = this.mDrawingElementArray;
            if (this.isFocused() && this.mGraph.isSymbolsVisibleAroundFocusedObjects()) {
                final int num = array.length;
                if (num <= MAX_NUMBER_OF_ANCHORS) {
                    for (int ii = 0; ii < num; ii++) {
                        ColorMapRectangle rect = (ColorMapRectangle) array[ii];
                        Point2D pos = new Point2D.Float();
                        pos.setLocation(rect.getX() + rect.getWidth() / 2.0, rect.getY() + rect.getHeight() / 2.0);
                        SGUtilityForFigureElementJava2D.drawAnchorAsFocusedObject(pos, g2d);
                    }
                } else {
                    int div = num / MAX_NUMBER_OF_ANCHORS;
                    int cnt = 0;
                    while (true) {
                        ColorMapRectangle rect = (ColorMapRectangle) array[cnt];
                        Point2D pos = new Point2D.Float();
                        pos.setLocation(rect.getX() + rect.getWidth() / 2.0, rect.getY() + rect.getHeight() / 2.0);
                        SGUtilityForFigureElementJava2D.drawAnchorAsFocusedObject(pos, g2d);
                        cnt += div;
                        if (cnt >= num) {
                            break;
                        }
                    }
                }
            }
        }

        return true;
    }

    /**
     * Returns whether this group set contains the given point.
     * 
     * @param x
     *          the x coordinate
     * @param y
     *          the y coordinate
     * @return
     *          true if this element group contains the given point
     */
    public boolean contains(final int x, final int y) {
    	if (this.mGridMode) {
            Rectangle2D gRect = this.mGraph.getGraphRect();
            Rectangle2D uRect = gRect.createIntersection(this.mBounds);
            return uRect.contains(x, y);
    	} else {
            // if a given point is out of the graph rectangle, returns false;
            Rectangle2D gRect = this.mGraph.getGraphRect();
            if (!gRect.contains(x, y)) {
                return false;
            }
            return super.contains(x, y);
    	}
    }
    
    /**
     * Write properties to the Element object.
     * 
     * @param el
     *          the Element object
     */
    public boolean writeProperty(final Element el) {
        if (super.writeProperty(el) == false) {
            return false;
        }
        el.setAttribute(KEY_VISIBLE, Boolean.toString(this.mVisibleFlag));
        return true;
    }

    /**
     * Read properties from the Element object.
     * 
     * @param el
     *          the Element object
     */
    public boolean readProperty(final Element el) {
        if (super.readProperty(el) == false) {
            return false;
        }
        return this.mGraph.readProperty(this, el);
    }

    @Override
    public float getX(int index) {
        return this.getLocation(index).x;
    }

    @Override
    public float getY(int index) {
        return this.getLocation(index).y;
    }
    
    /**
     * Retruns the location at a given index.
     * 
     * @param index
     *           the index
     * @return the location
     */
    public SGTuple2f getLocation(int index) {
        if (this.mGroupSet instanceof SGElementGroupSetInGraphSXYZ) {
            SGElementGroupSetInGraphSXYZ gs = (SGElementGroupSetInGraphSXYZ) this.mGroupSet;
            return gs.getLocation(index);
        } else {
            throw new Error("Not supported.");
        }
    }

}
