package jp.riken.brain.ni.samuraigraph.figure.java2d;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;

import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.figure.SGILineConstants;

/**
 * A class with static methods for Java2D.
 * 
 */
public class SGUtilityJava2D implements SGILineConstants {

    /**
     * 
     * @param list
     * @return
     */
    public static Rectangle2D getBoundingBox(final ArrayList list) {
        ArrayList rectList = new ArrayList();
        for (int ii = 0; ii < list.size(); ii++) {
            Object obj = list.get(ii);
            if (obj == null) {
                throw new NullPointerException("obj==null");
            }
            if (!(obj instanceof SGIDrawingElementJava2D)) {
                throw new IllegalArgumentException(
                        "!(obj instanceof SGIDrawingElementJava2D)");
            }

            SGIDrawingElementJava2D el = (SGIDrawingElementJava2D) obj;
            rectList.add(el.getElementBounds());
        }

        Rectangle2D rectAll = SGUtility.createUnion(rectList);

        return rectAll;
    }

    /**
     * 
     * @param itr
     * @return
     */
    public static ArrayList getSegmentList(final PathIterator itr) {
        ArrayList list = new ArrayList();
        while (!itr.isDone()) {
            final float[] array = new float[6];
            final int type = itr.currentSegment(array);

            if (type == PathIterator.SEG_MOVETO
                    || type == PathIterator.SEG_LINETO) {
                list.add(new Point2D.Float(array[0], array[1]));
            } else if (type == PathIterator.SEG_QUADTO) {
                list.add(new Point2D.Float(array[0], array[1]));
                list.add(new Point2D.Float(array[2], array[3]));
            } else if (type == PathIterator.SEG_CUBICTO) {
                list.add(new Point2D.Float(array[0], array[1]));
                list.add(new Point2D.Float(array[2], array[3]));
                list.add(new Point2D.Float(array[4], array[5]));
            }

            itr.next();
        }

        return list;
    }

    /**
     * Draw a circle.
     * 
     * @param x
     *            the x-coordinate
     * @param y
     *            the y-coordinate
     * @param size
     *            size of circle
     * @param lineWidth
     *            width of line
     * @param lineColor
     *            color of line
     * @param color
     *            color of body
     * @param g2d
     *            graphics context
     */
    public static void drawCircle(final float x, final float y, final float size,
            final float lineWidth, final Color lineColor, final Color color,
            final Graphics2D g2d) {
        Shape circle = new Ellipse2D.Float(x - 0.5f * size, y - 0.5f * size,
                size, size);
        drawSymbol(circle, lineWidth, lineColor, color, g2d);
    }

    /**
     * Draw a circle.
     * 
     * @param pos
     *            location
     * @param size
     *            size of circle
     * @param lineWidth
     *            width of line
     * @param lineColor
     *            color of line
     * @param color
     *            color of body
     * @param g2d
     *            graphics context
     */
    public static void drawCircle(final Point2D pos, final float size,
            final float lineWidth, final Color lineColor, final Color color,
            final Graphics2D g2d) {
        final float x = (float) pos.getX();
        final float y = (float) pos.getY();
        drawCircle(x, y, size, lineWidth, lineColor, color, g2d);
    }

    /**
     * Draw a square.
     * 
     * @param pos
     *            location
     * @param size
     *            size of square
     * @param lineWidth
     *            width of line
     * @param lineColor
     *            color of line
     * @param color
     *            color of body
     * @param g2d
     *            graphics context
     */
    public static void drawSquare(final Point2D pos, final float size,
            final float lineWidth, final Color lineColor, final Color color,
            final Graphics2D g2d) {
        final float x = (float) pos.getX();
        final float y = (float) pos.getY();
        Shape circle = new Rectangle2D.Float(x - 0.5f * size, y - 0.5f * size,
                size, size);
        drawSymbol(circle, lineWidth, lineColor, color, g2d);
    }
    
    /**
     * Draw a triangle.
     * 
     * @param pos
     *            location
     * @param size
     *            size of triangle
     * @param lineWidth
     *            width of line
     * @param lineColor
     *            color of line
     * @param color
     *            color of body
     * @param g2d
     *            graphics context
     */
    public static void drawTriangle(final Point2D pos, final float size,
            final float lineWidth, final Color lineColor, final Color color,
            final Graphics2D g2d) {
        final float x = (float) pos.getX();
        final float y = (float) pos.getY();
        final float half = 0.5f*size;
        int x1 = (int)Math.floor(x - 0.95f*half);
        int x2 = (int)x;
        int x3 = (int)Math.ceil(x + 0.95f*half);
        int y1 = (int)Math.floor(y - 0.55f*half);
        int y2 = (int)Math.ceil(y + 1.10f*half);
        int y3 = y1;
        int[] xpoints = { x1, x2, x3 };
        int[] ypoints = { y1, y2, y3 };
        int npoints = 3;
        Shape triangle = new Polygon(xpoints, ypoints, npoints);
        drawSymbol(triangle, lineWidth, lineColor, color, g2d);
    }

    /**
     * Draw a symbol.
     * 
     * @param sh
     * @param lineWidth
     * @param lineColor
     * @param color
     * @param g2d
     */
    public static void drawSymbol(final Shape sh, final float lineWidth,
            final Color lineColor, final Color color, final Graphics2D g2d) {
        g2d.setPaint(color);
        g2d.fill(sh);
        g2d.setPaint(lineColor);
        g2d.setStroke(new BasicStroke(lineWidth));
        g2d.draw(sh);
    }

    /**
     * Returns a new Stroke object.
     * 
     * @param type -
     *            the line type of a BasicStroke
     * @param width -
     *            the width of a BasicStroke
     * @param cap -
     *            the decoration of the ends of a BasicStroke
     * @param join -
     *            the decoration applied where path segments meet
     * @param miterLimit -
     *            the limit to trim the miter join. The miterlimit must be
     *            greater than or equal to 1.0f.
     * @param dashPhase -
     *            the offset to start the dashing pattern
     * @return a BasicStroke object
     */
    public static BasicStroke getBasicStroke(final int type, final float width,
            final int cap, final int join, final float miterLimit,
            final float dashPhase) {
        BasicStroke bs = null;
        switch (type) {
        // solid line
        case LINE_TYPE_SOLID: {
            bs = new BasicStroke(width, cap, join);
            break;
        }

        // broken line
        case LINE_TYPE_BROKEN: {
            final float[] dash = { 2.0f * width, width };
            bs = new BasicStroke(width, cap, join, miterLimit, dash, dashPhase);
            break;
        }

        // dotted line
        case LINE_TYPE_DOT: {
            final float[] dash = { width };
            bs = new BasicStroke(width, cap, join, miterLimit, dash, dashPhase);
            break;
        }

        // dashed line
        case LINE_TYPE_DASHED: {
            final float[] dash = { 4.0f * width, width, width, width };
            bs = new BasicStroke(width, cap, join, miterLimit, dash, dashPhase);
            break;
        }

        // dashed line
        case LINE_TYPE_DOUBLE_DASHED: {
            final float[] dash = { 4.0f * width, width, width, width, width,
                    width };
            bs = new BasicStroke(width, cap, join, miterLimit, dash, dashPhase);
            break;
        }

        default: {
            throw new IllegalArgumentException("Line type is invalid.");
        }
        }

        return bs;
    }

    /**
     * Returns the gradient of a segment between given two points.
     * 
     * @return the gradient in units of radian
     */
    public static float getGradient(final float startX, final float startY,
            final float endX, final float endY) {
        final float dx = endX - startX;
        final float dy = endY - startY;
        final float adx = Math.abs(dx);
        final float ady = Math.abs(dy);
        if (adx < Float.MIN_VALUE && ady < Float.MIN_VALUE) {
            return 0.0f;
        }

        float grad;
        if (adx < Float.MIN_VALUE) {
            if (dy > 0.0f) {
                grad = 0.50f * (float) Math.PI;
            } else {
                grad = 1.50f * (float) Math.PI;
            }
        } else if (ady < Float.MIN_VALUE) {
            if (dx > 0.0f) {
                grad = 0.0f;
            } else {
                grad = (float) Math.PI;
            }
        } else {
            final float tangent = dy / dx;
            if (dx > 0.0f) {
                if (dy > 0.0f) {
                    grad = (float) Math.atan(tangent);
                } else {
                    grad = (float) (2.0 * Math.PI + Math.atan(tangent));
                }
            } else {
                grad = (float) (Math.PI + Math.atan(tangent));
            }
        }

        return grad;
    }

    //
    // public static void main( String[] args )
    // {
    // final float[] xArray = { 1, -1, -1, 1 };
    // final float[] yArray = { 1, 1, -1, -1 };
    // for( int ii=0; ii<xArray.length; ii++ )
    // {
    // final float _angle = getAngle( 0, 0, xArray[ii], yArray[ii] );
    // final float _degree = _angle/(float)Math.PI;
    // System.out.println(ii+" "+_degree);
    // }
    // System.out.println();
    // }

}
